#include "clause.hpp"
#include "inline.hpp"
#include "references.hpp"
#include "solver.hpp"
#include "watch.hpp"

#include <cinttypes>

namespace TabularaSAT {

#ifndef QUIET

void Solver::update_trail_average () {

  assert (level);
  size_t assigned = trail.size ();
  assert (assigned >= control[1].trail);
  assigned -= control[1].trail;
  assert (assigned >= assigned_eliminated);
  assigned -= assigned_eliminated;
#ifndef NCHRONO
  assert (assigned >= out_of_order_units);
  assigned -= out_of_order_units;
#endif
  size_t remaining = statistics.active_variables;
  assert (assigned <= remaining);

  double filled = percent (assigned, remaining);
  averages ().trail += filled;

  debug ("trail = %zu, "
         "control[level=%u].trail = %u, "
         "control[1].trail = %u, "
         "assigned_eliminated = %u, "
#ifndef NCHRONO
         "out_of_order_units = %u, "
#endif
         "assigned = %zu, "
         "active = %zu, "
         "filled = %g %%, "
         "average = %g %%",
         trail.size (), level, control[level].trail, control[1].trail,
         assigned_eliminated,
#ifndef NCHRONO
         out_of_order_units,
#endif
         assigned, remaining, filled, (double) averages ().trail);
}

#endif

// Data maintained by the core analysis function to derive the first UIP
// clause and actually updated in the 'analyze_literal' function below.

struct Analyzer {
#ifndef NGLUE
  unsigned glue = 0; // Number of pulled in levels.
#endif
  unsigned jump = 0; // Backjump level.
  unsigned open = 0; // Remaining on current level.
#ifndef NOTFS
  unsigned reason_size = 0;    // Reason clause actual size.
  unsigned resolvent_size = 0; // Virtual resolvent size.
#endif
};

// We factored out this code which analyzes a literal.  Without tagged
// reasons (and virtual binary clauses) we could keep it in the
// 'analyze_conflict' function below, but we want to reuse it for the tagged
// binary clause reason and also for analyzing the falsified literal.

inline void Solver::analyze_literal (Analyzer &analyzer, unsigned lit) {
  unsigned idx = index (lit);
  unsigned l = levels[idx];
  if (!l)
    return;
#ifndef NOTFS
  analyzer.reason_size++;
#endif
  Flags &f = flags[idx];
  if (f.seen)
    return;
#ifndef NOTFS
  analyzer.resolvent_size++;
#endif
  debug ("analyzing literal %s", debug_literal (lit));
  assert (values[lit] < 0);
  analyzed.push_back (idx);
  f.seen = true;
  if (l == level) {
    analyzer.open++;
    return;
  }
  clause.push_back (lit);
  if (l > analyzer.jump) {
    analyzer.jump = l;
    if (clause.size () > 2)
      std::swap (clause[1], clause.back ()); // Watch higher level.
  }
  Frame &frame = control[l];
  if (frame.pulled++)
    return;
  pulled.push_back (l);
#ifndef NGLUE
  analyzer.glue++;
#endif
}

#ifndef NBUMPREASONS

// The following two functions analyzing reason side literals have as main
// effect to save unseen literals in reason clauses of literals from the new
// learned clause on the 'analyzed' stack together with the other seen
// literals which are all later going to be bumped.

void Solver::analyze_reason_side_literal (unsigned lit) {
  unsigned idx = index (lit);
  Flags &f = flags[idx];
  if (f.seen)
    return;
  unsigned l = levels[idx];
  if (!l)
    return;
#ifndef QUIET
  statistics.reason_side_bumped++;
#endif
  debug ("analyzing reason side literal %s", debug_literal (lit));
  assert (values[lit] < 0);
  analyzed.push_back (idx);
  f.seen = true;
}

// It turns out that reason-side bumping is also harmful for formulas with
// many decisions per conflict and in particular if the number of
// additionally bumped reason literals is way larger than the original seen
// literals.  To avoid this negative effects we limit reason side bumping.

void Solver::analyze_reason_side_literals () {

  if (!OPTION (bump_reason_side))
    return;

  if (stable && !OPTION (bump_reason_side_in_stable_mode))
    return;

  if (!stable && !OPTION (bump_reason_side_in_focused_mode))
    return;

  uint64_t &count = delay[stable].bump_reason_side.count;

  if (count) {
    debug ("bumping reason side literals "
           "still delayed %" PRIu64 " more times",
           count);
    count--;
    return;
  }

  {
    double decision_rate = averages ().decision_rate;
    double rate_limit = OPTION (bump_reason_side_rate);
    if (decision_rate > rate_limit) {
      debug ("decision rate %g exceeds reason side bumping limit %g",
             decision_rate, rate_limit);
      return;
    }
  }

  debug ("analyzing %zu reason side literals", clause.size ());

  size_t original_analyzed_size = analyzed.size ();
  size_t analyzed_limit =
      original_analyzed_size * OPTION (bump_reason_side_limit);

  debug ("original analyzed size %zu limit %zu", original_analyzed_size,
         analyzed_limit);

  for (auto lit : clause) {
    unsigned idx = index (lit);
    Reference reason = reasons[idx];
    if (reason == UNIT_REASON)
      continue;
    if (reason == DECISION_REASON)
      continue;
    if (tagged (reason)) {
      unsigned other = untag (reason);
      analyze_reason_side_literal (other);
    } else {
      Clause *c = dereference_clause (reason);
      INC_COUNT (search_ticks);
      for (auto other : *c) {
        if (other == lit)
          continue;
        analyze_reason_side_literal (other);
        if (analyzed.size () > analyzed_limit)
          break;
      }
    }
    if (analyzed.size () > analyzed_limit)
      break;
  }

  uint64_t &current = delay[stable].bump_reason_side.current;

  if (analyzed.size () > analyzed_limit) {
    debug ("reason side analyzed literals limit %zu exceeded",
           analyzed_limit);
    for (size_t i = original_analyzed_size; i != analyzed.size (); i++) {
      unsigned idx = analyzed[i];
      assert (flags[idx].seen);
      flags[idx].seen = false;
      debug ("unanalyzing reason side %s", debug_variable (idx));
    }
    analyzed.resize (original_analyzed_size);
    current++;
    debug ("bumping reason side delay increased to %" PRIu64, current);
  } else {
    debug ("analyzed %zu additional reason side literals",
           analyzed.size () - original_analyzed_size);
    if (current) {
      current /= 2;
      if (!current)
        debug ("bumping reason side delay back to zero");
      else
        debug ("bumping reason side delay halved to %" PRIu64, current);
    } else
      debug ("keeping zero bumping reason side delay");
  }

  count = current;
  debug ("next reason side bumping delay %" PRIu64, count);
}

#endif

void Solver::clear_analyzed () {
  debug ("clearing %zu analyzed literals", analyzed.size ());
  for (auto idx : analyzed)
    flags[idx].seen = false;
  analyzed.clear ();

  debug ("clearing %zu analyzed levels", pulled.size ());
  for (auto l : pulled)
    control[l].pulled = 0;
  pulled.clear ();
}

#ifndef NALLUIPS

void Solver::learn_all_uips (unsigned uip) {
  debug ("learning additional UIPs on assignment level one");
  assert (analyzed.empty ());
  unsigned uip_idx = index (uip);
  unsigned uip_level = levels[uip_idx];
  assert (uip_level == level);
#ifdef LOGGING
  unsigned uips = 0;
#endif
  for (;;) {
    assert (values[uip] > 0);
    assert (levels[uip_idx] == uip_level);
    Reference reason = reasons[uip_idx];
    if (reason == DECISION_REASON)
      break;
    unsigned pos = position[uip_idx];
    unsigned *t = &trail[pos];
    unsigned open = 0;
    do {
      if (tagged (reason)) {
        assert (binary_reference (reason));
        unsigned other = untag (reason);
        debug_binary (uip, other, "analyzing %s reason",
                      debug_literal (uip));
        assert (values[other] < 0);
        unsigned other_idx = index (other);
        assert (levels[other_idx] == uip_level);
        if (!flags[other_idx].seen) {
          flags[other_idx].seen = true;
          analyzed.push_back (other_idx);
          open++;
        }
      } else {
        assert (large_reference (reason));
        INC_COUNT (search_ticks);
        assert (large_reference (reason));
        Clause *c = dereference_clause (reason);
        debug_clause (c, "analyzing %s reason", debug_literal (uip));
        for (auto other : *c) {
          if (other == uip)
            continue;
          assert (values[other] < 0);
          unsigned other_idx = index (other);
          unsigned other_level = levels[other_idx];
          if (!other_level)
            continue;
          assert (other_level <= uip_level);
          if (other_level != uip_level)
            goto DONE;
          if (flags[other_idx].seen)
            continue;
          flags[other_idx].seen = true;
          analyzed.push_back (other_idx);
          open++;
        }
      }
      assert (open);
      open--;
      do {
        assert (t > &trail[0]);
        uip = *--t;
        uip_idx = index (uip);
      } while (!flags[uip_idx].seen);
      reason = reasons[uip_idx];
    } while (open);
    debug ("found additional UIP %s", debug_literal (uip));
    unsigned not_uip = negate (uip);
    trace_unit ('a', not_uip);
    units.push_back (not_uip);
#ifndef QUIET
    statistics.additional_learned_uips++;
#endif
#ifdef LOGGING
    uips++;
#endif
  }
DONE:;
  for (auto idx : analyzed)
    flags[idx].seen = false;
  analyzed.clear ();
  debug ("found %u additional UIPs on decision level one", uips);
}

#endif

void Solver::analyze_conflict () {

  start ("analyze", 4);

#ifndef NOTFS

#ifdef LOGGING
  size_t analysis_round = 1;
#endif

RESTART_CONFLICT_ANALYSIS: // TODO Turn this in a nice loop, maybe?

#endif

  assert (conflict != INVALID_REFERENCE);

  assert (clause.empty ());
  assert (analyzed.empty ());
  assert (conflict != DECISION_REASON);
  assert (conflict != UNIT_REASON);

#if !defined(NCHRONO) || !defined(NOTFS)

  unsigned conflict_size = 0;

  if (OPTION (on_the_fly_strengthen) ||
      OPTION (chronological_backtracking)) {

    // Compute maximum assignment level of all conflicting literals and
    // backtrack to that decision level first.  This is necessary to support
    // chronological backtracking as well as on-the-fly subsumption which
    // both might yield out-of-order assignments.

    unsigned conflict_level; // Maximum level in conflict.
    unsigned jump_level;     // Next bigger level in conflict.

    // Out-of-order assignments can make conflicts forcing.

    unsigned forced_unit = INVALID_LITERAL;

    if (tagged (conflict)) {
      unsigned other = untag (conflict);
      unsigned other_level = levels[index (other)];
      unsigned falsified_level = levels[index (falsified_literal)];
      conflict_level = std::max (other_level, falsified_level);
      jump_level = std::min (other_level, falsified_level);
      if (other_level < conflict_level)
        forced_unit = falsified_literal;
      else if (falsified_level < conflict_level)
        forced_unit = other, conflict = tag_binary (falsified_literal);
    } else {

      conflict_level = 0;
      jump_level = 0;

      unsigned literals_on_conflict_level = 0;

      Clause *c = dereference_clause (conflict);
      for (auto lit : *c) {
        unsigned lit_level = levels[index (lit)];
        if (lit_level)
          conflict_size++;
        if (lit_level > conflict_level) {
          jump_level = conflict_level;
          conflict_level = lit_level;
          literals_on_conflict_level = 1;
          forced_unit = lit;
        } else {
          if (lit_level == conflict_level)
            literals_on_conflict_level++;
          if (lit_level > jump_level)
            jump_level = lit_level;
        }
      }

      if (literals_on_conflict_level > 1)
        forced_unit = INVALID_LITERAL;
      else
        assert (forced_unit != INVALID_LITERAL);
    }

    if (!conflict_level) { // Conflict on root level.
      inconsistent = true;
      trace_empty ('a');
      stop ("analyze");
      return;
    }

    if (forced_unit != INVALID_LITERAL) {
      assert (conflict_level);
      debug ("conflict already forces unit %s on decision level %u",
             debug_literal (forced_unit), conflict_level);
      debug ("possible backjump level %u", jump_level);
      if (conflict_level - jump_level > OPTION (backjump_limit)) {
        debug ("forced chronological backtracking to decision level %u "
               "instead of %u",
               conflict_level - 1, jump_level);
#ifndef QUIET
        statistics.chronological_forced++;
#endif
        jump_level = conflict_level - 1;
      }
      backtrack (jump_level);
      assign (forced_unit, conflict);
      stop ("analyze");
      return;
    }

    if (conflict_level == level)
      debug ("conflict level %u matches decision level", conflict_level);
    else {
      assert (conflict_level < level);
      debug ("conflict level %u smaller decision level", conflict_level);
      backtrack (conflict_level);
    }

    // Use better watches for the conflict in case we find some.

    if (!tagged (conflict)) {
      Clause *c = dereference_clause (conflict);
      unsigned *literals = c->literals;
      unsigned size = c->size, swapped = 0;
      unsigned best[2], saved[2];
      for (size_t i = 0; i != 2; i++) {
        saved[i] = literals[best[i] = i];
        for (size_t j = 2; j != size; j++)
          if (better_watch (literals[j], literals[best[i]]))
            best[i] = j, swapped++;
      }
      if (swapped) {
        unwatch_literal (saved[0], conflict);
        unwatch_literal (saved[1], conflict);
        for (unsigned i = 0; i != 2; i++)
          if (best[i] != i)
            std::swap (literals[i], literals[best[i]]);
        watch_literal (literals[0], literals[1], conflict);
        watch_literal (literals[1], literals[0], conflict);
      }
    }
  }

#endif

  if (!level) { // Conflict on root level.
    inconsistent = true;
    trace_empty ('a');
    stop ("analyze");
    return;
  }

  assert (level);

#ifndef QUIET

  // Here starts the standard conflict analysis which is independent of the
  // chronological backtracking and on-the-fly subsumption handling above.

  averages ().level += level; // Conflict level average.

  // We lump together all analyzed clauses (even if we have multiple rounds
  // of analysis due to on-the-fly strengthening) as one conflict analysis.

  statistics.analyzed_conflicts++;

#endif

  // Update the trail size at conflicts for reporting progress.

  update_trail_average ();

  // Update best and target phases.

#ifndef NUPDATEPHASES
  update_phases ();
#endif

  // We make room for 'not_uip' as the first UIP will stay in the learned
  // clause and is guaranteed to be from the current (decision) level (after
  // we backtracked to the conflict level if necessary above).  The
  // value of the first UIP will be flipped and is going to be watched in
  // case a large clause is learned.  The other watch is selected and put at
  // 'clause[0]' during analyzing reasons in 'analyze_literal'.

  clause.push_back (0);

  // We have factored out several counters that are updated while we analyze
  // literals into this 'Analyzer' object to pass them around easily.

  Analyzer analyzer;

  unsigned uip = negate (falsified_literal);

  {
    // Analyzing a reason amounts to resolving it with the current (virtual)
    // resolvent.   Initially the conflict is analyzed, by adding all its
    // literals to this virtual resolvent.  As this is not a real resolution
    // step we start with counting the number of resolved clauses with '-1'.

#ifndef NOTFS
    unsigned resolved = -1;
#endif
    Reference reason = conflict; // Treat conflict as reason.
    size_t t = trail.size ();    // Walk trail backwards.

    if (tagged (reason))
      analyze_literal (analyzer, falsified_literal);

    for (;;) {

      assert (reason != DECISION_REASON);
      assert (reason != UNIT_REASON);

      // Resolve and analyze the 'reason' clause which initially is the
      // conflict and in later rounds the reason of 'uip'.

#ifndef NOTFS
      resolved++;
      analyzer.reason_size = 0;
#endif
      if (tagged (reason)) {
        assert (binary_reference (reason));
        unsigned other = untag (reason);
        debug_binary (uip, other, "analyzing");
        analyze_literal (analyzer, other);
      } else {
        INC_COUNT (search_ticks);
        assert (large_reference (reason));
        Clause *c = dereference_clause (reason);
        debug_clause (c, "analyzing");
        bump_clause (c);
#ifndef NPROMOTE
        if (OPTION (promote_clauses))
          promote_clause (c);
#endif
        for (auto lit : *c)
          analyze_literal (analyzer, lit);

          // In essence we are now done analyzing the reason but if
          // on-the-fly self-subsuming resolution is enabled, we can shrink
          // the reason clause if the resolvent is smaller than the reason.

#ifndef NOTFS

        if (OPTION (on_the_fly_strengthen) && 1 < analyzer.resolvent_size &&
            analyzer.resolvent_size < analyzer.reason_size) {

          // This 'shrinking' is called 'on-the-fly strengthening'.
#ifndef QUIET
          statistics.on_the_fly_strengthened++;
#endif
          assert (analyzer.reason_size == analyzer.resolvent_size + 1);

          debug_clause (c,
                        "removing %s due to "
                        "on-the-fly self-subsuming resolution in",
                        debug_literal (uip));

          clear_analyzed ();
          clause.clear ();

          copy_clause (c, uip);
          assert (clause.size () == analyzer.resolvent_size);

          Reference saved = conflict;

          if (analyzer.resolvent_size == 2) {
            strengthen_to_binary_clause (c);
            conflict = tag_binary (clause[1]);
            assert (tagged (conflict));
            falsified_literal = clause[0];
          } else {
            strengthen_to_large_clause (c);
            conflict = reason;
            falsified_literal = INVALID_LITERAL;
          }

          // All variables in an irredundant shrunken clause are considered
          // to be added now as the shrunken clause might trigger additional
          // clause subsumptions in the next subsumption round.  This is
          // done in the 'strengthen_...' functions above.  In addition we
          // consider the removed variable as removed to potentially trigger
          // additional variable attempts on the removed variable.

          if (!c->redundant)
            mark_removed (uip);

          // If on-the-fly subsumption happens after the first resolution
          // the shrunken clause might also subsume the conflict clause.

          if (resolved == 1 && analyzer.resolvent_size < conflict_size) {

            assert (!tagged (saved));
            assert (analyzer.resolvent_size + 1 == conflict_size);
            Clause *d = dereference_clause (saved);

            // This subsumption is called 'on-the-fly subsumption'.
#ifndef QUIET
            statistics.on_the_fly_subsumed++;
#endif
            mark_clause_garbage (d, "on-the-fly subsumed");

            unwatch_literal (d->literals[0], saved);
            unwatch_literal (d->literals[1], saved);

            if (analyzer.resolvent_size > 2 && c->redundant &&
                !d->redundant) {
              c->redundant = false;
              update_position (c);
              DEC_COUNT (redundant_clauses);
              INC_COUNT (irredundant_clauses);
              debug_clause (c, "subsuming clause changed to be");
            }
          }

#ifdef LOGGING
          analysis_round++;

          debug ("entering conflict analysis round %zu", analysis_round);
          if (tagged (conflict))
            debug_binary (clause[0], clause[1], "new conflict");
          else
            debug_reference (conflict, "new conflict");
#endif
          clause.clear ();

          goto RESTART_CONFLICT_ANALYSIS;
        }

#endif
      }

#ifndef NOTFS
      assert (analyzer.resolvent_size);
      analyzer.resolvent_size--;
#endif
      unsigned uip_idx;
      do {
        assert (t);
        uip = trail[--t];
        uip_idx = index (uip);
      } while (!flags[uip_idx].seen || levels[uip_idx] != level);

      assert (analyzer.open);
      if (!--analyzer.open)
        break;

      reason = reasons[uip_idx];
    }
  }

  debug ("first UIP %s", debug_literal (uip));
  debug ("original jump level %u", analyzer.jump);
#ifndef NGLUE
  assert (analyzer.glue == pulled.size ());
  debug ("glucose level %u (LBD/glue)", analyzer.glue);
  averages ().glue_slow += analyzer.glue;
  averages ().glue_fast += analyzer.glue;
#ifndef QUIET
  if (analyzer.glue <= OPTION (tier1_max_glue))
    statistics.learned_tier1_clauses++;
  else if (analyzer.glue <= OPTION (tier2_max_glue))
    statistics.learned_tier2_clauses++;
#endif
#endif

#ifndef NBUMPREASONS
  {
    uint64_t current = GET_COUNT (decisions);
    uint64_t previous = last.decisions_at_conflict[stable];
    uint64_t delta = current - previous;
    last.decisions_at_conflict[stable] = current;
    averages ().decision_rate += delta;
    debug ("decision previous=%zu current=%zu delta=%zu average=%g",
           previous, current, delta, (double) averages ().decision_rate);
  }
#endif

  assert (uip != INVALID_LITERAL);
  unsigned not_uip = negate (uip);
  clause[0] = not_uip;

  debug_vector (clause, "deduced first UIP");
  size_t deduced_clause_size = clause.size ();
#ifndef QUIET
  statistics.deduced_literals += deduced_clause_size;
#endif
  if (deduced_clause_size > 2) {
#ifndef NSHRINK
    if (OPTION (shrink_learned_clause))
      shrink_deduced_clause ();
    else
#endif
        if (OPTION (minimize_learned_clause))
      minimize_deduced_clause ();
  }

  size_t learned_clause_size = clause.size ();
#ifndef QUIET
  statistics.learned_literals += learned_clause_size;
  statistics.learned_clauses++;
#endif
  iterating = (learned_clause_size == 1);
  assert (learned_clause_size > 0);

  assert (level > analyzer.jump);
#ifndef NCHRONO
  if (OPTION (chronological_backtracking) &&
      level - analyzer.jump > OPTION (backjump_limit)) {
    debug ("forcing chronological backtracking "
           "instead of jumping over %u level",
           level - analyzer.jump);
    analyzer.jump = level - 1;
  }
#endif
#ifndef QUIET
  if (analyzer.jump + 1 == level)
    statistics.chronological_backtracking++;
#endif
#ifndef NBUMPREASONS
  analyze_reason_side_literals ();
#endif
  bump_analyzed_variables ();
  clear_analyzed ();

  trace_temporary ('a');

  if (learned_clause_size == 1) {
#ifndef QUIET
    statistics.learned_units++;
#endif
#ifndef NALLUIPS
    if (OPTION (all_uips))
      learn_all_uips (uip);
#endif
  }

  backtrack (analyzer.jump);

  Reference learned_clause;
  if (learned_clause_size == 1)
    learned_clause = UNIT_REASON;
  else if (learned_clause_size == 2)
    learned_clause = new_binary ();
  else {
    if (clauses.size () >= MAXIMUM_POSITION_OF_CLAUSE) {
      std::vector<unsigned> saved;
      std::swap (clause, saved);
      report_progress ('/');
      reduce_redundant_clauses (true);
      if (clauses.size () >= MAXIMUM_POSITION_OF_CLAUSE) {
        report_progress ('U');
#ifndef NUSED
        for (auto c : clauses)
          if (c->redundant)
            c->used = 0;
#endif
        reduce_redundant_clauses (true);
        if (clauses.size () >= MAXIMUM_POSITION_OF_CLAUSE) {
          report_progress ('G');
#ifndef NGLUE
          for (auto c : clauses)
            if (c->redundant)
              c->glue = MAXIMUM_GLUE;
#endif
          reduce_redundant_clauses (true);
        }
      }
      std::swap (clause, saved);
    }
#ifndef NGLUE
    learned_clause = new_clause (true, analyzer.glue);
#else
    learned_clause = new_clause (true);
#endif
  }

  assign (not_uip, learned_clause);

  clause.clear ();

  // TODO describe why this is here ...

  if (!units.empty ()) {
    for (auto unit : units)
      assign (unit, UNIT_REASON);
    units.clear ();
  }

  stop ("analyze");
}

}; // namespace TabularaSAT
