#ifndef NSWITCH
#include "radix.hpp"
#include <algorithm>
#endif
#include "solver.hpp"

namespace TabularaSAT {

#ifndef NSWITCH

struct rank_stamp {
  Queue &queue;
  rank_stamp (Queue &q) : queue (q) {}
  unsigned operator() (unsigned idx) const { return queue.stamped (idx); }
};

struct smaller_stamp {
  Queue &queue;
  smaller_stamp (Queue &q) : queue (q) {}
  bool operator() (unsigned a, unsigned b) const {
    return queue.stamped (a) < queue.stamped (b);
  }
};

void Solver::bump_focused_variables () {

  if (analyzed.size () < OPTION (radix_sort_limit))
    std::sort (analyzed.begin (), analyzed.end (), smaller_stamp (queue));
  else
    radix_sort_unsigned (analyzed, rank_stamp (queue));

  for (auto idx : analyzed) {
    queue.dequeue (idx);
    queue.enqueue (idx, !values[literal (idx)]);
    debug ("bumped %s stamped %u", debug_variable (idx),
           queue.stamped (idx));
  }
}

#endif

#ifndef NACTIVITY

void Solver::rescale_clause_activities () {
  double factor = 1.0 / CLAUSE_ACTIVITY_LIMIT;
  if (first_redundant != INVALID_POSITION) {
    Clause *begin = clauses[first_redundant];
    Clause *end = *clauses.end ();
    for (Clause *c = begin; c != end; c = c->next ())
      if (c->redundant)
        c->activity.as_float *= factor;
  }
  clause_activity_increment *= factor;
}

#endif

void Solver::bump_stable_variables () {
  for (auto idx : analyzed) {
    bumper.bump (idx);
    debug ("bumped %s with new score %g", debug_variable (idx),
           bumper.score (idx));
  }
  bumper.bump ();
}

void Solver::bump_analyzed_variables () {
#ifndef QUIET
  {
    size_t bumped = analyzed.size ();
    debug ("bumped %zu variables", bumped);
    statistics.bumped_variables += bumped;
  }
#endif
#ifndef NSWITCH
  if (!stable)
    bump_focused_variables ();
  else
#endif
    bump_stable_variables ();
#ifndef NACTIVITY
  clause_activity_increment *= clause_activity_factor;
  if (clause_activity_increment > CLAUSE_ACTIVITY_LIMIT)
    rescale_clause_activities ();
#endif
}

void Solver::bump_clause (Clause *c) {
  if (c->garbage)
    return;
  if (!c->redundant)
    return;
#ifndef NUSED
  c->used = 1 + (c->glue <= OPTION (tier2_max_glue));
#endif
#ifndef NACTIVITY
  c->activity.as_float += clause_activity_increment;
  if (c->activity.as_float > CLAUSE_ACTIVITY_LIMIT)
    rescale_clause_activities ();
#endif
  debug_clause (c, "bumped");
}

}; // namespace TabularaSAT
