#include "connect.hpp"
#include "clause.hpp"
#include "inline.hpp"
#include "references.hpp"
#include "solver.hpp"
#include "watch.hpp"

#include <cinttypes>

namespace TabularaSAT {

void Solver::flush_large_watches_and_satisfied_binary_clauses () {

  assert (watching);
  assert (!level);

  size_t duplicated = 0;
  size_t flushed = 0;
  size_t satisfied = 0;

  for (auto watches : matrix) {
    unsigned lit = watches->lit;
    if (lit == INVALID_LITERAL)
      continue;
    signed char lit_value = values[lit];
    auto begin = watches->begin ();
    auto end = watches->end ();
    auto q = begin, p = q;
    while (p != end) {
      Reference ref = *p++;
      if (!tagged (ref)) {
        p++, flushed++;
        continue;
      }
      unsigned other = untag (ref);
      signed char other_value = values[other];
#ifndef NDEBUG
      if (lit_value < 0)
        assert (other_value > 0);
      if (other_value < 0)
        assert (lit_value > 0);
#endif
      if (lit_value > 0 || other_value > 0) {
        if (lit > other) {
          mark_binary_garbage (lit, other, "flushing satisfied");
          satisfied++;
        }
        continue;
      }
      signed char mark = marks[other];
      if (mark) {
        if (lit > other) {
          mark_binary_garbage (lit, other, "flushing duplicated");
#ifndef QUIET
          statistics.removed_duplicated_binaries++;
#endif
          duplicated++;
        }
        continue;
      }
      marks[other] = 1;
      *q++ = ref;
    }
    watches->resize (q - begin);

    // As we are flushing and filling references frequently during initial
    // preprocessing or multiple rounds of subsumption or variable
    // elimination during inprocessing it is a bad idea to force compacting
    // the references arena all the time by 'clear_and_mark_as_garbage'.

    for (auto ref : *watches) {
      unsigned other = untag (ref);
      marks[other] = marks[negate (other)] = 0;
    }
  }
#ifndef QUIET
  message (2, "flushed %zu satisfied and %zu duplicated binary clauses",
           satisfied, duplicated);
  message (2, "flushed %zu large watches", flushed);

  if (duplicated || verbosity > 1)
    report_progress ('2');
#else
  (void) duplicated;
  (void) satisfied;
  (void) flushed;
#endif
}

void Solver::connecting_all_large_clauses () {
  assert (!level);
  assert (!watching);
  size_t satisfied = 0, connected = 0;
  for (auto c : clauses) {
    if (c->redundant)
      continue;
    if (c->garbage)
      continue;
    if (satisfied_clause (c)) {
      satisfied++;
      mark_clause_garbage (c, "satisfied");
    } else {
      connected++;
      Reference ref = reference_clause (c);
      for (auto lit : *c)
        connect_clause (lit, ref);
    }
  }
#ifndef QUIET
  message (2, "connected %zu large clauses and skipping %zu satisfied",
           connected, satisfied);
#else
  (void) satisfied;
  (void) connected;
#endif
}

void Solver::transition_to_dense_connections () {
  assert (!level);
  assert (watching);
  message (1, "transitioning to dense connections");
  flush_large_watches_and_satisfied_binary_clauses ();
  watching = false;
  connecting_all_large_clauses ();
}

// In inprocessing variable elimination after learning some clauses we add
// resolvents at the end of the arena after these learned clauses.  This
// makes the traversal of the arena more expensive than necessary if the
// traversal can be restricted to some subset.  In particular during
// reduction, which happens more frequently than other repeated
// procedures, this is particularly unfortunate, as clause data-base
// reduction can focus (most of the time) on redundant clauses (actually
// even 'tier2' clauses). This effect is pronounced by the effectiveness
// of variable elimination if many variables are eliminated leading to
// many irredundant clauses being added after redundant clauses.

// Other sources for producing irredundant clauses in the middle of
// redundant clause is on-the-fly double self-subsuming resolution during
// conflict clause analysis and subsumption during vivification (search
// for occurrences of '...->redundant = false').

// Therefore the following procedure moves the redundant clauses again to
// the end of the arena.  It is called before watching all clauses again
// (after variable elimination, subsumption, and local search) which first
// happens increasingly rarely compared to reduction and further is anyhow
// considered more expensive than copying the redundant clauses (twice)
// and the irredundant clauses sitting after the first redundant clause.

void Solver::move_redundant_clauses () {

  if (!OPTION (move_redundant_clauses))
    return;

  if (first_redundant >= last_irredundant) {
    debug ("no need to move redundant clauses");
    return;
  }

  {
    double fraction = OPTION (wasted_accessed) / 100.0;
    uint64_t redundant = GET_COUNT (redundant_clauses);
    uint64_t limit = fraction * redundant;
    bool smaller = (wasted_accessed < limit);
    message (2,
             "wasted accessed irredundant clauses %" PRIu64 " %s "
             "limit %" PRIu64 " = %.g * %" PRIu64,
             wasted_accessed, smaller ? "smaller than" : "reached", limit,
             fraction, redundant);
    if (smaller)
      return;
  }

  wasted_accessed = 0; // Wait for enough wasted accesses again.
#ifndef QUIET
  statistics.moved_redundant_clauses++;
#endif
  message (2, "last irredundant position %zu after first redundant %zu",
           last_irredundant, first_redundant);

  Clause *begin_redundant = clauses[first_redundant];
  Clause *end_clauses = *clauses.end ();
  Clause *end_irredundant = clauses[last_irredundant];
  assert (end_irredundant <= end_clauses);
  if (end_irredundant < end_clauses)
    end_irredundant = end_irredundant->next ();

  // We are conservative in terms of the number of bytes needed for the copy
  // space and just allocate the size of the mixed part (first redundant to
  // last irredundant clause).  With 'malloc' we will anyhow only access the
  // data needed (relying on page allocation an virtual memory done right).

  size_t copy_space_bytes =
      (char *) end_irredundant - (char *) begin_redundant;

  Clause *copy_space = (Clause *) malloc (copy_space_bytes);
  assert (copy_space_bytes);
  if (!copy_space)
    fatal ("out-of-memory allocating redundant clauses copy space");

#ifndef NDEBUG
  char *end_copy_space = (char *) copy_space + copy_space_bytes;
#endif
  message (2,
           "copying redundant clauses into %zu Bytes (%.0f MB) copy space",
           copy_space_bytes, copy_space_bytes / (double) (1u << 20));

  size_t redundant_clauses = 0;
  size_t irredundant_clauses = 0;
  Clause *last_irredundant_clause = 0;

  Clause *dst1 = begin_redundant; // Overwrite original clauses.
  Clause *dst2 = copy_space;      // Write into copy space.
  Clause *src, *next_src;         // Read from original or copy space.

  // First go over the mixed part (from first redundant to the last
  // irredundant clause), move the irredundant clauses downward and copy the
  // redundant clauses into the copy space.  Skip garbage clauses and
  // shrink shortened clauses on-the-fly.

  for (src = begin_redundant; src != end_irredundant; src = next_src) {

    next_src = src->next ();

    if (src->garbage)
      continue;

    size_t src_bytes;
    if (src->shortened) {
      src->shortened = false;
      src_bytes = (char *) &src->literals[src->size] - (char *) src;
    } else
      src_bytes = (char *) next_src - (char *) src;

    if (src->redundant) {
      assert ((char *) dst2 + src_bytes <= end_copy_space);
      memcpy (dst2, src, src_bytes);
      dst2 = (Clause *) ((char *) dst2 + src_bytes);
      redundant_clauses++;
    } else {
      memmove (dst1, src, src_bytes);
      last_irredundant_clause = dst1;
      dst1 = (Clause *) ((char *) dst1 + src_bytes);
      irredundant_clauses++;
    }
  }

#ifndef QUIET
  size_t redundant_bytes = (char *) dst2 - (char *) copy_space;
  size_t irredundant_bytes = copy_space_bytes - redundant_bytes;
#endif

  if (last_irredundant_clause)
    last_irredundant = clauses.position (last_irredundant_clause);
  else
    last_irredundant = first_redundant;

  debug ("new last irredundant position %zu", (size_t) last_irredundant);

  first_redundant = clauses.position (dst1);
  debug ("new first redundant position %zu", (size_t) first_redundant);

  // Second copy back the redundant clauses from the copy space into the
  // remaining original space in the clause arena right before all the other
  // remaining redundant clauses.

#ifndef NGLUE
  first_tier2 = INVALID_POSITION;
  unsigned tier1 = OPTION (tier1_max_glue);
#endif
  for (src = copy_space; src != dst2; src = next_src) {
    assert (src->redundant);
    assert (!src->garbage);
    next_src = src->next ();
    size_t src_bytes = (char *) next_src - (char *) src;
#ifndef NGLUE
    if (first_tier2 == INVALID_POSITION && tier1 < src->glue) {
      first_tier2 = clauses.position (dst1);
      debug ("new first redundant tier2 position %zu",
             (size_t) first_tier2);
    }
#endif
    memcpy (dst1, src, src_bytes);
    dst1 = (Clause *) ((char *) dst1 + src_bytes);
  }

  free (copy_space);

  // With garbage or shortened clauses we now have some space left in the
  // mixed part and have to move the remaining redundant clauses too.

  assert (dst1 <= end_irredundant);
  if (first_tier2 == INVALID_POSITION || dst1 < end_irredundant) {
    for (src = end_irredundant; src != end_clauses; src = next_src) {

      next_src = src->next ();

      if (src->garbage)
        continue;

      size_t src_bytes;
      if (src->shortened) {
        src->shortened = false;
        src_bytes = (char *) &src->literals[src->size] - (char *) src;
      } else
        src_bytes = (char *) next_src - (char *) src;

      assert (src->redundant);
#ifndef NGLUE
      if (first_tier2 == INVALID_POSITION && tier1 < src->glue) {
        first_tier2 = clauses.position (dst1);
        debug ("new first redundant tier2 position %zu",
               (size_t) first_tier2);
      }
#endif
      memmove (dst1, src, src_bytes);
      dst1 = (Clause *) ((char *) dst1 + src_bytes);
      redundant_clauses++;
    }
    clauses.resize (dst1);
  }

#ifndef NGLUE
  if (first_tier2 == INVALID_POSITION)
    message (2, "no redundant tier2 clause left");
#endif

#ifndef QUIET
  message (2, "moved %zu redundant clauses %.0f MB %.0f%%",
           redundant_clauses, redundant_bytes / (double) (1 << 20),
           percent (redundant_bytes, copy_space_bytes));
  message (2, "moved %zu irredundant clauses %.0f MB %.0f%%",
           irredundant_clauses, irredundant_bytes / (double) (1 << 20),
           percent (irredundant_bytes, copy_space_bytes));
#else
  (void) redundant_clauses;
  (void) irredundant_clauses;
#endif
}

void Solver::watch_all_large_clauses () {

  move_redundant_clauses (); // Should be way cheaper than rewatching.

  assert (!level);
  assert (watching);
  size_t watched = 0;
  for (auto c : clauses) {
    if (c->garbage)
      continue;
    watch_clause (c);
    watched++;
  }
#ifndef QUIET
  message (2, "watching %zu large clauses again", watched);
#else
  (void) watched;
#endif
}

void Solver::flush_connected_large_clauses () {
  assert (!watching);
  size_t flushed = 0;
  for (auto occurrences : matrix) {
    if (occurrences->lit == INVALID_LITERAL)
      continue;
    auto begin = occurrences->begin ();
    auto end = occurrences->end ();
    auto q = begin, p = q;
    Reference ref;
    while (p != end)
      if (tagged (ref = *p++))
        *q++ = ref;
      else
        flushed++;
    occurrences->resize (q - begin);
    if (occurrences->empty ())
      matrix.clear_and_mark_as_garbage (occurrences);
  }
#ifndef QUIET
  message (2, "flushed %zu connected large clause references", flushed);
#else
  (void) flushed;
#endif
}

void Solver::transition_to_sparse_watching () {
  assert (!level);
  assert (!watching);
  message (1, "transitioning back to sparse watching");
  flush_connected_large_clauses ();
  watching = true;
  watch_all_large_clauses ();
}

}; // namespace TabularaSAT
