#include "eliminate.hpp"
#include "references.hpp"
#include "schedule.hpp"
#include "solver.hpp"

#include <cinttypes>

namespace TabularaSAT {

#ifndef NELIMINATIONHEAP

inline double Solver::elimination_score (unsigned idx, unsigned type) {
  unsigned lit = literal (idx);
  unsigned not_lit = negate (lit);
  double pos = references (lit)->size ();
  double neg = references (not_lit)->size ();
  double product = pos * neg;
  double sum = pos + neg;
  double res;
  if (type == 3)
    res = product - sum;
  else if (type == 2)
    res = product + sum;
  else if (type == 1)
    res = product;
  else
    res = sum;
  return res;
}

struct better_elimination_candidate {
  Solver &solver;
  const bool irrelevant;
  const bool irrelevant_bumper;
#ifndef NQUEUE
  const bool irrelevant_queue;
#endif
  const unsigned scoring;
  const bool use_scores;
  better_elimination_candidate (Solver &s)
      : solver (s), irrelevant (SOLVER_OPTION (eliminate_irrelevant)),
        irrelevant_bumper (irrelevant &&
#ifndef NSWITCH
                           solver.stable &&
#endif
                           solver.bumper.capacity ()),
#ifndef NQUEUE
        irrelevant_queue (irrelevant &&
#ifndef NSWITCH
                          !solver.stable &&
#endif
                          solver.queue.capacity ()),
#endif
        scoring (SOLVER_OPTION (elimination_scoring)),
        use_scores (SOLVER_OPTION (elimination_scores)) {
  }
  inline bool operator() (unsigned a, unsigned b) {
    if (irrelevant_bumper) {
      auto s = solver.bumper.score (a);
      auto t = solver.bumper.score (b);
      if (s < t)
        return true;
      if (s > t)
        return false;
    }
#ifndef NQUEUE
    if (irrelevant_queue) {
      auto s = solver.queue.stamped (a);
      auto t = solver.queue.stamped (b);
      if (s < t)
        return true;
      if (s > t)
        return false;
    }
#endif
    if (use_scores) {
      double s = solver.elimination_score (a, scoring);
      double t = solver.elimination_score (b, scoring);
      if (s < t)
        return true;
      if (s > t)
        return false;
    }
    return a < b;
  }
};

#endif

bool Solver::eliminating_variables () {
  if (!OPTION (inprocessing))
    return false;
  if (!OPTION (eliminate_variables))
    return false;
  if (disabled.elimination)
    return false;
  if (GET_COUNT (conflicts) < limits.eliminate)
    return false;
  if (OPTION (trigger_elimination)) {
    if (!trigger.elimination) {
      debug ("waiting for next reduction before elimination");
      trigger.elimination = 1;
      return false;
    } else if (trigger.elimination < 2)
      return false;
  }
  return true;
}

#ifndef NELIMINATIONHEAP

void Solver::gather_elimination_candidates (Eliminator &eliminator) {
  eliminator.heap.resize (variables);
#ifndef QUIET
  size_t scheduled = 0;
#endif
  for (auto idx : variables)
    if (flags[idx].active) {
      assert (!values[literal (idx)]);
      eliminator.heap.update (idx, better_elimination_candidate (*this));
      eliminator.heap.push (idx, better_elimination_candidate (*this));
#ifndef QUIET
      scheduled++;
#endif
    }
  message (1, "gathered %zu active variable elimination candidates %.0f%%",
           scheduled, percent (scheduled, variables));
}

void Solver::schedule_variable (Eliminator &eliminator, unsigned idx) {
  assert (OPTION (elimination_heap));
  if (!flags[idx].active)
    return;
  eliminator.heap.update (idx, better_elimination_candidate (*this));
  if (!eliminator.heap.contains (idx))
    eliminator.heap.push (idx, better_elimination_candidate (*this));
}

inline void Solver::schedule_literal (Eliminator &eliminator,
                                      unsigned lit) {
  assert (OPTION (elimination_heap));
  schedule_variable (eliminator, index (lit));
}

inline void Solver::schedule_clause (Eliminator &eliminator, Clause *c) {
  assert (OPTION (elimination_heap));
  for (auto lit : *c)
    schedule_literal (eliminator, lit);
}

#endif

void Solver::eliminate_clauses_of_literal (Eliminator &eliminator,
                                           unsigned lit, bool push) {
  auto occurrences = references (lit);
  auto begin = occurrences->begin ();
  auto end = occurrences->end ();
  auto p = begin;
  while (p != end) {
    Reference ref = *p++;
    if (tagged (ref)) {
      unsigned other = untag (ref);
      mark_binary_garbage (lit, other, "eliminated");
#ifndef NELIMINATIONHEAP
      if (OPTION (elimination_heap))
        schedule_literal (eliminator, other);
#endif
      if (!push)
        continue;
      reconstruction_stack.push_back (INVALID_LITERAL);
      reconstruction_stack.push_back (lit);
      reconstruction_stack.push_back (other);
      debug_literals (2, &*reconstruction_stack.end () - 2, "saved");
    } else {
      Clause *c = dereference_clause (ref);
      if (c->garbage)
        continue;
      mark_clause_garbage (c, "eliminated");
#ifndef NELIMINATIONHEAP
      if (OPTION (elimination_heap))
        schedule_clause (eliminator, c);
#endif
      if (!push)
        continue;
      reconstruction_stack.push_back (INVALID_LITERAL);
      reconstruction_stack.push_back (lit);
      for (auto other : *c)
        if (other != lit)
          reconstruction_stack.push_back (other);
      debug_literals (c->size, &*reconstruction_stack.end () - c->size,
                      "saved");
    }
  }
}

void Solver::eliminate_clauses_of_variable (Eliminator &eliminator,
                                            unsigned pivot_idx) {

  Flags &f = flags[pivot_idx];
  assert (f.active);
  assert (!f.eliminated);
  assert (!f.fixed);
#ifndef NSUBSTITUTE
  assert (!f.substituted);
#endif
  f.eliminated = true;
#ifndef QUIET
  statistics.eliminated_variables++;
#endif
  deactivate_variable (pivot_idx);

  debug ("eliminating clauses with %s", debug_variable (pivot_idx));

  unsigned pivot = literal (pivot_idx);
  unsigned not_pivot = negate (pivot);

  auto pos_occs = references (pivot);
  auto neg_occs = references (not_pivot);

  if (pos_occs->size () > neg_occs->size ()) {
    debug ("pushing %s clauses on reconstruction stack",
           debug_literal (not_pivot));
    eliminate_clauses_of_literal (eliminator, pivot, false);
    eliminate_clauses_of_literal (eliminator, not_pivot, true);
    reconstruction_stack.push_back (INVALID_LITERAL);
    reconstruction_stack.push_back (pivot);
  } else {
    debug ("pushing %s clauses on reconstruction stack",
           debug_literal (pivot));
    eliminate_clauses_of_literal (eliminator, pivot, true);
    eliminate_clauses_of_literal (eliminator, not_pivot, false);
    reconstruction_stack.push_back (INVALID_LITERAL);
    reconstruction_stack.push_back (not_pivot);
  }

  debug_literals (1, &*reconstruction_stack.end () - 1, "saved");
  matrix.clear_and_mark_as_garbage (pos_occs);
  matrix.clear_and_mark_as_garbage (neg_occs);

  debug ("eliminated %s", debug_variable (pivot_idx));
}

void Solver::mark_eliminated_redundant_clauses_as_garbage (unsigned round) {
  assert (!watching);
  if (first_redundant == INVALID_POSITION)
    return;
  Clauses::iterator begin = clauses[first_redundant], end = clauses.end ();
#ifndef QUIET
  size_t eliminated = 0;
#endif
  for (Clauses::iterator i = begin; i != end; ++i) {
    Clause *c = *i;
    if (c->garbage)
      continue;
    if (!c->redundant)
      continue;
    bool eliminate = false;
    for (auto lit : *c)
      if (flags[index (lit)].eliminated) {
        eliminate = true;
        break;
      }
    if (!eliminate)
      continue;
    mark_clause_garbage (c, "eliminated");
#ifndef QUIET
    eliminated++;
#endif
  }
  message (2, "eliminated %zu redundant clauses in round %u", eliminated,
           round);
}

void Solver::flush_eliminated_binary_clauses () {
  assert (!watching);
#ifndef QUIET
  size_t flushed = 0;
#endif
  for (auto occurrences : matrix) {
    unsigned lit = occurrences->lit;
    if (lit == INVALID_LITERAL)
      continue;
    unsigned idx = index (lit);
    Flags &f = flags[idx];
    if (f.eliminated) {
      assert (occurrences->empty ());
      continue;
    }
    auto begin = occurrences->begin ();
    auto end = occurrences->end ();
    auto q = begin, p = q;
    while (p != end) {
      Reference ref = *q++ = *p++;
      if (!tagged (ref))
        continue;
      unsigned other = untag (ref);
      if (!flags[index (other)].eliminated)
        continue;
#ifndef QUIET
      flushed++;
#endif
      q--;
    }
    occurrences->resize (q - begin);
  }
  message (2, "eliminated and flushed %zu binary occurrences", flushed);
}

bool Solver::eliminate_variables () {

  if (level && !backtrack_to_root_level_and_propagate ())
    return false;

  check_all ();

  start_simplifier ("eliminate");
  assert (!eliminating);
  eliminating = true;

  EliminationScheduler elimination_scheduler (this);
  elimination_scheduler.increment_counter ();

  uint64_t limit;
  if (preprocessing)
    limit = UNLIMITED;
  else {
    uint64_t ticks = elimination_scheduler.get_ticks_limit ();
    limit = GET_COUNT (elimination_ticks) + ticks;
  }

  message (1, "starting elimination with %" PRIu64 " variables removed",
           statistics.removed_variables);

  transition_to_dense_connections ();

  unsigned max_rounds = OPTION (elimination_rounds);
  bool done = false, progress = false;
  unsigned eliminated = UINT_MAX;

  assert (max_rounds < UINT_MAX);
  for (unsigned round = 1; !done && round <= max_rounds; round++) {

    if (inconsistent || terminated)
      break;

#ifndef QUIET
    message (2, "starting elimination round %u", round);
    uint64_t start = statistics.resolutions;
#endif
    bool completed = false;
    eliminated = 0;

    Eliminator eliminator;
#ifndef NELIMINATIONHEAP
    bool use_heap = OPTION (elimination_heap);
    if (use_heap)
      gather_elimination_candidates (eliminator);
#endif
    unsigned idx = UINT_MAX;

    for (;;) {

      if (inconsistent || terminated)
        break;

      if (GET_COUNT (elimination_ticks) >= limit) {
        assert (!completed);
        message (2, "elimination ticks limit hit in round %u", round);
        done = true;
        break;
      }

#ifndef NELIMINATIONHEAP
      if (use_heap) {
        if (eliminator.heap.empty ())
          break;
        idx = *eliminator.heap;
        eliminator.heap.pop (better_elimination_candidate (*this));
      } else
#endif
      {
        if (++idx == variables)
          break;
      }

      Flags &f = flags[idx];
      if (!f.active)
        continue;
      if (!f.eliminate)
        continue;
      f.eliminate = false;
      if (!generate_resolvents (eliminator, idx))
        continue;

      add_resolvents (eliminator, idx);
#if !defined(QUIET) && !defined(NGATES)
      if (eliminator.extracted) {
        statistics.eliminated_gates++;
        (*eliminator.extracted)++;
      }
#endif
      if (inconsistent)
        break;
      eliminate_clauses_of_variable (eliminator, idx);
      progress = true;
      eliminated++;

      garbage_collect_clauses_during_variable_elimination ();
    }

#ifndef QUIET
    const int postponed_verbosity_level = 2;
    size_t remain = 0, total = 0;
#endif
    completed = true;

    for (auto idx : variables) {
      Flags &f = flags[idx];
      if (!f.active)
        continue;
      if (f.eliminate) {
#ifndef QUIET
        remain++;
#endif
        completed = false;
#ifndef QUIET
        if (verbosity < postponed_verbosity_level)
#endif
          break;
      }
#ifndef QUIET
      total++;
#endif
    }

    message (postponed_verbosity_level,
             "postponed %zu remaining elimination candidates %.0f%% in "
             "round %u",
             remain, percent (remain, total), round);

    message (1, "eliminated %u variables in round %u with bound %zu",
             eliminated, round, elimination_bound);

#ifndef QUIET
    uint64_t resolutions = statistics.resolutions - start;
#endif
    if (completed) {
      message (2,
               "complete variable elimination round after %" PRIu64
               " resolutions",
               resolutions);
      if (elimination_bound < OPTION (maximum_elimination_bound)) {
        elimination_bound = elimination_bound ? 2 * elimination_bound : 1;
        message (2, "new elimination bound %zu", elimination_bound);
        for (auto idx : variables)
          flags[idx].eliminate = true;
        report_progress ('^');
      } else {
        message (2, "maximum elimination bound %zu already reached",
                 elimination_bound);
        message (2, "disabling elimination until irredundant clause is "
                    "removed or shrunken");
        disabled.elimination = true;
        done = true;
      }
    } else
      message (2,
               "incomplete variable elimination round %u after %" PRIu64
               " resolutions",
               round, resolutions);

    if (eliminated) {
      flush_eliminated_binary_clauses ();
      mark_eliminated_redundant_clauses_as_garbage (round);
    }

    if (eliminated && !collecting_clauses) {
      message (2, "forcing garbage collection");
      collecting_clauses = true;
    }
    bool report = eliminated;
#ifndef QUIET
    if (verbosity > 0)
      report = true;
#endif
    if (report)
      report_progress ('e');
  }

  transition_to_sparse_watching ();
  elimination_scheduler.set_next_conflict_limit ();
  trigger.elimination = 0;

  assert (eliminating);
  eliminating = false;
  stop_simplifier ("eliminate");
  propagate_after_simplification ();

  check_all ();

  return progress;
}

}; // namespace TabularaSAT
