// This is the hottest function of the solver where it spends most of the
// time and thus the whole solver and its heuristics are build around the
// idea to reduce the time spent in this function.   It requires
// sophisticated data structures and needs careful implementation.

#ifdef SEARCH_PROPAGATE
#define ONLY_IF_VIVIFYING(...) /**/
#define PROPAGATE_FUNCTION search_propagate
#endif

#ifdef VIVIFY_PROPAGATE
#define ONLY_IF_VIVIFYING(...) __VA_ARGS__
#define PROPAGATE_FUNCTION vivify_propagate
#endif

#ifdef WARMING_PROPAGATE
#define ONLY_IF_VIVIFYING(...) /**/
#define PROPAGATE_FUNCTION warming_propagate
#endif

#ifdef PROBING_PROPAGATE
#define ONLY_IF_VIVIFYING(...) /**/
#define PROPAGATE_FUNCTION failed_literal_propagate
#endif

bool Solver::PROPAGATE_FUNCTION (ONLY_IF_VIVIFYING (Clause *ignore)) {

#ifndef NVIVIFY
#ifdef VIVIFY_PROPAGATE
  assert (vivifying);
#else
  assert (!vivifying);
#endif
#endif

#ifndef NWALK
#ifdef WARMING_PROPAGATE
  assert (warming);
#else
  assert (!warming);
#endif
#endif

#ifndef NPROBE
#ifdef PROBING_PROPAGATE
  assert (probing);
#else
  assert (!probing);
#endif
#endif

  bool res = true;

#ifndef WARMING_PROPAGATE
  uint64_t ticks = 0;
#endif

  conflict = INVALID_POSITION;
  falsified_literal = INVALID_LITERAL;
  auto &delayed = delayed_large_watches;
  unsigned *propagated = trail.propagated;

  while (res && propagated != trail.end ()) {

    unsigned lit = *propagated++;
    unsigned not_lit = negate (lit);
    debug ("propagating %s", debug_literal (lit));

    assert (delayed.empty ());

    auto watches = references (not_lit);
    auto begin_watches = watches->begin ();
    auto end_watches = watches->end ();
    auto p = begin_watches, q = p;
#ifndef WARMING_PROPAGATE
    ticks += ticks_references (watches);
#endif
    while (p != end_watches) {

      Reference watch = *q++ = *p++;
      unsigned blocking = untag (watch);
      signed char blocking_value = values[blocking];

      if (tagged (watch)) { // Virtual binary clause.

        Reference reason = tag_binary (not_lit);

        if (blocking_value > 0) {

          // As below with 'NREPLACEMENTCACHE' we keep this option
          // compile-time only and here disabled by default.  Making it
          // run-time configurable option (still disabled by default) gives
          // a non-negligible slow-down.  So use './configure
          // --improve-reasons' to enable.

#if !defined(PROBING_PROPAGATE) && defined(IMPROVEREASONS)

          unsigned blocking_idx = index (blocking);
          Reference blocking_reason = reasons[blocking_idx];

          if (large_reference (blocking_reason)) {

            unsigned other = lit;
            unsigned other_idx = index (other);
            Reference other_reason = reasons[other_idx];
            unsigned other_position = position[other_idx];
            assert (other_position + 1 == propagated);
            unsigned other_level = levels[other_idx];

            unsigned blocking_position = position[blocking_idx];
            unsigned blocking_level = levels[blocking_idx];

            unsigned twice = 0;
            if (binary_reference (other_reason)) {
              reason = other_reason;
              other = untag (other_reason);
              other_idx = index (other);
              other_reason = reasons[other_idx];
              other_position = position[other_idx];
              assert (levels[other_idx] == other_level);
              twice = 1;
            }

            if (blocking_position > other_position &&
                blocking_level == other_level) {
              debug_binary (
                  blocking, negate (other), "improving %s reason %s to",
                  debug_literal (blocking), twice ? "twice" : "once");
              (void) twice;
              statistics.improved_reasons++;
              reasons[blocking_idx] = reason;
              levels[blocking_idx] = other_level;
            }
          }
#endif
          continue;
        }

        if (blocking_value < 0) {
          debug_binary (not_lit, blocking, "conflicting");
#ifdef WARMING_PROPAGATE
          continue;
#else
          falsified_literal = blocking;
          conflict = reason;
          res = false;
          break;
#endif
        }

#ifdef PROBING_PROPAGATE
        failed_literal_assign (blocking, reason);
#else
        assign (blocking, reason);
#endif
        continue;
      }

      // Also by default keep the clause reference beside the blocking
      // literal which for binary clauses is kept anyhow.

      Reference ref = *q++ = *p++;

      // Again if the blocking literal is satisfied just continue with the
      // next watched clause.  This test can unfortunately not be pulled out
      // as we have to copy the clause reference for large clauses.

      if (blocking_value > 0)
        continue;

      Clause *c = dereference_clause (ref); // Should still be cheap.
#ifdef VIVIFY_PROPAGATE
      if (c == ignore) {
        assert (vivifying);
        continue;
      }
#endif

      // We could add here the test 'if (c->garbage) continue;' but this
      // would then be the hot-spot of the solver and rarely trigger
      // (almost all of the clauses are non-garbage). To avoid this awkward
      // situation (hot-spot at a branch almost never taken) we prefer to
      // eagerly remove watches to clauses marked garbage.

      assert (!c->garbage);
      unsigned *literals = c->literals;

      // The first time we access the actual clause data is the hot-spot of
      // the solver in almost all runs.  This is accounted by a 'tick'.
#ifndef WARMING_PROPAGATE
      ticks++;
#endif

      // This is a fancy branchless version of obtaining the other watched
      // literal without knowing whether the negation 'not_lit' of the
      // propagated literal sits at position zero or one.

      unsigned other = literals[0] ^ literals[1] ^ not_lit;

      assert (other != not_lit);
      assert (literals[0] == other || literals[1] == other);
      assert (literals[0] == not_lit || literals[1] == not_lit);

      // The other watched literals is not necessarily the blocking literal
      // and thus it can be satisfied even if the blocking literal is not.
      // In that case replace the later with the former in the watch list.

      signed char other_value = values[other];
      if (other_value > 0) {
        q[-2] = tag_large (other);
        continue;
      }

      // The following feature is only made configurable at compile-time as
      // it actually gives some non-negligible overhead if checked for being
      // enabled at run-time (in the order of 1% overall slow-down if
      // enabled as a run-time-option which we do not really want to explore
      // for 'vanilla').  Thus we can keep it 'on/enabled' by default.  Use
      // './configure --no-searched' to disable this feature.

#ifndef NREPLACEMENTCACHE
      {
        // Now we search for a 'replacement' of 'not_lit' as watch first
        // from the saved 'searched' to the 'end' and then if no replacement
        // has been found yet from 'literals + 2' to 'searched'.  This idea
        // of saving the position where the last replacement literal was
        // found gives an accumulated improvement in overall propagation
        // complexity.

        unsigned *end_literals = literals + c->size;
        unsigned *searched = literals + c->replaced;
        unsigned *l, replacement;

        for (l = searched; l != end_literals; l++)
          if (values[replacement = *l] >= 0) {
          FOUND_REPLACEMENT:
            debug_clause (c, "unwatching %s in", debug_literal (not_lit));
            literals[0] = other, literals[1] = replacement, *l = not_lit;
            delayed.push_back (
                LiteralBlockingReference (replacement, other, ref));
            c->replaced = l - literals; // Save 'searched'.
            q -= 2;
            goto CONTINUE_WITH_NEXT_REFERENCE;
          }

        for (l = literals + 2; l != searched; l++)
          if (values[replacement = *l] >= 0)
            goto FOUND_REPLACEMENT;
      }
#else
      {
        unsigned *end_literals = literals + c->size;
        unsigned *l, replacement;

        for (l = literals + 2; l != end_literals; l++)
          if (values[replacement = *l] >= 0) {
            debug_clause (c, "unwatching %s in", debug_literal (not_lit));
            literals[0] = other, literals[1] = replacement, *l = not_lit;
            delayed.push_back (
                LiteralBlockingReference (replacement, other, ref));
            q -= 2;
            goto CONTINUE_WITH_NEXT_REFERENCE;
          }
      }
#endif

      // If no replacement has been found and the other watch is false then
      // we found a conflicting clause.

      if (other_value < 0) {
        debug_reference (ref, "conflicting");
#ifdef WARMING_PROPAGATE
        continue;
#else
        conflict = ref;
        res = false;
        break;
#endif
      }

      // Otherwise, if no replacement was found and the other watch is
      // neither satisfied nor falsified, then we found a unit instead and
      // assign the other watched literal with this clause as reason.

#ifdef PROBING_PROPAGATE
      failed_literal_assign (other, ref);
#else
      assign (other, ref);
#endif
    CONTINUE_WITH_NEXT_REFERENCE:;
    }

    // In case of conflict the propagation loop above bails out early and we
    // need to copy the rest of the watches as is.

    while (p != end_watches)
      *q++ = *p++;

    if (q != end_watches)
      watches->resize (q - begin_watches);

    if (!delayed.empty ()) {
      for (auto &w : delayed)
        watch_literal (w.lit, w.blocking, w.ref);
      delayed.clear ();
    }
  }

#ifndef QUIET
  uint64_t propagations = propagated - trail.propagated;
#endif
  trail.propagated = propagated;
#ifndef QUIET
  statistics.propagations += propagations;
#endif
#ifdef VIVIFY_PROPAGATE
  assert (vivifying);
  assert (!warming);
#ifndef NPROBE
  assert (!probing);
#endif
  ADD_COUNT (vivification_ticks, ticks);
  ADD_STATS (vivification_propagations, propagations);
#endif
#ifdef PROBING_PROPAGATE
#ifndef NVIVIFY
  assert (!vivifying);
#endif
#ifndef NWALK
  assert (!warming);
#endif
  assert (probing);
  ADD_COUNT (failed_literal_ticks, ticks);
  ADD_STATS (failed_literal_propagations, propagations);
#endif

#ifdef SEARCH_PROPAGATE
#ifndef QUIET
  statistics.search_propagations += propagations;
#endif
  ADD_COUNT (search_ticks, ticks);
#ifndef QUIET
#ifndef NSWITCH
  if (stable) {
    statistics.stable_ticks += ticks;
    if (!res)
      statistics.stable_conflicts++;
  } else {
    statistics.focused_ticks += ticks;
    if (!res)
      statistics.focused_conflicts++;
  }
#endif
#endif
  if (!res)
    INC_COUNT (conflicts);
#endif

  return res;
}

#undef ONLY_IF_VIVIFYING
#undef PROPAGATE_FUNCTION
