#include "connect.hpp"
#include "cover.hpp"
#include "references.hpp"
#include "schedule.hpp"
#include "solver.hpp"
#include "ticks.hpp"

#include <algorithm>
#include <cinttypes>

namespace TabularaSAT {

bool Solver::subsuming_irredundant_clauses () {
  if (!OPTION (inprocessing))
    return false;
  if (!OPTION (subsume_clauses))
    return false;
  if (disabled.subsumption)
    return false;
  if (GET_COUNT (conflicts) < limits.subsume)
    return false;
  return true;
}

struct smaller_size {
  const Clauses &clauses;
  smaller_size (const Clauses &cs) : clauses (cs) {}
  bool operator() (unsigned p, unsigned q) {
    const Clause *c = clauses[p];
    const Clause *d = clauses[q];
    if (c->size < d->size)
      return true;
    if (c->size > d->size)
      return false;
    return p < q;
  }
};

struct rank_smaller_size {
  const Clauses &clauses;
  rank_smaller_size (const Clauses &cs) : clauses (cs) {}
  unsigned operator() (Position pos) { return clauses[pos]->size; };
};

void Solver::gather_subsumption_candidates (
    std::vector<Position> &candidates) {
  assert (!level);
  size_t satisfied = 0, potential = 0;
  unsigned size_limit = OPTION (subsumed_clause_size);

  for (auto c : clauses) {
    Position pos = clauses.position (c);
    if (pos > last_irredundant)
      break;
    if (c->garbage)
      continue;
    if (c->redundant)
      continue;
    if (c->size > size_limit)
      continue;
    unsigned subsume = 0;
    for (auto lit : *c) {
      signed char value = values[lit];
      if (value < 0)
        continue;
      if (value > 0) {
        mark_clause_garbage (c, "satisfied");
        satisfied++;
        break;
      }
      if (flags[index (lit)].subsume)
        subsume++;
    }
    if (c->garbage)
      continue;
    potential++;
    if (!c->subsume && subsume < 2)
      continue;
    debug_clause (c, "subsumption candidate");
    candidates.push_back (pos);
  }

#ifndef QUIET
  message (2, "found %zu satisfied irredundant large clauses", satisfied);
  size_t gathered = candidates.size ();
  message (1, "gathered %zu subsumption candidates %.0f%%", gathered,
           percent (gathered, potential));
#else
  (void) satisfied;
  (void) potential;
#endif
  if (candidates.size () < OPTION (radix_sort_limit))
    std::sort (candidates.begin (), candidates.end (),
               smaller_size (clauses));
  else
    radix_sort_generic<Position, Position, unsigned, rank_smaller_size> (
        candidates, sorter, rank_smaller_size (clauses));
}

bool Solver::strengthen_clause (Clause *c, unsigned remove) {
  assert (!level);
  assert (clause.empty ());
  assert (OPTION (strengthen_clauses));
  for (auto lit : *c) {
    if (lit == remove)
      continue;
    signed char value = values[lit];
    assert (value <= 0);
    if (value)
      continue;
    clause.push_back (lit);
  }
  debug_vector (clause, "temporary strengthened");
  unsigned old_size = c->size;
  unsigned new_size = clause.size ();
  cover (!new_size);
  if (new_size == 1) {
    unsigned unit = clause[0];
    debug_clause (c, "ignoring hidden unit %s", debug_literal (unit));
    units.push_back (unit);
    trace_unit ('a', unit);
    clause.clear ();
    return false;
  }
#ifndef QUIET
  statistics.strengthened++;
#endif
  assert (new_size < old_size);
  debug_clause (c, "strengthening");
  trace_temporary ('a');
  if (new_size == 2) {
    new_binary ();
    mark_clause_garbage (c, "strengthened");
  } else {
#ifndef NREPLACEMENTCACHE
    c->replaced = 2;
#endif
    memcpy (c->literals, &clause[0], 4 * new_size);
    if (!c->shortened) {
      c->shortened = true;
      assert (c->literals[old_size - 1] != INVALID_LITERAL);
      c->literals[old_size - 1] = INVALID_LITERAL;
    }
    c->size = new_size;
    debug_clause (c, "strengthened");
    mark_removed (remove);
    mark_added (c);
  }
  clause.clear ();
  return true;
}

bool Solver::subsume_clause (Clause *candidate, unsigned &remove,
                             Clause *&double_subsumed) {
  assert (!level);
  assert (!candidate->garbage);
  remove = INVALID_LITERAL;
  double_subsumed = 0;
  debug_clause (candidate, "trying to forward subsume");
  candidate->subsume = false;
#ifndef QUIET
  statistics.subsume_checks++;
#endif
  unsigned marked = 0;
  for (auto lit : *candidate) {
    signed char value = values[lit];
    assert (value <= 0);
    if (value < 0)
      continue;
    unsigned not_lit = negate (lit);
    assert (!marks[lit]), assert (!marks[not_lit]);
    marks[lit] = 1, marks[not_lit] = -1;
    marked++;
  }
  const size_t occurrence_limit = OPTION (subsumption_occurrence_limit);
  bool strengthen = OPTION (strengthen_clauses);
  bool subsumed = false;
  uint64_t ticks = 1;
  for (unsigned sign = 0; sign != 2; sign++) {
    if (sign && !strengthen)
      break;
    for (auto lit : *candidate) {
      auto occurrences = references (lit ^ sign);
      ticks++;
      if (occurrences->size () > occurrence_limit)
        continue;
      ticks += ticks_references (occurrences);
      for (auto ref : *occurrences) {
        if (!tagged (ref)) {
          Clause *c = dereference_clause (ref);
          if (c == candidate)
            continue;
          ticks++;
          if (c->garbage)
            continue;
#ifndef QUIET
          statistics.subsume_checks++;
#endif
          unsigned negated = 0, found = 0;
          subsumed = true;
          for (auto other : *c) {
            signed char value = values[other];
            if (value < 0)
              continue;
            assert (!value);
            signed char mark = marks[other];
            if (mark)
              found++;
            if (mark > 0)
              continue;
            if (mark < 0) {
              if (strengthen && !negated++) {
                remove = negate (other);
                continue;
              }
            }
            remove = INVALID_LITERAL;
            subsumed = false;
            break;
          }
          if (remove != INVALID_LITERAL) {
            assert (strengthen);
            debug_clause (c, "self-subsuming antecedent");
            if (found == marked)
              double_subsumed = c;
            subsumed = false;
            goto DONE;
          }
          if (!subsumed)
            continue;
          debug_clause (c, "subsuming");
          goto DONE;
        } else if (!sign) {
          unsigned other = untag (ref);
          assert (!values[other]);
          signed char mark = marks[other];
          if (mark > 0) {
            debug_binary (lit, other, "subsuming");
            subsumed = true;
            goto DONE;
          }
          if (strengthen && mark < 0) {
            debug_binary (lit, other,
                          "self-subsuming resolution antecedent");
            remove = negate (other);
            assert (!subsumed);
            goto DONE;
          }
        } // 'else { ... }' would need complex units handling code.
      }
    }
  }
DONE:
  ADD_COUNT (subsumption_ticks, ticks);

  for (auto lit : *candidate)
    marks[lit] = marks[negate (lit)] = 0;

  if (subsumed) {
#ifndef QUIET
    statistics.subsumed_clauses++;
#endif
    assert (!candidate->garbage);
    mark_clause_garbage (candidate, "subsumed");
  }

  assert (strengthen || remove == INVALID_LITERAL);

  return subsumed;
}

void Solver::connect_minimum_occurring_literal (Clause *c) {

  unsigned res = INVALID_LITERAL;
  size_t count = ~(size_t) 0;

  for (auto lit : *c) {
    if (values[lit])
      continue;
    const auto occurrences = references (lit);
    unsigned size = occurrences->size ();
    if (size >= count)
      continue;
    count = size;
    res = lit;
  }

  assert (res != INVALID_LITERAL);

  debug_clause (c, "minimum literal %s with %zu occurrences in",
                debug_literal (res), count);

  if (count > OPTION (subsumption_occurrence_limit))
    return;

  Reference ref = reference_clause (c);
  connect_clause (res, ref);
}

bool Solver::subsume_clauses () {

  if (level && !backtrack_to_root_level_and_propagate ())
    return false;

  check_all ();

  start_simplifier ("subsume");

  SubsumptionScheduler subsumption_scheduler (this);
  subsumption_scheduler.increment_counter ();

  uint64_t limit;
  if (preprocessing)
    limit = UNLIMITED;
  else {
    uint64_t ticks = subsumption_scheduler.get_ticks_limit ();
    limit = GET_COUNT (subsumption_ticks) + ticks;
  }

  message (2, "starting subsumption after %" PRIu64 " variables added",
           statistics.added_variables);

  assert (watching);
  flush_large_watches_and_satisfied_binary_clauses ();
  watching = false;

  unsigned max_rounds = OPTION (subsumption_rounds);
  bool progress = false;
  bool done = false;

  assert (max_rounds < UINT_MAX);
  for (unsigned round = 1; !done && round <= max_rounds; round++) {

    message (2, "starting subsumption round %u", round);

    size_t subsumed = 0, strengthened = 0;

    std::vector<Position> candidates;
    gather_subsumption_candidates (candidates);
    size_t total = candidates.size ();
    if (!total)
      break;

    for (auto idx : variables)
      flags[idx].subsume = false;

    size_t remain = 0;

    for (auto pos : candidates) {
      if (terminated)
        done = true;
      Clause *c = clauses[pos];
      if (!done && GET_COUNT (subsumption_ticks) >= limit) {
        message (2, "subsumption ticks limit hit in round %u", round);
        done = true;
      }
      if (done)
        c->subsume = true, remain++;
      else {
        unsigned remove;
        Clause *double_subsumed;
        if (subsume_clause (c, remove, double_subsumed))
          subsumed++;
        else {
          if (remove != INVALID_LITERAL && strengthen_clause (c, remove)) {
            strengthened++;
            if (double_subsumed) {
              subsumed++;
              mark_clause_garbage (double_subsumed,
                                   "double self-subsuming antecedent");
            }
          }
          if (!c->garbage)
            connect_minimum_occurring_literal (c);
        }
      }
    }

    flush_connected_large_clauses ();
    assert (!watching);

    message (1, "strengthened %zu clauses in round %u", strengthened,
             round);
    message (1, "subsumed %zu clauses in round %u", subsumed, round);

    if (remain) {
      assert (done);
      message (2, "%zu subsumption candidates remain %.0f%%", remain,
               percent (remain, total));
    } else
      message (2, "complete subsumption round %u", round);

    if (subsumed)
      collecting_clauses = true;
    if (strengthened)
      collecting_clauses = true;
    if (GET_COUNT (garbage_clauses))
      collecting_clauses = true;
    bool report = subsumed;
#ifndef QUIET
    if (verbosity > 0)
      report = true;
#endif
    if (report)
      report_progress ('s');
    if (subsumed || strengthened)
      progress = true;
    else {
      if (!remain) {
        assert (!done);
        message (2, "disabling subsumption until irredundant clause is "
                    "added or shrunken");
        disabled.subsumption = true;
      }
      break;
    }
  }

  watching = true;
  watch_all_large_clauses ();

  check_all ();

  subsumption_scheduler.set_next_conflict_limit ();

  stop_simplifier ("subsume");
  propagate_after_simplification ();

  return progress;
}

}; // namespace TabularaSAT
