#include "inline.hpp"
#include "solver.hpp"

#include <cinttypes>

namespace TabularaSAT {

bool Solver::switching_search_mode () {
  if (!OPTION (switch_search_mode))
    return false;
  uint64_t limit;
  if (GET_COUNT (switched_modes) & 1)
    limit = GET_COUNT (search_ticks);
  else
    limit = GET_COUNT (conflicts);
  if (limit <= limits.mode)
    return false;
  return true;
}

void Solver::switch_search_mode () {

  if (level && !backtrack_to_root_level_and_propagate ())
    return;

  if (stable)
    report_progress (']'), stop ("stable");
  else
    report_progress ('}'), stop ("focused");

  // No 'struct Switcher : Scheduler' as mode switching is too complex and
  // needs careful dedicated accounting and setting limits.

  if (++GET_COUNT (switched_modes) & 1)
    INC_COUNT (switched_into_second_mode);

  assert ((GET_COUNT (switched_modes) - 1) / 2 + 1 ==
          GET_COUNT (switched_into_second_mode));

  uint64_t ticks =
      GET_COUNT (search_ticks) - last.search_ticks_at_mode_switch;
#ifndef QUIET
  {
    uint64_t old_delta_conflicts =
        GET_COUNT (conflicts) - last.conflicts_at_mode_switch;
    message (1,
             "last %s mode took %" PRIu64 " conflicts and %" PRIu64
             " ticks (%.2f ticks per conflict)",
             mode (), old_delta_conflicts, ticks,
             average (ticks, old_delta_conflicts));
  }
#endif
  last.search_ticks_at_mode_switch = GET_COUNT (search_ticks);
  last.conflicts_at_mode_switch = GET_COUNT (conflicts);

  stable = !stable;

  message (1,
           "switching %" PRIu64 " to %s mode at %" PRIu64
           " ticks and %" PRIu64 " conflicts",
           statistics.switched_modes, mode (), statistics.search_ticks,
           statistics.conflicts);

  if (stable)
    reluctant.reset ();
  set_restart_limit ();

  if (stable) {
    if (bumper.capacity ())
      for (auto idx : variables)
        if (flags[idx].active && !values[literal (idx)] &&
            !bumper.contains (idx))
          bumper.push (idx);
  } else {
    if (queue.capacity ())
      for (auto idx : variables)
        if (flags[idx].active && !values[literal (idx)])
          queue.update (idx);
  }

  last.decisions_at_conflict[stable] = GET_COUNT (decisions);

  uint64_t second_mode = GET_COUNT (switched_into_second_mode);
  if (GET_COUNT (switched_modes) & 1) {
    if (stable) {
      double bias = OPTION (stable_bias);
      if (second_mode < bias) {
        assert (second_mode);
        double factor = bias / second_mode;
        message (2, "stable ticks limit by bias factor %g", factor);
        ticks *= factor;
      }
    }
    limits.mode = GET_COUNT (search_ticks) + ticks;
    message (2,
             "new %s mode switching limit at %" PRIu64 " after %" PRIu64
             " ticks",
             mode (), limits.mode, ticks);
  } else {
    uint64_t base = OPTION (mode_switching_interval);
    uint64_t new_delta_conflicts = nlogpow (second_mode, 6) * base;
    limits.mode = GET_COUNT (conflicts) + new_delta_conflicts;
    message (2,
             "new %s mode switching limit at %" PRIu64 " after %" PRIu64
             " conflicts",
             mode (), limits.mode, new_delta_conflicts);
  }

  if (stable)
    report_progress ('['), start ("stable");
  else
    report_progress ('{'), start ("focused");
}

}; // namespace TabularaSAT
