#ifndef _tabularasat_ticks_hpp_INCLUDED
#define _tabularasat_ticks_hpp_INCLUDED

#include <cstdlib>

namespace TabularaSAT {

// The concept of 'ticks' is a refined versions of 'mems' by Donald Knuth.
// While 'mems' correspond to the number of pointer accesses to machine
// words stored in memory (let us say one machine word is 8 bytes on a
// 64-bit machine) we try to approximate the number of accessed cache
// lines instead, actually more precisely the expected number of level one
// cache line misses.  Both can serve as an approximation of virtual running
// time which is both deterministic and portable.  In our context we only
// use ticks to balance and limit the relative running time in various
// routines in a deterministic way.  It is as such only considered a
// relative time and an instance local metric and has not the goal to
// provide a global virtual absolute time.

// The need for 'ticks' instead of 'mems' arose already in our earlier
// solvers such as 'Kissat', where we first used 'mems' to limit for
// instance the number of resolutions during variable elimination relative
// to the time spent in propagation during search.  It turned out that for
// instances with many binary clauses, not counting the traversal of those
// virtual binary clause literals ignores too much actual effort in search.
// Counting every binary clause literal in such traversals on the other hand
// accounts too much time for traversal, as those virtual binary clause
// literals reside mostly in the same cache line.

// Concretely, during propagation we simply count the number of large clause
// dereferences as one tick as well as the number of cache lines spanning
// the traversed watches stacks.  Accessing 'values' of literals is not
// counted. In order to not make the counts dependent on pointer addresses
// we simply assume that the start of such a watcher stack is cache line
// size aligned (which is an under-approximation).  Furthermore, as we want
// to keep the program deterministic for both 8-byte as well as compact
// 4-byte references, we assume that each references needs 4-bytes (also an
// under-approximation for non-compact references).  Finally we fix the size
// of the assumed cache lines to 128 bytes.

inline size_t ticks_spanning (size_t size) { return (size + 31) >> 5; }

inline size_t ticks_references (const References *v) {
  return ticks_spanning (v->size ());
}

}; // namespace TabularaSAT

#endif
