#include "references.hpp"
#include "solver.hpp"

#ifdef NCOMPACT
#include <cinttypes>
#endif

namespace TabularaSAT {

#ifndef NDEBUG

size_t Solver::compute_binary_clauses () {
  size_t res = 0;
  for (auto watches : matrix) {
    unsigned lit = watches->lit;
    if (lit == INVALID_LITERAL)
      continue;
    auto p = watches->begin ();
    auto end = watches->end ();
    while (p != end) {
      Reference ref = *p++;
      if (tagged (ref)) {
        unsigned other = untag (ref);
        if (lit < other)
          res++;
      } else
        p += watching;
    }
  }
  return res;
}

size_t Solver::compute_redundant_clauses () {
  size_t res = 0;
  for (auto c : clauses)
    if (c->redundant && !c->garbage)
      res++;
  return res;
}

size_t Solver::compute_irredundant_clauses () {
  size_t res = 0;
  for (auto c : clauses)
    if (!c->redundant && !c->garbage)
      res++;
  return res;
}

Position Solver::find_first_redundant_clause () {
  for (auto c : clauses)
    if (c->redundant && !c->garbage)
      return clauses.position (c);
  return INVALID_POSITION;
}

#ifndef NGLUE

Position Solver::find_first_tier2_clause () {
  unsigned tier1 = OPTION (tier1_max_glue);
  for (auto c : clauses)
    if (c->redundant && !c->garbage && c->glue > tier1)
      return clauses.position (c);
  return INVALID_POSITION;
}

#endif

Position Solver::find_last_irredundant_clause () {
  size_t res = 0;
  for (auto c : clauses)
    if (!c->redundant && !c->garbage)
      res = clauses.position (c);
  return res;
}

void Solver::check_clauses () {
  debug ("arena of size %zu", clauses.size ());
  size_t binary = compute_binary_clauses ();
  size_t redundant = compute_redundant_clauses ();
  size_t irredundant = compute_irredundant_clauses ();
  Position found_redundant = find_first_redundant_clause ();
  Position found_irredundant = find_last_irredundant_clause ();
  debug ("computed %zu binary clauses", binary);
  debug ("computed %zu redundant clauses", redundant);
  debug ("computed %zu irredundant clauses", irredundant);
  debug ("found first redundant clause at position %" POSFMT,
         found_redundant);
  debug ("found last irredundant clause at position %" POSFMT,
         found_irredundant);
  assert (redundant == statistics.redundant_clauses);
  assert (binary == statistics.binary_clauses);
  assert (irredundant == statistics.irredundant_clauses);
  assert (first_redundant <= found_redundant);
  assert (last_irredundant >= found_irredundant);
#ifndef NGLUE
  Position found_tier2 = find_first_tier2_clause ();
  debug ("found first tier2 clause at position %" POSFMT, found_tier2);
  assert (first_tier2 <= found_tier2);
#endif
}

void Solver::check_variables () {
  size_t active = 0;
  size_t inactive = 0;
  size_t activated = 0;
  size_t unactivated = 0;
  size_t eliminated = 0;
  size_t fixed = 0;
#ifndef NSUBSTITUTE
  size_t substituted = 0;
#endif
  for (auto idx : variables) {
    Flags &f = flags[idx];
    if (f.active) {
      active++;
      assert (f.activated);
      activated++;
      assert (!f.eliminated);
      assert (!f.fixed);
#ifndef NSUBSTITUTE
      assert (!f.substituted);
#endif
    } else {
      unsigned found = 0;
      if (f.eliminated)
        eliminated++, found++;
      if (f.fixed)
        fixed++, found++;
#ifndef NSUBSTITUTE
      if (f.substituted)
        substituted++, found++;
#endif
      if (f.activated) {
        inactive++;
        activated++;
        assert (found == 1);
      } else {
        unactivated++;
        assert (!found);
      }
    }
  }
  size_t sum = unactivated + active + inactive;
  assert (sum == variables);
  assert (activated + unactivated == sum);

  sum = eliminated + fixed;
#ifndef NSUBSTITUTE
  sum += substituted;
#endif
  assert (sum == inactive);
}

#endif

}; // namespace TabularaSAT
