#include "eliminate.hpp"
#include "references.hpp"
#include "solver.hpp"

namespace TabularaSAT {

bool Solver::find_equivalence_gate (Eliminator &eliminator,
                                    unsigned pivot) {
  if (!OPTION (extract_equivalences))
    return false;
  unsigned not_pivot = negate (pivot);
  References *pos_occs = references (pivot);
  References *neg_occs = references (not_pivot);
  std::vector<Reference> *gate = eliminator.gate;
  std::vector<Reference> *nogate = eliminator.nogate;
  for (auto ref : *pos_occs) {
    if (!tagged (ref))
      continue;
    unsigned other = untag (ref);
    marks[negate (other)] = -1;
  }
  unsigned equivalent = INVALID_LITERAL;
  for (auto ref : *neg_occs) {
    if (!tagged (ref))
      continue;
    unsigned other = untag (ref);
    if (marks[other] >= 0)
      continue;
    equivalent = other;
    break;
  }
  if (equivalent != INVALID_LITERAL) {
    debug ("found equivalence gate %s = %s", debug_literal (pivot),
           debug_literal (equivalent));
    Reference pos_ref = tag_binary (negate (equivalent));
    Reference neg_ref = tag_binary (equivalent);
    gate[0].push_back (pos_ref);
    gate[1].push_back (neg_ref);
    for (auto ref : *pos_occs)
      if (ref != pos_ref)
        nogate[0].push_back (ref);
    for (auto ref : *neg_occs)
      if (ref != neg_ref)
        nogate[1].push_back (ref);
  }
  for (auto ref : *pos_occs) {
    if (!tagged (ref))
      continue;
    unsigned other = untag (ref);
    marks[negate (other)] = 0;
  }
  return equivalent != INVALID_LITERAL;
}

bool Solver::find_and_gate (Eliminator &eliminator, unsigned pivot) {

  if (!OPTION (extract_and_gates))
    return false;

  unsigned not_pivot = negate (pivot);
  References *pos_occs = references (pivot);
  References *neg_occs = references (not_pivot);
  std::vector<Reference> *pos_gate = eliminator.gate + 0;
  std::vector<Reference> *neg_gate = eliminator.gate + 1;
  std::vector<Reference> *pos_nogate = eliminator.nogate + 0;
  std::vector<Reference> *neg_nogate = eliminator.nogate + 1;

  assert (pos_gate->empty ());
  assert (neg_gate->empty ());
  assert (pos_nogate->empty ());
  assert (neg_nogate->empty ());

  size_t max_size = INVALID_SIZE;

  for (auto pos_ref : *pos_occs) {
    if (tagged (pos_ref))
      continue;
    Clause *c = dereference_clause (pos_ref);
    if (c->garbage)
      continue;

    bool satisfied = false;
    size_t marked = 0;
    for (auto other : *c) {
      if (other == pivot)
        continue;
      signed value = values[other];
      if (value < 0)
        continue;
      if (value > 0) {
        satisfied = true;
        break;
      }
      marks[negate (other)] = -1;
      marked++;
    }

    if (!satisfied) {

      if (max_size == INVALID_SIZE) {
        max_size = 0;
        for (auto neg_ref : *neg_occs)
          if (tagged (neg_ref))
            max_size++;
      }

      if (marked <= max_size) {
        size_t found = 0;
        assert (neg_gate->empty ());
        assert (neg_nogate->empty ());
        for (auto neg_ref : *neg_occs) {
          if (tagged (neg_ref)) {
            unsigned other = untag (neg_ref);
            if (marks[other] < 0) {
              marks[other] = 0;
              neg_gate->push_back (neg_ref);
              assert (found < marked);
              found++;
            } else
              neg_nogate->push_back (neg_ref);
          } else
            neg_nogate->push_back (neg_ref);
        }
        if (found == marked) {
          assert (neg_gate->size () == found);
          assert (neg_nogate->size () == neg_occs->size () - found);
#ifdef LOGGING
          if (logging ()) {
            debug_prefix ();
            printf ("found AND gate %s =", debug_literal (pivot));
            for (auto other : *c)
              if (other != pivot)
                printf (" %s", debug_literal (negate (other)));
            debug_suffix ();
            debug_clause (c, "AND gate base");
          }
#endif
          pos_gate->push_back (pos_ref);
          break;
        }
        neg_gate->clear ();
        neg_nogate->clear ();
      }

    } else
      debug_clause (c, "satisfied");

    for (auto other : *c)
      if (other != pivot)
        marks[negate (other)] = 0;
  }

  if (pos_gate->empty ())
    return false;

  Reference base_clause_ref = (*pos_gate)[0];
  for (auto pos_ref : *pos_occs)
    if (pos_ref != base_clause_ref)
      pos_nogate->push_back (pos_ref);

  return true;
}

bool Solver::find_gate (Eliminator &eliminator, unsigned pivot) {
#ifndef QUIET
  eliminator.extracted = 0;
#endif
  if (!OPTION (extract_gates))
    return false;

  eliminator.gate[0].clear ();
  eliminator.gate[1].clear ();
  eliminator.nogate[0].clear ();
  eliminator.nogate[1].clear ();

  if (find_equivalence_gate (eliminator, pivot)) {
#ifndef QUIET
    eliminator.extracted = &statistics.eliminated_equivalences;
#endif
    return true;
  }

  if (find_and_gate (eliminator, pivot)) {
#ifndef QUIET
    eliminator.extracted = &statistics.eliminated_ands;
#endif
    return true;
  }

  return false;
}

}; // namespace TabularaSAT
