#include "solver.hpp"

namespace TabularaSAT {

State Solver::run_initial_preprocessing_loop () {

  if (inconsistent)
    return UNSATISFIABLE;

  if (!OPTION (preprocessing))
    return UNKNOWN;

  typedef bool (Solver::*Preprocessor) ();
  std::vector<Preprocessor> preprocessors;

#ifndef NTRANSITIVE
  if (OPTION (transitive_reduction))
    preprocessors.push_back (&Solver::transitive_reduction);
#endif
#ifndef NPROBE
  if (OPTION (failed_literal_probing))
    preprocessors.push_back (&Solver::probe_failed_literals);
#endif
#ifndef NVIVIFY
  if (OPTION (vivify_clauses))
    preprocessors.push_back (&Solver::vivify_clauses);
#endif
#ifndef NSUBSTITUTE
  if (OPTION (substitute_variables))
    preprocessors.push_back (&Solver::substitute_variables);
#endif
  if (OPTION (subsume_clauses))
    preprocessors.push_back (&Solver::subsume_clauses);
  if (OPTION (eliminate_variables))
    preprocessors.push_back (&Solver::eliminate_variables);

  if (preprocessors.empty ())
    return UNKNOWN;

  assert (!preprocessing);
  start ("preprocessing");
  start ("simplify");

  unsigned max_rounds = OPTION (preprocessing_rounds);
  message (1,
           "running %zu enabled simplification routines during initial "
           "inprocessing for %u rounds",
           preprocessors.size (), max_rounds);
  bool progress = true;

  assert (max_rounds < UINT_MAX);
  unsigned &round = preprocessing;

  for (round = 1; progress && round <= max_rounds; round++) {

    message (1, "starting preprocessing round %u", round);
    progress = false;

    for (auto preprocessor : preprocessors) {
      if ((this->*preprocessor) ())
        progress = true;
      if (inconsistent || terminated)
        break;
    }

    if (!inconsistent && collecting_clauses)
      collect_clauses ();

    message (1, "finished preprocessing round %u", round);
    if (inconsistent || terminated)
      break;
  }

  assert (preprocessing);
  preprocessing = 0;

  stop ("simplify");
  stop ("preprocessing");

  if (OPTION (inprocessing))
    initialize_inprocessing_limits ();

  return inconsistent ? UNSATISFIABLE : UNKNOWN;
}

}; // namespace TabularaSAT
