#include "radix.hpp"
#include "schedule.hpp"
#include "solver.hpp"

#include <algorithm>
#include <cinttypes>

namespace TabularaSAT {

void Solver::increase_kept_learned_clauses () {
#ifndef QUIET
  statistics.increased_kept++;
#endif
#ifndef VANILLA
  assert (OPTION (reduce_vanilla));
#endif
  double learned_factor = OPTION (learned_factor) / 100.0;
  limits.reduce.clauses *= learned_factor;
  message (1, "new kept learned clauses limit of %" PRIu64 " clauses",
           limits.reduce.clauses);

  double increase_factor = OPTION (increase_factor) / 100.0;
  limits.reduce.interval *= increase_factor;
  limits.reduce.conflicts = GET_COUNT (conflicts);
  limits.reduce.conflicts += limits.reduce.interval;
  message (
      1, "new kept learned clauses increase limit at %" PRIu64 " conflicts",
      limits.reduce.conflicts);

  report_progress ('+');
}

bool Solver::reducing_redundant_clauses () {
  if (!OPTION (reduce))
    return false;
#ifndef VANILLA
  if (!OPTION (reduce_vanilla))
    return limits.reduce.conflicts < GET_COUNT (conflicts);
#endif
  if (limits.reduce.conflicts <= GET_COUNT (conflicts))
    increase_kept_learned_clauses ();

  uint64_t limit = limits.reduce.clauses;
  uint64_t redundant = GET_COUNT (redundant_clauses);
  size_t assigned = trail.size () - trail_root_level () - level;
  return redundant > limit + assigned;
}

void Solver::gather_reduce_candidates (std::vector<Position> &candidates) {

#ifndef NUSED
  const bool keep_used = OPTION (keep_used);
#endif
#ifndef NGLUE
  const bool keep_tier1 = OPTION (keep_tier1);
  size_t first = keep_tier1 ? first_tier2 : first_redundant;
#else
  size_t first = first_redundant;
#endif

  assert (first != INVALID_POSITION);
  assert (first <= clauses.size ());
  Clauses::iterator begin = clauses[first];
  auto end = clauses.end ();

  size_t irredundant = 0, accessed = 0;

  for (auto i = begin; i != end; ++i) {
    Clause *c = *i;
    accessed++;
    if (c->reason)
      continue;
    if (c->garbage)
      continue;
    if (!c->redundant) {
      irredundant++;
      continue;
    }
#ifndef NUSED
    if (c->used) {
      c->used--;
      if (keep_used)
        continue;
    }
#endif
#ifndef NGLUE
    if (keep_tier1 && c->glue <= OPTION (tier1_max_glue))
      continue;
#endif
    Position pos = clauses.position (c);
    candidates.push_back (pos);
    debug_clause (c, "gathered reduce candidate");
  }

  // Triggers 'move_redundant_clauses' if enabled and many irredundant
  // clause accesses have been accumulated (see 'connect.cpp').

  wasted_accessed += irredundant;

#ifndef QUIET
  if (irredundant)
    message (1, "wasted accessing %zu irredundant clauses %.0f%%",
             irredundant, percent (irredundant, accessed));
  else
    message (1, "no irredundant clause traversed");

  message (1, "gathered %zu redundant reduce candidates",
           candidates.size ());
#else
  (void) accessed;
#endif
}

struct less_relevant_redundant_clause {
  const Clauses &clauses;
#ifndef NGLUE
  const bool glue_relevant;
#endif
#ifndef NACTIVITY
  const bool activity_relevant;
#endif
#ifndef NSIZERELEVANT
  const bool size_relevant;
#endif
  less_relevant_redundant_clause (Solver &solver)
      : clauses (solver.clauses)
#ifndef NGLUE
        ,
        glue_relevant (SOLVER_OPTION (glue_relevant))
#endif
#ifndef NACTIVITY
        ,
        activity_relevant (SOLVER_OPTION (activity_relevant))
#endif
#ifndef NSIZERELEVANT
        ,
        size_relevant (SOLVER_OPTION (size_relevant))
#endif
  {
  }
  bool operator() (Position p, Position q) {
    const Clause *c = clauses[p];
    const Clause *d = clauses[q];
#ifndef NGLUE
    if (glue_relevant) {
      if (c->glue > d->glue)
        return true;
      if (c->glue < d->glue)
        return false;
    }
#endif
#ifndef NACTIVITY
    if (activity_relevant) {
      if (c->activity.as_float < d->activity.as_float)
        return true;
      if (c->activity.as_float > d->activity.as_float)
        return false;
    }
#endif
#ifndef NSIZERELEVANT
    if (size_relevant) {
      if (c->size > d->size)
        return true;
      if (c->size < d->size)
        return false;
    }
#endif
    return p < q;
  }
};

#ifndef NGLUE

struct rank_larger_glue {
  const Clauses &clauses;
  rank_larger_glue (const Clauses &cs) : clauses (cs) {}
  unsigned operator() (Position pos) { return ~clauses[pos]->glue; };
};

#endif

#ifndef NACTIVITY

// Radix sorting needs to translate a float into an integral type. According
// to the IEEE 754 floating point standard comparing positive 4-byte
// (32-bit) floating point numbers of type 'float' gives the same result as
// as comparing their 'unsigned' representation, which use here.

// The compiler gives a 'breaking strict aliasing' warning when using a cast
// instead of the 'union' in the 'Clause' struct, which we want to avoid.

struct rank_smaller_activity {
  const Clauses &clauses;
  rank_smaller_activity (const Clauses &cs) : clauses (cs) {}
  unsigned operator() (Position pos) {
    return clauses[pos]->activity.as_unsigned;
  };
};

#endif

struct rank_larger_size {
  const Clauses &clauses;
  rank_larger_size (const Clauses &cs) : clauses (cs) {}
  unsigned operator() (Position pos) { return ~clauses[pos]->size; };
};

void Solver::mark_reduced_clauses (Position &first_marked) {

#ifndef NGLUE
  if (OPTION (keep_tier1) && first_tier2 == INVALID_POSITION) {
    message (2, "no redundant tier2 clause");
    return;
  }
#endif
  if (first_redundant == INVALID_POSITION) {
    message (2, "no redundant clause");
    return;
  }

  std::vector<Position> candidates;
  gather_reduce_candidates (candidates);
  if (candidates.empty ()) {
    message (2, "no reduce candidate found");
    return;
  }

  if (candidates.size () < OPTION (radix_sort_limit))
    std::sort (candidates.begin (), candidates.end (),
               less_relevant_redundant_clause (*this));
  else {
#ifndef NSIZERELEVANT
    if (OPTION (size_relevant))
      radix_sort_generic<Position, Position, unsigned, rank_larger_size> (
          candidates, sorter, rank_larger_size (clauses));
#endif
#ifndef NACTIVITY
    if (OPTION (activity_relevant))
      radix_sort_generic<Position, Position, unsigned,
                         rank_smaller_activity> (
          candidates, sorter, rank_smaller_activity (clauses));
#endif
#ifndef NGLUE
    if (OPTION (glue_relevant))
      radix_sort_generic<Position, Position, unsigned, rank_larger_glue> (
          candidates, sorter, rank_larger_glue (clauses));
#endif
  }

  double fraction = OPTION (reduce_fraction) / 100.0;
  size_t total = candidates.size ();
  size_t target = fraction * total;

  message (2, "reduce target of %zu clauses %.0f%% of %zu candidates",
           target, fraction, total);

  size_t first_reduced = INVALID_POSITION;
  size_t marked = 0;

  for (auto pos : candidates) {
    if (marked == target)
      break;
    auto c = clauses[pos];
    if (pos < first_reduced)
      first_reduced = pos;
    mark_clause_garbage (c, "less relevant");
    marked++;
  }

  if (first_reduced != INVALID_POSITION) {
    message (2, "first reduced clause at position %zu", first_reduced);
    if (first_reduced < first_marked)
      first_marked = first_reduced;
  }

  message (1, "marked %zu less relevant clauses as garbage", marked);
}

void Solver::reduce_redundant_clauses (bool force) {

  check_all ();

  start ("reduce", 2);

#ifndef VANILLA
  if (!OPTION (reduce_vanilla)) {
    ReduceScheduler reduce_scheduler (this);
    reduce_scheduler.increment_counter (force ? "forced " : "");
    if (force)
      message (2, "keeping reduce limit at %" PRIu64 " conflicts",
               limits.reduce.conflicts);
    else
      reduce_scheduler.set_next_conflict_limit ();
  } else
#endif
  {
    INC_COUNT (reductions);
    message (2,
             "reduction %" PRIu64 " of learned clauses at %" PRIu64
             " conflicts",
             statistics.reductions, statistics.conflicts);
  }
#ifndef QUIET
  if (force)
    statistics.forced_reductions++;
  else
    statistics.unforced_reductions++;
#endif
  if (level)
    mark_reason_clauses (true); // Protect reasoning clauses.

  // Gather reduce, garbage and simplified clauses.

  Position first_marked = INVALID_POSITION;
  mark_reduced_clauses (first_marked);
  mark_simplified_clauses (first_marked);

  // If we found clauses to collect, then collect them and flush watches.

  if (first_marked != INVALID_POSITION) {
    debug ("first marked clause at position %" POSFMT, first_marked);
    copy_collect_map_and_flush (first_marked);
  }

  if (level)
    mark_reason_clauses (false); // Release reason clause protection.

  // Remember how many units have been used for simplification
  // and assume all garbage clauses are collected.

  last.fixed_variables_at_reduce_or_collect = GET_COUNT (fixed_variables);
  SET_COUNT (garbage_clauses, 0);
  collecting_clauses = false;

#ifndef NVIVIFY
  if (trigger.vivification == 1) {
    debug ("triggering next vivification");
    trigger.vivification = 2;
  }
#endif

  if (trigger.elimination == 1) {
    debug ("triggering next elimination");
    trigger.elimination = 2;
  }

  stop ("reduce");

#ifndef QUIET
  if (verbosity > 0 ||
#ifndef VANILLA
      !OPTION (reduce_vanilla) ||
#endif
      is_power_of_two (statistics.reductions))
    report_progress ('-');
#endif
  check_all ();
}

}; // namespace TabularaSAT
