#include "references.hpp"
#include "inline.hpp"
#include "solver.hpp"

#include <cstring>

namespace TabularaSAT {

void Solver::resize_references_arena (References *new_end) {
#ifndef QUIET
  size_t old_size = matrix.arena.size ();
#endif

  matrix.arena.resize (new_end);
  if (matrix.arena.shrink_to_fit ()) {
#ifndef QUIET
    size_t new_capacity = matrix.arena.capacity ();
    pretty_memory_buffer buffer;
    message (1, "references arena capacity decreased to %zu words %s",
             new_capacity, pretty_memory (buffer, 4 * new_capacity));
#endif
  }
#ifndef QUIET
  {
    size_t new_size = matrix.arena.size ();
    size_t reclaimed_words = old_size - new_size;
    pretty_memory_buffer buffer;
    message (1, "reclaimed %zu words %s in references arena %.0f%%",
             reclaimed_words, pretty_memory (buffer, 4 * reclaimed_words),
             percent (reclaimed_words, old_size));
  }
#endif
}

void Matrix::clear_and_mark_as_garbage (References *refs) {
  unsigned lit = refs->lit;
  if (lit == INVALID_LITERAL)
    return;
#ifndef NDEBUG
  assert (lit < offsets.size ());
  size_t old_offset = offsets[lit];
  assert (arena[old_offset] == refs);
#endif
  refs->lit = INVALID_LITERAL;
  size_t words = References::words (refs->_capacity);
  garbage += words;
  offsets[lit] = 0;
}

#ifndef NDEBUG

void Solver::check_references () {
  References *begin = *matrix.arena.begin ();
  References *end = *matrix.arena.end ();
  assert (begin != end);
  References *refs = begin;
  assert (refs->lit == INVALID_LITERAL);
  assert (!refs->_size);
  assert (!refs->_capacity);
  size_t garbage = 0;
  References *next = refs->next ();
  assert (next <= end);
  for (refs = next; refs != end; refs = next) {
    assert (refs->_size <= refs->_capacity);
    next = refs->next ();
    assert (next <= end);
    unsigned lit = refs->lit;
    if (lit == INVALID_LITERAL) {
      unsigned words = References::words (refs->_capacity);
      assert ((unsigned *) refs + words == (unsigned *) next);
      garbage += words;
    } else {
      assert (lit < matrix.offsets.size ());
      size_t offset = (unsigned *) refs - (unsigned *) begin;
      assert (matrix.offsets[lit] == offset);
    }
  }
  assert (matrix.garbage == garbage);
  for (auto lit : literals) {
    assert (lit < matrix.offsets.size ());
    size_t offset = matrix.offsets[lit];
    assert (offset < matrix.arena.size ());
    assert (!offset || matrix.arena[offset]->lit == lit);
  }
}

#endif

// Garbage collection of references is triggered by 'enlarge_references'
// below, so independently of collecting garbage clauses and flushing
// watches in 'flush_and_map_watches' which in turn is called during
// garbage collecting clauses from 'reduce_redundant_clauses' and
// 'garbage_collect_clauses_during_variable_elimination'.  The reason for
// having this separate references garbage collection here is that for
// several large benchmarks the arena became full of garbage references
// during variable elimination without triggering garbage collection of
// clauses.  It is however easy to trigger and collect garbage references
// locally here whenever the arena has too many garbage references.

inline bool Solver::collecting_references () {
  size_t total_references = matrix.arena.size ();
  if (total_references < OPTION (references_garbage_minimum))
    return false;
  double fraction = OPTION (references_garbage_limit) / 100.0;
  size_t limit = fraction * total_references;
  return matrix.garbage > limit;
}

void Solver::collect_references () {

  debug ("collecting %zu garbage references %.0f%%", matrix.garbage,
         percent (matrix.garbage, matrix.arena.size ()));

  INC_STATS (collected_references);
  check_references ();

  start ("rollect", 2);

  References *begin_arena = *matrix.arena.begin ();
  References *end_arena = *matrix.arena.end ();
  References *dst_refs = begin_arena->next (); // Skip empty references.
  assert (dst_refs <= end_arena);

  for (References *src_refs = dst_refs, *next_src_refs;
       src_refs != end_arena; src_refs = next_src_refs) {
    next_src_refs = src_refs->next ();
    unsigned lit = src_refs->lit;
    if (lit == INVALID_LITERAL)
      continue;
    size_t size = src_refs->_size;
    size_t new_offset;
    if (size) {
      new_offset = (unsigned *) dst_refs - (unsigned *) begin_arena;
      dst_refs->lit = lit;
      dst_refs->_capacity = dst_refs->_size = size;
      size_t bytes = size * sizeof (Reference);
      memmove (dst_refs->data, src_refs->data, bytes);
      References *next = (References *) ((char *) &dst_refs->data + bytes);
      assert (next == dst_refs->next ());
      dst_refs = next;
    } else
      new_offset = 0;
    matrix.offsets[lit] = new_offset;
  }

  resize_references_arena (dst_refs);
  matrix.garbage = 0;

  stop ("rollect");
#ifndef QUIET
  if (verbosity > 0)
    report_progress ('G');
#endif

  check_references ();
}

// Our explicit memory allocator for 'References' is similar to the one in
// 'Kissat' and in principle simply moves full 'References' to the end of
// the arena after doubling their size leaving a hole of garbage references
// behind (marking the reference by setting its literal to
// 'INVALID_LITERAL').

// The main difficulty (shared with 'Kissat') with this approach in contrast
// to using 'std::vector<Reference>' for 'References' is that the arena
// might need to be resized through 'realloc' and move in memory.  Therefore
// any pointer to a 'References' stack during pushing any reference on
// another or the same 'References' stack needs to be adapted.

// This is in particular an issue when iterating 'References' of the
// negation of a propagated literal and pushing the clause reference to the
// 'References' of a replacement watched literal. That situation is solved
// by delaying pushing the references until traversal is finished.

References *Solver::enlarge_references (unsigned lit, References *old_refs,
                                        bool need_two) {
  if (collecting_references ()) {
    collect_references ();
    old_refs = references (lit);
  }

  assert (references (lit) == old_refs);

  References *next_refs = old_refs->next ();
  References *old_begin_arena = *matrix.arena.begin ();
  References *old_end_arena = *matrix.arena.end ();

  // We are checking whether the next references stack is garbage and if so
  // just will reuse its memory (which will increase the capacity by at
  // least 4 elements but not necessarily double it.  We also have to be
  // careful not to overshoot the end of the arena and further care about
  // the first shared empty references stack (with invalid literal).

  if (next_refs != old_end_arena && next_refs->lit == INVALID_LITERAL) {

    size_t reclaimed = References::words (next_refs->_capacity);
    assert (matrix.garbage >= reclaimed);

    // If we are not the first references stack we simply extend by reusing
    // the memory of the next references (including its three header words).

    if (old_refs->lit != INVALID_LITERAL) {
      size_t next_refs_bytes = References::bytes (next_refs->_capacity);
      assert (!(next_refs_bytes & (sizeof (Reference) - 1)));
      old_refs->_capacity += next_refs_bytes / sizeof (Reference);
      assert (old_refs->size () + 1 + need_two <= old_refs->capacity ());
      matrix.garbage -= reclaimed;
      return old_refs;
    }

    // In this corner case the old references point to the shared empty
    // references at the beginning of the arena (with invalid literal).

    assert (old_refs == old_begin_arena);
    assert (!old_refs->capacity ());
    assert (!old_refs->size ());

    // And we try to switch to the next (second) references stack.

    if (!need_two || next_refs->_capacity >= 2) {
      assert (next_refs->_capacity); // Only first stack has zero capacity.
      next_refs->lit = lit;          // So just reuse the next stack.
      next_refs->_size = 0;
      matrix.offsets[lit] = matrix.arena.position (next_refs);
      matrix.garbage -= reclaimed;
      return next_refs;
    }

    // If we need capacity for two more references (as we are trying to
    // watch a literal and thus need to push both blocking literal and a
    // clause reference to the same references stack) fall through.
  }

  unsigned old_size = old_refs->size ();
  unsigned old_capacity = old_refs->capacity ();

  // First determine the new capacity of the enlarged references stack.

  unsigned new_capacity = 0;
  if (!old_capacity)
    new_capacity = 1 + need_two;
  else if (old_capacity == 1)
    new_capacity = 2 + need_two;
  else if (old_capacity < (1u << 31)) {
    new_capacity = 2 * old_capacity;
    assert (old_capacity + 2 <= new_capacity);
  } else {

    // In order to use 4-byte '_size' and '_capacity' fields reducing the
    // memory of each references stack by 8 bytes we have a hard limit on
    // the size of these stacks of 'UINT_MAX = 2^32 - 1' which on the other
    // hand is (almost) 16 GB of watches in one references stack.

    // This limit might be exceeded for huge CNFs with many (2^31 or more)
    // large (say ternary) clauses with the same literal though which on the
    // other hand requires even that the 'clauses' field in a corresponding
    // DIMACS file for such a CNF exceeds 'INT_MAX'.

    // Also learning (and keeping) more than 2 billion clauses with the same
    // watched literal might hit the limit, which is on the other hand is
    // even more unlikely.

    new_capacity = UINT_MAX;
    if (new_capacity - 1 - need_two < old_capacity)
      fatal ("maximum capacity of %u words of references stack exceeded",
             new_capacity);
  }

#ifndef QUIET
  size_t old_arena_capacity = matrix.arena.capacity ();
#endif
  size_t new_words = References::words (new_capacity);
  References *new_refs = 0;

  bool extend_only = old_capacity && next_refs == old_end_arena;
  size_t old_words = References::words (old_capacity);

  if (extend_only) {

    // If the old references stack sits at the end of the arena we can
    // simply increase the arena and avoid copying the stack (except if it
    // is the shared zero capacity references stack at the beginning of the
    // arena). The offset of the literal also does not change in this case.

    assert (old_refs != old_begin_arena);
    size_t delta_words = new_words - old_words;
    (void) matrix.arena.allocate (delta_words);

  } else {

    // Allocate a bigger references stack in the arena (after the current
    // end of the arena) and map the literal to its new offset.

    assert (old_refs == old_begin_arena || old_refs->lit == lit);
    new_refs = matrix.arena.allocate (new_words);
    matrix.offsets[lit] = matrix.arena.position (new_refs);
  }

  References *new_begin_arena = *matrix.arena.begin ();
#ifndef QUIET
  {
    size_t new_arena_capacity = matrix.arena.capacity ();
    if (old_arena_capacity < new_arena_capacity) {
      pretty_memory_buffer buffer;
      message (2, "garbage references of %zu words and %s %.0f%%",
               matrix.garbage, pretty_memory (buffer, matrix.garbage * 4),
               percent (matrix.garbage, matrix.arena.size ()));
      message (1, "references capacity increased to %zu words %s ",
               new_arena_capacity,
               pretty_memory (buffer, new_arena_capacity * 4));
    }
  }
#endif

  if (!extend_only) {

    // If 'realloc' called in 'matrix.arena.allocate' moved the arena memory
    // we need to adapt the old 'References' pointed to with 'old_refs'.

    if (old_begin_arena != new_begin_arena) {
      std::ptrdiff_t diff =
          (char *) new_begin_arena - (char *) old_begin_arena;
      old_refs = (References *) (diff + (char *) old_refs);
    }

    // Leave the old references as garbage references behind by marking its
    // literal as invalid.  Note that the shared empty references stack at
    // the beginning of the arena always has an invalid literal field.

    if (old_capacity) {
      matrix.garbage += old_words;
      old_refs->lit = INVALID_LITERAL;
    }

    new_refs->lit = lit;
    new_refs->_capacity = new_capacity;
    new_refs->_size = old_size;

    if (old_size)
      memcpy (new_refs->data, old_refs->data,
              old_size * sizeof (Reference));

  } else {
    assert (!new_refs);
    if (new_begin_arena != old_begin_arena) {
      size_t old_offset =
          (unsigned *) old_refs - (unsigned *) old_begin_arena;
      assert (matrix.offsets[lit] == old_offset);
      new_refs = (References *) ((unsigned *) new_begin_arena + old_offset);
    } else
      new_refs = old_refs;
    new_refs->_capacity = new_capacity;
    assert (new_refs->next () == *matrix.arena.end ());
  }

  assert (new_refs == references (lit));

  return new_refs;
}

}; // namespace TabularaSAT
