#include "schedule.hpp"
#include "solver.hpp"

namespace TabularaSAT {

bool Solver::rephasing () {
  if (!OPTION (rephase))
    return false;
  if (!stable)
    return false;
  return GET_COUNT (conflicts) > limits.rephase;
}

void Solver::rephase_original () {
  signed char original_phase = OPTION (initial_phase) ? 1 : -1;
  for (auto idx : variables)
    phases[idx].target = original_phase;
  report_progress ('O');
}

void Solver::rephase_inverted () {
  signed char inverted_phase = OPTION (initial_phase) ? -1 : 1;
  for (auto idx : variables)
    phases[idx].target = inverted_phase;
  report_progress ('I');
}

void Solver::rephase_best () {
  signed char best;
  for (auto idx : variables)
    if ((best = phases[idx].best))
      phases[idx].target = best;
  best_assigned = 0;
  report_progress ('B');
}

#ifndef NWALK

void Solver::rephase_walk () {
  local_search ();
  report_progress ('W');
}

#endif

void Solver::rephase () {

  if (level && !backtrack_to_root_level_and_propagate ())
    return;

  stop_search ();
  start ("rephase", 2);

  RephaseScheduler rephase_scheduler (this);
  rephase_scheduler.increment_counter ();

  std::vector<char> schedule;

  if (OPTION (rephase_original)) {
    schedule.push_back ('O');
    if (OPTION (rephase_best))
      schedule.push_back ('B');
#ifndef NWALK
    if (OPTION (rephase_walk))
      schedule.push_back ('W');
#endif
  }

  if (OPTION (rephase_inverted)) {
    schedule.push_back ('I');
    if (OPTION (rephase_best))
      schedule.push_back ('B');
#ifndef NWALK
    if (OPTION (rephase_walk))
      schedule.push_back ('W');
#endif
  }

  if (schedule.empty ()) {
    if (OPTION (rephase_best))
      schedule.push_back ('B');
#ifndef NWALK
    if (OPTION (rephase_walk))
      schedule.push_back ('W');
#endif
  }

  if (!schedule.empty ()) {
    uint64_t rephasings = GET_COUNT (rephasings);
    unsigned pos = rephasings % schedule.size ();
    char scheduled = schedule[pos];
    switch (scheduled) {
    case 'B':
      rephase_best ();
      break;
    case 'I':
      rephase_inverted ();
      break;
    case 'O':
      rephase_original ();
      break;
#ifndef NWALK
    case 'W':
      rephase_walk ();
      break;
#endif
    }
    target_assigned = 0;
  }

  rephase_scheduler.set_next_conflict_limit ();

  stop ("rephase");
  start_search ();
}

}; // namespace TabularaSAT
