#include "inline.hpp"
#include "solver.hpp"

#include <cinttypes>

namespace TabularaSAT {

bool Solver::restarting () {
  if (!OPTION (restart))
    return false;
  if (!level)
    return false;
  Limits &l = limits;
  if (l.restart >= GET_COUNT (conflicts))
    return false;
#ifndef NGLUE
  if (stable && !OPTION (stable_restart_filter))
    return true;
  if (!stable && !OPTION (focused_restart_filter))
    return true;
  unsigned percent = OPTION (restart_margin);
  double scale = 1.0 + percent / 100.0;
  double slow = averages ().glue_slow;
  double fast = averages ().glue_fast;
  double padded = scale * slow;
  bool res = padded < fast;
  message (3,
           "%srestarting at conflict %" PRIu64
           " as %.2f = %.2f * slow %.2f %s fast glue %.2f",
           res ? "" : "not-", statistics.conflicts, padded, scale, slow,
           res ? "<" : ">=", fast);
  return res;
#else
  return true;
#endif
}

#ifndef NSWITCH

void Solver::set_focused_restart_limit (bool initial) {
  uint64_t interval = OPTION (filtered_restart_interval);
  uint64_t restarts = GET_COUNT (restarts);
  if (restarts)
    interval += logn (restarts) - 1;
  limits.restart = GET_COUNT (conflicts) + interval;
  if (initial)
    message (2, "initial focused restart limit at %" PRIu64 " conflicts",
             limits.restart);
  else
    message (3,
             "new focused restart limit at %" PRIu64 " after %" PRIu64
             " conflicts",
             limits.restart, interval);
}

#endif

void Solver::set_stable_restart_limit (bool initial) {
  uint64_t u = reluctant.u, v = reluctant.v;
  if ((u & -u) == v)
    u++, v = 1;
  else
    v *= 2;
  uint64_t interval = v;
  interval *= OPTION (reluctant_restart_interval);
  reluctant.u = u, reluctant.v = v;
  limits.restart = GET_COUNT (conflicts) + interval;
  if (initial)
    message (2, "initial stable restart limit at %" PRIu64 " conflicts",
             limits.restart);
  else
    message (3,
             "new stable restart limit at %" PRIu64 " after %" PRIu64
             " conflicts",
             limits.restart, interval);
}

void Solver::set_restart_limit (bool initial) {
#ifndef NSWITCH
  if (!stable)
    set_focused_restart_limit (initial);
  else
#endif
    set_stable_restart_limit (initial);
}

#ifndef NREUSETRAIL

unsigned Solver::reuse_trail () {

  if (!OPTION (reuse_trail))
    return 0;
  if (stable && !OPTION (reuse_trail_stable_mode))
    return 0;
  if (!stable && !OPTION (reuse_trail_focused_mode))
    return 0;

  assert (level);
  assert (unassigned);

  unsigned decision = INVALID_VARIABLE;
  unsigned res = level;

  size_t end = trail.size ();

  double decision_score = HEAP_SCORE_LIMIT;
  unsigned decision_stamp = UINT_MAX;

  for (size_t i = level; res && i; i--) {
    size_t begin = control[i].trail;
    assert (begin < end);
    for (size_t j = begin; res && j != end; j++) {
      unsigned lit = trail[j];
      unsigned idx = index (lit);
      unsigned lit_level = levels[idx];
      if (!lit_level) {
        res = 0;
        break;
      }
      if (lit_level > res)
        continue;
      if (lit_level < i)
        res = lit_level;
      if (stable) {
        if (decision == INVALID_VARIABLE) {
          decision = decide_variable_in_stable_mode ();
          decision_score = bumper.score (decision);
        }
        unsigned lit_score = bumper.score (idx);
        if (decision_score < lit_score)
          res = lit_level - 1;
      } else {
        if (decision == INVALID_VARIABLE) {
          decision = decide_variable_in_focused_mode ();
          decision_stamp = queue.stamped (decision);
        }
        unsigned lit_stamp = queue.stamped (idx);
        if (decision_stamp < lit_stamp)
          res = lit_level - 1;
      }
    }
    end = begin;
  }

  if (res) {
    message (3, "reusing trail level %u", res);
#ifndef QUIET
    statistics.reused_trails++;
    statistics.reused_levels += res;
#endif
  }

  return res;
}

#endif

void Solver::restart () {
  INC_COUNT (restarts);
  message (3, "restart %" PRIu64 " at conflict %" PRIu64,
           GET_COUNT (restarts), GET_COUNT (conflicts));
#ifndef NREUSETRAIL
  unsigned restart_level = reuse_trail ();
#else
  unsigned restart_level = 0;
#endif
  if (restart_level < level)
    backtrack (restart_level);
  set_restart_limit ();
}

}; // namespace TabularaSAT
