#ifndef _map_hpp_INCLUDED
#define _map_hpp_INCLUDED

#include <cassert>
#include <cstdint>
#include <cstdlib>

namespace TabularaSAT {

void fatal_error (const char *where, const char *, ...)
    __attribute__ ((format (printf, 2, 3)));

static const uint64_t DELETED_HASH_ENTRY = (uint64_t) 0 - 1;

class Map {
  size_t _size = 0;
  size_t _deleted = 0;
  size_t _capacity = 0;

  struct Bucket {
    uint64_t key, data;
  };

  Bucket *table = 0;

#ifndef NDEBUG
  static bool is_power_of_two (size_t n) { return n && !(n & (n - 1)); }
#endif

  size_t reduce (uint64_t e) const {
    uint64_t res = e;
    if (_capacity >= (1ull << 32))
      res ^= res >> 32;
    if (_capacity >= (1ull << 16))
      res ^= res >> 16;
    if (_capacity >= (1ull << 8))
      res ^= res >> 8;
    if (_capacity >= (1ull << 4))
      res ^= res >> 4;
    assert (is_power_of_two (_capacity));
    return res & (_capacity - 1);
  }

  size_t hash1 (uint64_t e) const {
    return reduce (e * 6156156156156156155ull);
  }

  size_t hash2 (uint64_t e) const {
    uint64_t res = reduce (e * 7657657657657657655ull);
    if (!(res & 1) && ++res == _capacity)
      res = 1;
    assert (res & 1), assert (res < _capacity);
    return res;
  }

  template <class Condition>
  Bucket &find (uint64_t e, Condition condition) {
    size_t i = hash1 (e);
    Bucket *b = table + i;
#ifndef NDEBUG
    auto start = i;
#endif
    if (!condition (b->key)) {
      size_t delta = hash2 (e);
      do {
        i += delta;
        if (i >= _capacity)
          i -= _capacity;
        assert (i != start);
        assert (i < _capacity);
        b = table + i;
      } while (!condition (b->key));
    }
    return *b;
  }

  struct empty_word {
    bool operator() (uint64_t key) {
      assert (key != DELETED_HASH_ENTRY);
      return !key;
    }
  };

  struct free_word {
    bool operator() (uint64_t key) {
      return !key || key == DELETED_HASH_ENTRY;
    }
  };

  struct matching_word {
    uint64_t match;
    matching_word (uint64_t m) : match (m) {
      assert (match), assert (match != DELETED_HASH_ENTRY);
    }
    bool operator() (uint64_t key) {
      assert (key);
      return key == match;
    }
  };

  void resize (size_t new_capacity) {
    size_t old_capacity = _capacity;
    Bucket *old_table = table;
    _capacity = new_capacity;
    if (new_capacity) {
      table = (Bucket *) calloc (new_capacity, sizeof *table);
      if (!table)
        fatal_error ("Map::resize ()", "out-of-memory");
    } else
      table = 0;
    for (size_t h = 0; h != old_capacity; h++) {
      Bucket &old_bucket = old_table[h];
      if (!old_bucket.key || old_bucket.key == DELETED_HASH_ENTRY)
        continue;
      Bucket &new_bucket = find (old_bucket.key, empty_word ());
      new_bucket = old_bucket;
    }
    if (old_table)
      free (old_table);
    _deleted = 0;
  }

  void enlarge () {
    if (_size >= 3 * (_capacity / 4))
      resize (_capacity ? 2 * _capacity : 1);
  }

  void shrink () {
    if (_size < _capacity / 4)
      resize (_capacity / 2);
  }

public:
  ~Map () {
    if (table)
      free (table);
  }

  bool empty () const { return !_size; }
  size_t size () const { return _size; }

  void insert (uint64_t key, uint64_t data) {
    assert (key), assert (key != DELETED_HASH_ENTRY);
    enlarge ();
    Bucket &bucket = find (key, free_word ());
    if (bucket.key == DELETED_HASH_ENTRY)
      assert (_deleted), _deleted--;
    bucket = {key, data};
    _size++;
  }

  void overwrite (uint64_t key, uint64_t data) {
    assert (key), assert (key != DELETED_HASH_ENTRY);
    Bucket &bucket = find (key, matching_word (key));
    assert (bucket.key == key);
    bucket.data = data;
  }

  uint64_t remove (uint64_t key) {
    assert (key), assert (key != DELETED_HASH_ENTRY);
    Bucket &bucket = find (key, matching_word (key));
    assert (bucket.key == key);
    bucket.key = DELETED_HASH_ENTRY;
    uint64_t data = bucket.data;
    assert (_size);
    _size--;
    _deleted++;
    shrink ();
    return data;
  }
};

}; // namespace TabularaSAT

#endif
