#ifndef _queue_hpp_INCLUDED
#define _queue_hpp_INCLUDED

#include <climits>
#include <vector>

namespace TabularaSAT {

void fatal_error (const char *where, const char *, ...)
    __attribute__ ((format (printf, 2, 3)));

// Variable-Move-To-Front (VMTF) decision variable queue with time stamping.

#define QUEUE_INVALID UINT_MAX

class Queue {

  struct Link {
    unsigned stamp;
    unsigned prev = QUEUE_INVALID, next = QUEUE_INVALID;
  };

  unsigned stamp = 0, search = QUEUE_INVALID;
  unsigned first = QUEUE_INVALID, last = QUEUE_INVALID;
  std::vector<Link> links;

  void restamp () {
    stamp = 0;
    for (unsigned idx = first, next; idx != QUEUE_INVALID; idx = next) {
      Link &link = links[idx];
      link.stamp = ++stamp;
      next = link.next;
    }
  }

public:
  bool empty () const { return search == QUEUE_INVALID; }

  size_t expensive_size () const {
    size_t res = 0;
    for (unsigned idx = first, next; idx != QUEUE_INVALID; idx = next)
      next = links[idx].next, res++;
    return res;
  }

  size_t capacity () const { return links.size (); }

  unsigned stamped (unsigned idx) const {
    assert (idx < capacity ());
    return links[idx].stamp;
  }

  unsigned operator* () const {
    assert (!empty ());
    return search;
  }

  void pop () {
    assert (!empty ());
    assert (search < capacity ());
    search = links[search].prev;
  }

  void enqueue (unsigned idx, bool update = false) {
    assert (idx != QUEUE_INVALID);
    assert (idx < capacity ());
    if (last == QUEUE_INVALID)
      first = idx;
    else
      links[last].next = idx;
    Link &link = links[idx];
    link.prev = last;
    link.next = QUEUE_INVALID;
    link.stamp = ++stamp;
    if (!stamp)
      restamp ();
    last = idx;
    if (update || search == QUEUE_INVALID)
      search = idx;
  }

  bool contains (unsigned idx) {
    assert (idx != QUEUE_INVALID);
    if (last == idx)
      return true;
    if (links.capacity () <= idx)
      return false;
    Link &link = links[idx];
    return link.next != QUEUE_INVALID;
  }

  void dequeue (unsigned idx) {
    assert (contains (idx));
    assert (idx != QUEUE_INVALID);
    assert (idx < capacity ());
    Link &link = links[idx];
    if (link.next == QUEUE_INVALID)
      last = link.prev;
    else
      links[link.next].prev = link.prev;
    if (link.prev == QUEUE_INVALID)
      first = link.next;
    else
      links[link.prev].next = link.next;
    if (search == idx) {
      search = link.prev;
      if (search == QUEUE_INVALID)
        search = link.next;
    }
    link.prev = link.next = QUEUE_INVALID;
  }

  void resize (unsigned variables) { links.resize (variables); }

  void update (unsigned idx) {
    assert (idx != QUEUE_INVALID);
    assert (idx < capacity ());
    if (search == QUEUE_INVALID || links[search].stamp < links[idx].stamp)
      search = idx;
  }
};

}; // namespace TabularaSAT

#endif
