// clang-format off

static const char * copyright[] = {
"2022-2023 Armin Biere, University of Freiburg, Germany"
};

static const char *basic_usage_header =
"usage: %s [ <option> ... ] [ <dimacs> [ <proof> ] ]\n"
;

static const char *basic_usage_body =

"with the list of basic options\n"
"\n"
"  -a | --ascii             use ASCII variant of DRAT proof format\n"
"  -h | --help              print this command line option summary\n"
#ifdef LOGGING
"  -l | --logging           print very verbose logging information\n"
#endif                     
"  -n | --no-witness        do not print witness if satisfiable\n"
#ifndef QUIET
"  -q | --quiet             disable all messages\n"
#endif
"  -s | --statistics        print full statistics\n"
#ifndef QUIET
"  -v | --verbose           increase verbosity\n"
#endif
"\n"
#ifndef QUIET
"  -p <level>     | --profile=<level>  set profiling level\n"
"                   --no-profile       disable internal profiling\n"
#endif
"  -c <conflicts> | --conflicts=<n>    set conflicts limit\n"
"                   --no-conflicts     set conflict limit to zero\n"
"  -t <seconds>   | --time=<seconds>   set time limit\n"
#ifndef NOPTIONS
"\n"
"  -A | --all-options       print all configurations and options\n"
#endif
;

#ifndef NOPTIONS

static const char *all_options_infix =

"  -B | --banner | --build  print banner, build and version information\n"
"  -E | --embedded-options  print options embedded as DIMACS comments\n"
"  -L | --list-options      list without arguments for autocompletion\n"
"  -R | --option-ranges     print option value ranges for CNF fuzzing\n"
"  -S | --print-sizes       print sizes of internal data-structures\n"
"  -V | --version           print version number\n"

;

#endif

static const char * basic_usage_footer =

"and '<dimacs>' the input file in DIMACS format, optionally compressed\n"
"with 'gzip', 'bzip2' or 'xz'.  If no input file is specified the solver\n"
"reads from '<stdin>'.  If the '<proof>' is given a DRAT (actually DRUP)\n"
"proof is written to that file unless the path name '-' is used in which\n"
"case it is written to '<stdout>'.  In this case the output format is the\n"
"ASCII version of DRAT, which can also be enforced by specifying '--ascii'.\n"

;

#ifndef NOPTIONS

static const char *all_options_suffix =

"Without '=' and no argument the value of the internal option is set to '1'\n"
"and to '0' if the name is preceded with 'no'.  For instance '--no-rephase'\n"
"switches off rephasing.  Further 'true' and 'false' are treated as synonyms\n"
"for '1' and '0' respectively.\n"

;

#endif

// clang-format on

#include "shared.hpp"

#include <cctype>
#include <climits>
#include <csignal>
#include <cstdarg>
#include <cstdlib>
#include <cstring>

extern "C" {

#include <stdio.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
};

namespace TabularaSAT {

#ifndef NOPTIONS

void Shared::print_all_option_ranges () {
  for (auto &o : iterate_options ())
    printf ("%s %u %u %u\n", o.name, o.minimum, o.initial, o.maximum);
}

void Shared::print_all_embedded_options () {
  for (auto &o : iterate_options ())
    printf ("c --%s=%u\n", o.name, o.initial);
}

void Shared::print_all_option_usage () {
  printf (basic_usage_header, one_word_name ());
  fputc ('\n', stdout);
  fputs (basic_usage_body, stdout);
  fputs (all_options_infix, stdout);
  fputc ('\n', stdout);
  fputs (basic_usage_footer, stdout);
  printf ("\nPredefined list of configurations (option sets):\n\n");
  for (auto configuration : iterate_configurations ())
    printf ("  --%-17s %s\n", configuration->name (),
            configuration->description ());
  printf ("\nThe full list of internal options is as follows:\n\n");
  size_t len = 0;
  for (int i = 0; i != 2; i++) {
    for (auto &option : iterate_options ()) {

      char type[32];
      if (option.minimum == option.maximum)
        snprintf (type, sizeof type, "%u", option.minimum);
      else if (option.minimum == 0 && option.maximum == 1)
        strcpy (type, "bool");
      else if (option.maximum == INT_MAX)
        snprintf (type, sizeof type, "%u..INT_MAX", option.minimum);
      else
        snprintf (type, sizeof type, "%u..%u", option.minimum,
                  option.maximum);

      char name[80];
      assert (strlen (name) + strlen (type) + 2 < sizeof name);
      snprintf (name, sizeof name, "%s=%s", option.name, type);

      if (i) {
        fputs ("  --", stdout);
        fputs (name, stdout);
        size_t tmp = strlen (name);
        assert (tmp <= len);
        for (size_t i = tmp; i != len; i++)
          fputc (' ', stdout);
        fputs ("  ", stdout);
        fputs (option.description, stdout);
        fputs (" (default ", stdout);
        char value[16];
        if (option.initial == INT_MAX)
          strcpy (value, "INT_MAX");
        else if (option.maximum == 1 && option.initial)
          strcpy (value, "true");
        else if (option.maximum == 1 && !option.initial)
          strcpy (value, "false");
        else
          snprintf (value, sizeof value, "%u", option.value);
        fputs (value, stdout);
        fputs (")\n", stdout);
      } else {
        size_t tmp = strlen (name);
        if (tmp > len)
          len = tmp;
      }
    }
  }
  fputc ('\n', stdout);
  fputs (all_options_suffix, stdout);
}

void Shared::print_all_option_values () {

  size_t printed_option_lines = 0;

  for (auto &option : iterate_options ()) {
    if (option.value != option.initial) {
      if (!printed_option_lines++)
        fputs ("c\n", stdout);
      printf ("c using option '--%s=%u' (different from default %u)\n",
              option.name, option.value, option.initial);
    }
#ifndef QUIET
    else if (verbosity) {
      printed_option_lines++;
      printf ("c using option '--%s=%u' (same as default %u)\n",
              option.name, option.value, option.initial);
    }
#endif
  }
}

static char *normalize_option (const char *name) {
  char *res;
  {
    const char *p = name;
    while (*p == '-')
      p++;
    res = strdup (p);
  }
  {
    for (char *p = res, ch; (ch = *p); p++)
      if (ch == '=') {
        *p = 0;
        break;
      }
  }
  return res;
}

#define INFINITE_DISTANCE (~((size_t) 0))

static size_t edit_distance (const char *a, const char *b, size_t limit) {
  char c = *a, d = *b;
  if (!c && !d)
    return 0;
  if (c && c == d)
    return edit_distance (a + 1, b + 1, limit);
  if (!limit)
    return INFINITE_DISTANCE;
  size_t res = INFINITE_DISTANCE, tmp;
  if (c) {
    tmp = edit_distance (a + 1, b, limit - 1);
    if (tmp < res)
      res = tmp;
  }
  if (d) {
    tmp = edit_distance (a, b + 1, limit - 1);
    if (tmp < res)
      res = tmp;
  }
  if (c && d) {
    tmp = edit_distance (a + 1, b + 1, limit - 1);
    if (tmp < res)
      res = tmp;
  }
  if (res == INFINITE_DISTANCE)
    return INFINITE_DISTANCE;
  return res + 1;
}

static void parse_shared_options (std::vector<char *> &shared_options,
                                  const char *start) {
  for (const char *p = start; *p; p++) {
    if (p[0] == ' ' && p[1] == ' ' && p[2] == '-') {
      p += 3;
      while (*p != '-')
        assert (*p), assert (*p != '\n'), p++;
      p += 2;
      const char *n = p;
      while (*p == '-' || ('a' <= *p && *p <= 'z') ||
             ('0' <= *p && *p <= '9'))
        p++;
      size_t name_len = p - n;
      char *name = (char *) malloc (name_len + 1);
      memcpy (name, n, name_len);
      name[name_len] = 0;
      while (*p != '\n')
        assert (*p), p++;
      const char *d = p - 1;
      while (assert (n < d), d[-1] != ' ' || d[-2] != ' ')
        d--;
      size_t description_len = p - d;
      char *description = (char *) malloc (description_len + 1);
      memcpy (description, d, description_len);
      description[description_len] = 0;
      shared_options.push_back (name);
      shared_options.push_back (description);
    } else {
      while (*p != '\n')
        assert (*p), p++;
    }
  }
}

void Shared::find_matching_options (const char *name,
                                    std::vector<const char *> &res) {

  char *normalized = normalize_option (name);

  if (shared_options.empty ()) {
    parse_shared_options (shared_options, basic_usage_body);
    parse_shared_options (shared_options, all_options_infix);
  }

  const size_t limit = 3;
  size_t minimum_distance = INFINITE_DISTANCE;

  for (size_t i = 0; i != shared_options.size (); i += 2) {
    const char *name = shared_options[i];
    const char *description = shared_options[i + 1];
    size_t tmp = edit_distance (normalized, name, limit);
    if (strlen (normalized) == tmp)
      continue;
    if (strlen (name) == tmp)
      continue;
    if (tmp == INFINITE_DISTANCE)
      continue;
    if (tmp > minimum_distance)
      continue;
    if (tmp < minimum_distance) {
      minimum_distance = tmp;
      res.clear ();
    }
    res.push_back (name);
    res.push_back (description);
  }

  for (auto c : iterate_configurations ()) {
    size_t tmp = edit_distance (normalized, c->name (), limit);
    if (tmp == INFINITE_DISTANCE)
      continue;
    if (tmp > minimum_distance)
      continue;
    if (tmp < minimum_distance) {
      minimum_distance = tmp;
      res.clear ();
    }
    res.push_back (c->name ());
    res.push_back (c->description ());
  }

  for (auto &o : iterate_options ()) {
    size_t tmp = edit_distance (normalized, o.name, limit);
    if (tmp == INFINITE_DISTANCE)
      continue;
    if (tmp > minimum_distance)
      continue;
    if (tmp < minimum_distance) {
      minimum_distance = tmp;
      res.clear ();
    }
    res.push_back (o.name);
    res.push_back (o.description);
  }

  free (normalized);
}

#endif

static void print_option_completions (const char *str) {
  int ch, last = ' ';
  for (const char *p = str; (ch = *p); last = ch, p++) {
    if (ch == '-' && last == ' ') {
      const char *q = p;
      while (ch != ' ' && ch != '=') {
        assert (ch);
        assert (('0' <= ch && ch <= '9') || ('a' <= ch && ch <= 'z') ||
                ('A' <= ch && ch <= 'Z') || ch == '-');
        fputc (ch, stdout);
        ch = *++p;
      }
      fputc ('\n', stdout);
      if (ch == '=') {
        while (q != p)
          fputc (*q++, stdout);
        fputs ("=\n", stdout);
      }
    }
  }
}

void Shared::print_all_option_completions () {
  print_option_completions (basic_usage_body);
#ifndef NOPTIONS
  print_option_completions (all_options_infix);
  for (auto &option : iterate_options ()) {
    printf ("--%s\n", option.name);
    if (option.maximum > 1)
      printf ("--%s=\n", option.name);
    if (!option.minimum)
      printf ("--no-%s\n", option.name);
  }
  for (auto configuration : iterate_configurations ())
    printf ("--%s\n", configuration->name ());
#endif
}

Shared::Shared () {}

Shared::~Shared () {

  free (levels);
  free (values);
  free (marks);
#ifndef NOPTIONS
  for (auto s : shared_options)
    free (s);
#endif
}

void Shared::set_verbosity (int new_verbosity) {
#ifndef QUIET
  verbosity = new_verbosity;
#else
  (void) new_verbosity;
#endif
}
void Shared::force_ascii_proof_format () { ascii = true; }

#ifdef LOGGING // Start of section with debugging / logging code.

void Shared::debug_prefix () {
  assert (logging ());
  printf ("c DEBUG %u ", level);
}

void Shared::debug_suffix () {
  assert (logging ());
  fputc ('\n', stdout);
  fflush (stdout);
}

void Shared::debug (const char *fmt, ...) {
  if (!logging ())
    return;
  debug_prefix ();
  va_list ap;
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  debug_suffix ();
}

const char *Shared::debug_literal (unsigned lit) {
  assert (lit < literals);
  Buffer &buffer = buffers.next ();
  char *res = buffer.chars;
  int exported = export_literal (lit);
  snprintf (res, sizeof buffer.chars, "%u(%d)", lit, exported);
  int value = values[lit];
  if (value) {
    size_t res_len = strlen (res);
    assert (res_len < sizeof buffer.chars);
    size_t remaining = sizeof buffer.chars - res_len;
    char *tmp = res + res_len;
    unsigned idx = index (lit);
    unsigned level = levels[idx];
    if (level != INVALID_LEVEL) {
      snprintf (tmp, remaining, "@%u", level);
      size_t tmp_len = strlen (tmp);
      assert (tmp_len < remaining);
      remaining -= tmp_len;
      tmp += tmp_len;
    }
    snprintf (tmp, remaining, "=%d", value);
  }
  assert (strlen (res) < sizeof buffer.chars);
  return res;
}

const char *Shared::debug_variable (unsigned idx) {
  unsigned lit = literal (idx);
  const char *tmp = debug_literal (lit);
  Buffer &buffer = buffers.next ();
  char *res = buffer.chars;
  snprintf (res, sizeof buffer.chars, "variable %u literal %s", idx, tmp);
  assert (strlen (res) < sizeof buffer.chars);
  return res;
}

void Shared::debug_literals (size_t size, unsigned *literals,
                             const char *fmt, ...) {
  if (!logging ())
    return;
  debug_prefix ();
  va_list ap;
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  printf (" size %zu clause", size);
  for (size_t i = 0; i != size; i++)
    printf (" %s", debug_literal (literals[i]));
  debug_suffix ();
}

void Shared::debug_vector (std::vector<unsigned> &v, const char *fmt, ...) {
  if (!logging ())
    return;
  debug_prefix ();
  va_list ap;
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  printf (" size %zu clause", v.size ());
  for (auto lit : v)
    printf (" %s", debug_literal (lit));
  debug_suffix ();
}

#endif // End of section with debugging / logging code.

const char *Shared::decode_state (int state) {
  switch (state) {
  case READY:
    return "READY";
  case ADDING:
    return "ADDING";
  case SATISFIABLE:
    return "SATISFIABLE";
  case UNSATISFIABLE:
    return "UNSATISFIABLE";
  }
  return "INVALID";
}

void Shared::transition (State new_state) {
  debug ("transition to state '%s' from state '%s'",
         decode_state (new_state), decode_state (state));
  state = new_state;
}

void Shared::resize (unsigned new_variables) {
  if (variables >= new_variables)
    return;
  variables.resize (new_variables);
  literals.resize (2 * variables);
  if (capacity < new_variables) {
    size_t old_capacity = capacity;
    if (old_capacity) {
      capacity = 2 * old_capacity;
      while (capacity < new_variables)
        capacity *= 2;
    } else
      capacity = new_variables;

    {
      unsigned *old_levels = levels;
      levels = (unsigned *) calloc (capacity, sizeof *levels);
      if (old_levels) {
        memcpy (levels, old_levels, old_capacity * sizeof *levels);
        free (old_levels);
      }
    }

    {
      signed char *old_values = values;
      values = (signed char *) calloc (2 * capacity, sizeof *values);
      if (old_values) {
        memcpy (values, old_values, 2 * old_capacity * sizeof *values);
        free (old_values);
      }
    }

    {
      signed char *old_marks = marks;
      marks = (signed char *) calloc (2 * capacity, sizeof *marks);
      if (old_marks) {
        memcpy (marks, old_marks, 2 * old_capacity * sizeof *marks);
        free (old_marks);
      }
    }

    if (!levels || !values || !marks)
      fatal ("out-of-memory adjusting shared variable data");
  }
  adjust_to_increased_variables ();
  debug ("resized to %u variables", (unsigned) variables);
}

unsigned Shared::import_literal (int elit) {
  assert (elit);
  assert (elit != INT_MIN);
  int eidx = abs (elit);
  unsigned iidx = eidx - 1;
  if (iidx >= variables)
    resize (eidx);
  unsigned ilit = literal (iidx);
  if (elit < 0)
    ilit = negate (ilit);
  debug ("imported external literal %d as internal literal %u", elit, ilit);
  return ilit;
}

int Shared::export_literal (unsigned internal_literal) const {
  assert (internal_literal < literals);
  unsigned internal_index = index (internal_literal);
  assert (internal_index < (unsigned) INT_MAX);
  int external_literal = internal_index + 1;
  if (sign (internal_literal))
    external_literal = -external_literal;
  return external_literal;
}

#ifndef QUIET

const char *Shared::pretty_memory (pretty_memory_buffer buffer,
                                   size_t bytes) {
  double kb = bytes / (double) (1u << 10);
  double mb = bytes / (double) (1u << 20);
  double gb = bytes / (double) (1u << 30);
  size_t size = SIZE_PRETTY_MEMORY_BUFFER;
  if (kb < 1)
    snprintf (buffer, size, "%zu bytes", bytes);
  else if (mb < 1)
    snprintf (buffer, size, "%zu bytes %.1f KB", bytes, kb);
  else if (gb < 1)
    snprintf (buffer, size, "%zu bytes %.1f MB", bytes, mb);
  else
    snprintf (buffer, size, "%zu bytes %.1f GB", bytes, gb);
  return buffer;
}

void Shared::message (int level, const char *fmt, ...) {
  if (verbosity < level)
    return;
  fputs ("c ", stdout);
  va_list ap;
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
  fflush (stdout);
}

#endif

void Shared::error (const char *fmt, ...) {
  fputs (one_word_name (), stderr);
  fputs (": error: ", stderr);
  va_list ap;
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
  exit (1);
}

void Shared::invalid_option (const char *arg) {
  assert (arg[0] == '-');
  fputs (one_word_name (), stderr);
  fprintf (stderr, ": error: invalid option '%s': ", arg);
#ifndef NOPTIONS
  {
    std::vector<const char *> matching;
    find_matching_options (arg, matching);
    if (matching.empty ())
      fputs ("use", stderr);
    else {
      fputs ("closest matching ", stderr);
      assert (!(matching.size () & 1));
      fputs (matching.size () == 2 ? "is" : "are", stderr);
      fputs (":\n\n", stderr);
      size_t name_length = 0;
      for (size_t i = 0; i != matching.size (); i += 2)
        name_length = std::max (name_length, strlen (matching[i]));
      char fmt[32];
      snprintf (fmt, sizeof fmt, "  --%%-%zus  %%s\n", name_length);
      for (size_t i = 0; i != matching.size (); i += 2)
        fprintf (stderr, fmt, matching[i], matching[i + 1]);
      fputs ("\nUse", stderr);
    }
  }
  fputs (" '-h' for usage or '-A' to print all options.\n", stderr);
#else
  fputs ("use '-h' for usage.\n", stderr);
#endif
  fflush (stderr);
  exit (1);
}

void Shared::fatal (const char *fmt, ...) {
  fputs (one_word_name (), stderr);
  fputs (": fatal error: ", stderr);
  va_list ap;
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
  abort ();
}

void fatal_error (const char *where, const char *fmt, ...) {
  fprintf (stderr, "tabularasat: fatal error in '%s': ", where);
  va_list ap;
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
  abort ();
}

#define size_write_buffer (1u << 20)

struct Trace {
  Shared &shared;
  FILE *file;
  size_t bytes = 0;
  struct {
    size_t added = 0, deleted = 0;
  } lines;
  char int_buffer[16];
  struct {
    size_t pos = 0;
    char chars[size_write_buffer];
  } write_buffer;
  Trace (Shared &s, FILE *f) : shared (s), file (f) {
    int_buffer[sizeof int_buffer - 1] = 0;
  }
  void flush_buffer () {
    size_t write = write_buffer.pos;
    if (!write)
      return;
    size_t written = fwrite (write_buffer.chars, 1, write, file);
    if (written != write)
      shared.fatal ("flushing %zu bytes in proof trace buffer failed",
                    bytes);
    write_buffer.pos = 0;
    bytes += written;
  }
  ~Trace () {
    flush_buffer ();
#ifndef QUIET
    size_t total = lines.added + lines.deleted;
    shared.message (0, "wrote %zu bytes (%.0f MB) to proof file", bytes,
                    bytes / (double) (1u << 20));
    shared.message (0, "wrote %zu proof lines (%zu added, %zu deleted)",
                    total, lines.added, lines.deleted);
#endif
  }
  inline void write_char (unsigned ch) __attribute__ ((always_inline)) {
    if (write_buffer.pos == size_write_buffer)
      flush_buffer ();
    write_buffer.chars[write_buffer.pos++] = ch;
    bytes++;
  }
  inline void write_space () { write_char (' '); }
  inline void write_str (const char *str) {
    for (const char *p = str; *p; p++)
      write_char (*p);
  }
  inline void write_int (int i) __attribute__ ((always_inline)) {
    if (i) {
      char *p = int_buffer + sizeof int_buffer - 1;
      assert (!*p);
      assert (i != INT_MIN);
      unsigned tmp = abs (i);
      while (tmp) {
        *--p = '0' + (tmp % 10);
        tmp /= 10;
      }
      if (i < 0)
        *--p = '-';
      write_str (p);
    } else
      write_char ('0');
  }
};

void Shared::set_proof (FILE *file) {
  if (!file)
    error ("zero proof file argument in 'set_proof");
  if (proof)
    error ("can not call 'set_proof' twice");
  if (state != READY)
    error ("'set_proof' called in '%s' state", decode_state (state));
  if (added_original_clauses)
    error ("'set_proof' called after adding clauses");
  proof = new Trace (*this, file);
}

void Shared::flush_proof () {
  assert (proof);
  delete proof;
  proof = 0;
}

inline void Shared::trace_only_literals (size_t size,
                                         const unsigned *literals,
                                         unsigned except) {
  assert (proof);
  unsigned lit;
  if (ascii) {
    for (const unsigned *p = literals, *end = p + size; p != end; p++)
      if ((lit = *p) != except)
        proof->write_int (export_literal (lit)), proof->write_space ();
    proof->write_str ("0\n");
  } else {
    for (const unsigned *p = literals, *end = p + size; p != end; p++) {
      if ((lit = *p) == except)
        continue;
      unsigned tmp = lit + 2;
      while (tmp & ~0x7f)
        proof->write_char ((tmp & 0x7f) | 0x80), tmp >>= 7;
      proof->write_char (tmp);
    }
    proof->write_char (0);
  }
}

void Shared::trace_literals (char type, size_t size,
                             const unsigned *literals, unsigned except) {
  assert (type == 'a' || type == 'd');
  if (!proof)
    return;
  if (!ascii)
    proof->write_char (type);
  else if (type == 'd')
    proof->write_str ("d ");
  trace_only_literals (size, literals, except);
  if (type == 'a')
    proof->lines.added++;
  else
    proof->lines.deleted++;
}

void Shared::add (int elit) {
  if (elit == INT_MIN || abs (elit) > MAXIMUM_EXTERNAL_VARIABLE_INDEX)
    error ("invalid literal in 'add (%d)'", elit);
  if (state != READY && state != ADDING)
    error ("can not call 'add (%d)' in '%s' state", elit,
           decode_state (state));
  if (state != ADDING)
    transition (ADDING);
#ifndef NDEBUG
  original.push_back (elit);
#endif
  if (elit) {
    unsigned ilit = import_literal (elit);
    imported.push_back (ilit);
  } else {
    added_original_clauses++;
    assert (clause.empty ());
    bool tautological = false;
    for (auto ilit : imported) {
      signed char value = values[ilit];
      if (value > 0) {
        debug_vector (imported, "original[%zu] %s satisfied unsimplified",
                      added_original_clauses, debug_literal (ilit));
        tautological = true;
        break;
      }
      if (value < 0)
        continue;
      unsigned iidx = index (ilit);
      signed char mark = marks[iidx];
      if (sign (ilit))
        mark = -mark;
      if (mark < 0) {
        debug_vector (imported, "original[%zu] tautological unsimplified",
                      added_original_clauses);
        tautological = true;
        break;
      }
      if (mark > 0)
        continue;
      marks[iidx] = sign (ilit) ? -1 : 1;
      clause.push_back (ilit);
    }
    for (auto ilit : clause)
      marks[index (ilit)] = 0;
    if (tautological)
      trace_vector ('d', imported);
    else {
      debug_vector (imported, "original[%zu] imported",
                    added_original_clauses);
      add_simplified_irredundant_clause ();
      if (proof && clause.size () != imported.size ())
        trace_temporary ('a'), trace_vector ('d', imported);
    }
    imported.clear ();
    clause.clear ();
    transition (READY);
  }
}

int Shared::solve () {
  if (state == ADDING)
    error ("can not call 'solve ()' without terminating clause with '0'");
  if (state != READY && state != SATISFIABLE)
    error ("invalid '%s' state calling 'solve ()'", decode_state (state));
  State res = solve_internally ();
  if (res != state)
    transition (res);
  debug ("'solve ()' call returns %d ('%s')", state, decode_state (state));
  terminated = false;
  return res;
}

int Shared::value (int elit) {
  if (!elit || elit == INT_MIN ||
      abs (elit) > MAXIMUM_EXTERNAL_VARIABLE_INDEX)
    error ("invalid literal in 'value (%d)'", elit);
  int eidx = abs (elit);
  unsigned iidx = eidx - 1;
  if (iidx >= variables)
    return 0;
  unsigned ilit = literal (iidx);
  if (elit < 0)
    ilit = negate (ilit);
  assert (ilit < literals);
  signed char value = values[ilit];
  return !value ? 0 : value < 0 ? -elit : elit;
}

#define size_read_buffer (1u << 20)

struct Parser {
  Shared &shared;
  FILE *file;
  size_t bytes = 0;
  size_t lineno = 0;
  size_t litlineno = 0;
  int &variables, last = '\n';
  size_t clauses = 0;
  struct {
    size_t pos = 0, end = 0;
    char chars[size_read_buffer];
  } buffer;
  Parser (Shared &s, FILE *f, int &m)
      : shared (s), file (f), variables (m) {}

  inline size_t fill_buffer () __attribute__ ((always_inline)) {
    buffer.pos = 0;
    buffer.end = fread (buffer.chars, 1, size_read_buffer, file);
    return buffer.end;
  }

  inline int read_buffer () __attribute__ ((always_inline)) {
    if (buffer.pos == buffer.end && !fill_buffer ())
      return EOF;
    bytes++;
    return buffer.chars[buffer.pos++];
  }

  inline int get_char () __attribute__ ((always_inline)) {
    if (last == '\n')
      lineno++;
    int res = read_buffer ();
    if (res == '\r') {
      res = read_buffer ();
      if (res != '\n')
        res = '\r';
    }
    return last = res;
  }

  inline bool fast_isdigit (int ch) __attribute__ ((always_inline)) {
    return '0' <= ch && ch <= '9';
  }

  inline bool get_int (int &res, int &ch) __attribute__ ((always_inline)) {
    litlineno = lineno;
    int sign = (ch == '-' ? -1 : 1);
    if (ch == '-' && (ch = get_char ()) == '0')
      return false;
    if (!fast_isdigit (ch))
      return false;
    res = ch - '0';
    while (fast_isdigit (ch = get_char ())) {
      if (INT_MAX / 10 < res)
        return false;
      res *= 10;
      int digit = ch - '0';
      if (INT_MAX - digit < res)
        return false;
      res += digit;
    }
    res *= sign;
    return true;
  }

  bool get_size (size_t &res, int &ch) {
    litlineno = lineno;
    if (!fast_isdigit (ch))
      return false;
    res = ch - '0';
    const size_t size_t_max = ~(size_t) 0;
    while (fast_isdigit (ch = get_char ())) {
      if (size_t_max / 10 < res)
        return false;
      res *= 10;
      int digit = ch - '0';
      if (size_t_max - digit < res)
        return false;
      res += digit;
    }
    return true;
  }

  const char *error (const char *fmt) {
    size_t size = sizeof shared.parse_error;
    snprintf (shared.parse_error, size, "line %zu: %s", lineno, fmt);
    return shared.parse_error;
  }

  const char *invalid_literal () {
    assert (litlineno), lineno = litlineno;
    return error ("invalid literal");
  }

  const char *end_of_file_in_comment () {
    return error ("end-of-file in comment");
  }

  const char *no_new_line_after_carriage_return () {
    return error ("carriage return without new line");
  }
  const char *parse ();
};

static bool string_to_unsigned (const char *str, unsigned &res) {
  const char *p = str;
  char ch = *p++;
  if (!isdigit (ch))
    return false;
  unsigned tmp = ch - '0';
  ch = *p++;
  if (ch) {
    if (!isdigit (ch))
      return false;
    if (!tmp)
      return false;
    tmp = 10 * tmp + (ch - '0');
    while ((ch = *p++)) {
      if (!isdigit (ch))
        return false;
      if (UINT_MAX / 10 < tmp)
        return false;
      tmp *= 10;
      unsigned digit = ch - '0';
      if (UINT_MAX - digit < tmp)
        return false;
      tmp += digit;
    }
  }
  res = tmp;
  return true;
}

#ifndef NOPTIONS

static bool match_simple_long_option (const char *arg, const char *name) {
  return arg[0] == '-' && arg[1] == '-' && !strcmp (arg + 2, name);
}

#endif

static bool match_long_option (const char *arg, const char *name,
                               unsigned &res) {
  if (arg[0] != '-' || arg[1] != '-')
    return false;
  if (arg[2] == 'n' && arg[3] == 'o' && arg[4] == '-' &&
      !strcmp (arg + 5, name)) {
    res = 0;
    return true;
  }
  const char *a = arg + 2;
  const char *n = name;
  while (*a && *a == *n)
    a++, n++;
  if (*n)
    return false;
  if (!*a) {
    res = 1;
    return true;
  }
  if (*a++ != '=')
    return false;
  if (!strcmp (a, "true")) {
    res = 1;
    return true;
  }
  if (!strcmp (a, "false")) {
    res = 0;
    return true;
  }
  return string_to_unsigned (a, res);
}

#ifndef NOPTIONS

Configuration *Shared::parse_long_configuration (const char *arg) {
  for (auto configuration : iterate_configurations ()) {
    if (match_simple_long_option (arg, configuration->name ())) {
      configuration->set ();
      return configuration;
    }
  }
  return 0;
}

bool Shared::parse_long_option (const char *arg) {
  for (auto &option : iterate_options ()) {
    unsigned res;
    if (!match_long_option (arg, option.name, res))
      continue;
    if (res < option.minimum)
      return false;
    if (res > option.maximum)
      return false;
    option = res;
    return true;
  }
  return false;
}

#endif

const char *Parser::parse () {
  int ch, comments = 0;
#if !defined(NDEBUG) && !defined(NOPTIONS)
  std::vector<char> buffer;
#endif
  while ((ch = get_char ()) == 'c') {
    comments++;
#if defined(NDEBUG) || defined(NOPTIONS)
    while ((ch = get_char ()) != '\n')
      if (ch == '\r')
        return no_new_line_after_carriage_return ();
      else if (ch == EOF)
        return end_of_file_in_comment ();
#else
    assert (buffer.empty ());
    while ((ch = get_char ()) != '\n') {
      if (ch == '\r')
        return no_new_line_after_carriage_return ();
      else if (ch == EOF)
        return end_of_file_in_comment ();
      if (ch != ' ' && ch != '\t')
        buffer.push_back (ch);
    }
    buffer.push_back (0);
    const char *arg = &buffer[0];
    shared.parse_long_option (arg);
    buffer.clear ();
#endif
  }
  if (ch == EOF) {
    if (comments) {
      assert (lineno > 1), lineno--;
      return error ("expected header after comments");
    } else
      return error ("empty input file");
  }
  if (ch != 'p')
    return error ("expected header or comment");
  for (const char *p = " cnf "; *p; p++)
    if (get_char () != *p)
      return error ("invalid header");
  if (!get_int (variables, ch = get_char ()) || variables < 0)
    return error ("expected number of variables");
  if (ch != ' ')
    return error ("expected space after variables");
  if (variables > MAXIMUM_EXTERNAL_VARIABLE_INDEX)
    return error ("maximum number of variables exceeded");
  if (!get_size (clauses, ch = get_char ()))
    return error ("expected number of clauses");
  if (ch != '\n')
    do {
      if (ch == EOF)
        return error ("end-of-file in header before end-of-line");
      else if (ch == '\r')
        return no_new_line_after_carriage_return ();
      else if (ch != ' ' && ch != '\t')
        return error ("invalid character after header before end-of-line");
      ch = get_char ();
    } while (ch != '\n');
#ifndef QUIET
  shared.message (0, "parsed header 'p cnf %d %zu'", variables, clauses);
#endif
  shared.resize (variables);
  size_t parsed = 0;
  int elit = 0;
  while ((ch = get_char ()) != EOF) {
    if (shared.terminated)
      return 0;
    if (ch == 'c') {
    SKIP_COMMENT:
      while ((ch = get_char ()) != '\n')
        if (ch == '\r')
          return no_new_line_after_carriage_return ();
        else if (ch == EOF)
          return end_of_file_in_comment ();
      continue;
    }
    if (ch == '\r')
      return no_new_line_after_carriage_return ();
    if (ch == ' ' || ch == '\t' || ch == '\n')
      continue;
    if (ch != '-' && !isdigit (ch))
      return error ("expected literal or comment");
    if (!get_int (elit, ch))
      return invalid_literal ();
    if (ch != 'c' && ch != ' ' && ch != '\t' && ch != '\n')
      return error ("expected white space after literal");
    if (!clauses)
      return error ("no clause expected");
    if (parsed == 1 && clauses == 1)
      return error ("more than one clause");
    if (parsed == clauses)
      return error ("too many clauses");
    assert (elit != INT_MIN);
    if (abs (elit) > variables)
      return invalid_literal ();
    parsed += !elit;
    shared.add (elit);
    if (ch == 'c')
      goto SKIP_COMMENT;
  }
  if (elit) {
    assert (litlineno), lineno = litlineno;
    return error ("zero missing after last literal");
  }
  if (parsed != clauses)
    return error ("clause missing");
#ifndef QUIET
  shared.message (0, "read and parsed %zu bytes (%.0f MB)", bytes,
                  bytes / (double) (1u << 20));
  if (parsed == 1)
    shared.message (0, "successfully parsed single clause");
  else
    shared.message (0, "successfully parsed all %zu clauses", parsed);
#endif
  return 0;
}

const char *Shared::parse (FILE *file, int &maximum_variable) {
  start ("parse");
  Parser *parser = new Parser (*this, file, maximum_variable);
  const char *res = parser->parse ();
  delete parser;
  stop ("parse");
  return res;
}

double Shared::process_time () {
  struct rusage u;
  double res;
  if (getrusage (RUSAGE_SELF, &u))
    return 0;
  res = u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  return res;
}

size_t Shared::maximum_resident_set_size (void) {
  struct rusage u;
  if (getrusage (RUSAGE_SELF, &u))
    return 0;
  return ((size_t) u.ru_maxrss) << 10;
}

#ifdef __APPLE__

extern "C" {
#include <mach/task.h>
mach_port_t mach_task_self (void);
};

size_t Shared::current_resident_set_size (void) {
  struct task_basic_info info;
  mach_msg_type_number_t count = TASK_BASIC_INFO_COUNT;
  if (KERN_SUCCESS != task_info (mach_task_self (), TASK_BASIC_INFO,
                                 (task_info_t) &info, &count))
    return 0;
  return info.resident_size;
}

#else

size_t Shared::current_resident_set_size (void) {
  char path[48];
  snprintf (path, sizeof path, "/proc/%d/statm", (int) getpid ());
  FILE *file = fopen (path, "r");
  if (!file)
    return 0;
  size_t dummy, rss;
  int scanned = fscanf (file, "%zu %zu", &dummy, &rss);
  fclose (file);
  return scanned == 2 ? rss * sysconf (_SC_PAGESIZE) : 0;
}

#endif

void Shared::print_resources () {
  double mb = maximum_resident_set_size () / (double) (1 << 20);
  printf ("c %-44s %13.2f MB\n", "maximum-resident-set-size:", mb);
  printf ("c %-44s %13.2f seconds\n", "process-time:", process_time ());
}

#ifndef QUIET

void Profiles::flush (Profiles::Profiling &profiling, double now) {
  assert (profiling.start <= now);
  double delta = now - profiling.start;
  profiling.start = now;
  profiles[profiling.idx].time += delta;
}

size_t Profiles::find (const char *name) {
  size_t size = profiles.size ();
  for (size_t i = 0; i != size; i++)
    if (!strcmp (profiles[i].name, name))
      return i;
  profiles.push_back (Profile (name));
  return size;
}

void Profiles::start (const char *name) {
  size_t idx = find (name);
  double now = Shared::process_time ();
  started.push_back (Profiling (idx, now, name));
}

void Profiles::stop (const char *name) {
  assert (!started.empty ());
  Profiling &profiling = started.back ();
  assert (profiling.idx < profiles.size ());
  assert (!strcmp (profiles[profiling.idx].name, name));
  double now = Shared::process_time ();
  flush (profiling, now);
  started.pop_back ();
  (void) name;
}

void Profiles::print (double now) {
  for (auto &profiling : started)
    flush (profiling, now);
  size_t size = profiles.size ();
  for (unsigned i = 0; i != size; i++)
    for (unsigned j = i + 1; j != size; j++)
      if (profiles[i].time < profiles[j].time ||
          (profiles[i].time == profiles[j].time &&
           strcmp (profiles[i].name, profiles[j].name) < 0))
        std::swap (profiles[i], profiles[j]);
  for (auto &profile : profiles)
    printf ("c %11.2f  %6.2f%%  %s\n", profile.time,
            now ? 100 * profile.time / now : 0, profile.name);
  fputs ("c -----------------------------------\n", stdout);
  printf ("c %11.2f  100.00%%  total\n", now);
}

#endif

#ifndef NDEBUG

// Checking the model on the original formula is extremely useful for
// testing and debugging.  This 'checker' aborts if an unsatisfied clause is
// found and prints the clause on '<stderr>' for debugging purposes.

void Shared::check_model () {
  debug ("checking model");
  bool satisfied = false;
  size_t start = 0, count = 1;
  for (size_t i = 0; i != original.size (); i++) {
    int lit = original[i];
    if (lit) {
      if (!satisfied && value (lit) == lit)
        satisfied = true;
    } else if (!satisfied) {
      fprintf (stderr, "%s: unsatisfied clause[%zu]:\n", one_word_name (),
               count);
      for (size_t j = start; j != i; j++)
        fprintf (stderr, "%d ", original[j]);
      fputs ("0\n", stderr);
      fflush (stderr);
      abort ();
      exit (1);
    } else {
      count++;
      satisfied = false;
      start = i + 1;
    }
  }
}

#endif

class Line {
  char buffer[80];
  size_t size = 0;

public:
  void put_char (char ch) {
    assert (size < sizeof buffer);
    buffer[size++] = ch;
  }

  void put_str (const char *s) {
    while (*s)
      put_char (*s++);
  }

  void put_int (int i) {
    char tmp[16];
    snprintf (tmp, sizeof tmp, " %d", i);
    size_t len = strlen (tmp);
    if (size + len > 76) {
      flush ();
      put_char ('v');
    }
    put_str (tmp);
  }

  void flush () {
    if (!size)
      return;
    for (size_t i = 0; i != size; i++)
      fputc (buffer[i], stdout);
    fputc ('\n', stdout);
    size = 0;
  }
};

// Printing the model in the format of the SAT competition, e.g.,
//
//   v -1 2 3 0
//
// Always prints a full assignments even if not all values are set.

void Shared::print_model (int maximum_variable, FILE *file) {
  Line line;
  line.put_char ('v');
  int idx = 0;
  assert (0 <= maximum_variable);
  while (idx++ != maximum_variable) {
    int tmp = value (idx);
    if (tmp)
      line.put_int (tmp == idx ? idx : -idx);
  }
  line.put_str (" 0");
  line.flush ();
}

static bool has_suffix (const char *a, const char *b) {
  size_t l = strlen (a), k = strlen (b);
  return l >= k && !strcmp (a + l - k, b);
}

static bool looks_like_compressed (const char *name) {
  if (has_suffix (name, ".gz"))
    return true;
  if (has_suffix (name, ".bz2"))
    return true;
  if (has_suffix (name, ".xz"))
    return true;
  return false;
}

static bool looks_like_cnf (const char *name) {
  if (has_suffix (name, ".cnf"))
    return true;
  if (has_suffix (name, ".dimacs"))
    return true;
  return false;
}

static bool exists_file (const char *path) {
  struct stat statbuf;
  return !stat (path, &statbuf);
}

static int bz2sig[] = { 0x42, 0x5A, 0x68, EOF };
static int gzsig[] = { 0x1F, 0x8B, EOF };
static int xzsig[] = { 0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00, 0x00, EOF };

static bool matches_signature (const char *path, const int *sig) {
  FILE *tmp = fopen (path, "r");
  if (!tmp)
    return false;
  bool res = true;
  for (const int *p = sig; res && (*p != EOF); p++)
    res = (getc (tmp) == *p);
  fclose (tmp);
  return res;
}

static FILE *read_from_pipe (const char *fmt, const char *path) {
  size_t size = strlen (fmt) + strlen (path);
  char *cmd = new char[size];
  snprintf (cmd, size, fmt, path);
  FILE *res = popen (cmd, "r");
  delete[] cmd;
  return res;
}

void Shared::print_banner (const char *prefix) {
  printf ("%s%s\n", prefix, one_line_description ());
  printf ("%sVersion %s", prefix, version ());
  if (*gitid ())
    printf (" %s", gitid ());
  putc ('\n', stdout);
  for (const char **p = copyright;
       p != copyright + sizeof (copyright) / sizeof (*copyright); p++)
    printf ("%sCopyright (c) %s\n", prefix, *p);
  printf ("%sCompiled with %s\n", prefix, build ());
  fflush (stdout);
}

int Shared::main (int argc, char **argv) {

  bool witness_printing_enabled = true;
#ifndef NOPTIONS
  Configuration *configuration = 0;
#endif
  const char *proof_name = 0, *dimacs_name = 0;
  int close_dimacs_file = 2, close_proof_file = 0;
  FILE *dimacs_file = 0, *proof_file = 0;

  for (int i = 1; i != argc; i++) {
    const char *arg = argv[i];
    if (!strcmp (arg, "-a") || !strcmp (arg, "--ascii")) {
      ascii = true;
#ifndef NOPTIONS
    } else if (!strcmp (arg, "-A") || !strcmp (arg, "--all-options")) {
      print_all_option_usage ();
      return 0;
#endif
    } else if (!strcmp (arg, "-B") || !strcmp (arg, "--banner") ||
               !strcmp (arg, "--build")) {
      print_banner ("");
      return 0;
#ifndef NOPTIONS
    } else if (!strcmp (arg, "-E") || !strcmp (arg, "--embedded-options")) {
      print_all_embedded_options ();
      return 0;
#endif
    } else if (!strcmp (arg, "-L") || !strcmp (arg, "--list-options")) {
      print_all_option_completions ();
      return 0;
#ifndef NOPTIONS
    } else if (!strcmp (arg, "-R") || !strcmp (arg, "--option-ranges")) {
      print_all_option_ranges ();
      return 0;
#endif
    } else if (!strcmp (arg, "-S") || !strcmp (arg, "--print-sizes")) {
      print_sizes ("");
      return 0;
    } else if (!strcmp (arg, "-h") || !strcmp (arg, "--help")) {
      printf (basic_usage_header, one_word_name ());
      fputc ('\n', stdout);
      fputs (basic_usage_body, stdout);
      fputc ('\n', stdout);
      fputs (basic_usage_footer, stdout);
      return 0;
    } else if (!strcmp (arg, "-l") || !strcmp (arg, "--logging"))
#ifdef LOGGING
      verbosity = INT_MAX;
#else
      error ("compiled without logging (use './configure --logging')");
#endif
    else if (!strcmp (arg, "-n") || !strcmp (arg, "--no-witness"))
      witness_printing_enabled = false;
#ifndef QUIET
    else if (!strcmp (arg, "-q") || !strcmp (arg, "--quiet"))
      verbosity = -1;
#endif
    else if (!strcmp (arg, "-s") || !strcmp (arg, "--statistics"))
      full_statistics = true;
#ifndef QUIET
    else if (!strcmp (arg, "-v") || !strcmp (arg, "--verbose"))
      verbosity += (verbosity >= 0 && verbosity != INT_MAX);
#endif
    else if (!strcmp (arg, "-V") || !strcmp (arg, "--version")) {
      printf ("%s\n", version ());
      return 0;
#ifndef QUIET
    } else if (!strcmp (arg, "-p")) {
      if (++i == argc)
        error ("argument to '-p' missing");
      arg = argv[i];
      if (!string_to_unsigned (arg, profile))
        error ("expected positive number in '-p %s'", arg);
    } else if (match_long_option (arg, "profile", profile)) {
#endif
    } else if (!strcmp (arg, "-c")) {
      if (++i == argc)
        error ("argument to '-c' missing");
      arg = argv[i];
      if (!string_to_unsigned (arg, conflict_limit))
        error ("expected non-negative number in '-c %s'", arg);
      conflict_limit_set = true;
    } else if (!strcmp (arg, "--no-conflicts"))
      conflict_limit = 0, conflict_limit_set = true;
    else if (match_long_option (arg, "conflicts", conflict_limit))
      conflict_limit_set = true;
    else if (!strcmp (arg, "-t")) {
      if (++i == argc)
        error ("argument to '-t' missing");
      arg = argv[i];
      if (!string_to_unsigned (arg, time_limit) || !time_limit)
        error ("expected positive number in '-t %s'", arg);
    } else if (!strcmp (arg, "--no-time"))
      error ("invalid zero seconds time limit '%s'", arg);
    else if (match_long_option (arg, "time", time_limit)) {
      if (!time_limit)
        error ("expected positive number in '%s'", arg);
#ifndef NOPTIONS
    } else if (arg[0] == '-' && arg[1] == '-') {
      Configuration *other_configuration = parse_long_configuration (arg);
      if (other_configuration) {
        if (other_configuration == configuration)
          error ("configuration '%s' specified twice", arg);
        if (configuration)
          error ("can not use two configurations '--%s' and '%s'",
                 configuration->name (), arg);
        configuration = other_configuration;
      } else if (!parse_long_option (arg))
        invalid_option (arg);
#endif
    } else if (arg[0] == '-' && arg[1])
      invalid_option (arg);
    else if (proof_name)
      error ("too many arguments '%s', '%s' and '%s' (try '-h')",
             dimacs_name, proof_name, arg);
    else if (dimacs_name)
      proof_name = arg;
    else
      dimacs_name = arg;
  }

  if (proof_name) {
    if (!strcmp (proof_name, "-")) {
      proof_name = "<stdout>";
      proof_file = stdout;
      ascii = true;
    } else if (looks_like_compressed (proof_name))
      error ("can not write compressed proof file '%s'", proof_name);
    else if (looks_like_cnf (proof_name))
      error ("not over-writing '%s' with proof which looks like a CNF file",
             proof_name);
    else if (!(proof_file = fopen (proof_name, "w")))
      error ("can not write to proof file '%s'", proof_name);
    else
      close_proof_file = 1;
  }

  if (!dimacs_name || !strcmp (dimacs_name, "-")) {
    dimacs_name = "<stdin>";
    close_dimacs_file = 0;
    dimacs_file = stdin;
  } else if (!exists_file (dimacs_name))
    error ("could not find '%s' file", dimacs_name);
  else if (has_suffix (dimacs_name, ".gz") &&
           matches_signature (dimacs_name, gzsig))
    dimacs_file = read_from_pipe ("gunzip -c -d %s", dimacs_name);
  else if (has_suffix (dimacs_name, ".bz2") &&
           matches_signature (dimacs_name, bz2sig))
    dimacs_file = read_from_pipe ("bunzip2 -c -d %s", dimacs_name);
  else if (has_suffix (dimacs_name, ".xz") &&
           matches_signature (dimacs_name, xzsig))
    dimacs_file = read_from_pipe ("xz -c -d %s", dimacs_name);
  else if (!(dimacs_file = fopen (dimacs_name, "r")))
    error ("could not open and read '%s'", dimacs_name);
  else
    close_dimacs_file = 1;

#ifndef QUIET
  if (verbosity >= 0) {
    print_banner ("c ");
    printf ("c\n");
    fflush (stdout);
  }
#endif

  {
    size_t printed_option_lines = 0;

#ifndef NOPTIONS
    if (configuration) {
      message (0,
               "using configuration '--%s' "
               "(use '--all-options' for more information)",
               configuration->name ());
      printed_option_lines++;
    }
#endif

    if (proof_file) {
      set_proof (proof_file);
      message (0, "enabled proof tracing to '%s'", proof_name);
      printed_option_lines++;
    }

    catch_signals ();
    if (time_limit) {
      message (0, "time limit set to %u seconds", time_limit);
      set_alarm (time_limit);
      printed_option_lines++;
    }

    if (conflict_limit_set) {
      message (0, "conflict limit set to %u conflicts", conflict_limit);
      printed_option_lines++;
    }
#ifndef QUIET
    if (verbosity >= 0 && printed_option_lines)
      fputs ("c\n", stdout);
#else
    (void) printed_option_lines;
#endif
  }

  message (0, "reading from '%s'", dimacs_name);
  int maximum_variable;
  const char *failed = parse (dimacs_file, maximum_variable);
  if (close_dimacs_file == 1)
    fclose (dimacs_file);
  if (close_dimacs_file == 2)
    pclose (dimacs_file);
  if (failed) {
    fprintf (stderr, "%s: parse error in '%s': %s\n", one_word_name (),
             dimacs_name, failed);
    fflush (stderr);
    exit (1);
  }

  int res = UNKNOWN;
  if (terminated)
    message (0, "time limit hit during parsing");
  else {
#if !defined(QUIET) && !defined(NOPTIONS)
    if (verbosity >= 0)
      print_all_option_values ();
#endif
    res = solve ();
  }

  if (proof_file) {
    flush_proof ();
    if (close_proof_file)
      fclose (proof_file);
#ifndef QUIET
    if (verbosity >= 0)
      fputs ("c\n", stdout);
#endif
  }

  if (res == 10) {
#ifndef NDEBUG
    check_model ();
#endif
    printf ("s SATISFIABLE\n");
    fflush (stdout);
    if (witness_printing_enabled)
      print_model (maximum_variable, stdout);
  } else if (res == 20)
    printf ("s UNSATISFIABLE\n");
  fflush (stdout);

  if (time_limit)
    reset_alarm ();
  reset_signals ();
#ifndef QUIET
  if (verbosity >= 0) {
    fputs ("c\n", stdout);
    print_full_statistics ();
    printf ("c\nc exit code %d\n", res);
    fflush (stdout);
  }
#else
  if (full_statistics) {
    print_full_statistics ();
    fflush (stdout);
  }
#endif

  return res;
}

void Shared::print_full_statistics () {
  double now = process_time ();
#ifndef QUIET
  if (profile)
    profiles.print (now), fputs ("c\n", stdout);
#endif
  print_statistics (now);
#ifndef QUIET
  fputs ("c\n", stdout);
#endif
  print_resources ();
}

// Print statistics even if killed after receiving a signal.  This is not
// completely safe as we lazily use standard non-reentrant functions such as
// 'printf'.  Occasionally these might fail (with the result that the
// signal handling code hangs and/or does not produce any output).  We also
// need access to a (single) global solver instances, which is register with
// the signal handlers through 'catch_signals' below.

volatile Shared *Shared::singleton;

#define SIGNALS \
  SIGNAL (SIGABRT) \
  SIGNAL (SIGBUS) \
  SIGNAL (SIGILL) \
  SIGNAL (SIGINT) \
  SIGNAL (SIGSEGV) \
  SIGNAL (SIGTERM)

#define SIGNAL(SIG) static void (*saved_##SIG##_handler) (int);
SIGNALS

void Shared::reset_signals () {
  assert (singleton == this);
  singleton = 0;
#undef SIGNAL
#define SIGNAL(SIG) signal (SIG, saved_##SIG##_handler);
  SIGNALS
}

void Shared::catch_signal (int sig) {
  Shared *solver = (Shared *) singleton;
  solver->reset_signals ();
#ifndef QUIET
  if (solver && solver->verbosity >= 0) {
    printf ("c\nc caught signal %d\nc\n", sig);
    solver->print_full_statistics ();
    printf ("c\nc raising signal %d\n", sig);
    fflush (stdout);
  }
#else
  if (solver && solver->full_statistics) {
    solver->print_full_statistics ();
  }
#endif
  raise (sig);
}

static void (*saved_SIGALRM_handler) (int);

void Shared::reset_alarm () {
  assert (singleton == this);
  signal (SIGALRM, saved_SIGALRM_handler);
}

void Shared::catch_alarm (int sig) {
  assert (sig == SIGALRM);
  Shared *solver = (Shared *) singleton;
  solver->reset_alarm ();
  solver->terminate_solving ();
  (void) sig;
  return;
}

void Shared::catch_signals () {
  assert (!singleton);
  singleton = this;
#undef SIGNAL
#define SIGNAL(SIG) saved_##SIG##_handler = signal (SIG, catch_signal);
  SIGNALS
}

void Shared::set_alarm (unsigned seconds) {
  assert (singleton);
  assert (this == singleton);
  saved_SIGALRM_handler = signal (SIGALRM, catch_alarm);
  alarm (seconds);
}

}; // namespace TabularaSAT
