#ifndef _Shared_hpp_included
#define _Shared_hpp_included

/*------------------------------------------------------------------------*/

#include <cassert>
#include <climits>
#include <cstdio>
#include <cstring>
#include <vector>

/*------------------------------------------------------------------------*/

#define VOID ((void) 0)

/*------------------------------------------------------------------------*/

namespace TabularaSAT {

// The 'Shared' class below contains code shared among all solver variants
// without any internal data-structures and includes a generic interface.

/*------------------------------------------------------------------------*/

// These constants are used as internal state encoding as well as result
// values for the 'solve' function of a solver.

enum State {
  READY = 0,          // Initial solver state.
  ADDING = 1,         // Adding clause.
  UNKNOWN = 0,        // Unknown solver result.
  SATISFIABLE = 10,   // All clauses are satisfied.
  UNSATISFIABLE = 20, // Determined unsatisfiability.
};

// For Invalid literal and invalid decision level constants.  We assume that
// the compiler adheres to two-complement semantics for 'unsigned' with
// precise underflow semantics (as 'C/C++' guarantees for 'unsigned').

// Even though these constant are all the same we introduce separate
// symbolic compiler constants for more precise code and documentation.

const unsigned INVALID_LEVEL = ((unsigned) 0) - 1;
const unsigned INVALID_LITERAL = ((unsigned) 0) - 1;
const unsigned INVALID_VARIABLE = ((unsigned) 0) - 1;

// We reserve one bit for bit-stuffing and thus in principle only support
// 2^30 variables (instead of 2^31).  We further have to reserve one
// additional value above the literal range to distinguish tagged literals
// from the decision reason constant 'UINT_MAX = 0xffffffff' and the unit
// reason constant '0xfffffffe' one less (for 32-bit clause references).
//
// Therefore we require
//
//     (((MAXIMUM_INTERNAL_LITERAL << 1)) | 1) < 0xfffffffe
//
// which gives (well the '... | 1' tagging does not really matter here)
//
//     MAXIMUM_INTERNAL_LITERAL  < 0x7fffffff
//
// thus
//
//     MAXIMUM_INTERNAL_LITERAL == 0x7ffffffd
//
// as the maximum literal is negated and thus odd.
//
// Further note that for 64-bit references we would not really need to
// reserve space for those two virtual reasons but in order to keep the
// maximum supported variable number the same for 32-bit and 64-bit variants
// we remove those two values from the literal range.  The bit needed for
// bit-stuffing reduces the available range by a factor of two which is much
// more severe (but still gives more than a billion variables).

const unsigned MAXIMUM_INTERNAL_LITERAL = (1u << 31) - 3; // 0x7ffffffd

// The following two constants are a consequence of the argument above.

const unsigned MAXIMUM_INTERNAL_VARIABLE = (1u << 30) - 2;
const int MAXIMUM_EXTERNAL_VARIABLE_INDEX = (1 << 30) - 1;

// The individual solver variants only have to implement the following
// abstract virtual member functions. We factored out this interface for
// documentation purposes only.  The parent class of an actual solver
// variant should instead be the 'Shared' class defined below.

struct Interface {
  virtual ~Interface () {}

  // Short name and description of the solver variant.
  //
  virtual const char *one_word_name () = 0;
  virtual const char *one_line_description () = 0;

  // Notify the internal solver variant to adjust its own internal
  // data-structures to the new increased number of 'variables' and
  // 'literals' triggered by 'import_literal'.
  //
  virtual void adjust_to_increased_variables () = 0;

  // Add the temporary simplified clause as irredundant clause. This
  // function is virtual to keep the code for importing literals resizing
  // data-structures as well as simplifying the clause code shared among all
  // solver variants even though 'resizing' needs the 'resize' call back.
  //
  virtual void add_simplified_irredundant_clause () = 0;

  // Solve and return '10=SATISFIED' and '20=UNSATISFIED' as above.
  //
  virtual State solve_internally () = 0;

  // At the end of solving 'main' wants to print solver specific statistics.
  //
  virtual void print_statistics (double current_process_time) {}

#ifndef NOPTIONS

  // Allows to traverse over all options and configurations;
  //
  virtual struct Options &iterate_options () = 0;
  virtual struct Configurations &iterate_configurations () = 0;

#endif

  // Print sizes of internal data-structures.

  virtual void print_sizes (const char *prefix) {}
};

/*------------------------------------------------------------------------*/

// This 'Range' class allows to easily iterate over all variables and all
// literals with the following range-based for-loop idioms:
//
//    for (auto idx : variables)
//      do_something_with (idx);
//
//    for (auto lit : literals) {
//      do_something_with (lit);
//
// We only use it for the 'variables' and 'literals' fields.

class Range {

  class iterator {
    unsigned element;

  public:
    iterator (unsigned i) : element (i) {}
    void operator++ () {
      assert (element != UINT_MAX);
      element++;
    }
    const unsigned &operator* () const { return element; }
    friend bool operator!= (const iterator &a, const iterator &b) {
      return a.element != b.element;
    }
  };

  unsigned bound = 0; // The range is from '0' to 'bound-1'.

public:
  void resize (unsigned new_bound) {
    assert (bound <= new_bound);
    bound = new_bound;
  }

  operator unsigned () const { return bound; }

  iterator end () const { return iterator (bound); }
  iterator begin () const { return iterator (0); }
};

/*------------------------------------------------------------------------*/

#ifndef QUIET

// Code for internal performance profiling (used as 'Solver::profiles').

class Profiles {

  struct Profile {
    const char *name;
    double time;

    Profile (const char *n) : name (n), time (0) {}
  };

  struct Profiling {
    size_t idx;
    double start;
    const char *name;

    Profiling (size_t i, double s, const char *n)
        : idx (i), start (s), name (n) {}
  };

  std::vector<Profile> profiles;
  std::vector<Profiling> started;

  void flush (Profiling &, double);
  size_t find (const char *);

public:
  void start (const char *);
  void stop (const char *);

  const char *back () const {
    if (started.empty ())
      return 0;
    return started.back ().name;
  }

  void print (double now);
};

#endif

/*------------------------------------------------------------------------*/

#ifndef NOPTIONS

// Configurations set several options used to simulate certain solver
// variants (in particular a 'vanilla' variant to mimic MiniSAT).

struct Configuration {
  virtual const char *name () = 0;
  virtual const char *description () = 0;
  virtual void set () = 0;
  virtual ~Configuration () {}
};

// Used to provide an iterator over all solver configurations.

struct Configurations {
  virtual Configuration **begin () = 0;
  virtual Configuration **end () = 0;
};

#endif

/*------------------------------------------------------------------------*/

#ifndef NOPTIONS

// Options are solver specific and kept abstract here (in 'shared.[ch]pp').

struct Option {

  unsigned value, initial, minimum, maximum;
  const char *name, *description;

  Option (const char *n, unsigned i, unsigned l, unsigned h, const char *d)
      : value (i), initial (i), minimum (l), maximum (h), name (n),
        description (d) {
    assert (minimum <= initial);
    assert (initial <= maximum);
    assert (maximum <= (unsigned) INT_MAX);
  }

  operator unsigned () { return value; }

  void operator= (unsigned new_value) {
    if (new_value < minimum)
      new_value = minimum;
    if (new_value > maximum)
      new_value = maximum;
    value = new_value;
  }

  struct iterator {
    virtual Option *begin () = 0;
    virtual Option *end () = 0;
  };
};

// Used to provide an iterator over all solver options.

struct Options {
  virtual Option *begin () = 0;
  virtual Option *end () = 0;
};

#endif

/*------------------------------------------------------------------------*/

#ifdef LOGGING

// Fixed non-resizable buffers for logging literals and variables.

struct Buffer {
  char chars[128];
  size_t size = 0;
};

struct Buffers {
  Buffer buffers[4];
  size_t current = 0;

  Buffer &next () {
    Buffer &res = buffers[current++];
    if (current == sizeof buffers / sizeof *buffers)
      current = 0;
    return res;
  }
};

#endif

/*------------------------------------------------------------------------*/

struct Trace;

// This is the actual parent class for all solver variants.

class Shared : Interface {
public:
  static unsigned sign (unsigned lit) { return lit & 1; }
  static unsigned index (unsigned lit) { return lit / 2; }
  static unsigned negate (unsigned lit) { return lit ^ 1; }
  static unsigned literal (unsigned idx) { return 2 * idx; }

protected:             // Code used by each solver variant.
  State state = READY; // Actually of type 'State'.
  bool terminated = false;

  static const char *decode_state (int);
  void transition (State new_state);

  Range variables; // Number of variables.
  Range literals;  // Number of literals (2*variables).

  // The data structures below are allocated until this capacity and only
  // need to be resized if the 'variables' exceeds it.

  size_t capacity = 0;     // Size of 'levels', 'marks', and 'values'.
  unsigned *levels = 0;    // Maps variables to assignment level.
  signed char *marks = 0;  // Mark literals/variables for parsing.
  signed char *values = 0; // Maps literals to values (-1,0,1).

  // The current decision level is used by the shared debugging functions
  // below to include it in the debugging messages.  Similarly we use the
  // 'levels' array for debugging literals too.

  unsigned level = 0;

  signed char fixed (int lit) {
    signed char res = values[lit];
    if (!res)
      return 0;
    if (levels[index (lit)])
      return 0;
    return res;
  }

#ifndef NDEBUG
  std::vector<int> original; // Original zero separated clause literals.
#endif

  void resize (unsigned new_variables);

  // Import added literal in 'add (int)' in the range '[1,-1,2,-2,...]' to
  // internal literals in the range '[0,1,2,3,...]' and if necessary resize
  // data-structures of 'Shared' and through 'resize' of the solver variant.

  unsigned import_literal (int external_literal);

  // Maps internal literals '[0,1,2,3,...]' to external '[1,-1,2,-2,...]'.
  //
  int export_literal (unsigned internal_literal) const;

  std::vector<unsigned> clause;   // Simplified clause.
  std::vector<unsigned> imported; // Unsimplified clause.

  bool simplify_clause (); // Returns 'false' if tautological.

  size_t added_original_clauses = 0; // Number of added original clauses.

#ifndef QUIET
  int verbosity = 0;    // Negative if 'quiet'.
  unsigned profile = 2; // Profiling level.
#endif
  bool full_statistics = false;    // Even if '!verbosity'.
  bool conflict_limit_set = false; // Assume 'conflict_limit' set.
  unsigned conflict_limit = 0;     // The actual conflict limit.
  unsigned time_limit = 0;         // Zero if not set.

  // Proof tracing part.

  bool ascii = false;      // Use ASCII version of DRAT proof format.
  struct Trace *proof = 0; // Trace to a file if non-zero.
  friend struct Trace;

  inline void trace_only_literals (size_t size, const unsigned *, unsigned)
      __attribute__ ((always_inline))

      ;

  void trace_literals (char type, size_t size, const unsigned *literals,
                       unsigned except = INVALID_LITERAL);

  void trace_vector (char type, const std::vector<unsigned> &v) {
    trace_literals (type, v.size (), &v[0]);
  }

  void trace_empty (char type) { trace_literals (type, 0, (unsigned *) 0); }

  void trace_unit (char type, unsigned lit) {
    trace_literals (type, 1, &lit);
  }

  void trace_binary (char type, unsigned lit, unsigned other) {
    unsigned literals[2] = {lit, other};
    trace_literals (type, 2, literals);
  }

  void trace_temporary (char type) { trace_vector (type, clause); }

#ifndef QUIET

  // Internal profiling.

  Profiles profiles;

  void start (const char *name, unsigned level = 1) {
    if (verbosity >= 0 && profile >= level)
      profiles.start (name);
  }

  void stop (const char *name) {
    const char *back = profiles.back ();
    if (back && !strcmp (back, name))
      profiles.stop (name);
  }

#else
#define start(...) VOID
#define stop(...) VOID
#endif

#ifndef NOPTIONS

  bool parse_long_option (const char *arg);
  Configuration *parse_long_configuration (const char *arg);

#endif

#ifdef LOGGING

  Buffers buffers;

  bool logging () const { return verbosity == INT_MAX; }

  void debug_prefix (); // Start a debugging line.
  void debug_suffix (); // Finishes debugging line with new-line etc.

  // Generic 'printf' style debugging/logging function.  The code is only
  // included if 'LOGGING' is defined (with './configure -l' - the default
  // for './configure -g') and still needs to be enabled at run-time.

  // The '__attribute__ ...' checks correct format strings at compile-time.

  void debug (const char *fmt, ...) __attribute__ ((format (printf, 2, 3)));

  // The following three functions give variable and literal strings with
  // their assigned value and decision level (if assigned) and uses
  // 'values', 'levels' and 'buffer'.

  const char *debug_literal (unsigned lit);
  const char *debug_variable (unsigned idx);

  void debug_literals (size_t, unsigned *, const char *fmt, ...)
      __attribute__ ((format (printf, 4, 5)));

  void debug_vector (std::vector<unsigned> &, const char *fmt, ...)
      __attribute__ ((format (printf, 3, 4)));
#else

  // The debugging code is not compiled in if 'LOGGING' is undefined.  This
  // allows to make sure that it is completely ignored if not enabled.

#define debug(...) VOID
#define debug_literals(...) VOID
#define debug_vector(...) VOID

#endif

#ifndef QUIET

  // Generic messages which print some header and finish the message with a
  // new line.  The '__attribute__ ...' parts make it easy to find incorrect
  // format strings at compile time, not matching the given arguments.  The
  // 'level' for message specifies the requested 'verbosity' level.

  void message (int level, const char *fmt, ...)
      __attribute__ ((format (printf, 3, 4)));

#define SIZE_PRETTY_MEMORY_BUFFER 64
  typedef char pretty_memory_buffer[SIZE_PRETTY_MEMORY_BUFFER];
  const char *pretty_memory (pretty_memory_buffer buffer, size_t bytes);

#else

#define message(...) VOID

#endif

  // Produces an error message and exits the process.

  void error (const char *fmt, ...) __attribute__ ((format (printf, 2, 3)));

  // Prints and error message about this invalid option and exists.

  void invalid_option (const char *arg);

  // Produces a fatal error message and aborts the process raising a
  // 'SIGABRT' signal which in turn can be caught by signal handlers
  // allowing to produce additional information (such as statistics).

  void fatal (const char *fmt, ...) __attribute__ ((format (printf, 2, 3)));

  char parse_error[128];
  friend struct Parser;

public: // Common API for all solvers.
  Shared ();
  ~Shared ();

  void set_verbosity (int verbosity); // Change verbosity (-1=quiet).
  void force_ascii_proof_format ();   // Enforce ASCII DRAT format.
  void set_proof (FILE *file);        // Start proof tracing.
  void flush_proof ();                // Finish proof tracing.

  static const char *version (); // Global version number.
  static const char *gitid ();   // Return Git-ID.
  static const char *build ();   // Return compilation flags.

  void add (int); // Import and add literal.
                  // Clauses are terminated by '0'.

  int solve (); // External 'solve' function,
                // returns 10=SATISFIABLE or
                // 20=UNSATISFIABLE through
                // calling 'solve_internally'.

  // There is a way to asynchronously stop solving, which for instance is
  // used to flag time out to the solver if the time limit is reached.
  //
  virtual void terminate_solving () { terminated = true; }

  int value (int lit); // Returns 'lit' if assigned true,
                       // and '-lit' if assigned false,
                       // and '0' if unassigned.
#ifndef NDEBUG
  void check_model (); // Check model to satisfy clauses.
#endif

  // Print satisfying assignment in competition format with 'v ...' lines.

  void print_model (int max_var, FILE *);

  // Parse DIMACS file and add its clauses. Returns parse error.

  const char *parse (FILE *, int &max_var);

  // Print the banner even if compiled with 'QUIET' set.

  void print_banner (const char *prefix);
  void print_all_option_completions ();

#ifndef NOPTIONS

  // Print option lists of solver in various formats.

  void print_all_option_ranges ();
  void print_all_option_usage ();
  void print_all_option_values ();
  void print_all_embedded_options ();

  // Method to print closest matching option in case of mispelling.

  std::vector<char *> shared_options;
  void find_matching_options (const char *, std::vector<const char *> &);

#endif

  // Print statistics and resource usage.

  void print_resources ();
  void print_full_statistics ();

  // Resource usage gathering functions.

  static double process_time ();
  static size_t maximum_resident_set_size ();
  static size_t current_resident_set_size ();

  // Main of stand-alone solver.

  int main (int argc, char **argv);

private:
  void catch_signals ();
  void reset_signals ();

  void set_alarm (unsigned seconds);
  void reset_alarm ();

  static volatile Shared *singleton;  // To catch signals.
  static void catch_signal (int sig); // Actual signal catcher.
  static void catch_alarm (int sig);  // Actual alarm signal catcher.
};

/*------------------------------------------------------------------------*/

}; // namespace TabularaSAT

/*------------------------------------------------------------------------*/

#endif
