#include "references.hpp"
#include "schedule.hpp"
#include "solver.hpp"
#include "ticks.hpp"

#include <cinttypes>

namespace TabularaSAT {

bool Solver::failed_literal_probing () {
  if (!OPTION (inprocessing))
    return false;
  if (!OPTION (failed_literal_probing))
    return false;
  if (GET_COUNT (conflicts) < limits.probe)
    return false;
  return true;
}

bool Solver::propagates (unsigned lit) {
  assert (probing);
  assert (!values[lit]);
  unsigned not_lit = negate (lit);
  auto watches = references (not_lit);
  auto begin_watches = watches->begin ();
  auto end_watches = watches->end ();
  auto p = begin_watches;
  bool res = false;
  while (p != end_watches) {
    Reference watch = *p++;
    if (!tagged (watch)) {
      p++;
      continue;
    }
    unsigned blocking = untag (watch);
    signed char blocking_value = values[blocking];
    assert (blocking_value >= 0);
    if (blocking_value)
      continue;
    debug_binary (not_lit, blocking, "%s propagates", debug_literal (lit));
    res = true;
    break;
  }
  uint64_t ticks = ticks_spanning (p - begin_watches);
  if (!res) {
    p = begin_watches;
    while (p != end_watches) {
      Reference watch = *p++;
      if (tagged (watch))
        continue;
      Reference ref = *p++;
      unsigned blocking = untag (watch);
      signed char blocking_value = values[blocking];
      if (blocking_value > 0)
        continue;
      Clause *c = dereference_clause (ref);
      assert (!c->garbage);
      ticks++;
      unsigned *literals = c->literals;
      unsigned satisfied = 0, unassigned = 0;
      unsigned *end_literals = literals + c->size;
      for (unsigned *l = literals; l != end_literals; l++) {
        unsigned other = *l;
        if (other == not_lit)
          continue;
        signed char other_value = values[other];
        if (other_value > 0) {
          satisfied++;
          break;
        }
        if (other_value < 0)
          continue;
        if (unassigned++)
          break;
      }
      if (satisfied)
        continue;
      assert (unassigned >= 1);
      if (unassigned == 1) {
        debug_clause (c, "%s propagates", debug_literal (lit));
        res = true;
        break;
      }
    }
  }
  if (!res) {
    ADD_COUNT (failed_literal_ticks, ticks);
    debug ("%s does not propagate", debug_literal (lit));
  }
  return res;
}

bool Solver::probe_failed_literals () {

  assert (OPTION (failed_literal_probing));

  assert (OPTION (preprocessing) || OPTION (inprocessing));

  if (level && !backtrack_to_root_level_and_propagate ())
    return false;

  assert (!level);
  assert (watching);
  assert (!inconsistent);
  assert (!probing);

  probing = true;
  start_simplifier ("probe");
  FailedLiteralScheduler scheduler (this);
  scheduler.increment_counter ();

  uint64_t limit;

  if (preprocessing) {
    unsigned minimum_effort = OPTION (minimum_effort);
    unsigned factor = OPTION (failed_literal_preprocessing);
    limit = factor * minimum_effort;
    message (2, "failed literal effort %" PRIu64 " = %u * %u ticks", limit,
             factor, minimum_effort);
  } else
    limit = scheduler.get_ticks_limit ();

  limit += GET_COUNT (failed_literal_ticks);

  unsigned start_probe = next_failed_literal_probe;

  message (2, "begin failed literal at %.0f%% (literal %u)",
           percent (start_probe, literals), start_probe);

  assert (start_probe < literals);

  bool completed = false;
  unsigned failed = 0;
  size_t probed = 0;

  uint64_t *stamps = (uint64_t *) calloc (literals, sizeof *stamps);
  uint64_t stamp = 1;

  while (!completed && !terminated &&
         GET_COUNT (failed_literal_ticks) < limit) {

    unsigned probe = next_failed_literal_probe++;
    if (next_failed_literal_probe == literals) {
#ifndef QUIET
      statistics.failed_literal_completed++;
      message (1, "failed literal probing completed %" PRIu64 " times",
               statistics.failed_literal_completed);
#endif
      next_failed_literal_probe = 0;
    }

    if (active (probe) && stamps[probe] < stamp && propagates (probe)) {
      probed++;
      debug ("probing %s", debug_literal (probe));
      size_t start = trail.size ();
      assert (!level);
      decide_literal (probe);
      INC_STATS (failed_literal_decisions);
      if (failed_literal_propagate ()) {
        for (size_t i = start; i != trail.size (); i++)
          stamps[trail[i]] = stamp;
        backtrack (0);
      } else {
        debug ("failed literal %s", debug_literal (probe));
        INC_STATS (failed_literals);
        failed++, stamp++;
        backtrack (0);
        unsigned not_probe = negate (probe);
        trace_unit ('a', not_probe);
        assign (not_probe, UNIT_REASON);
        if (!failed_literal_propagate ()) {
          debug ("inconsistent propagation of negated failed literal");
          inconsistent = true;
          trace_empty ('a');
        }
        start_probe = probe & ~1;
      }
    }

    if (next_failed_literal_probe == start_probe)
      completed = true;
  }

  free (stamps);

#ifndef QUIET
  message (1, "probed %zu literals during %s failed literal probing",
           probed, completed ? "complete" : "incomplete");
  if (failed || verbosity > 0)
    message (1, "found %u failed literals during failed literal probing",
             failed);
#else
  (void) probed;
#endif

  scheduler.set_next_conflict_limit ();
  bool report = failed;
#ifndef QUIET
  if (verbosity > 0)
    report = true;
#endif
  if (report)
    report_progress ('f');
  stop_simplifier ("probe");
  probing = false;

  return failed;
}

}; // namespace TabularaSAT
