#include "eliminate.hpp"
#include "references.hpp"
#include "solver.hpp"
#include "ticks.hpp"

namespace TabularaSAT {

#define RESOLVENT_SIZE_EXCEEDED (~(size_t) 0)

template <class R>
void Solver::resolve (unsigned pivot, R *pos_occs, R *neg_occs,
                      size_t &resolved, const size_t resolved_limit,
                      std::vector<unsigned> &resolvents) {

  assert (resolved_limit < RESOLVENT_SIZE_EXCEEDED);
  assert (resolved_limit + 1 < RESOLVENT_SIZE_EXCEEDED);

  unsigned not_pivot = negate (pivot);

  auto pos_begin = pos_occs->begin ();
  auto neg_begin = neg_occs->begin ();

  auto pos_end = pos_occs->end ();
  auto neg_end = neg_occs->end ();

  const size_t size_limit = OPTION (resolvent_size);
  assert (size_limit < RESOLVENT_SIZE_EXCEEDED);

  for (int binary = 1; binary >= 0; binary--) {
    auto pos_it = pos_begin;
    while (pos_it != pos_end) {

      if (inconsistent)
        break;
      if (resolved > resolved_limit)
        break;

      Reference pos_ref = *pos_it++;
      if (binary && !tagged (pos_ref))
        continue;

      auto neg_it = neg_begin;
      while (neg_it != neg_end) {

        Reference neg_ref = *neg_it++;
        if (binary && !tagged (neg_ref))
          continue;

        INC_COUNT (elimination_ticks);
#ifndef QUIET
        statistics.resolutions++;
#endif
        bool tautological = false;
        size_t begin_resolvent = resolvents.size ();

        if (tagged (pos_ref)) {
          unsigned other = untag (pos_ref);
          debug_binary (pivot, other, "1st antecedent");
          signed char value = values[other];
          if (!value)
            resolvents.push_back (other);
          else if (value > 0)
            tautological = true;
        } else {
          Clause *c = dereference_clause (pos_ref);
          debug_clause (c, "1st antecedent");
          if (c->garbage) {
            tautological = true;
            continue;
          }
          for (auto other : *c) {
            if (other == pivot)
              continue;
            signed char value = values[other];
            if (value > 0) {
              mark_clause_garbage (c, "satisfied");
              tautological = true;
              break;
            }
            if (!value)
              resolvents.push_back (other);
          }
        }

        if (tautological) {
          debug ("1st antecedent already satisfied");
          neg_it = neg_end;
          resolvents.resize (begin_resolvent);
          continue;
        }

        size_t marked = resolvents.size ();
        for (size_t i = begin_resolvent; i != marked; i++) {
          unsigned lit = resolvents[i];
          marks[lit] = -1, marks[negate (lit)] = 1;
        }

        if (tagged (neg_ref)) {
          unsigned other = untag (neg_ref);
          debug_binary (not_pivot, other, "2nd antecedent");
          signed char value = values[other];
          if (!value)
            value = marks[other];
          if (value > 0)
            tautological = true;
          else if (!value)
            resolvents.push_back (other);
        } else {
          Clause *c = dereference_clause (neg_ref);
          debug_clause (c, "2nd antecedent");
          if (c->garbage)
            tautological = true;
          else {
            for (auto other : *c) {
              if (other == not_pivot)
                continue;
              signed char value = values[other];
              if (value > 0) {
                mark_clause_garbage (c, "satisfied");
                tautological = true;
                break;
              }
              if (!value) {
                value = marks[other];
                if (value > 0) {
                  tautological = true;
                  break;
                }
              }
              if (!value)
                resolvents.push_back (other);
            }
          }
        }

        for (size_t i = begin_resolvent; i != marked; i++) {
          unsigned lit = resolvents[i];
          marks[lit] = marks[negate (lit)] = 0;
        }

        size_t end_resolvent = resolvents.size ();
        size_t size = end_resolvent - begin_resolvent;
#ifndef NDEBUG
        for (size_t i = begin_resolvent; i != end_resolvent; i++) {
          unsigned lit = resolvents[i];
          unsigned idx = index (lit);
          assert (flags[idx].active);
          assert (lit != pivot);
          assert (lit != not_pivot);
        }
#endif
        if (tautological) {
          debug ("2nd antecedent satisfied or clashing");
          resolvents.resize (begin_resolvent);
        } else if (!size) {
          debug ("empty resolvent");
          inconsistent = true;
          trace_empty ('a');
          break;
        } else if (size == 1) {
          unsigned unit = resolvents[begin_resolvent];
          resolvents.resize (begin_resolvent);
          debug ("unit resolvent %s", debug_literal (unit));
          trace_unit ('a', unit);
          assign (unit, UNIT_REASON);
        } else if (size > size_limit) {
          debug ("resolvent of size %zu exceeds limit %zu", size,
                 size_limit);
          resolved = RESOLVENT_SIZE_EXCEEDED;
          break;
        } else {
          resolvents.push_back (INVALID_LITERAL);
          debug_literals (size, &resolvents[begin_resolvent],
                          "computed resolvent[%zu]", resolved);
          if (++resolved > resolved_limit)
            break;
        }
      }
    }
  }
}

inline size_t
Solver::actual_number_of_occurrences (References *occurrences) {
  assert (!watching);
  uint64_t ticks = ticks_references (occurrences);
  auto begin = occurrences->begin ();
  auto end = occurrences->end ();
  auto q = begin, p = q;
  const size_t occurrent_limit = OPTION (elimination_occurrence_limit);
  const size_t size_limit = OPTION (resolvent_size);
  size_t res = 0;
  while (p != end) {
    Reference ref = *q++ = *p++;
    if (tagged (ref)) {
      unsigned other = untag (ref);
      if (flags[index (other)].eliminated)
        q--;
      else if (values[other] <= 0 && res++ == occurrent_limit) {
        assert (res > occurrent_limit);
        break;
      }
    } else {
      ticks++;
      Clause *c = dereference_clause (ref);
      if (c->garbage) {
        q--;
      } else if (c->size > size_limit) {
        res = RESOLVENT_SIZE_EXCEEDED;
        assert (res > occurrent_limit);
        break;
      } else if (res++ == occurrent_limit) {
        assert (res > occurrent_limit);
        break;
      }
    }
  }
  while (p != end)
    *q++ = *p++;
  occurrences->resize (q - begin);
  ADD_COUNT (elimination_ticks, ticks);
  return res;
}

bool Solver::generate_resolvents (Eliminator &eliminator,
                                  unsigned pivot_idx) {
  unsigned pivot = literal (pivot_idx);
  unsigned not_pivot = negate (pivot);

  auto pos_occs = references (pivot);
  auto neg_occs = references (not_pivot);

  const size_t occurrence_limit = OPTION (elimination_occurrence_limit);
  size_t pos_size = actual_number_of_occurrences (pos_occs);
  if (pos_size > occurrence_limit)
    return false;

  size_t neg_size = actual_number_of_occurrences (neg_occs);
  if (neg_size > occurrence_limit)
    return false;

  debug ("pivot %s has %zu positive and %zu negative occurrences",
         debug_literal (pivot), pos_size, neg_size);

  size_t limit = pos_size + neg_size + elimination_bound;
  debug ("trying to eliminate pivot %s within %zu resolutions",
         debug_literal (pivot), limit);

  assert (!level);
  eliminator.resolvents.clear ();

  size_t resolved = 0;

  std::vector<unsigned> &resolvents = eliminator.resolvents;
#ifndef NGATES
  std::vector<Reference> *gate = eliminator.gate;
  std::vector<Reference> *nogate = eliminator.nogate;
  if (find_gate (eliminator, pivot)) {
    debug ("resolving only %s gate against non-gate clauses",
           debug_literal (pivot));
    resolve (pivot, &gate[0], &nogate[1], resolved, limit, resolvents);
    if (resolved <= limit)
      resolve (not_pivot, &gate[1], &nogate[0], resolved, limit,
               resolvents);
  } else if (find_gate (eliminator, not_pivot)) {
    debug ("resolving only %s gate against non-gate clauses",
           debug_literal (not_pivot));
    resolve (pivot, &gate[1], &nogate[0], resolved, limit, resolvents);
    if (resolved <= limit)
      resolve (not_pivot, &gate[0], &nogate[1], resolved, limit,
               resolvents);
  } else
#endif
    resolve (pivot, pos_occs, neg_occs, resolved, limit, resolvents);

#ifdef LOGGING
  if (resolved == RESOLVENT_SIZE_EXCEEDED)
    debug ("elimination of pivot %s exceeds resolvent size limit %u",
           debug_literal (pivot), (unsigned) OPTION (resolvent_size));
  else if (resolved > limit)
    debug ("elimination of pivot %s produces more than %zu resolvents",
           debug_literal (pivot), limit);
  else
    debug ("elimination of pivot %s produces only %zu resolvents",
           debug_literal (pivot), resolved);
#endif

  return !inconsistent && resolved <= limit;
}

void Solver::add_resolvents (Eliminator &eliminator, unsigned pivot_idx) {
  std::vector<unsigned> &resolvents = eliminator.resolvents;
  debug ("adding resolvents on %s from resolvents stack of size %zu",
         debug_variable (pivot_idx), resolvents.size ());

#ifdef LOGGING
  size_t added = 0;
#endif
  bool satisfied = false;
  assert (clause.empty ());

  auto begin = resolvents.begin ();
  auto end = resolvents.end ();
  auto resolvent = begin;

#ifdef LOGGING
  size_t processed = 0;
#endif
  size_t large_clauses_added = 0;

  for (auto p = begin; p != end; p++) {
    unsigned lit = *p;
    if (lit == INVALID_LITERAL) {
      size_t resolvent_size = p - resolvent;
      if (satisfied) {
        satisfied = false;
        debug_literals (resolvent_size, &*resolvent,
                        "satisfied resolvent[%zu]", processed);
      } else {
        size_t actual_size = clause.size ();
        if (!actual_size) {
          debug_literals (resolvent_size, &*resolvent,
                          "falsified resolvent[%zu]", processed);
          inconsistent = true;
          trace_empty ('a');
          break;
        } else if (actual_size == 1) {
          unsigned unit = clause[0];
          debug_literals (resolvent_size, &*resolvent,
                          "unit %s resolvent[%zu]", debug_literal (unit),
                          processed);
          trace_temporary ('a');
          assign (unit, UNIT_REASON);
        } else {
          if (actual_size < resolvent_size)
            debug_literals (resolvent_size, &*resolvent,
                            "shrinking resolvent[%zu]", processed);
          if (actual_size == 2)
            new_binary ();
          else {
            new_clause ();
            large_clauses_added++;
          }
          trace_temporary ('a');
#ifdef LOGGING
          added++;
#endif
        }
      }
      resolvent = p + 1;
      clause.clear ();
#ifdef LOGGING
      processed++;
#endif
    } else if (!satisfied) {
      signed value = values[lit];
      if (value > 0)
        satisfied = true;
      else if (!value)
        clause.push_back (lit);
    }
  }

  if (GET_COUNT (redundant_clauses))
    wasted_accessed += large_clauses_added;

  debug ("added %zu resolvents on %s", added, debug_variable (pivot_idx));
#ifndef LOGGING
  (void) pivot_idx;
#endif
}

}; // namespace TabularaSAT
