#include "solver.hpp"
#include "radix.hpp"
#include "schedule.hpp"

#include <cinttypes>

namespace TabularaSAT {

Solver::Solver ()
#ifndef NOPTIONS
    : configurations (this->options)
#endif
{
  control.push_back (Frame (INVALID_LITERAL, 0));
}

Solver::~Solver () {
  if (variables) {
    free (reasons);
    free (position);
    free (phases);
    free (flags);
  }
}

void Solver::adjust_to_increased_variables () {

  if (!variables)
    return;

    // We currently only support one-shot solving and size adjusting!

#ifndef QUIET
  assert (!statistics.active_variables);
#endif
  assert (!reasons);
  assert (!position);

  reasons = (Reference *) calloc (variables, sizeof *reasons);
  position = (unsigned *) calloc (variables, sizeof *position);

  assert (!phases);
  assert (!flags);

  phases = (Phase *) calloc (variables, sizeof *phases);
  flags = (Flags *) calloc (variables, sizeof *flags);

  if (!reasons || !position || !phases || !flags)
    fatal ("out-of-memory adjusting internal variable data");

  trail.adjust (variables);

  matrix.resize (literals);

  // The 'queue' and 'heap' are initialized lazily in 'decide_literal'.
}

inline bool Solver::conflict_limit_hit () {
  return conflict_limit_set && conflict_limit <= GET_COUNT (conflicts);
}

State Solver::run_conflict_driven_clause_learning_loop () {
  for (;;)
    if (inconsistent)
      return UNSATISFIABLE;
    else if (!search_propagate ())
      analyze_conflict ();
    else if (iterating)
      report_propagation_of_learned_unit_clause ();
    else if (!unassigned)
      return SATISFIABLE;
    else if (terminated)
      return UNKNOWN;
    else if (conflict_limit_hit ())
      return UNKNOWN;
    else if (reducing_redundant_clauses ())
      reduce_redundant_clauses ();
    else if (restarting ())
      restart ();
#ifndef NREPHASE
    else if (rephasing ())
      rephase ();
#endif
#ifndef NSWITCH
    else if (switching_search_mode ())
      switch_search_mode ();
#endif
#ifndef NTRANSITIVE
    else if (reducing_binary_implication_graph ())
      transitive_reduction ();
#endif
#ifndef NPROBE
    else if (failed_literal_probing ())
      probe_failed_literals ();
#endif
#ifndef NVIVIFY
    else if (vivifying_clauses ())
      vivify_clauses ();
#endif
#ifndef NSUBSTITUTE
    else if (substituting_variables ())
      substitute_variables ();
#endif
    else if (subsuming_irredundant_clauses ())
      subsume_clauses ();
    else if (eliminating_variables ())
      eliminate_variables ();
    else if (collecting_clauses)
      collect_clauses ();
    else
      decide_literal ();
}

void Solver::initialize_inprocessing_limits () {
#ifndef NTRANSITIVE
  if (OPTION (transitive_reduction)) {
    TransitiveReductionScheduler transitive_reduction_scheduler (this);
    transitive_reduction_scheduler.set_initial_conflict_limit ();
  }
#endif
#ifndef NVIVIFY
  if (OPTION (vivify_clauses)) {
    VivificationScheduler vivification_scheduler (this);
    vivification_scheduler.set_initial_conflict_limit ();
  }
#endif
#ifndef NSUBSTITUTE
  if (OPTION (substitute_variables)) {
    SubstitutionScheduler substitution_scheduler (this);
    substitution_scheduler.set_initial_conflict_limit ();
  }
#endif
  if (OPTION (subsume_clauses)) {
    SubsumptionScheduler subsumption_scheduler (this);
    subsumption_scheduler.set_initial_conflict_limit ();
  }
  if (OPTION (eliminate_variables)) {
    EliminationScheduler elimination_scheduler (this);
    elimination_scheduler.set_initial_conflict_limit ();
  }
}

void Solver::print_sizes (const char *prefix) {
  printf ("%s'sizeof (Solver) = %zu' and 'sizeof (Statistics) = %zu'\n",
          prefix, sizeof (Solver), sizeof (Statistics));
  printf ("%s'sizeof (Clause) = %zu' and 'Clause::bytes (3) = %zu'\n",
          prefix, sizeof (Clause), Clause::bytes (3));
  printf ("%s'sizeof (Position) = %zu' and 'sizeof (Reference) = %zu'\n",
          prefix, sizeof (Position), sizeof (Reference));
  printf ("%s'sizeof (References) = %zu'\n", prefix, sizeof (References));
}

void Solver::initialize_internal_solving () {
  assert (!initialized);
#ifndef QUIET
  if (verbosity > 0)
    print_sizes ("c ");
#endif
#ifndef NSWITCH
  {
    stable = OPTION (initial_mode_stable);
    message (1, "starting in %s mode", mode ());
  }
#endif

#ifndef NGLUE
#ifndef NOPTIONS
  if (OPTION (tier2_max_glue) < OPTION (tier1_max_glue)) {
    OPTION (tier2_max_glue) = OPTION (tier1_max_glue);
    message (0, "forced '--tier2-max-glue=%u",
             (unsigned) OPTION (tier2_max_glue));
  }
#else
  assert (OPTION (tier2_max_glue) >= OPTION (tier1_max_glue));
#endif
#endif

  if (OPTION (reduce)) {

#ifndef VANILLA
    if (!OPTION (reduce_vanilla)) {
      ReduceScheduler reduce_scheduler (this);
      reduce_scheduler.set_initial_conflict_limit ();
    } else
#endif
    {

      uint64_t irredundant = GET_COUNT (irredundant_clauses);
      uint64_t binary = GET_COUNT (binary_clauses);
      double factor = OPTION (initial_learned_fraction) / 100.0;
      uint64_t minimum = OPTION (minimum_learned);
      uint64_t limit = factor * (irredundant + binary);
      limit = std::max (minimum, limit);
      limits.reduce.clauses = limit;

      message (1,
               "initial kept learned clauses limit of %" PRIu64 " clauses",
               limit);

      uint64_t initial_increase = OPTION (initial_increase);
      limits.reduce.interval = initial_increase;
      limits.reduce.conflicts = initial_increase;
      message (1,
               "initial kept learned clauses increase limit of %" PRIu64
               " conflicts",
               initial_increase);
    }
  }

  set_restart_limit (true);

#ifndef NREPHASE
  if (OPTION (rephase)) {
    RephaseScheduler rephase_scheduler (this);
    rephase_scheduler.set_initial_conflict_limit ();
  }
#endif
  {
    double alpha = OPTION (variable_score_alpha) / 1000.0;
    double decay = 1 - alpha;
    bumper.set_decay (decay);
    message (1,
             "VSIDS decay set to %g "
             "(alpha specified as %g = %u/1000)",
             decay, alpha, (unsigned) OPTION (variable_score_alpha));
  }
#ifndef NACTIVITY
  {
    double alpha = OPTION (clause_activity_alpha) * 1e-6;
    double decay = 1 - alpha;
    message (1,
             "clause activity decay set to %g "
             "(alpha specified as %g = %u/1000)",
             decay, alpha, (unsigned) OPTION (clause_activity_alpha));
    assert (decay);
    clause_activity_factor = 1 / decay;
  }
#endif
#ifndef NSWITCH
  if (OPTION (switch_search_mode)) {
    assert (!statistics.switched_modes);
    assert (!statistics.switched_into_second_mode);
    limits.mode = OPTION (mode_switching_interval);
    message (1, "initial %s mode switching limit of %" PRIu64 " conflicts",
             mode (), limits.mode);
  }
#endif
  if (!OPTION (preprocessing) && OPTION (inprocessing))
    initialize_inprocessing_limits ();

  initialized = true;
}

void Solver::reconstruct_solution () {
  {
    size_t assigned = 0;
    signed char intial_phase = OPTION (initial_phase) ? 1 : -1;
    signed char not_intial_phase = -intial_phase;
    for (auto idx : variables) {
      unsigned lit = literal (idx);
      if (values[lit])
        continue;
      debug ("forcing assign %s", debug_literal (lit));
#ifndef NDEBUG
      Flags &f = flags[idx];
      if (f.activated) {
        assert (!f.active);
        assert (f.eliminated || f.substituted);
      }
#endif
      unsigned not_lit = negate (lit);
      values[lit] = intial_phase;
      assert (!values[not_lit]);
      values[not_lit] = not_intial_phase;
      assigned++;
    }
#ifndef QUIET
    message (2, "forced assignment of %zu eliminated variabels", assigned);
#else
    (void) assigned;
#endif
  }
  auto end = reconstruction_stack.end ();
  auto begin = reconstruction_stack.begin ();
  unsigned last = INVALID_LITERAL;
  bool satisfied = false;
  unsigned flipped = 0;
  auto p = end;
  while (p != begin) {
    unsigned lit = *--p;
    if (lit == INVALID_LITERAL) {
      if (satisfied)
        satisfied = false;
      else {
        assert (last != INVALID_LITERAL);
        values[last] = 1, values[negate (last)] = -1;
        debug ("flipped eliminated %s", debug_literal (last));
        flipped++;
      }
    } else {
      signed char value = values[lit];
      if (value > 0)
        satisfied = true;
    }
    last = lit;
  }
#ifndef QUIET
  message (2, "flipped %u literals during reconstruction", flipped);
#else
  (void) flipped;
#endif
}

State Solver::solve_internally () {
  State res;
  if (inconsistent)
    res = UNSATISFIABLE;
  else if (!variables)
    res = SATISFIABLE;
  else {
    if (!initialized)
      initialize_internal_solving ();
    report_progress ('*');
    if (!search_propagate ()) {
      analyze_conflict ();
      assert (inconsistent);
      res = UNSATISFIABLE;
    } else {
      res = run_initial_preprocessing_loop ();
      if (!res && !terminated && !conflict_limit_hit ()) {
        start_search ();
#ifndef NSWITCH
        report_progress (stable ? '[' : '{');
#endif
        res = run_conflict_driven_clause_learning_loop ();
#ifndef NSWITCH
        report_progress (stable ? ']' : '}');
#endif
        stop_search ();
      }
    }
  }
#ifndef QUIET
  if (verbosity >= 0) {
    report_progress (res == 10 ? '1' : res == 20 ? '0' : '?');
    fputs ("c\n", stdout), fflush (stdout);
  }
#endif
  if (res == 10)
    reconstruct_solution ();
  return res;
}

}; // namespace TabularaSAT

int main (int argc, char **argv) {
  TabularaSAT::Solver solver;
  int res = solver.main (argc, argv);
  return res;
}
