#ifndef _tabularasat_solver_hpp_INCLUDED
#define _tabularasat_solver_hpp_INCLUDED

// First include the local compile-time definitions for this variant.

#include "variant.hpp"

// Then include the shared (compile-time) dependencies this implies.

#include "dependencies.hpp" // Implied compile-time options (next).

// Now include all the TabularaSAT 'shared' header files which do not
// depend on the particular solver instance below.

#include "arena.hpp"
#include "heap.hpp"
#ifndef NQUEUE
#include "queue.hpp"
#endif
#include "radix.hpp"
#ifndef NWALK
#include "random.hpp"
#endif
#include "shared.hpp"

// Then include all the solver header files parameterized by the
// macros/definitions of 'variant.hpp' and then 'dependencies.hpp'.

#include "configurations.hpp"
#include "options.hpp"
#include "statistics.hpp"

// Finally include all the standard header files.

#include <cstdint>
#include <cstdio>
#include <vector>

namespace TabularaSAT {

#ifndef NACTIVITY

// Avoid floating point overflow in updating clause activities.

static const double CLAUSE_ACTIVITY_LIMIT = 1e20;

#endif

// Infinite ticks limits during preprocessing.

static const uint64_t UNLIMITED = ~(uint64_t) 0;

#ifndef NGLUE

// As discussed below the maximum stored glue is limited.  One can still
// learn clauses with larger glue (even though it is unlikely), but when a
// larger glue is saved it is truncated to 'MAXIMUM_GLUE'.

static const unsigned MAXIMUM_GLUE = (1u << 23) - 1;

#endif

// Clauses are the core data structure for large non-binary clauses.

// In order to save space and achieve a compact memory layout for clauses we
// are using bit-fields, which requires to limit the maximum glue to the
// 'MAXIMUM_GLUE' above to fit the number of bits ('23') for the 'glue'
// field left.  This memory layout saves 8 bytes for each large clause.

// On the other hand there are also at least 8 bytes for watches or 12 bytes
// needed for the occurrence lists for original clauses during variable
// elimination.  Still this compact versions not only saves some modest
// space but also is faster (arguably by decreasing cache access pressure in
// propagation as that hot-spot computation is memory bound).

struct Clause {

#ifdef LOGGING
  uint64_t id; // For debugging only.
#endif

  unsigned size; // Actual size of 'literals'.

  // Bit-fields let the compiler pack the following Boolean flags including
  // the 2-bit used field and the 23-bit glue field into 4 bytes.  Without
  // bit-fields this gives 3 unsigned words as we have 7 Boolean flags which
  // take up one bit each and 2 more bits for 'used'.

  unsigned garbage : 1;   // Marked as garbage clause.
  unsigned reason : 1;    // Active reason clause.
  unsigned redundant : 1; // Learned redundant clause.
  unsigned shortened : 1; // Allocated size actually smaller.
  unsigned shrink : 1;    // Contains root-level falsified literal.
  unsigned subsume : 1;   // Untried subsumption candidate.
#ifndef NVIVIFY
  unsigned vivify : 1; // Untried vivification candidate.
#endif
#ifndef NUSED
  unsigned used : 2; // Used since last reduction: 0,1,2.
#endif
#ifndef NGLUE
  unsigned glue : 23; // Glucose level (LBD).
#endif

  // Here starts the second 8 byte block.

#ifndef NREPLACEMENTCACHE
  unsigned replaced; // Last replacement position during propagation.
#endif

#ifndef NACTIVITY

  union {
    float as_float;       // Clause activity as tie breaker.
    unsigned as_unsigned; // For radix sort treat is as 'unsigned'.
  } activity;

#endif

  unsigned literals[]; // Actually 'literals[size]'.

  // Provides simple iteration of the literals of the clause.

  unsigned *begin () { return literals; }
  unsigned *end () { return literals + size; }

  // Support for the low-level clause arena memory allocators.

  Clause *next () {
    unsigned *p = literals + size;
    if (shortened)
      while (*p++ != INVALID_LITERAL)
        ;
    return (Clause *) p;
  }

  // The number of words has to match the number of unsigned words 'n'
  // needed for allocation as given to the arena 'allocate ()' function.
  // It is determined by first computing the number of needed bytes which is
  // pretty compiler dependent as well as whether we have an 'id' or not.

  static size_t bytes (unsigned size) {
    return (size_t) & ((Clause *) 0)->literals[size];
  }

  static size_t words (unsigned size) {
    size_t bytes = Clause::bytes (size);
    assert (!(bytes & 3)); // Has to be aligned to 4-bytes (32-bits).
    return bytes >> 2;     // Otherwise this would truncate bits.
  }
};

// Large clauses are stored in a memory arena and referenced through their
// position (offset) from the start of the arena.  This not only allows
// compact memory allocation under full control of the solver but also gives
// deterministic sorting using those references (instead of pointers) and
// avoids the use of stable sorting routines.  Binary clauses are virtual
// and only reside in watch/occurrence lists.

typedef class Arena<Clause> Clauses;

// The starting position of a clause in the clause arena, which is the
// offset in terms of words from the start of the memory arena, can either
// be encoded with a 32-bit number or with a 64-bit number.  The former
// almost reaches a factor of two in memory reduction for CNFs with many
// binary clauses, while the latter on the other hand allows to handle CNFs
// with a large number of non-binary clauses.

// These arena positions are encoded as references which need to be tagged
// to distinguish them from binary clause references. Thus the number of
// references is even half the number of possible positions.  We give those
// positions the same type even though they can use one bit less.

// Another benefit of 32-bit references is of course that the array
// allocated for the 'References' stack below needs less memory too.

#ifndef NCOMPACT

typedef unsigned Position;
typedef unsigned Reference;

#define POSFMT "u" // Format string '%' suffix for 'printf'.

#else

typedef size_t Position;
typedef size_t Reference;

#define POSFMT "zu" // Format string '%' suffix for 'printf'.

#endif

// The following constants could all be the same but for readability and
// precision of the code we distinguish them explicitly by name.

static const Position INVALID_POSITION = ((Position) 0) - 1;
static const Reference INVALID_REFERENCE = ((Reference) 0) - 1;
static const size_t INVALID_SIZE = ((size_t) 0) - 1;

// We are going to use a stack of references allocated in an arena.

struct References {

  unsigned lit;       // Invalid if garbage.
  unsigned _capacity; // Larger equal to '_size' (not a power of two).
  unsigned _size;     // The actual size of 'data'.
  Reference data[];   // The actual references / watcher data.

  size_t size () const { return _size; }
  size_t capacity () const { return _capacity; }

  bool full () const { return _size == _capacity; }
  bool empty () const { return !_size; }

  Reference *begin () { return data; }
  Reference *end () { return data + _size; }

  // Returns the next 'References' after 'this' in the 'arena.

  References *next () { return (References *) (data + _capacity); }

  void clear () { _size = 0; }

  // Can only used for shrinking the size in-place.

  void resize (size_t new_size) {
    assert (new_size <= _size);
    _size = new_size;
  }

  // The number of words (32-bit thus 4-byte) needed for the given size.

  static size_t bytes (unsigned size) {
    return (size_t) & ((References *) 0)->data[size];
  }

  static size_t words (unsigned size) {
    size_t bytes = References::bytes (size);
    assert (!(bytes & 3)); // Has to be aligned to 4-bytes (32-bits).
    return bytes >> 2;     // Otherwise this would truncate bits.
  }
};

struct Matrix {
  size_t garbage = 0;
  Arena<References> arena;
  std::vector<size_t> offsets;

  References &operator[] (unsigned lit) {
    assert (lit < offsets.size ());
    size_t offset = offsets[lit];
    return *arena[offset];
  }

  Matrix () {
    size_t n = References::words (0);
    References *v = arena.allocate (n);
    v->lit = INVALID_LITERAL;
    v->_size = v->_capacity = 0;
  }

  size_t size () { return offsets.size (); };
  void resize (size_t new_size) { offsets.resize (new_size); }

  void clear_and_mark_as_garbage (References *);

  Arena<References>::iterator begin () { return arena.begin (); }
  Arena<References>::iterator end () { return arena.end (); }
};

// The stack of clause references used to store watches and occurrences.

typedef References References;

// Unit and decision reasons are references with specific constants.  We
// need to distinguish unit from decision reasons due to chronological
// backtracking which might assign units on higher decision levels with of
// course still assignment level zero.  We use 'INVALID_REFERENCE' and one
// less which are checked never ever to point to the start of a clause.

static const Reference DECISION_REASON = INVALID_REFERENCE;
static const Reference UNIT_REASON = INVALID_REFERENCE - 1;

// The largest position of a clause in the arena and tagged with zero needs
// to be smaller than the 'UNIT_REASON = 0xfff..ffe' thus the actual
// position smaller than '(UNIT_REASON-1)/2 = 0x7ff..fff'.  This is same
// computation as for literals in 'shared.hpp' except that we work here in
// principle with both 64-bit and 32-bit variants and also can end with an
// even number (so '0x7ff..ffe').

static const Position MAXIMUM_POSITION_OF_CLAUSE =
    (UNIT_REASON - 1) / 2 - 1;

// Flags of an individual variable.

struct Flags {
  bool active : 1;      // Activated in the past and still active.
  bool activated : 1;   // Has been used and thus was activated.
  bool eliminate : 1;   // Try to eliminate this removed variable.
  bool eliminated : 1;  // Variable eliminated and thus inactive.
  bool fixed : 1;       // Variable fixed to a constant at root-level.
  bool minimizable : 1; // Cache that this variable is minimizable.
  bool poisoned : 1;    // Cache that this variable is not minimizable.
  bool seen : 1;        // Variable is analyzed during conflict analysis.
#ifndef NSHRINK
  bool shrinkable : 1; // Mark this variable as shrinkable.
#endif
#ifndef NSUBSTITUTE
  bool substituted : 1; // Variable substituted thus inactive.
#endif
  bool subsume : 1; // Try to subsume this added variable.
};

struct Phase {

#ifndef NREPHASE

  // Most consistent assignment reset and reused in best phase rephasing.

  signed char best;

#endif

  // Saved phase during assignment is fall back in stable mode and the main
  // phase decision heuristic in focused mode.

  signed char saved;

#ifndef NTARGET

  // Maximum consistent assignment since last rephase takes priority in
  // deciding the phase (value) during stable mode over saved phase.

  signed char target;

#endif
};

// Unbiased exponential moving averages (check the ADAM paper for details).

struct Average {

  double value;  // The corrected current value of the average.
  double alpha;  // The learning rate (close to zero).
  double beta;   // Cached 'beta = (1 - alpha)'.
  double biased; // The current average biased towards zero.
  double exp;    // Set to 'beta^n' after adding n-th 'y' value.

  Average (double a)
      : value (0), alpha (a), beta (1 - a), biased (0), exp (1) {}

  operator double () { return value; }

  void operator+= (double y) {
    biased += alpha * (y - biased);
    exp *= beta;
    value = biased / (1 - exp);
  }
};

// Limits for scheduling functions.

struct Limits {
  uint64_t eliminate = 0;
  uint64_t mode = 0;
  struct {
    uint64_t conflicts = 0;
    uint64_t clauses = 0;  // For vanilla style reduce scheduling.
    uint64_t interval = 0; // For vanilla style reduce scheduling.
  } reduce;
#ifndef NPROBE
  uint64_t probe = 0;
#endif
#ifndef NREPHASE
  uint64_t rephase = 0;
#endif
  uint64_t restart = 0;
#ifndef NSUBSTITUTE
  uint64_t substitute = 0;
#endif
  uint64_t subsume = 0;
#ifndef NTRANSITIVE
  uint64_t transitive = 0;
#endif
#ifndef NVIVIFY
  uint64_t vivify = 0;
#endif
};

// Saving statistics for computing what happened since the last time a
// particular function, i.e. variable elimination, was called.  Most
// important is saving the 'search_ticks' so that the effort spend in such a
// function can be limit relative to the effort spent in plain CDCL search
// since the last time the function was executed.

struct Last {
#ifndef NSWITCH
  uint64_t conflicts_at_mode_switch = 0;
#endif
#ifndef NBUMPREASONS
  uint64_t decisions_at_conflict[2] = {0, 0}; // 0=focused, 1=stable
#endif
  uint64_t fixed_variables_at_reduce_or_collect = 0;
  uint64_t search_ticks_at_elimination = 0;
#ifndef NPROBE
  uint64_t search_ticks_at_failed_literal_probing = 0;
#endif
#ifndef NSWITCH
  uint64_t search_ticks_at_mode_switch = 0;
#endif
  uint64_t search_ticks_at_subsumption = 0;
#ifndef NTRANSITIVE
  uint64_t search_ticks_at_transitive_reduction = 0;
#endif
#ifndef NVIVIFY
  uint64_t search_ticks_at_vivification = 0;
#endif
#ifndef NWALK
  uint64_t search_ticks_at_walk = 0;
#endif
};

#ifndef NDELAY

struct Delays {
  struct {
    uint64_t count = 0, current = 0;
  } bump_reason_side;
};

#endif

struct Disabled {
  bool elimination = false;
  bool subsumption = false;
};

struct Trigger {
  unsigned char elimination = 0;
#ifndef NVIVIFY
  unsigned char vivification = 0;
#endif
};

// Decision control stack frame.

struct Frame {
  unsigned decision;   // Decision literal of frame.
  unsigned pulled = 0; // Needed to determine levels during analysis.
#ifndef NPROMOTE
  unsigned promote = 0; // Used to recompute glues of a single clause.
#endif
  unsigned trail; // Trail height when entering frame.

  Frame (unsigned d = 0, unsigned t = 0) : decision (d), trail (t) {}
};

struct Averages {

  // Slow and fast exponential moving average 'alphas' (so similar to
  // learning rates).  Unfortunately we can not define them as constants as
  // otherwise the member initialization below does not compile.

  // The glue and decision rate averages are actually needed, while the
  // other two ('level' and 'trail') are only used for reporting statistics.

#define SLOW_ALPHA 1e-5
#define FAST_ALPHA 3e-2

#ifndef NGLUE
  Average glue_slow = Average (SLOW_ALPHA);
  Average glue_fast = Average (FAST_ALPHA);
#endif
#ifndef QUIET
  Average level = Average (SLOW_ALPHA);
  Average trail = Average (SLOW_ALPHA);
#endif
#ifndef NBUMPREASONS
  Average decision_rate = Average (SLOW_ALPHA);
#endif
};

struct reluctant {
  uint64_t u = 1, v = 1;
  void reset () { u = v = 1; }
};

// Data structures to save delayed not yet watched literals.

struct LiteralBlocking {
  unsigned lit;
  unsigned blocking;
  LiteralBlocking (unsigned l, unsigned b) : lit (l), blocking (b) {}
};

struct LiteralBlockingReference : LiteralBlocking {
  Reference ref;
  LiteralBlockingReference (unsigned l, unsigned b, Reference r)
      : LiteralBlocking (l, b), ref (r) {}
};

class Trail {
  unsigned *_begin = 0, *_end = 0, *_allocated = 0;

public:
  unsigned *propagated = 0;
  ~Trail () { free (_begin); }
  size_t size () const { return _end - _begin; }
  size_t capacity () const { return _allocated - _begin; }
  unsigned &operator[] (size_t i) {
    assert (i < size ());
    return _begin[i];
  }
  unsigned *begin () { return _begin; }
  unsigned *end () { return _end; }
  void push_back (unsigned lit) {
    assert (_end < _allocated);
    *_end++ = lit;
  }
  void adjust (size_t new_capacity) {
    size_t old_capacity = capacity ();
    assert (old_capacity <= new_capacity);
    if (old_capacity == new_capacity)
      return;
    assert (new_capacity);
    size_t old_size = size ();
    size_t old_propagated = propagated - _begin;
    _begin = (unsigned *) realloc (_begin, new_capacity * sizeof *_begin);
    if (!_begin)
      fatal_error ("Trail::resize", "out-of-memory");
    _end = _begin + old_size;
    _allocated = _begin + new_capacity;
    propagated = _begin + old_propagated;
  }
  void resize (size_t i) {
    assert (i <= size ());
    _end = _begin + i;
  }
};

#ifndef NWALK

// Forward declaration for walking data.

struct Walker;

#endif

// The actual solver code starts here.

class Solver : public Shared {

#ifndef NSWITCH
  bool stable = true; // Stable or focused search mode.
#endif
  bool collecting_clauses = false; // Force garbage collection.
  bool eliminating = false;        // Not searching but eliminating.
  bool inconsistent = false;       // Found or learned empty clause.
  bool initialized = false;        // Limits need to be initialized.
  bool iterating = false;          // Learned unit clause.
#ifndef NPROBE
  bool probing = false; // Not searching but failed literal probing.
#endif
#ifndef NVIVIFY
  bool vivifying = false; // Not searching but vivifying.
#endif
#ifndef NWALK
  bool warming = false; // Not searching bug warming up.
#endif
  bool watching = true; // Watching or connecting clauses.

  unsigned preprocessing = 0; // Preprocessing round.

  std::vector<Frame> control; // One frame per decision level.

  Trail trail;
  unsigned unassigned = 0; // Remaining unassigned active variables.

#ifndef QUIET
#ifndef NCHRONO
  unsigned out_of_order_units = 0; // Due to chronological backtracking.
#endif
  unsigned assigned_eliminated = 0; // These are assigned as decisions.
#endif

  // During propagation we replace watched literals while traversing the
  // watched literal references, which sit in an arena that might move when
  // watching and thus pushing a watch to any references stack.  Therefore
  // we can not immediately watch replacement literals but need to wait
  // until such an iteration finished (see 'propagate.hpp' for details).

  std::vector<LiteralBlockingReference> delayed_large_watches;

#ifndef NTARGET

  // These two implement target phases.

  unsigned best_assigned = 0;
  unsigned target_assigned = 0;

#endif

  Clauses clauses; // The clause data memory arena.

  Matrix matrix;               // Maps literals to references.
  std::vector<unsigned> units; // Found delayed units.

  Reference *reasons = 0; // Maps variables to reasons.
  unsigned *position = 0; // Position of variable on trail.

  Phase *phases = 0; // Maps variables to phases.
  Flags *flags = 0;  // Maps variables to flags.

#ifndef NQUEUE
  Queue queue; // Decision priority queue in focused mode.
#endif
  Bumper bumper; // Decision priority heap in stable mode.

  std::vector<unsigned> analyzed; // Variables analyzed.
  std::vector<unsigned> pulled;   // Decision levels analyzes.

  std::vector<unsigned> minimizable; // Cached minimizable variables.
  std::vector<unsigned> poisoned;    // Cached poisoned variables.
#ifndef NSHRINK
  std::vector<unsigned> shrinkable; // Variables marked shrinkable.
#endif
#ifndef NPROMOTE
  std::vector<unsigned> promote; // Levels for promotion.
#endif

  std::vector<char> sorter;

  template <class Ranker>
  void radix_sort_unsigned (std::vector<unsigned> &v, Ranker r) {
    radix_sort_generic<unsigned, unsigned, unsigned, Ranker> (v, sorter, r);
  }

  // Stack to reconstruct models after variable elimination.

  std::vector<unsigned> reconstruction_stack;

#ifndef NWALK
  Random random;
#endif

  Last last;
  Limits limits;
  Trigger trigger;
#ifndef NDELAY
  Delays delay[2];
#endif
  Disabled disabled;

#ifndef NSWITCH
  Averages _averages[2];
  Averages &averages () { return _averages[stable]; }
#else
  Averages _averages[1];
  Averages &averages () { return _averages[0]; }
#endif

  Statistics statistics;
#ifndef NOPTIONS
  SolverOptions options;
  SolverConfigurations configurations;
#endif

#ifdef LOGGING
  void debug_clause (Clause *, const char *, ...)
      __attribute__ ((format (printf, 3, 4)));
  void debug_binary (unsigned, unsigned, const char *, ...)
      __attribute__ ((format (printf, 4, 5)));
  void debug_clauses ();
#else
#define debug_clause(...) VOID
#define debug_binary(...) VOID
#endif

#define debug_reference(REF, ...) \
  debug_clause (dereference_clause (REF), __VA_ARGS__)

  void mark_removed (unsigned lit) {
    unsigned idx = index (lit);
    Flags &f = flags[idx];
    if (!f.active)
      return;
    if (f.eliminate)
      return;
    debug ("marking removed and to eliminate %s", debug_literal (lit));
#ifndef QUIET
    statistics.removed_variables++;
#endif
    f.eliminate = true;
    if (disabled.elimination) {
      message (2, "enabled elimination again");
      disabled.elimination = false;
    }
  }

  void mark_removed (Clause *c) {
    for (auto lit : *c)
      mark_removed (lit);
  }

  void mark_added (unsigned lit) {
    unsigned idx = index (lit);
    Flags &f = flags[idx];
    if (!f.active)
      return;
    if (f.subsume)
      return;
    debug ("marking added and to subsume %s", debug_literal (lit));
#ifndef QUIET
    statistics.added_variables++;
#endif
    f.subsume = true;
    if (disabled.subsumption) {
      message (2, "enabled subsumption again");
      disabled.subsumption = false;
    }
  }

  void mark_added (Clause *c) {
    for (auto lit : *c)
      mark_added (lit);
  }

  bool active (unsigned lit) { return flags[index (lit)].active; }

  Reference new_binary ();
  void bump_clause (Clause *);
#ifndef NGLUE
  void adjust_glue_to_updated_size (Clause *);
#endif
#ifndef NACTIVITY
  void rescale_clause_activities ();
  double clause_activity_increment = 1;
  double clause_activity_factor = 1.001;
#endif
  Reference reference_clause (Clause *) __attribute__ ((always_inline));
  Clause *dereference_clause (Reference ref)
      __attribute__ ((always_inline));
  void resize_references_arena (References *new_end);
  bool collecting_references ();
  void collect_references ();
  References *enlarge_references (unsigned lit, References *,
                                  bool need_two);
  void push_reference (unsigned lit, References *&, Reference ref,
                       bool expect_another) __attribute__ ((always_inline));
  void push_reference (unsigned lit, Reference ref, bool expect_another)
      __attribute__ ((always_inline));
  References *references (unsigned lit) __attribute__ ((always_inline));
#ifndef NGLUE
  Reference new_clause (bool redundant = false, unsigned glue = 0);
#else
  Reference new_clause (bool redundant = false);
#endif
  void connect_clause (unsigned lit, Reference);
  void connect_binary (unsigned lit, unsigned other);
  void watch_literal (unsigned lit, unsigned other, Reference);
  void unwatch_binary (unsigned lit, unsigned other);
  void unwatch_literal (unsigned lit, Reference);
  void watch_clause (Clause *);
  void unwatch_clause (Clause *);
  void copy_clause (Clause *, unsigned except = INVALID_LITERAL);
  bool better_watch (unsigned lit, unsigned other);
  void watch_binary (unsigned lit, unsigned other);
  void strengthen_to_large_clause (Clause *);
  void strengthen_to_binary_clause (Clause *);

  void trace_clause (char type, Clause *c) {
    trace_literals (type, (size_t) c->size, c->literals);
  }

  void assign (unsigned lit, Reference reason);
  void unassign (unsigned lit);
#ifndef NPROBE
  void failed_literal_assign (unsigned lit, Reference reason);
#endif

  unsigned decide_variable_in_stable_mode ();
#ifndef NSWITCH
  unsigned decide_variable_in_focused_mode ();
#endif
  signed char decide_phase (unsigned);
  unsigned decide_variable ();
  void decide_literal (unsigned decision = INVALID_LITERAL);

  Reference conflict;
  unsigned falsified_literal;
  void assign_and_propagate_units ();
  void propagate_after_simplification ();
  void backtrack (unsigned new_level);
  bool backtrack_to_root_level_and_propagate ();
  void report_propagation_of_learned_unit_clause ();

  bool search_propagate ();
#ifndef NPROBE
  bool propagates (unsigned lit);
  bool failed_literal_propagate ();
#endif
#ifndef NVIVIFY
  bool vivify_propagate (Clause *ignore);
#endif
#ifndef NWALK
  bool warming_propagate ();
#endif
#ifndef NPROMOTE
  void promote_clause (Clause *);
#endif

  friend struct larger_level;
#ifndef NUPDATEPHASES
  void update_phases ();
#endif
#ifndef NSHRINK
  void shrink_deduced_clause ();
#endif
  void minimize_deduced_clause ();
  void reset_minimized_and_poisoned ();
  bool minimize_literal (unsigned lit, unsigned depth = 0);
  void analyze_literal (struct Analyzer &, unsigned lit);
  void clear_analyzed ();
#ifndef NALLUIPS
  void learn_all_uips (unsigned);
#endif
#ifndef QUIET
  void update_trail_average ();
#else
#define update_trail_average() VOID
#endif
  void analyze_conflict ();
#ifndef NBUMPREASONS
  void analyze_reason_side_literal (unsigned lit);
  void analyze_reason_side_literals ();
#endif
  void bump_stable_variables ();
#ifndef NBUMPREASONS
  void bump_focused_variables ();
#endif
  void bump_analyzed_variables ();
#ifndef QUIET
  void report_progress (char type);
#else
#define report_progress(...) ((void) (0))
#endif
  bool satisfied_clause (Clause *);
  void mark_clause_garbage (Clause *, const char *reason);
  void mark_binary_garbage (unsigned lit, unsigned other,
                            const char *reason);

  size_t first_redundant = INVALID_POSITION;
  size_t first_tier2 = INVALID_POSITION;
  size_t last_irredundant = 0;
  void reset_positions ();
  void update_position (Clause *);
  void finalize_position_update ();

  friend struct Effort;
  friend struct Scheduler;
  friend struct PowerLogScalingScheduler;

  friend struct ReduceScheduler;
  friend struct less_relevant_redundant_clause;

  size_t trail_root_level () {
    return level ? control[1].trail : trail.size ();
  }

  void mark_reason_clauses (bool mark);
  void gather_reduce_candidates (std::vector<Position> &);
  void mark_reduced_clauses (Position &first_reduced);
  void mark_satisfied_and_shrunken_clauses (Position &first_marked);
  void mark_simplified_clauses (Position &first_simplified);
  void copy_and_collect_clauses (Position first_swept,
                                 std::vector<Position> &);
  void flush_and_map_watches (Position first_swept, bool simplify,
                              const std::vector<Position> &);
  void map_reasons (Position first_swept, const std::vector<Position> &);
  void copy_collect_map_and_flush (Position first_swept);
  void increase_kept_learned_clauses ();
  bool reducing_redundant_clauses ();
  void reduce_redundant_clauses (bool force = false);
  void collect_clauses ();

  bool restarting ();
  struct reluctant reluctant;
  void set_focused_restart_limit (bool initial);
  void set_stable_restart_limit (bool initial);
  void set_restart_limit (bool initial = false);
  unsigned reuse_trail ();
  void restart ();

#ifndef NREPHASE
  friend struct RephaseScheduler;
  bool rephasing ();
  void rephase_best ();
  void rephase_walk ();
  void rephase_inverted ();
  void rephase_original ();
  void rephase ();
#endif

#ifndef NSWITCH
  bool switching_search_mode ();
  void switch_search_mode ();
#endif

  void initialize_internal_solving ();
  void initialize_inprocessing_limits ();
  State run_conflict_driven_clause_learning_loop ();
  State run_initial_preprocessing_loop ();
  bool conflict_limit_hit ();

  uint64_t wasted_accessed = 0;
  void move_redundant_clauses ();
  void watch_all_large_clauses ();
  void connecting_all_large_clauses ();
  void flush_connected_large_clauses ();
  void flush_large_watches_and_satisfied_binary_clauses ();
  void transition_to_sparse_watching ();
  void transition_to_dense_connections ();

#ifndef NWALK
  friend struct Walker;
  void warming_up_saved_phases ();
  void initialize_local_search (Walker &);
  double break_score (Walker &, unsigned lit);
  void break_clauses (Walker &, unsigned lit);
  void make_clauses (Walker &, unsigned lit);
  void flush_pending_minimimum_assignment (Walker &);
  void run_local_search (Walker &);
  void reset_local_search (Walker &);
  void local_search ();
#endif

#ifndef NSWITCH
  const char *mode () { return stable ? "stable" : "focused"; }
#endif
  void start_search () {
    start ("search");
#ifndef NSWITCH
    start (mode ());
#endif
  }

  void stop_search () {
#ifndef NSWITCH
    stop (mode ());
#endif
    stop ("search");
  }

  void start_simplifier (const char *simplifier, unsigned level = 1) {
    if (!preprocessing)
      stop_search (), start ("inprocessing"), start ("simplify");
    start (simplifier, level);
  }

  void stop_simplifier (const char *simplifier) {
    stop (simplifier);
    if (!preprocessing)
      stop ("simplify"), stop ("inprocessing"), start_search ();
  }

  friend struct Eliminator;
  friend struct EliminationScheduler;
  size_t elimination_bound = 0;
#ifndef NELIMINATIONHEAP
  friend struct better_elimination_candidate;
  double elimination_score (unsigned idx, unsigned type);
  void schedule_variable (struct Eliminator &, unsigned);
  void schedule_literal (struct Eliminator &, unsigned);
  void schedule_clause (struct Eliminator &, Clause *);
  void gather_elimination_candidates (struct Eliminator &);
#endif
#ifndef NGATES
  bool find_equivalence_gate (struct Eliminator &, unsigned pivot);
  bool find_and_gate (struct Eliminator &, unsigned pivot);
  bool find_gate (struct Eliminator &, unsigned pivot);
#endif
  size_t actual_number_of_occurrences (References *);
  void mark_eliminated_redundant_clauses_as_garbage (unsigned round);
  void flush_eliminated_binary_clauses ();
  template <class R>
  void resolve (unsigned pivot, R *, R *, size_t &resolved,
                const size_t limit, std::vector<unsigned> &resolvents);
  bool generate_resolvents (struct Eliminator &, unsigned idx);
  void add_resolvents (struct Eliminator &, unsigned idx);
  void eliminate_clauses_of_literal (Eliminator &, unsigned lit, bool push);
  void eliminate_clauses_of_variable (Eliminator &, unsigned idx);
  void garbage_collect_clauses_during_variable_elimination ();
  void reconstruct_solution ();
  bool eliminating_variables ();
  bool eliminate_variables ();

  friend struct SubsumptionScheduler;
  void gather_subsumption_candidates (std::vector<Position> &);
  bool strengthen_clause (Clause *, unsigned remove);
  bool subsume_clause (Clause *, unsigned &remove,
                       Clause *&double_subsumed);
  void connect_minimum_occurring_literal (Clause *);
  bool subsuming_irredundant_clauses ();
  bool subsume_clauses ();

#ifndef NSUBSTITUTE
  friend struct SubstitutionScheduler;
  bool substituting_variables ();
  unsigned decompose_binary_implication_graph (unsigned *reprs);
  void trace_adding_equivalences (unsigned *reprs);
  void mark_substituted (unsigned *reprs);
  void substitute_large_clauses (unsigned *reprs);
  void substitute_binary_clauses (unsigned *reprs);
  bool substitute_variables ();
#endif

#ifndef NTRANSITIVE
  friend struct TransitiveReductionScheduler;
  unsigned next_transitive_src = 0;
  bool reducing_binary_implication_graph ();
  bool transitive_reduction ();
#endif

#ifndef NPROBE
  friend struct FailedLiteralScheduler;
  unsigned next_failed_literal_probe = 0;
  bool failed_literal_probing ();
  bool probe_failed_literals ();
#endif

#ifndef NVIVIFY
  friend struct VivificationScheduler;
  bool vivifying_clauses ();
  unsigned *count_occurrences ();
  void gather_vivification_candidates (unsigned tier, unsigned *counts,
                                       std::vector<Clause *> &);
  void vivify_analyze_literal (unsigned lit);
  bool vivify_clause (unsigned tier, unsigned *counts, Clause *);
  void vivify_tier_clauses (unsigned tier, unsigned *counts,
                            uint64_t limit);
  bool vivify_clauses ();
#endif

  double scale ();

  // Code to check internal consistency at garbage collections etc.
#ifndef NDEBUG
  size_t compute_binary_clauses ();
  size_t compute_redundant_clauses ();
  size_t compute_irredundant_clauses ();
  Position find_first_redundant_clause ();
#ifndef NGLUE
  Position find_first_tier2_clause ();
#endif
  Position find_last_irredundant_clause ();
  void check_clauses ();
  void check_references ();
  void check_variables ();
  void check_all () {
    check_clauses ();
    check_references ();
    check_variables ();
  }
#else
#define check_all() VOID
#define check_clauses() VOID
#define check_references() VOID
#define check_variables() VOID
#endif

  // The virtual functions of 'Interface' called from 'Shared' are here.
  // The other functions above are all private.

  const char *one_word_name () { return solver_name; }
  const char *one_line_description () { return SOLVER_NAME " Solver"; }

  void adjust_to_increased_variables ();
  void activate_variable (unsigned idx);
  void deactivate_variable (unsigned idx);
  void activate_literal (unsigned lit);
  void add_simplified_irredundant_clause ();
  State solve_internally ();
  void print_sizes (const char *prefix);
  void print_statistics (double t);
#ifndef NOPTIONS
  Configurations &iterate_configurations () { return configurations; }
  Options &iterate_options () { return options; }
#endif
public:
  Solver ();
  ~Solver ();
};

}; // namespace TabularaSAT

#endif
