#include "clause.hpp"
#include "connect.hpp"
#include "references.hpp"
#include "schedule.hpp"
#include "solver.hpp"
#include "watch.hpp"

namespace TabularaSAT {

bool Solver::substituting_variables () {
  if (!OPTION (inprocessing))
    return false;
  if (!OPTION (substitute_variables))
    return false;
  return GET_COUNT (conflicts) >= limits.substitute;
}

unsigned Solver::decompose_binary_implication_graph (unsigned *reprs) {
  assert (watching);
  size_t size = literals;
  unsigned *found = (unsigned *) calloc (size, sizeof *found);
  unsigned *reaches = (unsigned *) malloc (size * sizeof *reaches);
  if (size)
    if (!found || !reaches)
      fatal ("out-of-memory composing binary implication graph");
  size_t equivalences = 0, components = 0;
  std::vector<unsigned> work, stack;
  unsigned count = 0;
  for (auto root_watches : matrix) {
    unsigned root = root_watches->lit;
    if (root == INVALID_LITERAL)
      continue;
    if (!active (root))
      continue;
    if (found[root])
      continue;
    assert (work.empty ());
    assert (stack.empty ());
    work.push_back (root);
    while (!work.empty ()) {
      unsigned lit = work.back ();
      if (lit == INVALID_LITERAL) {
        work.pop_back ();
        lit = work.back ();
        unsigned lit_found = found[lit], lit_reaches = lit_found;
        assert (lit_found);
        work.pop_back ();
        unsigned not_lit = negate (lit);
        const auto watches = references (not_lit);
        auto p = watches->begin (), end = watches->end ();
        while (p != end) {
          Reference ref = *p++;
          if (tagged (ref)) {
            unsigned other = untag (ref);
            if (!active (other))
              continue;
            assert (found[other]);
            unsigned other_reaches = reaches[other];
            if (other_reaches < lit_reaches)
              lit_reaches = other_reaches;
          } else
            p++;
        }
        assert (lit_reaches != INVALID_LITERAL);
        if (lit_reaches == lit_found) {
          components++;
          unsigned repr = lit, other;
          size_t stack_size = stack.size (), start = stack_size;
          while (assert (start), (other = stack[--start]) != lit) {
            reaches[other] = INVALID_LITERAL;
            if (other < repr)
              repr = other;
          }
          assert (repr != INVALID_LITERAL);
          size_t component_size = stack_size - start;
          if (component_size > 1) {
            debug ("component %zu of size %zu with representative %s "
                   "entered through found[%s] = %u",
                   components, component_size, debug_literal (repr),
                   debug_literal (lit), lit_found);
            unsigned not_repr = negate (repr);
            for (size_t i = start; i != stack_size; i++) {
              other = stack[i];
              reaches[other] = INVALID_LITERAL;
              if (other == repr)
                continue;
              debug ("found literal equivalence %s = %s",
                     debug_literal (other), debug_literal (repr));
              if (other == not_repr) {
                debug ("found inconsistent component");
                trace_unit ('a', repr);
                trace_empty ('a');
                inconsistent = true;
                goto DONE;
              }
              assert (reprs[other] == other);
              assert (repr < other);
              reprs[other] = repr;
#ifndef QUIET
              statistics.equivalent_literals++;
#endif
              equivalences++;
            }
          } else {
            reaches[lit] = INVALID_LITERAL;
            debug ("trivial component %zu of size one with found[%s] = %u",
                   components, debug_literal (lit), lit_found);
          }
          stack.resize (start);
        } else
          reaches[lit] = lit_reaches;
        unsigned not_root = negate (root);
        if (lit == not_root) {
          debug ("root literal %s implies its negation %s",
                 debug_literal (root), debug_literal (not_root));
          unsigned unit = not_root;
          debug ("adding unit %s", debug_literal (unit));
          units.push_back (unit);
          trace_unit ('a', unit);
        }
      } else if (found[lit])
        work.pop_back ();
      else {
        stack.push_back (lit);
        reaches[lit] = found[lit] = ++count;
        assert (count);
        work.push_back (INVALID_LITERAL);
        unsigned not_lit = negate (lit);
        const auto watches = references (not_lit);
        auto p = watches->begin (), end = watches->end ();
        while (p != end) {
          Reference ref = *p++;
          if (tagged (ref)) {
            unsigned other = untag (ref);
            if (!found[other] && active (other))
              work.push_back (other);
          } else
            p++;
        }
      }
    }
  }

  assert (!(equivalences & 1));

#ifndef QUIET
  if (verbosity > 0)
#endif
#if !defined(QUIET) || !defined(NDEBUG)
  {
    for (auto root : literals)
      if (active (root) && !found[root])
        components++;
    assert (!(components & 1));
  }
#else
  (void) components;
#endif
#ifndef QUIET
  message (2,
           "decomposed binary implication graph into %zu components %.0f%%",
           components, percent (components, literals));

  message (1, "found %zu equivalent literals (%zu variables)", equivalences,
           equivalences / 2);
#endif

DONE:

  free (reaches);
  free (found);

  return equivalences / 2;
}

void Solver::trace_adding_equivalences (unsigned *reprs) {
  if (inconsistent)
    return;
  if (!proof)
    return;
  for (auto lit : literals) {
    unsigned repr = reprs[lit];
    if (lit != repr)
      trace_binary ('a', negate (repr), lit);
  }
}

void Solver::mark_substituted (unsigned *reprs) {
  if (inconsistent)
    return;
  for (auto idx : variables) {
    unsigned lit = literal (idx);
    unsigned repr = reprs[lit];
    if (lit == repr)
      continue;
    Flags &f = flags[idx];
    assert (f.active);
    assert (!f.eliminated);
    assert (!f.fixed);
    assert (!f.substituted);
    f.substituted = true;
#ifndef QUIET
    statistics.substituted_variables++;
#endif
    deactivate_variable (idx);
    unsigned not_lit = negate (lit);
    unsigned not_repr = negate (repr);
    reconstruction_stack.push_back (INVALID_LITERAL);
    reconstruction_stack.push_back (not_lit);
    reconstruction_stack.push_back (repr);
    debug_literals (2, &*reconstruction_stack.end () - 2, "saved");
    reconstruction_stack.push_back (INVALID_LITERAL);
    reconstruction_stack.push_back (lit);
    debug_literals (1, &*reconstruction_stack.end () - 1, "saved");
    trace_binary ('d', not_repr, lit);
    trace_binary ('d', repr, not_lit);
  }
}

void Solver::substitute_binary_clauses (unsigned *reprs) {
  if (inconsistent)
    return;

  size_t substituted = 0, kept = 0, trivial = 0, flushed = 0;

  std::vector<LiteralBlocking> delayed;

  for (auto watches : matrix) {
    unsigned lit = watches->lit;
    if (lit == INVALID_LITERAL)
      continue;
    unsigned lit_repr = reprs[lit];
    auto begin = watches->begin ();
    auto end = watches->end ();
    auto q = begin, p = q;
    while (p != end) {
      Reference ref = *p++;
      if (tagged (ref)) {
        unsigned other = untag (ref);
        unsigned other_repr = reprs[other];
        if (lit_repr == other_repr) {
          if (lit < other) {
            unsigned unit = lit_repr;
            debug ("substitution of binary clause %s %s yields unit %s",
                   debug_literal (lit), debug_literal (other),
                   debug_literal (unit));
            debug ("adding unit %s", debug_literal (unit));
            units.push_back (unit);
            trace_unit ('a', unit);
            mark_binary_garbage (lit, other, "trivial after substitution");
          }
        } else if (lit_repr == negate (other_repr)) {
          if (lit < other) {
            debug ("substitution of binary clause %s %s yields "
                   "trivial clause %s %s",
                   debug_literal (lit), debug_literal (other),
                   debug_literal (lit_repr), debug_literal (other_repr));
            mark_binary_garbage (lit, other, "trivial after substitution");
            trivial++;
          }
        } else if (lit == lit_repr && other == other_repr) {
          debug_binary (lit, other, "keeping");
          *q++ = ref;
          kept++;
        } else {
          if (lit < other) {
            debug ("substitution of binary clause %s %s yields "
                   "substituted binary clause %s %s",
                   debug_literal (lit), debug_literal (other),
                   debug_literal (lit_repr), debug_literal (other_repr));
            trace_binary ('a', lit_repr, other_repr);
            trace_binary ('d', lit, other);
            substituted++;
          }
          if (lit == lit_repr)
            *q++ = tag_binary (other_repr);
          else {
            assert (lit_repr < lit);
            delayed.push_back (LiteralBlocking (lit_repr, other_repr));
          }
        }
      } else
        p++, flushed++;
    }
    watches->resize (q - begin);
  }

  for (auto &p : delayed)
    watch_binary (p.lit, p.blocking);

#ifndef QUIET
  message (2, "kept %zu binary clause references", kept);
  message (2, "substituted %zu binary clause references", substituted);
  message (2,
           "substitution of binary clauses "
           "yields %zu units and %zu trivial clauses",
           units.size (), trivial);
  message (2, "flushed %zu large clause references", flushed);
#else
  (void) kept;
  (void) substituted;
  (void) trivial;
  (void) flushed;
#endif
}

void Solver::substitute_large_clauses (unsigned *reprs) {

  if (inconsistent)
    return;

  assert (!level);
  assert (watching);

  size_t satisfied = 0, trivial = 0, binary = 0;
  size_t kept = 0, substituted = 0, total = 0;

  Clause *begin = *clauses.begin ();
  Clause *end = *clauses.end ();

  Clause *src = begin, *dst = begin;

  reset_positions ();
  while (src != end) {

    total++;

    bool changed = false;

    Clause *next_src = src->next ();
    if (src->garbage)
      goto CONTINUE;

    assert (clause.empty ());

    for (auto lit : *src) {
      signed char value = values[lit];
      if (value < 0) {
        changed = true;
        continue;
      }
      if (value > 0) {
        mark_clause_garbage (src, "satisfied");
#ifndef QUIET
        statistics.collected_clauses++;
#endif
        satisfied++;
        goto CONTINUE;
      }
      unsigned repr = reprs[lit];
      assert (!values[repr]);
      signed char mark = marks[repr];
      if (mark > 0) {
        changed = true;
        continue;
      }
      if (mark < 0) {
        mark_clause_garbage (
            src, "as substitution produces clashing literal in");
#ifndef QUIET
        statistics.collected_clauses++;
#endif
        trivial++;
        goto CONTINUE;
      }
      clause.push_back (repr);
      marks[repr] = 1;
      unsigned not_repr = negate (repr);
      assert (!marks[not_repr]);
      marks[not_repr] = -1;
      if (repr != lit)
        changed = true;
    }

    {
      size_t dst_size = clause.size ();

      if (dst_size == 1) {
        unsigned unit = clause[0];
        trace_unit ('a', unit);
        mark_clause_garbage (src,
                             "as substitution produces unit clause from");
        units.push_back (unit);
#ifndef QUIET
        statistics.collected_clauses++;
#endif
        unit++;
        goto CONTINUE;
      }

      if (dst_size == 2) {
        trace_temporary ('a');
        new_binary ();
        mark_clause_garbage (src,
                             "as substitution produces binary clause from");
#ifndef QUIET
        statistics.collected_clauses++;
#endif
        binary++;
        goto CONTINUE;
      }

      assert (dst_size > 2);

      if (changed) {
        substituted++;
        trace_temporary ('a');
        trace_clause ('d', src);
      } else
        kept++;

      *dst = *src;

      if (changed) {
#ifndef NREPLACEMENTCACHE
        dst->replaced = 2;
#endif
        dst->size = dst_size;
#ifndef NGLUE
        adjust_glue_to_updated_size (dst);
#endif
      }

      memcpy (dst->literals, &clause[0], dst_size * 4);

      watch_clause (dst);
      update_position (dst);

      dst->shortened = false;
      dst = dst->next ();
    }

  CONTINUE:

    for (auto lit : clause)
      marks[lit] = marks[negate (lit)] = 0;

    clause.clear ();
    src = next_src;
  }

  clauses.resize (dst);
  finalize_position_update ();

#ifndef QUIET
  message (2, "collected satisfied %zu clauses %.0f%%", satisfied,
           percent (satisfied, total));
  message (2, "collected %zu large clauses which became trivial  %.0f%%",
           trivial, percent (trivial, total));
  message (2, "collected %zu large clauses which became unit %.0f%%",
           units.size (), percent (units.size (), total));
  message (2, "collected %zu large clauses which became binary %.0f%%",
           binary, percent (binary, total));
  message (1, "substituted %zu large clauses %.0f%%", substituted,
           percent (substituted, total));
  message (1, "kept %zu large clauses %.0f%%", kept, percent (kept, total));
#else
  (void) satisfied;
  (void) total;
  (void) trivial;
  (void) binary;
  (void) substituted;
  (void) kept;
#endif
}

bool Solver::substitute_variables () {

  if (level && !backtrack_to_root_level_and_propagate ())
    return false;

  check_all ();

  assert (watching);
  assert (!level);
  assert (units.empty ());

  start_simplifier ("substitute");

  SubstitutionScheduler substitution_scheduler (this);
  substitution_scheduler.increment_counter ();

  unsigned *reprs = (unsigned *) malloc ((size_t) literals * sizeof *reprs);
  if (literals && !reprs)
    fatal ("out-of-memory allocating representatives array");

  bool progress = false;
  unsigned round = 1;

  for (;;) {

    message (2, "substitution round %u", round);

    for (auto lit : literals)
      reprs[lit] = lit;
    unsigned substituted = decompose_binary_implication_graph (reprs);
    trace_adding_equivalences (reprs);
    substitute_binary_clauses (reprs);
    substitute_large_clauses (reprs);
    mark_substituted (reprs);
    bool report = substituted || inconsistent || !units.empty ();
#ifndef QUIET
    if (verbosity > 0)
      report = true;
#endif
    if (report)
      report_progress ('d');
    if (!substituted)
      break;
    if (inconsistent)
      break;
    if (terminated)
      break;
    if (round >= OPTION (substitution_rounds))
      break;
    if (!units.empty ()) {
      assign_and_propagate_units ();
      if (inconsistent)
        break;
    }
    round++;
  }

  free (reprs);

  substitution_scheduler.set_next_conflict_limit ();

  stop_simplifier ("substitute");
  propagate_after_simplification ();

  check_all ();

  return progress;
}

}; // namespace TabularaSAT
