#include "inline.hpp"
#include "solver.hpp"

namespace TabularaSAT {

// Report result of propagating a learned unit clause.

void Solver::report_propagation_of_learned_unit_clause () {
  assert (iterating);
  iterating = false;
  report_progress ('i');
}

#ifndef QUIET

class Reporter {

  struct Cell {
    const char *name;
    int column;
  };

  bool header;

  static const unsigned capacity = 32;
  Cell cells[capacity];
  unsigned size = 0;

  char line[128];
  int column = 0;

public:
  Reporter (bool h) : header (h) {}

  void add (const char *name, const char *format, double d) {
    if (!column)
      line[column++] = '@';
    line[column++] = ' ';
    assert ((size_t) column < sizeof line);
    char *value = line + column;
    size_t remaining = sizeof line - column;
    snprintf (value, remaining, format, d);
    int length_value = strlen (value);
    assert (column + length_value < (int) sizeof line);
    if (header) {
      assert (size < capacity);
      Cell &cell = cells[size++];
      cell.name = name;
      int length_name = strlen (name);
      if (length_name <= length_value)
        cell.column = column + (length_value - length_name + 1) / 2;
      else
        cell.column = column - (length_name - length_value + 1) / 2;
    }
    column += length_value;
  }

  void print (char type) {
    if (header) {
      const int rows = 3;
      int column[rows], row = 0;
      char line[rows][128];
      for (int i = 0; i != rows; i++)
        column[i] = 0;
      for (Cell *c = cells; c != cells + size; c++) {
        if (column[row])
          line[row][column[row]++] = ' ';
        while (column[row] < c->column)
          line[row][column[row]++] = ' ';
        for (const char *p = c->name; *p; p++)
          line[row][column[row]++] = *p;
        if (++row == rows)
          row = 0;
      }
      fputs ("c\n", stdout);
      for (int i = rows - 1; i >= 0; i--) {
        fputs ("c ", stdout);
        for (int j = 0; j != column[i]; j++)
          fputc (line[i][j], stdout);
        fputc ('\n', stdout);
      }
      fputs ("c\n", stdout);
    }
    fputs ("c ", stdout);
    line[0] = type;
    fputs (line, stdout);
    fputc ('\n', stdout);
  }
};

void Solver::report_progress (char type) {

#ifndef QUIET
  if (verbosity < 0)
    return;
#endif

  Statistics &s = statistics;

  double seconds = process_time ();
  double MB = current_resident_set_size () / (double) (1 << 20);

  Reporter reporter (!(s.reported & 15));
  s.reported++;

  reporter.add ("seconds", "%0.2f", seconds);
  reporter.add ("MB", "%.0f", MB);
#ifndef QUIET
  double level = averages ().level;
  reporter.add ("level", "%.0f", level);
#endif
#ifndef NSWITCH
  if (OPTION (switch_search_mode))
    reporter.add ("switched", "%.0f", s.switched_modes);
#endif
#ifndef QUIET
#ifndef VANILLA
  if (OPTION (reduce_vanilla))
#endif
    reporter.add ("increased", "%.0f", s.increased_kept);
#endif
  reporter.add ("reductions", "%.0f", s.reductions);
  reporter.add ("restarts", "%.0f", s.restarts);
#ifndef NBUMPREASONS
#if 0
  double rate = averages[stable].decision_rate;
  reporter.add ("rate", "%.1f", rate);
#endif
#endif
  reporter.add ("conflicts", "%.0f", s.conflicts);
  reporter.add ("redundant", "%.0f", s.redundant_clauses);
  reporter.add ("binary", "%.0f", s.binary_clauses);
  reporter.add ("irredundant", "%.0f", s.irredundant_clauses);
#ifndef QUIET
  double trail = averages ().trail;
  reporter.add ("trail", "%.0f%%", trail);
#endif
#ifndef NGLUE
  double glue = averages ().glue_slow;
  reporter.add ("glue", "%.0f", glue);
#endif
#if 0
  double garbage = percent (matrix.garbage, matrix.arena.size ());
  reporter.add ("garbage-references", "%.0f%%", garbage);
#endif
#ifndef QUIET
  unsigned remaining = s.active_variables;
  reporter.add ("variables", "%.0f", remaining);
  reporter.add ("remaining", "%.0f%%", percent (remaining, variables));
#endif

  reporter.print (type), fflush (stdout);
}

#endif

}; // namespace TabularaSAT
