#include "schedule.hpp"
#include "solver.hpp"

#include <cinttypes>

namespace TabularaSAT {

void Scheduler::increment_counter (const char *prefix) {
  if (solver.preprocessing) {
    uint64_t *p = preprocessed ();
    if (p)
      *p += 1;
  }

  count () += 1;
#ifndef QUIET
  if (solver.verbosity >= 2) {
    if (solver.preprocessing)
      solver.message (2, "%s%s %" PRIu64 " during preprocessing round %u",
                      prefix, name (), count (), solver.preprocessing);
    else
      solver.message (2, "%s%s %" PRIu64 " at conflict %" PRIu64, prefix,
                      name (), count (),
                      (uint64_t) solver.statistics.conflicts);
  }
#endif
}

void Scheduler::set_initial_conflict_limit () {
#ifndef NDEBUG
  if (SOLVER_OPTION (preprocessing)) {
    uint64_t *p = preprocessed ();
    uint64_t d = p ? *p : 0;
    uint64_t c = count ();
    assert (c == d);
  } else
    assert (!count ());
#endif
  limit () = scale () * initial ();
#ifndef QUIET
  solver.message (1, "initial %s limit of %" PRIu64 " conflicts", name (),
                  limit ());
#endif
}

void Scheduler::set_next_conflict_limit () {
  if (solver.preprocessing)
    return;
  uint64_t *p = preprocessed ();
  uint64_t d = p ? *p : 0;
  uint64_t c = count ();
  assert (c >= d);
  c -= d;
  uint64_t delta = scale () * function (c) * interval ();
  limit () = GET_SOLVER_COUNT (conflicts) + delta;
#ifndef QUIET
  solver.message (2,
                  "new %s limit at %" PRIu64 " after %" PRIu64 " conflicts",
                  name (), limit (), delta);
#endif
}

uint64_t Effort::get_ticks_limit () {
  assert (!solver.preprocessing);
  uint64_t search_ticks = GET_SOLVER_COUNT (search_ticks);
  uint64_t delta_ticks = search_ticks - last ();
  last () = search_ticks;
  double fraction = effort () / 1000.0;
  uint64_t limit = fraction * delta_ticks;
  if (limit < SOLVER_OPTION (minimum_effort))
    limit = SOLVER_OPTION (minimum_effort);
#ifndef QUIET
  solver.message (2, "%s effort of %.1f%% and %" PRIu64 " ticks", name (),
                  fraction * 100, limit);
#endif
  return limit;
}

double Solver::scale () {
  uint64_t irredundant = GET_COUNT (irredundant_clauses);
  uint64_t binary = GET_COUNT (binary_clauses);
  uint64_t clauses = irredundant + binary;
  return logpow (clauses + !clauses, 2);
}

}; // namespace TabularaSAT
