#include "clause.hpp"
#include "cover.hpp"
#include "radix.hpp"
#include "references.hpp"
#include "schedule.hpp"
#include "solver.hpp"

#include <algorithm>
#include <cinttypes>

namespace TabularaSAT {

bool Solver::vivifying_clauses () {
  if (!OPTION (inprocessing))
    return false;
  if (!OPTION (vivify_clauses))
    return false;
  if (GET_COUNT (conflicts) < limits.vivify)
    return false;
  if (!trigger.vivification) {
    if (OPTION (trigger_vivification)) {
      debug ("waiting for next reduction before vivification");
      trigger.vivification = 1;
    } else
      trigger.vivification = 2;
  }
  return trigger.vivification == 2;
}

unsigned *Solver::count_occurrences () {

  unsigned *counts = (unsigned *) calloc (literals, sizeof *counts);
  if (literals && !counts)
    fatal ("out-of-memory allocating counters");

  for (auto watches : matrix) {
    unsigned lit = watches->lit;
    if (lit == INVALID_LITERAL)
      continue;
    auto begin = watches->begin (), end = watches->end ();
    for (auto p = begin; p != end; p++) {
      Reference ref = *p;
      if (tagged (ref)) {
        unsigned other = untag (ref);
        if (lit < other)
          increment_saturating (counts[lit]),
              increment_saturating (counts[other]);
      } else
        p++;
    }
  }

  for (auto c : clauses)
    for (auto lit : *c)
      if (!c->garbage)
        increment_saturating (counts[lit]);

  return counts;
}

#ifndef QUIET

static const char *tier_name (unsigned tier) {
  switch (tier) {
  case 1:
    return "redundant tier 1";
  case 2:
    return "redundant tier 2";
  default:
    assert (!tier);
    return "irredundant";
  }
}

#endif

void Solver::gather_vivification_candidates (
    unsigned tier, unsigned *counts, std::vector<Clause *> &candidates) {

  if (tier == 2 && first_tier2 == INVALID_POSITION)
    return;

  if (tier == 1 && first_redundant == INVALID_POSITION)
    return;

  unsigned tier1 = OPTION (tier1_max_glue);
  unsigned tier2 = OPTION (tier2_max_glue);

  size_t max_count = 0;
  for (auto lit : literals)
    if (counts[lit] > max_count)
      max_count = counts[lit];

  unsigned low = 0, high = 0;
  if (tier == 1)
    low = 0, high = tier1;
  if (tier == 2)
    low = tier1 + 1, high = tier2;

  if (tier)
    message (3, "%s vivification candidate glue ranges from %u to %u",
             tier_name (tier), low, high);

  unsigned *tmp[2];
  tmp[0] = (unsigned *) calloc (max_count + 1, sizeof tmp[0]);
  tmp[1] = (unsigned *) calloc (max_count + 1, sizeof tmp[1]);
  if (!tmp[0] || !tmp[1])
    fatal ("out-of-memory during sorting vivification candidates");

  // An earlier version solving 'GP_...' benchmarks from SAT Competition
  // 2022 spent 60% of vivification time on sorting the candidate clauses
  // (actually on accessing the clauses during a comparison functions).
  // Therefore we replaced generic sorting by a dedicated radix sort
  // implementation, which puts rescheduled clauses with their 'vivify'
  // bit first and then sorts by the maximal occurring literal in the
  // clause (cached in 'searched').  As radix sort is stable the clause
  // order and thus the age of the clause remains as tie breaker.

  // The earlier version also took 'size' and 'used' into account which
  // would require two more passes on the clauses.

  size_t rescheduled = 0, gathered = 0;

  Clause *begin, *end;
  if (tier == 1)
    begin = clauses[first_redundant], end = *clauses.end ();
  else if (tier == 2)
    begin = clauses[first_tier2], end = *clauses.end ();
  else {
    begin = *clauses.begin ();
    end = clauses[last_irredundant];
    if (end != *clauses.end ())
      end = end->next ();
  }

  // In the first round of radix sort we count the number of candidates
  // with a given 'vivify' bit and 'searched' count.  This also gives the
  // number of 'gathered' candidates.  In the second round we first
  // compute the starting position of the bucket of clauses with the same
  // 'vivify' bit and 'search' count and then simple put the clauses in
  // the vector at the right position increasing the position of the
  // buckets appropriately.

  for (unsigned counted = 0; counted != 2; counted++) {

    if (counted) {

      candidates.resize (gathered);

      assert (!tmp[0][0]);
      assert (!tmp[1][0]);

      unsigned pos = 0;
      for (int vivify = 1; vivify >= 0; vivify--)
        for (unsigned i = max_count; i; i--) {
          unsigned count = tmp[vivify][i];
          tmp[vivify][i] = pos;
          pos += count;
        }
    }

    // TODO check whether reverse tie breaking on clause order is better
    // (would require to traverse the clauses in reverse order simply).

    for (auto c = begin; c != end; c = c->next ()) {

      if (c->garbage)
        continue;
      if (tier) {
        if (!c->redundant)
          continue;
        if (c->glue < low)
          continue;
        if (c->glue > high)
          continue;
      } else if (c->redundant)
        continue;

      unsigned vivify = c->vivify;
      if (counted) {
        unsigned count = c->replaced;
        unsigned pos = tmp[vivify][count];
        assert (pos < candidates.size ());
        candidates[pos++] = c;
        tmp[vivify][count] = pos;
      } else {
        rescheduled += !!vivify;
        unsigned best_count = 0;
        unsigned best = INVALID_LITERAL;
        for (auto lit : *c) {
          unsigned lit_count = counts[lit];
          if (best == INVALID_LITERAL || lit_count > best_count)
            best = lit, best_count = lit_count;
        }
        assert (best != INVALID_LITERAL);
        debug_clause (c, "vivification candidate with best counts[%s] = %u",
                      debug_literal (best), best_count);
        c->replaced = best_count;

        assert (best_count <= max_count);
        tmp[c->vivify][best_count]++;

        assert (gathered < UINT_MAX);
        gathered++;
      }
    }
  }

  free (tmp[0]);
  free (tmp[1]);

#ifndef QUIET
  size_t total =
      tier ? statistics.redundant_clauses : statistics.irredundant_clauses;
  message (2, "rescheduling %zu %s vivification candidates %.0f%%",
           rescheduled, tier_name (tier), percent (rescheduled, gathered));
  message (1, "gathered %zu %s vivification candidates %.0f%%", gathered,
           tier_name (tier), percent (gathered, total));
#endif
  // Need to reset the cached best count misused 'searched' field and at
  // the same time mark all clauses as to be vivified next time if not
  // checked in this round.

  if (rescheduled)
    for (auto c : candidates)
      c->replaced = 2;
  else
    for (auto c : candidates)
      c->replaced = 2, c->vivify = true;
}

inline void Solver::vivify_analyze_literal (unsigned lit) {
  unsigned idx = index (lit);
  Reference reason = reasons[idx];
  if (reason == UNIT_REASON)
    return;
  Flags &f = flags[idx];
  if (f.seen)
    return;
  assert (values[lit] < 0);
  f.seen = true;
  analyzed.push_back (idx);
  debug ("analyzing vivification literal %s", debug_literal (lit));
  if (reason == DECISION_REASON)
    clause.push_back (lit);
}

struct more_counts {
  const unsigned *const counts;
  more_counts (const unsigned *cs) : counts (cs) {}
  bool operator() (unsigned a, unsigned b) {
    unsigned c = counts[a];
    unsigned d = counts[b];
    if (c > d)
      return true;
    if (c < d)
      return false;
    return a < b;
  }
};

struct rank_more_counts {
  const unsigned *const counts;
  rank_more_counts (const unsigned *cs) : counts (cs) {}
  unsigned operator() (unsigned a) { return ~counts[a]; }
};

bool Solver::vivify_clause (unsigned tier, unsigned *counts,
                            Clause *candidate) {

  assert (watching);
  assert (vivifying);
  assert (!inconsistent);

  debug_clause (candidate, "vivifying %s", tier_name (tier));
#ifndef QUIET
  statistics.vivify_checks++;
#endif
  candidate->vivify = false; // Mark as checked for vivification.

  // First check whether the candidate clause is root-level satisfied and
  // should be ignored and all is root-level unassigned literals on the
  // temporary clause stack.  At the same time compute check whether the
  // candidates is currently used as reason for a literal.

  unsigned used_as_reason_on_level = INVALID_LEVEL;
  Reference cand_ref = reference_clause (candidate);

  assert (clause.empty ());
  for (auto lit : *candidate) {
    signed char value = values[lit];
    unsigned idx = index (lit);
    unsigned lit_level = levels[idx];
    if (value && !lit_level) {
      if (value < 0)
        continue;
      assert (value > 0);
      mark_clause_garbage (candidate, "satisfied");
      unwatch_clause (candidate);
      clause.clear ();
      return false;
    }
    clause.push_back (lit);
    if (value <= 0)
      continue;
    if (used_as_reason_on_level != INVALID_LEVEL)
      continue;
    Reference reason = reasons[idx];
    if (reason != cand_ref)
      continue;
    assert (lit_level != INVALID_LEVEL);
    used_as_reason_on_level = lit_level;
  }

  // As we are trying to reuse decisions and propagations across several
  // candidate clauses (by just not backtracking) it might be the case
  // that the candidate clause is a reason of some of its literal.  If
  // this is the case we first have to backtrack before that assignment
  // level.

  if (used_as_reason_on_level != INVALID_LEVEL) {
    debug ("vivification candidate used as reason on level %u",
           used_as_reason_on_level);
    assert (used_as_reason_on_level);
    backtrack (used_as_reason_on_level - 1);
  }

  // The literal with maximum count is used as sorting criteria when
  // determining the order of candidate clauses.  Here we additionally
  // sort the literals of the clause (stored in the temporary clause) with
  // respect to their occurrence count.  This should both give more
  // propagation and higher chances of success as well as more sharing of
  // decisions and propagations across vivifying consecutive candidates.

  if (clause.size () < OPTION (radix_sort_limit))
    std::sort (clause.begin (), clause.end (), more_counts (counts));
  else {
    radix_sort_unsigned (clause, rank_less_unsigned ());
    radix_sort_unsigned (clause, rank_more_counts (counts));
  }

#ifdef LOGGING
  if (logging ()) {
    debug_prefix ();
    printf ("sorted vivification candidate clause[%" PRIu64 "]@%" PRIu64,
            candidate->id, clauses.position (candidate));
    for (auto lit : clause)
      printf (" %s#%u", debug_literal (lit), counts[lit]);
    debug_suffix ();
  }
#endif

  // When we try to reuse decisions and propagations we simply find the
  // longest matching prefix of the negations in the candidate clause
  // with the current decisions.  Note that counting in the clause starts
  // with zero while decision levels start at one.

  if (level)
    for (size_t i = 0; i < level && i != clause.size (); i++) {
      unsigned lit = clause[i];
      unsigned not_lit = negate (lit);
      assert (i + 1 < control.size ());
      unsigned decision = control[i + 1].decision;
      if (decision == not_lit)
        debug ("reusing vivification decision %s",
               debug_literal (decision));
      else {
        backtrack (i);
        break;
      }
    }

  // Here comes the actual vivification loop, which assumes the negation
  // of the literals in the clause as decisions and propagates.  If this
  // yields to a conflict or one of the literals is already satisfied we
  // move on to the conflict analysis below.  Counting the number of
  // falsified literals allows us to cheaply determine, whether the clause
  // can be shrunken without conflict analysis in case all clause literals
  // can be set to false and propagation succeeds too.

  unsigned satisfied_literal = INVALID_LITERAL; // Found satisfied literal.
  unsigned num_falsified = 0; // Skipped implied falsified literals.
  bool conflicting = false;   // Propagation result in a conflict.

  for (size_t i = 0; i != clause.size (); i++) {
    unsigned lit = clause[i];
    signed char value = values[lit];

    // We make sure that the first 'level' literals of the clause are
    // currently the same as the first 'level' literals (in the same
    // order) and thus we can reuse those decisions as well as their
    // propagation.

    if (i < level) {
#ifndef NDEBUG
      unsigned idx = index (lit);
      unsigned lit_level = levels[idx];
      assert (lit_level == i + 1);
      assert (lit_level <= level);
      Reference reason = reasons[idx];
      assert (reason == DECISION_REASON);
      assert (value < 0);
#endif
      continue;
    }

    // Found a literal which is negatively implied and thus we can shrink
    // the clause in case by at least such a literal.

    if (value < 0) {
      num_falsified++;
      continue;
    }

    // The clause contains a literal which is positively implied.  This is
    // for instance the case if another subsuming clause exists which
    // implies the last remaining literal in the clause not yet assumed.

    if (value > 0) {
      satisfied_literal = lit;
      break;
    }

    // The default case is that the candidate clause literal is still
    // unsatisfied and thus we assume its negation and propagate.

    unsigned not_lit = negate (lit);
    decide_literal (not_lit);

    // Propagate the assumption but ignore the candidate clause.

    if (!vivify_propagate (candidate)) {
      conflicting = true;
      break;
    }
  }

  assert (analyzed.empty ()); // Prepare conflict analysis.

  // If the analyzed/learned clause at the end turns out to be smaller
  // than the (reduced) candidate clause (without root-level falsified
  // literals) then we will shrink the candidate clause.

  unsigned old_size = clause.size ();
  clause.clear ();

  bool redundant = false; // Analysis resolved a redundant clause.
  bool vivified = true;   // Candidate vivified (result of this function)?

  // In case we can not shrink the clause but find a clause which subsumes
  // the candidate clause (or more important has exactly the same
  // literals) then we can still recycle the candidate clause.

  Reference subsuming_ref = INVALID_REFERENCE;

  // The actual redundancy check succeeds if a resolved clause contains
  // only literals from the candidate clause, which we implement by
  // marking the literals in the candidate clause.

  for (auto lit : *candidate)
    assert (!marks[lit]), marks[lit] = 1;

  // The subsumption check only needs to be applied to the conflicting
  // clause or to the reason of the satisfied literal.  In order to check
  // for the later we remember the index of the satisfied literal.

  unsigned satisfied_index = INVALID_VARIABLE;

  // Subsumption is tricky though if the candidate is irredundant, because
  // then we have to make sure to promote the subsuming clause to an
  // irredundant, if it was redundant.

  bool redundant_subsuming = false;

  // Conflict analysis only triggers if propagation failed or a satisfied
  // literal is found.  It is initialized by analyzing the conflict or
  // pushing the satisfied literal index on the 'analyzed' stack.  In both
  // cases we also should check for subsumption.  For conflicts it is
  // checked here and for satisfied literals during conflict analysis.

  if (conflicting) {

    debug_clause (candidate,
                  "vivification analysis after failed propagation for");

    if (tagged (conflict)) {
      assert (falsified_literal != INVALID_LITERAL);
      vivify_analyze_literal (falsified_literal);
      unsigned lit = untag (conflict);
      vivify_analyze_literal (lit);
      if (marks[falsified_literal] && marks[lit]) {
        debug_binary (falsified_literal, lit,
                      "conflicting vivify subsuming");
        assert (!redundant_subsuming);
        subsuming_ref = conflict;
      }
    } else {
      Clause *resolved = dereference_clause (conflict);
      INC_COUNT (vivification_ticks);
      subsuming_ref = conflict;
      assert (subsuming_ref != INVALID_REFERENCE);
      for (auto lit : *resolved) {
        vivify_analyze_literal (lit);
        if (subsuming_ref != INVALID_REFERENCE && !marks[lit] &&
            !fixed (lit))
          subsuming_ref = INVALID_REFERENCE;
      }
      if (subsuming_ref != INVALID_REFERENCE) {
        debug_clause (resolved, "conflicting vivify subsuming");
        assert (subsuming_ref == conflict);
        redundant_subsuming = resolved->redundant;
      } else if (resolved->redundant)
        redundant = true;
    }

    vivified = true;

  } else if (satisfied_literal != INVALID_LITERAL) {

    // Found one literal in the candidate clause, which is implied
    // positively by the negation of other literals in the clause.
    // Then we start the analysis by pushing this literal on the
    // 'analyzed' stack such that the analysis will start with its reason
    // clause.

    debug_clause (candidate, "vivification analysis after %s satisfied in",
                  debug_literal (satisfied_literal));
    clause.push_back (satisfied_literal);
    satisfied_index = index (satisfied_literal);
    assert (!flags[satisfied_index].seen);
    flags[satisfied_index].seen = true;
    analyzed.push_back (satisfied_index);

    vivified = true;

  } else if (num_falsified) {

    // Propagation of the negated literals of the clause did not yield a
    // conflict, none of is implied to true, but at least one to false.
    // Then at least all those implied to false can be removed and we also
    // do not need a conflict analysis.

    debug_clause (candidate,
                  "vivification analysis after %u literals implied "
                  "to be falsified in",
                  num_falsified);

    for (auto lit : *candidate) {
      assert (values[lit] < 0);
      unsigned idx = index (lit);
      Reference reason = reasons[idx];
      if (reason == DECISION_REASON)
        clause.push_back (lit);
    }

    vivified = true;

  } else {
    debug_clause (candidate, "vivification failed for");
    vivified = false;
  }

  // Perform conflict analysis if the conflict does not subsume the
  // candidate clause or a satisfying literal was found.  In contrast to
  // the standard CDCL conflict analysis we continue resolving reasons
  // until only decisions are left.  We use a simple BFS do to that
  // using the 'analyzed' stack anyhow needed to remember 'seen'
  // variables.

  if (subsuming_ref == INVALID_REFERENCE &&
      (conflicting || satisfied_literal != INVALID_LITERAL)) {

    size_t next = 0;
    while (next != analyzed.size ()) {
      unsigned idx = analyzed[next++];
      Reference reason = reasons[idx];
      assert (reason != UNIT_REASON);
      if (reason == DECISION_REASON)
        continue;
      if (tagged (reason)) {
        unsigned other = untag (reason);
        vivify_analyze_literal (other);
        if (idx == satisfied_index) {
          assert (subsuming_ref == INVALID_REFERENCE);
          if (marks[satisfied_literal] && marks[other]) {
            debug_binary (other, satisfied_literal,
                          "satisfied literal vivify subsuming");
            subsuming_ref = reason;
            break; // No need to continue analysis.
          }
        }
      } else {
        subsuming_ref = reason;
        assert (subsuming_ref != INVALID_REFERENCE);
        Clause *resolved = dereference_clause (reason);
        INC_COUNT (vivification_ticks);
        for (auto other : *resolved) {
          if (other == satisfied_literal)
            assert (marks[other]);
          else {
            vivify_analyze_literal (other);
            if (subsuming_ref != INVALID_REFERENCE && !marks[other] &&
                !fixed (other))
              subsuming_ref = INVALID_REFERENCE;
          }
        }
        if (subsuming_ref != INVALID_REFERENCE) {
          debug_clause (resolved, "satisfied literal vivify subsuming");
          assert (subsuming_ref == reason);
          redundant_subsuming = resolved->redundant;
          break; // No need to continue analysis.
        }
        if (resolved->redundant)
          redundant = true;
      }
    }
  }

  // Reset the conflict analysis.

  for (auto idx : analyzed)
    flags[idx].seen = false;
  analyzed.clear ();

  // Also reset the marks of the literals in the candidate clause for
  // subsumption checking.

  for (auto lit : *candidate)
    assert (marks[lit]), marks[lit] = 0;

  // We always prefer subsumption over shrinking.  Actually conflict
  // analysis is aborted if the subsumption check for the reason of the
  // satisfied literal in the candidate clause success.

  if (subsuming_ref != INVALID_REFERENCE) {

    // We only figure out that the candidate clause is subsumed if a
    // conflict occurs or a satisfied literal is found, in which case we
    // already declared that the candidate can be vivified.

    assert (vivified);

    mark_clause_garbage (candidate, "vivified subsumed");
    unwatch_clause (candidate);

    if (level)
      backtrack (0);

    if (!candidate->redundant) {

      // If the subsumed candidate clause is irredundant but the subsuming
      // not we need to make the subsuming clause irredundant.

      if (redundant_subsuming) {
        assert (!tagged (subsuming_ref));
        Clause *subsuming_clause = dereference_clause (subsuming_ref);
        assert (subsuming_clause->redundant);
        subsuming_clause->redundant = false;
        update_position (subsuming_clause);
        DEC_COUNT (redundant_clauses);
        INC_COUNT (irredundant_clauses);
      }

      // For an irredundant clause we want to mark literals removed and
      // added. This is achieved by marking all subsuming clause literals
      // and then checking the candidate clause for those marked literals.

      // Mark all literals of the subsuming clause.

      if (falsified_literal != INVALID_LITERAL) {
        assert (!marks[falsified_literal]);
        marks[falsified_literal] = 1;
      }
      if (satisfied_literal != INVALID_LITERAL) {
        assert (!marks[satisfied_literal]);
        marks[satisfied_literal] = 1;
      }
      if (tagged (subsuming_ref)) {
        unsigned lit = untag (subsuming_ref);
        assert (!marks[lit]);
        marks[lit] = 1;
      } else {
        Clause *subsuming_clause = dereference_clause (subsuming_ref);
        for (auto lit : *subsuming_clause)
          if (lit != satisfied_literal) {
            assert (!marks[lit]);
            marks[lit] = 1;
          }
      }

      // Now can mark the literals from the vivified clause as removed.

      for (auto lit : *candidate)
        if (!marks[lit])
          mark_removed (lit);

      // Unmark literals of the subsuming clause and mark them as added.

      if (falsified_literal != INVALID_LITERAL) {
        assert (marks[falsified_literal]);
        marks[falsified_literal] = 0;
        mark_added (falsified_literal);
      }
      if (satisfied_literal != INVALID_LITERAL) {
        assert (marks[satisfied_literal]);
        marks[satisfied_literal] = 0;
        mark_added (satisfied_literal);
      }
      if (tagged (subsuming_ref)) {
        unsigned lit = untag (subsuming_ref);
        assert (marks[lit]);
        marks[lit] = 0;
        mark_added (lit);
      } else {
        Clause *subsuming_clause = dereference_clause (subsuming_ref);
        for (auto lit : *subsuming_clause)
          if (lit != satisfied_literal) {
            assert (marks[lit]);
            marks[lit] = 0;
            mark_added (lit);
          }
      }
    }

  } else if (vivified) { // ... not subsuming but vivified ...

    // Target was to replace the candidate clause with the learned clause
    // from the analysis but only if the learned clause is smaller.

    unsigned new_size = clause.size ();

    if (new_size < old_size) {

      // Target reached: learned clause shorter than candidate.

      debug_vector (clause, "vivified %s", tier_name (tier));

      if (!candidate->redundant) {
        for (auto lit : clause)
          assert (!marks[lit]), marks[lit] = 1;
        for (auto lit : *candidate)
          if (!fixed (lit) && !marks[lit])
            mark_removed (lit);
        for (auto lit : clause)
          assert (marks[lit]), marks[lit] = 0;
      }

      // After shrinking backtrack to flush all the decisions as we have
      // at least one new propagation possibility that should be
      // considered.

      if (level)
        backtrack (0);

      if (!new_size) {
        debug ("vivified to empty clause");
        inconsistent = true;
        trace_empty ('a');
        cover ("vivified to empty clause");
      } else if (new_size == 1) {
#ifndef QUIET
        statistics.vivified_units++;
#endif
        unsigned unit = clause[0];
        debug ("new unit shrunken clause %s", debug_literal (unit));
        trace_unit ('a', unit);
        assign (unit, UNIT_REASON);
        if (!vivify_propagate (0)) {
          debug ("propagated unit results in inconsistency");
          inconsistent = true;
          trace_empty ('a');
        }
      } else if (new_size == 2)
        strengthen_to_binary_clause (candidate);
      else
        strengthen_to_large_clause (candidate);

    } else if (!candidate->redundant && !redundant) {

      // Irredundant clauses are turned into redundant clauses in case
      // they are implied by only irredundant clauses.  Afterwards we
      // still give them the glucose level (glue) zero so that they become
      // tier 1 clauses and are never recycled in learned clause reductions.

      debug_clause (candidate, "irredundant implied vivified");
      mark_removed (candidate);
      DEC_COUNT (irredundant_clauses);
      INC_COUNT (redundant_clauses);
      candidate->redundant = true;
      update_position (candidate);
      candidate->glue = 0; // Mark as not reducible.
      update_position (candidate);
      if (level && conflicting)
        backtrack (level - 1);
      debug_clause (candidate, "vivified irredundant clause to");

      vivified = true;

    } else {

      // Otherwise the candidate clause is redundant or the irredundant
      // candidate clause was shown to be implied by the formula, but
      // during that implication check a redundant clause was used.

      debug_vector (clause, "vivification failed on");
      assert (!num_falsified);
      if (level && conflicting)
        backtrack (level - 1);

      // Unfortunately we have to correct our earlier 'vivified' verdict.
      // Even though the candidate clause is implied by other clauses
      // conflict analysis did not give a shorter clause or the same
      // clause but a redundant clause was used during conflict analysis
      // while the candidate is irredundant.

      vivified = false;
    }
  }
#ifndef QUIET
  if (vivified) {
    assert (tier < 3);
    statistics.vivified_clauses.tier[tier]++;
    statistics.vivified_clauses.total++;
  }
#endif
  clause.clear ();
  assert (analyzed.empty ());
  (void) tier;

  return vivified;
}

void Solver::vivify_tier_clauses (unsigned tier, unsigned *counts,
                                  uint64_t limit) {

  if (tier && first_redundant == INVALID_POSITION) {
    message (3, "no %s vivification candidates", tier_name (tier));
    return;
  }

  assert (!level);
  assert (watching);
  assert (vivifying);

  message (3, "starting %s vivification", tier_name (tier));

  std::vector<Clause *> candidates;
  gather_vivification_candidates (tier, counts, candidates);

  size_t vivified = 0, checked = 0;

  for (size_t i = 0; i != candidates.size (); i++) {
    checked++;
    Clause *candidate = candidates[i];
    if (vivify_clause (tier, counts, candidate)) {
      vivified++;
      if (tier <= 1 && !candidate->garbage) {
        debug_clause (candidate, "rescheduling vivification candidate");
        candidates.push_back (candidate);
        candidate->vivify = true;
      }
    }
    if (inconsistent)
      break;
    if (GET_COUNT (vivification_ticks) > limit)
      break;
  }

  if (level)
    backtrack (0);

#ifndef QUIET
  size_t scheduled = candidates.size ();
  statistics.vivification_completed[tier] += (checked == scheduled);

  message (1, "vivified %zu %s clauses %.0f%% out of %zu checked %.0f%%",
           vivified, tier_name (tier), percent (vivified, checked), checked,
           percent (checked, scheduled));
#else
  (void) checked;
  (void) vivified;
#endif
  bool report = vivified;
#ifndef QUIET
  if (verbosity > 0)
    report = true;
#endif
  if (report)
    report_progress (tier == 1 ? 'u' : tier == 2 ? 'v' : 'w');
}

bool Solver::vivify_clauses () {

  assert (OPTION (vivify_clauses));
  assert (OPTION (preprocessing) || OPTION (inprocessing));

  if (level && !backtrack_to_root_level_and_propagate ())
    return false;

  assert (!level);
  assert (watching);
  assert (!inconsistent);
  assert (!vivifying);

  vivifying = true;
  start_simplifier ("vivify");
  VivificationScheduler vivification_scheduler (this);
  vivification_scheduler.increment_counter ();
  uint64_t total;

  unsigned tier1 = OPTION (vivify_tier1_clauses);
  unsigned tier2 = OPTION (vivify_tier2_clauses);
  unsigned redundant = tier1 + tier2;
  unsigned irredundant = OPTION (vivify_irredundant_clauses);

  if (preprocessing) {
    message (3, "vivification effort is completely focused on "
                "irredundant clauses during preprocessing");
    irredundant = 1, tier1 = tier2 = redundant = 0;
    unsigned minimum_effort = OPTION (minimum_effort);
    unsigned factor = OPTION (vivification_preprocessing);
    total = factor * minimum_effort;
    message (2, "preprocessing vivify effort %" PRIu64 " = %u * %u ticks",
             total, factor, minimum_effort);
  } else {

    // Only perform irredundant clause vivification during inprocessing if
    // there are enough redundant clauses to vivify.

    if (irredundant && redundant) {
      double percentage = OPTION (vivify_irredundant_limit) / 100.0;
      uint64_t limit = percentage * GET_COUNT (irredundant_clauses);
      if (GET_COUNT (irredundant_clauses) < limit) {
        message (3,
                 "no irredundant clause vivification yet"
                 " as redundant clauses %" PRIu64 " < %" PRIu64
                 " = %g * %" PRIu64 " irredundant clauses",
                 statistics.redundant_clauses, limit, percentage,
                 statistics.irredundant_clauses);
        irredundant = 0;
      }
    }

    total = vivification_scheduler.get_ticks_limit ();
  }

  uint64_t sum = irredundant + redundant;
  uint64_t vivified_before = statistics.vivified_clauses.COUNTER (total);

  if (sum) {

    unsigned *counts = count_occurrences ();
    uint64_t limit = GET_COUNT (vivification_ticks);

    if (tier1) {
      limit += average (tier1, sum) * total;
      vivify_tier_clauses (1, counts, limit);
    }

    if (tier2 && !inconsistent && !terminated) {
      limit += average (tier2, sum) * total;
      vivify_tier_clauses (2, counts, limit);
    }

    if (irredundant && !inconsistent && !terminated) {
      limit += average (irredundant, sum) * total;
      vivify_tier_clauses (0, counts, limit);
    }

    free (counts);
  }

  trigger.vivification = 0;
  vivification_scheduler.set_next_conflict_limit ();

  assert (vivifying);
  vivifying = false;

  stop_simplifier ("vivify");

  return statistics.vivified_clauses.COUNTER (total) > vivified_before;
}

}; // namespace TabularaSAT
