#include "inline.hpp"
#include "map.hpp"
#include "references.hpp"
#include "schedule.hpp"
#include "solver.hpp"
#include "ticks.hpp"

#include <cinttypes>

namespace TabularaSAT {

// Additional data for 'WalkSAT' style local search (aka 'Walker').

struct Walker : Effort {

  Walker (Solver *s) : Wrapper (s) {}
  const char *name () { return "local search walking"; }
  uint64_t &last () { return solver.last.search_ticks_at_walk; }
  unsigned effort () { return SOLVER_OPTION (walking_effort); }

  // Unsatisfied clauses are encoded as one 64-bit word, using either the
  // arena reference directly or packing both literals of a virtual binary
  // clause into 8 bytes.  These words encoding unsatisfied clauses are
  // then pushed on a stack, from which we select the 'next' unsatisfied
  // clause in the 'ProbSAT' style local search algorithm.

  // We further also use a hash table to map these 64-bit words to their
  // position in the stack in order to efficiently remove clauses which
  // become satisfied again (during updating clause status in the
  // 'make_clauses' function) from this stack (by swapping them with the
  // last clause word in the stack).

  // We originally used a random selection in the hash table which however
  // depends on the clause offsets and thus would give different behaviour
  // if the size of clauses changes (i.e., adding 'id' for debugging).

  std::vector<uint64_t> unsatisfied_clauses;
  Map clause_position;
  size_t next = 0;

  std::vector<double> break_score_table;
  size_t maximum_break_count;

  // This trail keeps a list of flipped literals in order to avoid
  // updating the full minimum assignment every time it is improved.
  // Instead we remember the trail position 'pending' when the best
  // assignment was achieved.  If we run out of space (trail becomes too
  // large) or the search terminates we still need to flip all those
  // literals in the saved assignment until 'pending'.  After flushing the
  // pending best assignment we set 'pending' to 'INVALID' to remember
  // that the saved assignment is already the best found so far.  The next
  // minimum will again copy the full assignment but then resume saving
  // flipped literals on the trail.

  std::vector<unsigned> trail;
  size_t pending = 0;

  // First start with encoding and decoding clauses into words.

  static uint64_t encode_large (Reference ref) {
    assert (!tagged (ref));
    uint64_t word = ref + 2;
    assert (ref < word);
    assert (!tagged (word));
    return word;
  }

  static Reference decode_large (uint64_t word) {
    assert (!tagged (word));
    assert (word > 1);
    Reference ref = word - 2;
    assert (!tagged (ref));
    return ref;
  }

  static uint64_t encode_binary (unsigned lit, unsigned other) {
    if (lit > other)
      std::swap (lit, other);
    assert (other < (1u << 31));
    uint64_t word = ((uint64_t) lit << 32) + (other << 1) + 1;
    assert (tagged (word));
    return word;
  }

  static void decode_binary (uint64_t word, unsigned &lit,
                             unsigned &other) {
    assert (tagged (word));
    lit = word >> 32;
    other = (word & 0xffffffff) >> 1;
  }

  // Add and remove clauses simultaneously to both stack and map.

  void add_encoded (uint64_t word) {
    size_t position = unsatisfied_clauses.size ();
    unsatisfied_clauses.push_back (word);
    clause_position.insert (word, position);
    assert (clause_position.size () == unsatisfied_clauses.size ());
  }

  void remove_encoded (uint64_t word) {
    size_t word_position = clause_position.remove (word);
    uint64_t last = unsatisfied_clauses.back ();
    unsatisfied_clauses.pop_back ();
    if (last != word) {
      clause_position.overwrite (last, word_position);
      unsatisfied_clauses[word_position] = last;
    }
    assert (clause_position.size () == unsatisfied_clauses.size ());
  }

  void add_unsatisfied_large (Reference ref) {
    uint64_t word = encode_large (ref);
    add_encoded (word);
  }

  void add_unsatisfied_binary (unsigned lit, unsigned other) {
    uint64_t word = encode_binary (lit, other);
    add_encoded (word);
  }

  void remove_unsatisfied_binary (unsigned lit, unsigned other) {
    uint64_t word = encode_binary (lit, other);
    remove_encoded (word);
  }

  void remove_unsatisfied_large (Reference ref) {
    uint64_t word = encode_large (ref);
    remove_encoded (word);
  }
};

static double cbtab[][2] = {{0.0, 2.00}, {3.0, 2.50}, {4.0, 2.85},
                            {5.0, 3.70}, {6.0, 5.10}, {7.0, 7.40}};

static double fit_break_score_base (double size) {
  const size_t num_cbvals = sizeof cbtab / sizeof *cbtab;
  size_t i = 0;
  while (i + 2 < num_cbvals &&
         (cbtab[i][0] > size || cbtab[i + 1][0] < size))
    i++;
  const double x2 = cbtab[i + 1][0], x1 = cbtab[i][0];
  const double y2 = cbtab[i + 1][1], y1 = cbtab[i][1];
  const double dx = x2 - x1, dy = y2 - y1;
  assert (dx);
  const double res = dy * (size - x1) / dx + y1;
  assert (res > 0);
  return res;
}

void Solver::initialize_local_search (Walker &walker) {
  assert (!level);

  // First go over all variables and set their value to to their decision
  // phase, by default set by 'warming_up_saved_phases' unless not all
  // variables have saved phases already.  The root-level assigned variables
  // are unassigned and reassigned after local search finishes.

  unsigned negative = 0, positive = 0;
  for (auto idx : variables) {
    unsigned lit = literal (idx);
#ifdef LOGGING
    levels[idx] = INVALID_LEVEL;
#endif
    signed char value = values[lit] ? 0 : decide_phase (idx);
    values[lit] = value;
    values[negate (lit)] = -value;
    if (!value)
      continue;
    phases[idx].target = value;
    if (value < 0)
      negative++;
    else
      positive++;
    debug ("local search assignment %s", debug_literal (lit));
  }
#ifndef QUIET
  message (2,
           "local search starts with "
           "%u negative and %u positive phases",
           negative, positive);
#else
  (void) negative;
  (void) positive;
#endif

  struct {
    size_t count = 0, falsified = 0, literals = 0;
  } binary, large, total;

  // Next determine unsatisfied binary clauses.

  for (auto occurrences : matrix) {
    unsigned lit = occurrences->lit;
    if (lit == INVALID_LITERAL)
      continue;
    signed char lit_value = values[lit];
    if (!lit_value)
      continue; // Skip binary clauses with root-level assigned literals!
    auto begin = occurrences->begin ();
    auto end = occurrences->end ();
    auto p = begin;
    while (p != end) {
      Reference ref = *p++;
      if (!tagged (ref))
        continue;
      unsigned other = untag (ref);
      if (lit > other)
        continue;
      signed char other_value = values[other];
      if (!other_value)
        continue; // Skip binary clauses with root-level assigned literals!
      binary.count++;
      if (lit_value > 0 || other_value > 0)
        continue;
      assert (lit_value < 0), assert (other_value < 0);
      debug_binary (lit, other, "falsified");
      walker.add_unsatisfied_binary (lit, other);
      binary.falsified++;
    }
  }
  message (2, "found %zu falsified binary clauses %.0f%% out of %zu",
           binary.falsified, percent (binary.falsified, binary.count),
           binary.count);

  // Then determine the unsatisfied large clauses.

  for (auto c : clauses) {
    if (c->redundant)
      continue;
    if (c->garbage)
      continue;
    unsigned satisfied = 0, falsified = 0;
    for (auto lit : *c) {
      signed char value = values[lit];
      if (value > 0)
        satisfied++;
      if (value < 0)
        falsified++;
    }
    large.count++;
    large.literals += falsified + satisfied;
    if (satisfied)
      continue;
    debug_clause (c, "falsified");
    Reference ref = reference_clause (c);
    walker.add_unsatisfied_large (ref);
    large.falsified++;
  }
  message (2, "found %zu falsified large clauses %.0f%% out of %zu",
           large.falsified, percent (large.falsified, large.count),
           large.count);

  binary.literals = 2 * binary.count;
  total.count = binary.count + large.count;
  total.literals = binary.literals + large.literals;
  double average_length = average (total.literals, total.count);

  message (2, "average clause length %.2f (%.2f large clauses)",
           average_length, average (large.literals, large.count));

  // Finally compute the break score table.

  double base = 2.0;
  if (OPTION (walking_fit) && (GET_COUNT (local_searches) & 1))
    base = fit_break_score_base (average_length);
  assert (base > 0);
  double epsilon = 1e-10;
  double factor = 1 / base;
  double epsilon2 = epsilon * epsilon;
  for (double score = factor; score >= epsilon2; score *= factor)
    walker.break_score_table.push_back ((score + epsilon) / 2);
  assert (!walker.break_score_table.empty ());
  walker.maximum_break_count = walker.break_score_table.size () - 1;
  message (2, "break score factor %g = 1/%g yields score table of size %zu",
           factor, base, walker.break_score_table.size ());
}

double Solver::break_score (Walker &walker, unsigned lit) {
  assert (values[lit] < 0);
  size_t maximum_break_count = walker.maximum_break_count;
  unsigned not_lit = negate (lit);
  const auto occurrences = references (not_lit);
  size_t break_count = 0;
  auto begin = occurrences->begin ();
  auto end = occurrences->end ();
  auto p = begin;
  uint64_t ticks = 0;
  while (p != end && break_count < maximum_break_count) {
    Reference ref = *p++;
    if (tagged (ref)) {
      unsigned other = untag (ref);
      if (values[other] < 0)
        break_count++;
    } else {
      ticks++;
      Clause *c = dereference_clause (ref);
      for (auto other : *c)
        if (other != not_lit && values[other] > 0)
          goto CONTINUE;
      break_count++;
    CONTINUE:;
    }
  }
  ticks += ticks_spanning (p - begin);
  ADD_COUNT (local_search_ticks, ticks);
  assert (break_count < walker.break_score_table.size ());
  double break_score = walker.break_score_table[break_count];
  debug ("literal %s would break %zu clauses and has break score %g",
         debug_literal (lit), break_count, break_score);
  return break_score;
}

void Solver::make_clauses (Walker &walker, unsigned lit) {
  debug ("making clauses after flipping %s", debug_literal (lit));
  assert (values[lit] > 0);
  const auto occurrences = references (lit);
  uint64_t ticks = ticks_references (occurrences);
#ifdef LOGGING
  size_t made = 0;
#endif
  for (auto ref : *occurrences) {
    if (tagged (ref)) {
      unsigned other = untag (ref);
      signed char value = values[other];
      assert (value);
      if (value > 0)
        continue;
      debug_binary (lit, other, "making");
      walker.remove_unsatisfied_binary (lit, other);
#ifdef LOGGING
      made++;
#endif
    } else {
      ticks++;
      Clause *c = dereference_clause (ref);
      bool satisfied = false;
      for (auto other : *c) {
        if (other == lit)
          continue;
        signed char value = values[other];
        if (value > 0) {
          satisfied = true;
          break;
        }
      }
      if (satisfied)
        continue;
      debug_clause (c, "making");
      walker.remove_unsatisfied_large (ref);
#ifdef LOGGING
      made++;
#endif
    }
  }
  ADD_COUNT (local_search_ticks, ticks);
  debug ("made %zu clause flipping %s", made, debug_literal (lit));
}

void Solver::break_clauses (Walker &walker, unsigned lit) {
  debug ("breaking clauses after flipping %s", debug_literal (lit));
  assert (values[lit] < 0);
  const auto occurrences = references (lit);
  uint64_t ticks = ticks_references (occurrences);
#ifdef LOGGING
  size_t broken = 0;
#endif
  for (auto ref : *occurrences) {
    if (tagged (ref)) {
      unsigned other = untag (ref);
      signed char value = values[other];
      assert (value);
      if (value > 0)
        continue;
      debug_binary (lit, other, "breaking");
      walker.add_unsatisfied_binary (lit, other);
#ifdef LOGGING
      broken++;
#endif
    } else {
      ticks++;
      Clause *c = dereference_clause (ref);
      bool satisfied = false;
      for (auto other : *c) {
        if (other == lit)
          continue;
        signed char value = values[other];
        if (value > 0) {
          satisfied = true;
          break;
        }
      }
      if (satisfied)
        continue;
      debug_clause (c, "breaking");
      walker.add_unsatisfied_large (ref);
#ifdef LOGGING
      broken++;
#endif
    }
  }
  ADD_COUNT (local_search_ticks, ticks);
  debug ("broken %zu clause flipping %s", broken, debug_literal (lit));
}

void Solver::flush_pending_minimimum_assignment (Walker &walker) {
  assert (walker.pending <= walker.trail.size ());
  auto begin = walker.trail.begin ();
  auto pending = begin + walker.pending;
  message (2, "exporting %zu pending flipped variables", pending - begin);
  for (auto p = begin; p != pending; ++p) {
    unsigned lit = *p, idx = index (lit);
    signed value = sign (lit) ? -1 : 1;
    phases[idx].target = value;
  }
}

void Solver::run_local_search (Walker &walker) {

  size_t initial = walker.unsatisfied_clauses.size (), minimum = initial;
  message (1, "initial unsatisfied clauses minimum %zu", minimum);

  uint64_t limit = walker.get_ticks_limit ();
  limit += GET_COUNT (local_search_ticks);

  uint64_t flipped = 0;

  std::vector<unsigned> literals;
  std::vector<double> scores;

  while (minimum && !terminated && GET_COUNT (local_search_ticks) < limit) {

    // First pick a random unsatisfied clause and save its
    // non-root-level-assigned literals (those with a value).

    size_t unsatisfied = walker.unsatisfied_clauses.size ();
    size_t pick = walker.next++ % unsatisfied;
    debug ("picked unsatisfied clause %zu out of %zu", pick, unsatisfied);
    uint64_t word = walker.unsatisfied_clauses[pick];

    literals.clear ();
    if (tagged (word)) {
      unsigned lit, other;
      walker.decode_binary (word, lit, other);
      assert (values[lit] < 0);
      assert (values[other] < 0);
      literals.push_back (lit);
      literals.push_back (other);
      debug_binary (lit, other, "picked");
    } else {
      Reference ref = walker.decode_large (word);
      Clause *c = dereference_clause (ref);
      debug_clause (c, "picked");
      for (auto lit : *c) {
        signed char value = values[lit];
        if (value)
          literals.push_back (lit);
      }
    }
    assert (literals.size () > 1);

    // Compute scores of all literals in the clause and sum them up.

    scores.clear ();
    double sum = 0;
    for (auto lit : literals) {
      double score = break_score (walker, lit);
      scores.push_back (score);
      sum += score;
    }
    assert (scores.size () == literals.size ());

    // Now sample in the sum of scores for the score of a literal.

    double sampled = random.next_double () * sum;
    assert (0 <= sampled), assert (sampled < sum);
    debug ("total break score sum %g sample %g", sum, sampled);
    double tmp = 0;
    size_t picked = 0;
    for (;;) {
      assert (picked < scores.size ());
      tmp += scores[picked];
      if (tmp >= sampled)
        break;
      picked++;
    }

    // Pick the literal with the sampled score.

    flipped++;
#ifndef QUIET
    statistics.flipped_literals++;
#endif
    unsigned flip = literals[picked];
    debug ("picked literal %s with score %g", debug_literal (flip),
           scores[picked]);
    debug ("flipping %s", debug_literal (flip));
    unsigned not_flip = negate (flip);
    values[flip] = 1, values[not_flip] = -1;

    // Save it on the walker trail.

    if (walker.pending != INVALID_SIZE &&
        walker.trail.size () > variables / 4) {
      flush_pending_minimimum_assignment (walker);
      walker.pending = INVALID_SIZE;
      walker.trail.clear ();
    }

    if (walker.pending != INVALID_SIZE)
      walker.trail.push_back (flip);

    // Update unsatisfied clauses (make=remove, break=add).

    make_clauses (walker, flip);
    break_clauses (walker, not_flip);

    debug ("reached %zu unsatisfied clauses after flipping %s",
           walker.unsatisfied_clauses.size (), debug_literal (flip));

    // Finally update the global assignment if a new minimum is reached.

    if (walker.unsatisfied_clauses.size () < minimum) {
      minimum = walker.unsatisfied_clauses.size ();
      message (2, "new unsatisfied clauses minimum %zu", minimum);

      if (walker.pending == INVALID_SIZE) {
        message (2, "exporting full local search assignment");
        for (auto idx : variables) {
          unsigned lit = literal (idx);
          signed char value = values[lit];
          if (value)
            phases[idx].target = value;
        }
        assert (walker.trail.empty ());
        walker.pending = 0;
      } else {
        walker.pending = walker.trail.size ();
        debug ("new minimum assignment at position %zu in trail",
               walker.pending);
      }
    }

    INC_COUNT (local_search_ticks);
  }

  if (walker.pending != INVALID_SIZE)
    flush_pending_minimimum_assignment (walker);
#ifndef QUIET
  unsigned improved = minimum < initial;
  statistics.improved_minimum += improved;
  message (1,
           "%s final unsatisfied clause minimum %zu after flipping %" PRIu64
           " literals",
           improved ? "improved" : "unchanged", minimum, flipped);
#else
  (void) flipped;
#endif
}

void Solver::reset_local_search (Walker &walker) {
  assert (!level);
  memset (&values[0], 0, literals);
  for (auto lit : trail) {
    values[lit] = 1, values[negate (lit)] = -1;
#ifdef LOGGING
    levels[index (lit)] = 0;
#endif
  }
}

void Solver::warming_up_saved_phases () {
  start ("warmup", 3);
  assert (!warming);
  warming = true;

  bool warm = true;
  for (auto idx : variables)
    if (!phases[idx].saved) {
      warm = false;
      break;
    }
  warm = false;

  if (warm)
    message (1, "already warmed up as all variables have saved phases");
  else {
    unsigned decisions = 0;
    while (unassigned) {
      decisions++;
      decide_literal ();
      warming_propagate ();
    }
    if (level)
      backtrack (0);
#ifndef QUIET
    message (1, "warmed up all variables explicitly with %u decisions",
             decisions);
#else
    (void) decisions;
#endif

#ifndef NDEBUG
    for (auto idx : variables)
      if (flags[idx].active)
        assert (phases[idx].saved);
#endif
  }

  assert (warming);
  warming = false;
  stop ("warmup");
}

void Solver::local_search () {
  start ("walk");
  INC_COUNT (local_searches);
  warming_up_saved_phases ();
  transition_to_dense_connections ();
  assert (!watching);
  Walker walker (this);
  initialize_local_search (walker);
  run_local_search (walker);
  reset_local_search (walker);
  transition_to_sparse_watching ();
  stop ("walk");
}

}; // namespace TabularaSAT
