#include "clause.hpp"
#include "connect.hpp"
#include "references.hpp"
#include "solver.hpp"
#include "watch.hpp"

namespace TabularaSAT {

void Solver::copy_clause (Clause *c, unsigned except) {
  assert (clause.empty ());
  for (auto lit : *c) {
    if (lit == except)
      continue;
    assert (values[lit] < 0);
    if (levels[index (lit)])
      clause.push_back (lit);
  }
}

// Determine whether 'a' is better watched than 'b'.  Use 'false' as
// tie-breaker as we want to avoid generating new watches. This makes this
// comparison not suited for sorting algorithms.

bool Solver::better_watch (unsigned a, unsigned b) {
  assert (watching);
  signed char u = values[a];
  signed char v = values[b];
  unsigned l = levels[index (a)];
  unsigned k = levels[index (b)];
  if (!u && !v)
    return false; // Beware: non-stable!
  else if (u > 0) {
    if (v <= 0)
      return true; // Prioritize satisfied.
    assert (v > 0);
    return l < k; // Prioritize lower level.
  } else {
    if (v > 0)
      return false; // Prioritize satisfied.
    if (u && !v)
      return true; // Prioritize unassigned.
    if (!u && v)
      return false; // Prioritize unassigned.
    assert (u < 0), assert (v < 0);
    return l > k; // Prioritize higher level.
  }
}

// The following function shortens a large clause in-place by assuming that
// the kept literals of the shrunken clause are given in the temporary
// clause and need to be copied to the existing large clause in the clause
// arena.  The shrunken clause will then have its 'shrunken' bit set
// indicating that it allocated memory actually extends further then its
// size. Thus there will be a gap between the shrunken clause and the next
// in the arena. The end of the gap is marked with 'INVALID_LITERAL'.

// When we shrink a clause while watching clauses we need to reestablish the
// watched clause invariant in case watched literals are removed.  This is
// particular tricky since removed literals might also be blocking literals.
// Thus it is best to just unwatch the clause and rewatch it again but that
// is done by the calling function.

void Solver::strengthen_to_large_clause (Clause *c) {
  assert (watching);
  assert (clause.size () > 2);

  debug_clause (c, "strengthening");

  trace_temporary ('a');
  trace_clause ('d', c);

  unwatch_clause (c);

  // Put the two better literals into the first two positions.

  auto begin = clause.begin ();
  auto end = clause.end ();
  for (auto p = begin; p != begin + 2; p++)
    for (auto q = begin + 2; q != end; q++)
      if (better_watch (*q, *p))
        std::swap (*p, *q);

  unsigned old_size = c->size;
  unsigned new_size = clause.size ();

  memcpy (c->literals, &clause[0], 4 * new_size);

  c->size = new_size;
#ifndef NGLUE
  adjust_glue_to_updated_size (c);
#endif
#ifndef NREPLACEMENTCACHE
  c->replaced = 2;
#endif

  if (!c->shortened) {
    c->shortened = true;
    assert (c->literals[old_size - 1] != INVALID_LITERAL);
    c->literals[old_size - 1] = INVALID_LITERAL;
  }

  if (!c->redundant)
    mark_added (c);

  watch_clause (c);

  debug_clause (c, "strengthened");
}

// The logic of strengthening to a binary clause is slightly simpler.

void Solver::strengthen_to_binary_clause (Clause *c) {
  assert (watching);
  assert (clause.size () == 2);
  debug_clause (c, "strengthening");
  unwatch_clause (c);
  trace_temporary ('a');
  (void) new_binary ();
  mark_clause_garbage (c, "strengthened to binary");
  debug_binary (clause[0], clause[1], "strengthened");
}

Reference Solver::new_binary () {
  assert (clause.size () == 2);
  unsigned lit = clause[0], other = clause[1];
  debug_binary (lit, other, "new binary clause");
  if (watching) {
    watch_binary (lit, other);
    watch_binary (other, lit);
  } else {
    connect_binary (lit, other);
    connect_binary (other, lit);
  }
  mark_added (lit);
  mark_added (other);
  INC_COUNT (binary_clauses);
#ifndef QUIET
  statistics.added_binary_clauses++;
  statistics.added_clauses++;
#endif
  return tag_binary (other);
}

#ifndef NGLUE

void Solver::adjust_glue_to_updated_size (Clause *c) {
  unsigned old_glue = c->glue;
  if (old_glue < c->size)
    return;
  c->glue = c->size - 1;
  assert (c->glue < old_glue);
  assert (c->glue < c->size);
}

#endif

Reference Solver::new_clause
#ifndef NGLUE
    (bool redundant, unsigned glue)
#else
    (bool redundant)
#endif
{
  size_t size = clause.size ();
  assert (3 <= size);
  assert (size <= MAXIMUM_INTERNAL_VARIABLE);

  size_t words = Clause::words (size);
  assert (words >= 4);

  size_t pos = clauses.size ();
  if (clauses.size () > MAXIMUM_POSITION_OF_CLAUSE)
    fatal ("maximum clause arena of %zu words exceeded "
           "(for such large formula configure and compile with "
           "'./configure --no-compact')",
           (size_t) MAXIMUM_POSITION_OF_CLAUSE);

#ifndef NGLUE
  if (glue > MAXIMUM_GLUE)
    glue = MAXIMUM_GLUE;
#endif

#ifndef QUIET
  size_t old_capacity = clauses.capacity ();
#endif
  Clause *c = clauses.allocate (words);
  Reference ref = tag_large (pos);
#ifndef QUIET
  size_t new_capacity = clauses.capacity ();
  if (old_capacity < new_capacity) {
    pretty_memory_buffer buffer;
    message (1, "clauses arena capacity increased to %zu words %s",
             new_capacity, pretty_memory (buffer, new_capacity * 4));
  }
#endif

#ifdef LOGGING
  c->id = statistics.added_large_clauses;
#endif
#ifndef QUIET
  statistics.added_large_clauses++;
  statistics.added_clauses++;
#endif
  if (redundant)
    INC_COUNT (redundant_clauses);
  else
    INC_COUNT (irredundant_clauses);

  c->garbage = false;
  c->reason = false;
  c->redundant = redundant;
  c->shortened = false;
  c->shrink = false;
  c->subsume = false;
#ifndef NVIVIFY
  c->vivify = false;
#endif
#ifndef NUSED
  c->used = 0;
#endif
#ifndef NGLUE
  c->glue = glue;
  assert (c->glue == glue); // Check bit-field big enough.
#endif
#ifndef NACTIVITY
  c->activity.as_float = 0;
#endif
  c->size = size;
#ifndef NREPLACEMENTCACHE
  c->replaced = 2;
#endif

  memcpy (c->literals, &clause[0], size * sizeof (unsigned));

  debug_clause (c, "new");

  if (!redundant)
    mark_added (c);
  else if (OPTION (bump_learned))
    bump_clause (c);

  update_position (c);

  if (watching)
    watch_clause (c);
  else
    for (auto lit : clause)
      connect_clause (lit, ref);

  assert (dereference_clause (ref) == c);

  return ref;
}

void Solver::deactivate_variable (unsigned idx) {
  debug ("deactivating %s", debug_variable (idx));
  Flags &f = flags[idx];
  assert (f.active);
  f.active = false;
#ifndef QUIET
  assert (statistics.active_variables);
  statistics.active_variables--;
#endif
  assert (!values[literal (idx)]);
  assert (unassigned);
  unassigned--;

#ifndef NQUEUE
  if (queue.contains (idx))
    queue.dequeue (idx);
#endif
  if (bumper.contains (idx))
    bumper.remove (idx);
}

inline void Solver::activate_variable (unsigned idx) {
  Flags &f = flags[idx];
  if (f.activated)
    return;
  debug ("activating %s", debug_variable (idx));
  assert (!f.active);
  assert (!f.eliminated);
  assert (!f.fixed);
#ifndef NSUBSTITUTE
  assert (!f.substituted);
#endif
  f.activated = f.active = true;
  f.eliminate = f.subsume = true;
  INC_COUNT (activated_variables);
#ifndef QUIET
  statistics.active_variables++;
#endif
  unassigned++;
  assert (unassigned);
}

inline void Solver::activate_literal (unsigned lit) {
  activate_variable (index (lit));
}

void Solver::add_simplified_irredundant_clause () {
#ifndef QUIET
  statistics.added_original_clauses++;
  statistics.added_clauses++;
#endif
  assert (!level);
  for (auto lit : clause)
    activate_literal (lit);
  size_t size = clause.size ();
  if (!size) {
    debug ("parsed empty clause");
    inconsistent = true;
  } else if (size == 1) {
    unsigned unit = clause[0];
    debug ("parsed unit clause %s", debug_literal (unit));
    assert (!level);
    assign (unit, UNIT_REASON);
  } else if (size == 2)
    new_binary ();
  else
    new_clause ();
}

#ifndef NPROMOTE

void Solver::promote_clause (Clause *c) {
  if (!c->redundant)
    return;
  unsigned old_glue = c->glue;
  if (old_glue <= OPTION (tier1_max_glue))
    return;
  unsigned limit = OPTION (tier2_max_glue);
  if (old_glue - 1 < limit)
    limit = old_glue - 1;
#ifndef NDEBUG
  for (auto lit : *c)
    assert (!control[levels[index (lit)]].promote);
#endif
  unsigned new_glue = 0;
  assert (promote.empty ());
  for (auto lit : *c) {
    assert (values[lit]);
    unsigned lit_level = levels[index (lit)];
    if (!lit_level)
      continue;
    Frame &frame = control[lit_level];
    if (frame.promote++)
      continue;
    promote.push_back (lit_level);
    if (new_glue++ == limit)
      break;
  }
  for (auto promote_level : promote)
    control[promote_level].promote = 0;
  promote.clear ();
  if (new_glue > limit)
    return;
  assert (new_glue < old_glue);
  assert (new_glue < c->size);
  c->glue = new_glue;
  debug_clause (c, "promoted glue %u to", old_glue);
#ifndef QUIET
  statistics.promoted_clauses++;
#endif
}

#endif

}; // namespace TabularaSAT
