#include "inline.hpp"
#include "radix.hpp"
#include "references.hpp"
#include "solver.hpp"

#include <algorithm>

namespace TabularaSAT {

struct larger_level {
  const Solver &solver;
  larger_level (const Solver &s) : solver (s) {}
  bool operator() (unsigned a, unsigned b) {
    unsigned i = Shared::index (a);
    unsigned j = Shared::index (b);
    unsigned k = solver.levels[i];
    unsigned l = solver.levels[j];
    if (k > l)
      return true;
    if (k < l)
      return false;
    unsigned s = solver.position[i];
    unsigned t = solver.position[j];
    assert (s != t);
    return s > t;
  }
};

struct rank_larger_literal_first {
  const unsigned *ranks;
  rank_larger_literal_first (const unsigned *r) : ranks (r) {}
  unsigned operator() (unsigned lit) { return ~ranks[Shared::index (lit)]; }
};

void Solver::shrink_deduced_clause () {
  start ("shrink", 4);
  if (clause.size () < OPTION (radix_sort_limit))
    std::sort (clause.begin (), clause.end (), larger_level (*this));
  else {
    radix_sort_unsigned (clause, rank_larger_literal_first (position));
    radix_sort_unsigned (clause, rank_larger_literal_first (&levels[0]));
  }
  debug_vector (clause, "starting to shrink deduced");
  if (clause.size () < OPTION (radix_sort_limit))
    std::sort (pulled.begin (), pulled.end ());
  else
    radix_sort_unsigned (pulled, rank_less_unsigned ());
  size_t end_block = clause.size ();
  unsigned shrunken = 0, minimized = 0;
  for (auto block_level : pulled) {
    unsigned pulled_at_block_level = control[block_level].pulled;
    if (pulled_at_block_level == 1) {
      end_block--;
      debug ("cant not shrink size 1 block[%zu]@%u containing only %s",
             end_block, block_level, debug_literal (clause[end_block]));
      continue;
    }
    size_t begin_block = end_block;
    unsigned max_trail = 0;
#ifdef LOGGING
    unsigned min_trail = UINT_MAX;
#endif
    assert (shrinkable.empty ());
    for (;;) {
      assert (begin_block);
      unsigned lit = clause[begin_block - 1];
      assert (values[lit] < 0);
      unsigned idx = index (lit);
      unsigned lit_level (levels[idx]);
      assert (lit_level >= block_level);
      if (lit_level > block_level)
        break;
      unsigned lit_trail = position[idx];
      if (lit_trail > max_trail)
        max_trail = lit_trail;
#ifdef LOGGING
      if (lit_trail < min_trail)
        min_trail = lit_trail;
#endif
      Flags &f = flags[idx];
      assert (!f.shrinkable);
      assert (!f.minimizable);
      assert (!f.poisoned);
      f.shrinkable = true;
      shrinkable.push_back (idx);
      begin_block--;
    }
    size_t size = end_block - begin_block;
    assert (size == pulled_at_block_level);
    assert (size == shrinkable.size ());
#ifdef LOGGING
    if (logging ()) {
      debug_prefix ();
      printf ("trying to shrink size %zu block[%zu..%zu]@%u "
              "on trail[%u..%u]",
              size, begin_block, end_block - 1, block_level, max_trail,
              min_trail);
      for (size_t i = begin_block; i != end_block; i++)
        printf (" %s", debug_literal (clause[i]));
      debug_suffix ();
    }
#endif
    unsigned t = max_trail + 1;
    unsigned uip = INVALID_LITERAL;
    assert (size <= UINT_MAX);
    unsigned open = size;
#ifndef NDEBUG
    unsigned trail_at_block_level = control[block_level].trail;
#endif
    for (;;) {
      assert (t > trail_at_block_level);
      uip = trail[--t];
      unsigned uip_idx = index (uip);
      unsigned uip_level = levels[uip_idx];
      if (uip_level != block_level)
        continue;
      if (!flags[uip_idx].shrinkable)
        continue;
      if (open == 1)
        break;
      debug ("trying to shrink %s open %u", debug_literal (uip), open);
      Reference reason = reasons[uip_idx];
      assert (reason != UNIT_REASON);
      if (reason == DECISION_REASON) {
        debug ("can not shrink decision %s", debug_literal (uip));
        assert (open > 1);
        goto DONE;
      }
      if (tagged (reason)) {
        unsigned other = untag (reason);
        debug_binary (uip, other, "trying to shrink along");
        assert (values[other] < 0);
        unsigned other_idx = index (other);
        assert (levels[other_idx] == block_level);
        open--;
        Flags &f = flags[other_idx];
        if (f.shrinkable)
          continue;
        f.shrinkable = true;
        shrinkable.push_back (other_idx);
        open++;
      } else {
        Clause *c = dereference_clause (reason);
        debug_clause (c, "trying to shrink along");
        INC_COUNT (search_ticks);
        for (auto other : *c) {
          if (other == uip)
            continue;
          assert (values[other] < 0);
          unsigned other_idx = index (other);
          unsigned other_level = levels[other_idx];
          if (!other_level)
            continue;
          Flags &f = flags[other_idx];
          if (f.shrinkable)
            continue;
          if (other_level < block_level) {
            if (minimize_literal (other, 1))
              continue;
            debug ("lower level %u literal %s not shrinkable", other_level,
                   debug_literal (other));
            assert (open > 1);
            goto DONE;
          }
          assert (other_level == block_level);
          f.shrinkable = true;
          shrinkable.push_back (other_idx);
          open++;
        }
        open--;
      }
    }
  DONE:
    if (open == 1) {
      unsigned not_uip = negate (uip);
      debug ("shrinking size %zu block[%zu..%zu]@%u "
             "to single literal %s",
             size, begin_block, end_block, block_level,
             debug_literal (not_uip));
      for (auto i = begin_block + 1; i != end_block; i++)
        clause[i] = INVALID_LITERAL;
      clause[begin_block] = not_uip;
      unsigned uip_idx = index (not_uip);
      if (!flags[uip_idx].seen) {
        flags[uip_idx].seen = true;
        analyzed.push_back (uip_idx);
      }
      for (auto idx : shrinkable) {
        Flags &f = flags[idx];
        assert (f.shrinkable);
        assert (!f.minimizable);
        assert (!f.poisoned);
        f.shrinkable = false;
        f.minimizable = true;
        minimizable.push_back (idx);
      }
#ifndef QUIET
      statistics.shrunken_literals += size - 1;
#endif
      shrunken += size - 1;
    } else {
      debug ("failed to shrink size %zu block[%zu..%zu]@%u", size,
             begin_block, end_block, block_level);
      assert (block_level > 1);
      for (auto idx : shrinkable) {
        Flags &f = flags[idx];
        assert (!f.poisoned);
        assert (!f.minimizable);
        assert (f.shrinkable);
        f.shrinkable = false;
      }
      if (size > 2) {
        for (size_t i = begin_block; i != end_block; i++) {
          unsigned lit = clause[i];
          if (minimize_literal (lit)) {
            debug ("removing minimizable literal %s", debug_literal (lit));
            clause[i] = INVALID_LITERAL;
#ifndef QUIET
            statistics.shrunken_literals++; // Count shrunken not minimized
#endif
            minimized++; // but locally as minimized.
          }
        }
      }
    }
    shrinkable.clear ();
    end_block = begin_block;
  }
  if (shrunken || minimized) {
    auto begin = clause.begin ();
    auto end = clause.end ();
    auto p = begin, q = p;
    while (p != end) {
      unsigned lit = *p++;
      if (lit != INVALID_LITERAL)
        *q++ = lit;
    }
    assert (q + shrunken + minimized == end);
    clause.resize (q - begin);
#ifdef LOGGING
    if (!minimized)
      debug_vector (clause, "shrunken by %u literals", shrunken);
    else if (!shrunken)
      debug_vector (clause, "minimized by %u literals", minimized);
    else
      debug_vector (clause, "shrunken by %u and minimized by %u", shrunken,
                    minimized);
#endif
  }
  reset_minimized_and_poisoned ();
  stop ("shrink");
}

}; // namespace TabularaSAT
