// vi:tw=0

#include "inline.hpp"
#include "solver.hpp"

#include <cinttypes>
#include <cstring>

namespace TabularaSAT {

Statistics::Statistics () { memset (this, 0, sizeof *this); }

inline void Statistics::print (bool full, const char *name, uint64_t count,
                               double ratio, const char *rest) {
  if (full || (count && name[0] != ' '))
    printf ("c %-30s %13" PRIu64 " %13.2f %s\n", name, count, ratio, rest);
}

#define PRINT_COUNT(NAME, SYMBOL, ...) \
  print (full, NAME, COUNTER (SYMBOL), __VA_ARGS__)

#ifndef QUIET
#define PRINT_STATS(NAME, SYMBOL, ...) \
  print (full, NAME, STATISTIC (SYMBOL), __VA_ARGS__)
#else
#define PRINT_STATS(...) /**/
#endif

void Statistics::print (bool full, double time_in_seconds,
                        uint64_t total_variables) {

  // Print statistics in alphabetic order.

  // clang-format off

#ifndef CHECK_STATISTICS_USAGE
  PRINT_COUNT ("activated-variables:", activated_variables, percent (activated_variables, total_variables), "% total");
  PRINT_STATS ("added-binary-clauses:", added_binary_clauses, percent (added_binary_clauses, added_clauses), "% added");
  PRINT_STATS ("added-large-clauses:", added_large_clauses, percent (added_large_clauses, added_clauses), "% added");
  PRINT_STATS ("added-original-clauses:", added_original_clauses, percent (added_original_clauses, added_clauses), "% added");
  PRINT_STATS ("added-variables:", added_variables, percent (added_variables, activated_variables), "% activated");
  PRINT_STATS ("bumped-variables:", bumped_variables, average (bumped_variables, analyzed_conflicts), "per analysis");
  PRINT_STATS ("chronological-backtracking:", chronological_backtracking, percent (chronological_backtracking, analyzed_conflicts), "% analyzed");
#ifndef NCHRONO
  PRINT_STATS ("  chronological-forced:", chronological_forced, percent (chronological_forced, chronological_backtracking), "% chronologial");
#endif
  PRINT_STATS ("collected-large-clauses:", collected_clauses, percent (collected_clauses, added_large_clauses), "% added");
  PRINT_STATS ("collected-references:", collected_references, average (conflicts, collected_references), "interval");
  PRINT_COUNT ("conflicts:", conflicts, average (conflicts, time_in_seconds), "per second");
  PRINT_STATS ("  analyzed-conflicts:", analyzed_conflicts, percent (analyzed_conflicts, conflicts), "% conflicts");
#ifndef NREPHASE
  PRINT_STATS ("    updated-best-phases:", updated_best_phases, percent (updated_best_phases, analyzed_conflicts), "% analyzed");
#endif
#ifndef NTARGET
  PRINT_STATS ("    updated-target-phases:", updated_target_phases, percent (updated_target_phases, analyzed_conflicts), "% analyzed");
#endif
#ifndef NSWITCH
  PRINT_STATS ("  focused-conflicts:", focused_conflicts, percent (focused_conflicts, conflicts), "% conflicts");
  PRINT_STATS ("  stable-conflicts:", stable_conflicts, percent (stable_conflicts, conflicts), "% conflicts");
#endif
  PRINT_COUNT ("decisions:", decisions, average (decisions, conflicts), "per conflict");
  PRINT_STATS ("deduced-literals:", deduced_literals, average (deduced_literals, learned_clauses), "average size");
  PRINT_STATS ("eliminated-variables:", eliminated_variables, percent (eliminated_variables, activated_variables), "% activated");
#ifndef NGATES
  PRINT_STATS ("  eliminated-gates:", eliminated_gates, percent (eliminated_gates, eliminated_variables), "% eliminated");
  PRINT_STATS ("    eliminated-ands:", eliminated_ands, percent (eliminated_ands, eliminated_gates), "% gates");
  PRINT_STATS ("    eliminated-equivalences:", eliminated_equivalences, percent (eliminated_equivalences, eliminated_gates), "% gates");
#endif
  PRINT_COUNT ("eliminations:", eliminations, average (conflicts, eliminations), "interval");
#ifndef NPROBE
  PRINT_COUNT ("failed-literal-probings:", failed_literal_probings, average (conflicts, failed_literal_probings), "interval");
  PRINT_STATS ("  failed-literal-completed:", failed_literal_completed, average (failed_literal_probings, failed_literal_completed), "probings");
  PRINT_STATS ("  failed-literal-decisions:", failed_literal_decisions, average (failed_literal_decisions, failed_literal_probings), "per probing");
  PRINT_STATS ("  failed-literal-propagations:", failed_literal_propagations, percent (failed_literal_propagations, propagations), "% propagations");
  PRINT_STATS ("failed-literals:", failed_literals, percent (failed_literals, failed_literal_decisions), "% decisions");
#endif
  PRINT_COUNT ("fixed-variables:", fixed_variables, percent (fixed_variables, activated_variables), "% activated");
#ifndef NWALK
  PRINT_STATS ("flipped-literals:", flipped_literals, average (flipped_literals, local_searches), "per local-search");
#endif
  PRINT_STATS ("forced-reductions:", forced_reductions, percent (forced_reductions, reductions), "% reductions");
  PRINT_STATS ("garbage-collections:", garbage_collections, average (conflicts, garbage_collections), "interval");
  PRINT_STATS ("increased-kept:", increased_kept, average (conflicts, increased_kept), "interval");
#ifndef NJUMPREASONS
  PRINT_STATS ("jumped-reasons:", jumped_reasons, percent (jumped_reasons, propagations), "% propagations");
#endif
#ifndef NWALK
  PRINT_STATS ("improved-minimum:", improved_minimum, percent (improved_minimum, local_searches), "% per local-search");
#endif
#ifdef IMPROVEREASONS
  PRINT_STATS ("improved-reasons:", improved_reasons, percent (improved_reasons, propagations), "% propagations");
#endif
  PRINT_STATS ("learned-clauses:", learned_clauses, average (learned_literals, learned_clauses), "average size");
#ifndef QUIET
#ifndef NALLUIPS
  PRINT_STATS ("  additional-uips:", additional_learned_uips, average (additional_learned_uips, learned_units), "per learned unit");
#endif
#endif
#ifndef NGLUE
  PRINT_STATS ("  learned-tier1:", learned_tier1_clauses, percent (learned_tier1_clauses, learned_clauses), "% learned");
  PRINT_STATS ("  learned-tier2:", learned_tier2_clauses, percent (learned_tier2_clauses, learned_clauses), "% learned");
#endif
  PRINT_STATS ("  learned-units:", learned_units, percent (learned_units, learned_clauses), "% learned");
#ifndef NPROMOTE
  PRINT_STATS ("  promoted-clauses:", promoted_clauses, percent (promoted_clauses, learned_clauses), "% learned");
#endif
#ifndef NWALK
  PRINT_COUNT ("local-searches:", local_searches, average (conflicts, local_searches), "interval");
#endif
  PRINT_STATS ("minimized-literals:", minimized_literals, percent (minimized_literals, deduced_literals), "% deduced");
#ifndef NOTFS
  PRINT_STATS ("on-the-fly-strengthened:", on_the_fly_strengthened, percent (on_the_fly_strengthened, learned_clauses), "% learned");
  PRINT_STATS ("  on-the-fly-subsumed:", on_the_fly_subsumed, percent (on_the_fly_subsumed, on_the_fly_strengthened), "% strengthened");
#endif
  PRINT_STATS ("propagations:", propagations, average (propagations, time_in_seconds), "per second");
#ifndef NBUMPREASONS
  PRINT_STATS ("reason-side-bumped:", reason_side_bumped, percent (reason_side_bumped, bumped_variables), "% bumped");
#endif
  PRINT_COUNT ("reductions:", reductions, average (conflicts, reductions), "interval");
  PRINT_STATS ("  moved-redundant-clauses:", moved_redundant_clauses, percent (moved_redundant_clauses, reductions), "% reductions");
  PRINT_STATS ("removed-duplicated-binaries:", removed_duplicated_binaries, percent (removed_duplicated_binaries, added_binary_clauses), "added");
  PRINT_STATS ("removed-variables:", removed_variables, percent (removed_variables, activated_variables), "% activated");
#ifndef NREPHASE
  PRINT_COUNT ("rephasings:", rephasings, average (conflicts, rephasings), "interval");
#endif
  PRINT_STATS ("resolutions:", resolutions, average (resolutions, eliminated_variables), "per eliminated");
  PRINT_COUNT ("restarts:", restarts, average (conflicts, restarts), "interval");
#ifndef NREUSETRAIL
  PRINT_STATS ("  reused-levels:", reused_levels, average (reused_levels, reused_trails), "per reused trail");
  PRINT_STATS ("  reused-trails:", reused_trails, percent (reused_trails, restarts), "% restarts");
#endif
  PRINT_STATS ("search-propagations:", search_propagations, percent (search_propagations, propagations), "% propagations");
  PRINT_STATS ("  search-ticks:", search_ticks, average (search_ticks, search_propagations), "per search-prop");
#ifndef NSWITCH
  PRINT_STATS ("    focused-ticks:", focused_ticks, percent (focused_ticks, search_ticks), "% ticks");
  PRINT_STATS ("    stable-ticks:", stable_ticks, percent (stable_ticks, search_ticks), "% ticks");
#endif
#ifndef NSHRINK
  PRINT_STATS ("shrunken-literals:", shrunken_literals, percent (shrunken_literals, deduced_literals), "% deduced");
#endif
  PRINT_STATS ("simplifications:", simplifications, percent (simplifications, reductions), "% reductions");
  PRINT_STATS ("strengthened-clauses:", strengthened, percent (strengthened, subsume_checks), "% subsume-checks");
#ifndef NSUBSTITUTE
  PRINT_COUNT ("substitutions:", substitutions, average (conflicts, substitutions), "interval");
  PRINT_STATS ("substituted-variables:", substituted_variables, percent (substituted_variables, activated_variables), "% activated");
#endif
  PRINT_STATS ("subsume-checks:", subsume_checks, average (subsume_checks, subsumptions), "per subsumption");
  PRINT_STATS ("subsumed-clauses:", subsumed_clauses, percent (subsumed_clauses, subsume_checks), "% subsume-checks");
  PRINT_COUNT ("subsumptions:", subsumptions, average (conflicts, subsumptions), "interval");
#ifndef NSWITCH
  PRINT_COUNT ("switched-modes:", switched_modes, average (conflicts, switched_modes), "interval");
#endif
#ifndef NTRANSITIVE
  PRINT_COUNT ("transitive-reductions:", transitive_reductions, average (conflicts, transitive_reductions), "interval");
  PRINT_STATS ("transitive-reduced:", transitive_reduced, percent (transitive_reduced, added_binary_clauses), "% binaries");
  PRINT_STATS ("  transitive-checks:", transitive_checks, average (transitive_checks, transitive_reductions), "per reduction");
  PRINT_STATS ("  transitive-completed:", transitive_completed, average (transitive_reductions, transitive_completed), "reductions");
  PRINT_STATS ("  transitive-failed:", transitive_failed, percent (transitive_failed, activated_variables), "% activated");
#endif
#ifndef NVIVIFY
  PRINT_COUNT ("vivifications:", vivifications, average (conflicts, vivifications), "interval");
  PRINT_STATS ("  irredundant-completed:", vivification_completed[0], average (vivifications, vivification_completed[0]), "vivifications");
  PRINT_STATS ("  tier1-completed:", vivification_completed[1], average (vivifications, vivification_completed[1]), "vivifications");
  PRINT_STATS ("  tier2-completed:", vivification_completed[2], average (vivifications, vivification_completed[2]), "vivifications");
  PRINT_STATS ("vivified-clauses:", vivified_clauses.total, percent (vivified_clauses.total, learned_clauses), "% learned");
  PRINT_STATS ("  vivified-irredundant:", vivified_clauses.tier[0], percent (vivified_clauses.tier[0], added_original_clauses), "% original");
  PRINT_STATS ("  vivified-tier1:", vivified_clauses.tier[1], percent (vivified_clauses.tier[1], learned_tier1_clauses), "% learned tier2");
  PRINT_STATS ("  vivified-tier2:", vivified_clauses.tier[2], percent (vivified_clauses.tier[2], learned_tier2_clauses), "% learned tier2");
  PRINT_STATS ("  vivified-units:", vivified_units, percent (vivified_units, vivified_clauses.total), "% per vivified");
  PRINT_STATS ("vivify-checks:", vivify_checks, average (vivify_checks, vivified_clauses.total), "per vivified");
  PRINT_STATS ("  vivify-propagations:", vivification_propagations, percent (vivification_propagations, propagations), "% propagations");
#endif
#endif

  // clang-format on
}

void Solver::print_statistics (double t) {
  bool full = full_statistics;
#ifndef QUIET
  if (verbosity > 0)
    full = true;
#endif
  statistics.print (full, t, variables);
}

}; // namespace TabularaSAT
