#ifndef _tabularasat_statistics_hpp_INCLUDED
#define _tabularasat_statistics_hpp_INCLUDED

#include <cstdint>

#include "dependencies.hpp"
#include "variant.hpp"

namespace TabularaSAT {

// Common statistics are split into 'counters' and plain 'statistics'.
// While the former are needed for scheduling functions and other heuristics
// the latter are only used for monitoring what the solver is doing and not
// activated when compiling with 'QUIET' defined (using './configure -q').

// For counters we use macros with the suffix '..._COUNT' and for statistics
// the macros with suffix '..._STATS'. There is also C++ symbol prefix for
// these which is constructted with the following two macros.

// If you want to detect bad usages of directly accessing counters define
// 'CHECK_STATISTICS_USAGE' in the makefile manually.

#ifdef CHECK_STATISTICS_USAGE

// Here the prefix for counter and statistics is different, which eases
// figuring out bad (explicit) usage, i.e., the counts and statistics should
// only be accessed through macros, which however requires to disable
// printing as we currently use many statistics without prefix there and
// changing that is too tedious (for now).

#define COUNTER(SYMBOL) count_##SYMBOL
#define STATISTIC(SYMBOL) statistic_##SYMBOL

#else

// In a transition or testing mode we simply do not use any prefix.

#define COUNTER(SYMBOL) SYMBOL
#define STATISTIC(SYMBOL) SYMBOL

#endif

// Counter macros are always active (as we need those counters).

#define DEFINE_COUNT(SYMBOL) uint64_t COUNTER (SYMBOL)
#define GET_COUNT(SYMBOL) (statistics.COUNTER (SYMBOL))
#define GET_SOLVER_COUNT(SYMBOL) (solver.statistics.COUNTER (SYMBOL))
#define INC_COUNT(SYMBOL) (statistics.COUNTER (SYMBOL)++)
#define DEC_COUNT(SYMBOL) \
  (assert (statistics.COUNTER (SYMBOL)), statistics.COUNTER (SYMBOL)--)
#define ADD_COUNT(SYMBOL, INC) (statistics.COUNTER (SYMBOL) += (INC))
#define SET_COUNT(SYMBOL, VALUE) (statistics.COUNTER (SYMBOL) = (VALUE))

#ifndef QUIET

// Statistics macros are active in non-quiet mode.

#define DEFINE_STATS(SYMBOL) uint64_t STATISTIC (SYMBOL)
#define GET_STATS(SYMBOL) (statistics.STATISTIC (SYMBOL))
#define GET_SOLVER_STATS(SYMBOL) (solver.statistics.STATISTIC (SYMBOL))
#define INC_STATS(SYMBOL) (statistics.STATISTIC (SYMBOL)++)
#define DEC_STATS(SYMBOL) \
  (assert (statistics.STATISTIC (SYMBOL)), statistics.STATISTIC (SYMBOL)--)
#define ADD_STATS(SYMBOL, INC) (statistics.STATISTIC (SYMBOL) += (INC))

#else

// NO 'GET_STATS' and no 'GET_SOLVER_STATS' here as a compiler error in case
// such a statistic is used outside of quiet code is exactly what we want.

#define DEFINE_STATS(SYMBOL)   /**/
#define INC_STATS(SYMBOL)      /**/
#define DEC_STATS(SYMBOL)      /**/
#define ADD_STATS(SYMBOL, INC) /**/

#endif

struct Statistics {

  DEFINE_COUNT (activated_variables);
  DEFINE_STATS (active_variables);
  DEFINE_STATS (added_binary_clauses);
  DEFINE_STATS (added_clauses);
  DEFINE_STATS (added_large_clauses);
  DEFINE_STATS (added_original_clauses);
  DEFINE_STATS (added_variables);
#ifndef NALLUIPS
  DEFINE_STATS (additional_learned_uips);
#endif
  DEFINE_STATS (analyzed_conflicts);
  DEFINE_COUNT (binary_clauses);
  DEFINE_STATS (bumped_variables);
  DEFINE_STATS (chronological_backtracking);
#ifndef NCHRONO
  DEFINE_STATS (chronological_forced);
#endif
  DEFINE_STATS (collected_clauses);
  DEFINE_STATS (collected_references);
  DEFINE_COUNT (conflicts);
  DEFINE_COUNT (decisions);
  DEFINE_STATS (deduced_literals);
#ifndef NGATES
  DEFINE_STATS (eliminated_ands);
  DEFINE_STATS (eliminated_equivalences);
  DEFINE_STATS (eliminated_gates);
#endif
  DEFINE_STATS (eliminated_variables);
  DEFINE_COUNT (eliminations);
  DEFINE_COUNT (elimination_ticks);
#ifndef NSUBSTITUTE
  DEFINE_STATS (equivalent_literals);
#endif
#ifndef NPROBE
  DEFINE_STATS (failed_literal_decisions);
  DEFINE_COUNT (failed_literal_probings);
  DEFINE_STATS (failed_literal_propagations);
  DEFINE_STATS (failed_literals);
  DEFINE_COUNT (failed_literal_ticks);
  DEFINE_STATS (failed_literal_completed);
#endif
  DEFINE_COUNT (fixed_variables);
#ifndef NWALK
  DEFINE_STATS (flipped_literals);
#endif
#ifndef NSWITCH
  DEFINE_STATS (focused_conflicts);
  DEFINE_STATS (focused_ticks);
#endif
  DEFINE_STATS (forced_reductions);
  DEFINE_COUNT (garbage_clauses);
  DEFINE_STATS (garbage_collections);
#ifndef NWALK
  DEFINE_STATS (improved_minimum);
#endif
#ifdef IMPROVEREASONS
  DEFINE_STATS (improved_reasons);
#endif
  DEFINE_STATS (increased_kept);
  DEFINE_COUNT (irredundant_clauses);
#ifndef NJUMPREASONS
  DEFINE_STATS (jumped_reasons);
#endif
  DEFINE_STATS (learned_clauses);
  DEFINE_STATS (learned_literals);
#ifndef NGLUE
  DEFINE_STATS (learned_tier1_clauses);
  DEFINE_STATS (learned_tier2_clauses);
#endif
  DEFINE_STATS (learned_units);
#ifndef NWALK
  DEFINE_COUNT (local_searches);
  DEFINE_COUNT (local_search_ticks);
#endif
  DEFINE_STATS (minimized_literals);
  DEFINE_STATS (moved_redundant_clauses);
#ifndef NOTFS
  DEFINE_STATS (on_the_fly_strengthened);
  DEFINE_STATS (on_the_fly_subsumed);
#endif
  DEFINE_COUNT (preprocessing_eliminations);
#ifndef NPROBE
  DEFINE_COUNT (preprocessing_probings);
#endif
  DEFINE_COUNT (preprocessing_subsumptions);
#ifndef NSUBSTITUTE
  DEFINE_COUNT (preprocessing_substitutions);
#endif
#ifndef NTRANSITIVE
  DEFINE_COUNT (preprocessing_transitive_reductions);
#endif
#ifndef NVIVIFY
  DEFINE_COUNT (preprocessing_vivifications);
#endif
#ifndef NPROMOTE
  DEFINE_STATS (promoted_clauses);
#endif
  DEFINE_STATS (propagations);
#ifndef NBUMPREASONS
  DEFINE_STATS (reason_side_bumped);
#endif
  DEFINE_COUNT (reductions);
  DEFINE_COUNT (redundant_clauses);
  DEFINE_STATS (removed_duplicated_binaries);
  DEFINE_STATS (removed_variables);
#ifndef NREPHASE
  DEFINE_COUNT (rephasings);
#endif
  DEFINE_STATS (reported);
  DEFINE_STATS (resolutions);
  DEFINE_COUNT (restarts);
#ifndef NREUSETRAIL
  DEFINE_STATS (reused_levels);
  DEFINE_STATS (reused_trails);
#endif
#ifndef NSHRINK
  DEFINE_STATS (shrunken_literals);
#endif
  DEFINE_STATS (search_propagations);
  DEFINE_COUNT (search_ticks);
  DEFINE_STATS (simplifications);
#ifndef NSWITCH
  DEFINE_STATS (stable_conflicts);
  DEFINE_STATS (stable_ticks);
#endif
  DEFINE_STATS (strengthened);
#ifndef NSUBSTITUTE
  DEFINE_STATS (substituted_variables);
  DEFINE_COUNT (substitutions);
#endif
  DEFINE_STATS (subsume_checks);
  DEFINE_STATS (subsumed_clauses);
  DEFINE_COUNT (subsumptions);
  DEFINE_COUNT (subsumption_ticks);
#ifndef NSWITCH
  DEFINE_COUNT (switched_into_second_mode);
  DEFINE_COUNT (switched_modes);
#endif
#ifndef NTRANSITIVE
  DEFINE_STATS (transitive_checks);
  DEFINE_STATS (transitive_completed);
  DEFINE_STATS (transitive_failed);
  DEFINE_STATS (transitive_reduced);
  DEFINE_COUNT (transitive_reductions);
#endif
  DEFINE_STATS (unforced_reductions);
#ifndef NREPHASE
  DEFINE_STATS (updated_best_phases);
#endif
#ifndef NTARGET
  DEFINE_STATS (updated_target_phases);
#endif
#ifndef NVIVIFY
  DEFINE_STATS (vivification_completed[3]); // 0=irr, 1=tier1, 2=tier2
  DEFINE_COUNT (vivifications);
  DEFINE_STATS (vivification_propagations);
  DEFINE_COUNT (vivification_ticks);
  struct {
#ifndef QUIET
    uint64_t tier[3];
#endif
    DEFINE_COUNT (total);
  } vivified_clauses;
  DEFINE_STATS (vivified_units);
  DEFINE_STATS (vivify_checks);
#endif

  Statistics ();

  void print (bool full, const char *, uint64_t, double, const char *);
  void print (bool full, double time_in_seconds, uint64_t total_variables);
};

#undef DEFINE_COUNT
#undef DEFINE_STATS

}; // namespace TabularaSAT

#endif
