#include "references.hpp"
#include "schedule.hpp"
#include "solver.hpp"
#include "ticks.hpp"

#include <cinttypes>

namespace TabularaSAT {

bool Solver::reducing_binary_implication_graph () {
  if (!OPTION (inprocessing))
    return false;
  if (!OPTION (transitive_reduction))
    return false;
  return GET_COUNT (conflicts) >= limits.transitive;
}

bool Solver::transitive_reduction () {

  assert (OPTION (transitive_reduction));
  assert (OPTION (preprocessing) || OPTION (inprocessing));

  if (level && !backtrack_to_root_level_and_propagate ())
    return false;

  assert (!level);
  assert (watching);
  assert (!inconsistent);

  start_simplifier ("transitive");
  TransitiveReductionScheduler scheduler (this);
  scheduler.increment_counter ();

  uint64_t limit;
  if (preprocessing) {
    unsigned minimum_effort = OPTION (minimum_effort);
    unsigned factor = OPTION (transitive_preprocessing);
    limit = factor * minimum_effort;
    message (2, "transitive reduction effort %" PRIu64 " = %u * %u ticks",
             limit, factor, minimum_effort);
  } else
    limit = scheduler.get_ticks_limit ();

  unsigned start_src = next_transitive_src;

  message (2, "begin transitive reduction at %.0f%% (literal %u)",
           percent (start_src, literals), start_src);

  assert (start_src < literals);
  bool completed = false;
  uint64_t reduced = 0;
  unsigned failed = 0;
  uint64_t ticks = 0;
  unsigned tried = 0;

  std::vector<unsigned> marked;

  while (!completed && !inconsistent && !terminated && ticks < limit) {

    tried++;

    unsigned src = next_transitive_src++;
    if (next_transitive_src == literals) {
#ifndef QUIET
      statistics.transitive_completed++;
      message (1, "transitive reduction completed %" PRIu64 " times",
               statistics.transitive_completed);
#endif
      next_transitive_src = 0;
    }

    if (active (src)) {
#ifndef QUIET
      debug ("transitive reduction of source literal %s",
             debug_literal (src));
      statistics.transitive_checks++;
#endif
      unsigned not_src = negate (src);

      auto src_watches = references (not_src);
      ticks += 1 + ticks_references (src_watches);
      auto begin_src = src_watches->begin ();
      auto end_src = src_watches->end ();

      uint64_t reduced_before = reduced;
      bool failing = false;

      for (auto p = begin_src; !failing && p != end_src; p++) {

        Reference ref = *p;
        if (!tagged (ref)) {
          p++;
          continue;
        }

        unsigned dst = untag (ref);
        if (dst < src)
          continue;
        if (!active (dst))
          continue;

        debug_binary (not_src, dst, "transitive reduction of");

        assert (!marks[src]);
        assert (!marks[not_src]);
        marks[src] = 1;
        marks[not_src] = -1;

        debug ("transitive assuming %s", debug_literal (src));

        assert (marked.empty ());
        marked.push_back (src);

        bool transitive = false;
        bool failed = false;
        size_t next = 0;

        while (!transitive && !failed && ticks < limit &&
               next != marked.size ()) {
          unsigned lit = marked[next++];
          debug ("transitive propagate %s", debug_literal (lit));
          unsigned not_lit = negate (lit);

          const auto lit_watches = references (not_lit);
          ticks += 1 + ticks_references (lit_watches);
          auto begin_lit = lit_watches->begin ();
          auto end_lit = lit_watches->end ();

          for (auto q = begin_lit; q != end_lit; q++) {
            if (p == q)
              continue;
            Reference ref = *q;
            if (!tagged (ref)) {
              q++;
              continue;
            }
            if (ref == INVALID_REFERENCE)
              continue;
            unsigned other = untag (ref);
            if (other == dst) {
              debug ("can reach %s from %s transitively too",
                     debug_literal (src), debug_literal (dst));

              transitive = true;
              break;
            }
            signed char mark = marks[other];
            if (mark < 0) {
              debug ("both %s and %s reachable from %s",
                     debug_literal (negate (other)), debug_literal (other),
                     debug_literal (src));
              failing = true;
              break;
            }
            if (mark > 0)
              continue;

            debug ("transitive implied %s", debug_literal (other));

            unsigned not_other = negate (other);
            assert (!marks[not_other]);
            assert (!marks[other]);
            marks[not_other] = -1;
            marks[other] = 1;

            marked.push_back (other);
          }
        }

        for (auto lit : marked)
          marks[lit] = marks[negate (lit)] = 0;
        marked.clear ();

        if (transitive) {
          mark_binary_garbage (not_src, dst, "transitive reduced");
          unwatch_binary (dst, not_src);
          *p = INVALID_REFERENCE;
#ifndef QUIET
          statistics.transitive_reduced++;
#endif
          reduced++;
        }
      }

      if (reduced_before < reduced) {
        auto q = begin_src;
        for (auto p = q; p != end_src; p++) {
          Reference ref = *p;
          if (ref == INVALID_REFERENCE)
            continue;
          *q++ = ref;
          if (!tagged (ref))
            *q++ = *++p;
        }
        src_watches->resize (q - begin_src);
      }

      if (failing) {
#ifndef QUIET
        statistics.transitive_failed++;
#endif
        debug ("failed transitive assumption %s", debug_literal (src));
        unsigned unit = not_src;
        trace_unit ('a', unit);
        assign (unit, UNIT_REASON);
        failed++;
        if (!search_propagate ()) {
          inconsistent = true;
          trace_empty ('a');
        }
      }
    }

    if (next_transitive_src == start_src)
      completed = true;
  }

  message (2, "end partial transitive reduction at %.0f%% (literal %u)",
           percent (next_transitive_src, literals), next_transitive_src);

  message (2, "%s transitive reduction of %.0f%% literals",
           tried >= literals ? "full" : "partial",
           percent (tried, literals));

  if (reduced)
    message (1, "removed %" PRIu64 " transitive binary clauses", reduced);

  if (failed)
    message (1, "found %u transitive failed literals", failed);

  scheduler.set_next_conflict_limit ();
  bool report = reduced || failed;
#ifndef QUIET
  if (verbosity > 0)
    report = true;
#else
  (void) tried;
#endif
  if (report)
    report_progress ('t');
  stop_simplifier ("transitive");

  return reduced || failed;
}

}; // namespace TabularaSAT
