#ifndef _arena_hpp_INCLUDED
#define _arena_hpp_INCLUDED

#include <cassert>
#include <cstdlib>

namespace TabularaSAT {

// Arena for allocating clauses and clause references (watches and
// occurrences) consecutively and referencing respectively accessing them
// through their unsigned word position in the arena.

// In order to allow iteration of 'C' objects in the arena, the client class
// 'C' (such as 'Clause' or 'References') needs to support 'next ()' which
// should point to the end of the allocated 'C' object. This also needs to
// be consistent with 'n' the number of words given as argument to 'allocate
// (size_t n)'.  There is also a corresponding 'words ()' function in
// 'Clause' and 'References' that has to match this 'n' too.

// As the 'arena' is indepdent from 'Shared' (and 'Solver') we have to use a
// separate indepdent out-of-memory fatal error message here too.

void fatal_error (const char *where, const char *, ...)
    __attribute__ ((format (printf, 2, 3)));

template <class C> class Arena {
  unsigned *_begin = 0, *_end = 0, *_allocated = 0;

public:
  class iterator {
    C *current;

  public:
    iterator (C *c) : current (c) {}
    void operator++ () { current = current->next (); }
    C *operator* () const { return current; }
    friend bool operator!= (const iterator &a, const iterator &b) {
      return a.current != b.current;
    }
  };

  ~Arena () { free (_begin); }

  size_t size () { return _end - _begin; }
  size_t capacity () { return _allocated - _begin; }

  C *allocate (size_t n) {
    size_t old_size = size ();
    size_t new_size = old_size + n;
    size_t old_capacity = capacity ();
    if (new_size > old_capacity) {
      size_t new_capacity = old_capacity ? 2 * old_capacity : 1;
      while (new_size > new_capacity)
        new_capacity *= 2;
      size_t new_bytes = new_capacity * sizeof *_begin;
      _begin = (unsigned *) realloc (_begin, new_bytes);
      if (!_begin)
        fatal_error ("Arena::allocate", "out-of-memory");
      _end = _begin + old_size;
      _allocated = _begin + new_capacity;
    }
    C *res = (C *) _end;
    _end += n;
    return res;
  }

  size_t position (C *c) {
    unsigned *w = (unsigned *) c;
    assert (_begin <= w), assert (w <= _end);
    return w - _begin;
  }

  void resize (C *new_end) {
    unsigned *new_word_end = (unsigned *) new_end;
    assert (_begin <= new_word_end);
    assert (new_word_end <= _end);
    _end = new_word_end;
  }

  bool shrink_to_fit () {
    size_t old_size = size ();
    size_t old_capacity = capacity ();
    if (old_size >= old_capacity / 4)
      return false;
    size_t new_capacity = old_capacity / 2;
    assert (old_size <= new_capacity);
    assert (new_capacity);
    size_t new_bytes = new_capacity * sizeof *_begin;
    _begin = (unsigned *) realloc (_begin, new_bytes);
    if (!_begin)
      fatal_error ("Arena::shrink_to_fit", "out-of-memory");
    _end = _begin + old_size;
    _allocated = _begin + new_capacity;
    return true;
  }

  // Access a 'C' object from its (word) position 'pos' in the arena.  Note
  // that 'pos' most of the time does not correspond to the "pos'th" object.

  C *operator[] (size_t pos) {
    unsigned *w = _begin + pos;
    C *res = (C *) w;
    assert (w == _end || res->next () <= (C *) _end);
    return res;
  }

  const C *operator[] (size_t pos) const {
    unsigned *w = _begin + pos;
    C *res = (C *) w;
    assert (w == _end || res->next () <= (C *) _end);
    return res;
  }

  iterator begin () { return iterator ((C *) _begin); }
  iterator end () { return iterator ((C *) _end); }
};

}; // namespace TabularaSAT

#endif
