#ifndef _heap_h_INCLUDED
#define _heap_h_INCLUDED

#include <cassert>
#include <climits>
#include <vector>

namespace TabularaSAT {

// Generic pairing heap implementation with floating priorities for elements
// encoded as unsigned integers (different from 'UINT_MAX').

#define HEAP_INVALID UINT_MAX

template <class Less> class Heap {

protected:
  struct Node {
    unsigned child = HEAP_INVALID;
    unsigned prev = HEAP_INVALID;
    unsigned next = HEAP_INVALID;
  };

  std::vector<Node> nodes;
  unsigned root = HEAP_INVALID;

private:
  unsigned merge (unsigned a, unsigned b, Less less) {
    if (a == HEAP_INVALID)
      return b;
    if (b == HEAP_INVALID)
      return a;
    unsigned parent, child;
    if (less (b, a))
      parent = b, child = a;
    else
      parent = a, child = b;
    auto &node_parent = nodes[parent];
    unsigned parent_child = node_parent.child;
    auto &node_child = nodes[child];
    node_child.next = parent_child;
    if (parent_child != HEAP_INVALID)
      nodes[parent_child].prev = child;
    node_child.prev = parent;
    node_parent.child = child;
    node_parent.prev = node_parent.next = HEAP_INVALID;
    return parent;
  }

  unsigned collapse (unsigned idx, Less less) {
    if (idx == HEAP_INVALID)
      return HEAP_INVALID;
    unsigned next = idx, tail = HEAP_INVALID;
    do {
      unsigned a = next;
      assert (a != HEAP_INVALID);
      auto &node_a = nodes[a];
      unsigned b = node_a.next;
      if (b == HEAP_INVALID) {
        node_a.prev = tail;
        tail = a;
        break;
      }
      next = nodes[b].next;
      unsigned tmp = merge (a, b, less);
      assert (tmp != HEAP_INVALID);
      nodes[tmp].prev = tail;
      tail = tmp;
    } while (next != HEAP_INVALID);

    unsigned res = HEAP_INVALID;
    while (tail != HEAP_INVALID) {
      unsigned prev = nodes[tail].prev;
      res = merge (res, tail, less);
      tail = prev;
    }
    return res;
  }

  void dequeue (unsigned idx) {
    assert (idx != HEAP_INVALID);
    auto &node_idx = nodes[idx];
    unsigned prev = node_idx.prev;
    unsigned next = node_idx.next;
    assert (prev != HEAP_INVALID);
    auto &node_prev = nodes[prev];
    node_idx.prev = HEAP_INVALID;
    if (node_prev.child == idx)
      node_prev.child = next;
    else
      node_prev.next = next;
    if (next != HEAP_INVALID)
      nodes[next].prev = prev;
  }

public:
  size_t expensive_size (unsigned idx) const {
    if (idx == HEAP_INVALID)
      return 0;
    size_t res = 1;
    auto &node = nodes[idx];
    for (unsigned child_idx = node.child, next_child_idx;
         child_idx != HEAP_INVALID; child_idx = next_child_idx) {
      auto &child = nodes[child_idx];
      next_child_idx = child.next;
      res += expensive_size (child_idx);
    }
    return res;
  }

  size_t expensive_size () const { return expensive_size (root); }

  void update (unsigned idx, Less less) {
    auto &node_idx = nodes[idx];
    unsigned tmp = root;
    if (tmp == idx)
      return;
    if (node_idx.prev == HEAP_INVALID)
      return;
    dequeue (idx);
    root = merge (tmp, idx, less);
  }

  unsigned operator* () const { return root; }

  bool empty () const { return root == HEAP_INVALID; }

  size_t capacity () const { return nodes.size (); }

  bool contains (unsigned idx) const {
    if (idx >= capacity ())
      return false;
    return root == idx || nodes[idx].prev != HEAP_INVALID;
  }

  void push (unsigned idx, Less less) {
    assert (idx < capacity ());
    assert (idx != HEAP_INVALID);
    assert (!contains (idx));
    nodes[idx].child = HEAP_INVALID;
    root = merge (root, idx, less);
    assert (contains (idx));
  }

  void resize (unsigned variables) {
    unsigned old_size = nodes.capacity (), idx = old_size;
    while (idx != variables)
      idx++, nodes.push_back (Node ());
  }

  void pop (Less less) {
    unsigned idx = root;
    assert (idx != HEAP_INVALID);
    unsigned child = nodes[idx].child;
    root = collapse (child, less);
  }

  void remove (unsigned idx, Less less) {
    assert (idx != HEAP_INVALID);
    assert (contains (idx));
    unsigned child = nodes[idx].child;
    if (root == idx)
      root = collapse (child, less);
    else {
      dequeue (idx);
      assert (nodes[idx].child == child);
      unsigned collapsed = collapse (child, less);
      root = merge (root, collapsed, less);
    }
    assert (!contains (idx));
  }
};

// Exponential VSIDS priority queue implemented as pairing heap.

#define HEAP_SCORE_LIMIT 1e150

class Bumper {

  double increment = 1.0;
  double factor = 1.05;

  std::vector<double> scores;

  struct larger_score {
    const std::vector<double> &scores;
    larger_score (const std::vector<double> &s) : scores (s) {}
    bool operator() (unsigned a, unsigned b) {
      assert (a < scores.size ()), assert (b < scores.size ());
      return scores[a] > scores[b];
    }
  };

  Heap<larger_score> heap;

  void rescore (double scale) {
    assert (scale > 0);
    double inverse = 1 / scale;
    for (auto &score : scores)
      score *= inverse;
    increment *= inverse;
  }

public:
  void set_decay (double d) {
    assert (0.5 <= d), assert (d <= 1.0);
    factor = 1 / d;
  }

  size_t capacity () const {
    assert (scores.size () == heap.capacity ());
    return scores.size ();
  }

  double score (unsigned idx) const {
    assert (idx < capacity ());
    return scores[idx];
  }

  void resize (unsigned variables) {
    unsigned old_size = scores.capacity (), idx = old_size;
    while (idx != variables)
      idx++, scores.push_back (0);
    heap.resize (variables);
  }

  void bump (unsigned idx) {
    assert (idx < scores.size ());
    assert (idx < heap.capacity ());
    double new_score = (scores[idx] += increment);
    heap.update (idx, larger_score (scores));
    if (new_score > HEAP_SCORE_LIMIT)
      rescore (new_score);
  }

  void bump () {
    increment *= factor;
    if (increment > HEAP_SCORE_LIMIT)
      rescore (increment);
  }

  unsigned operator* () const { return *heap; }

  bool contains (unsigned idx) const { return heap.contains (idx); }
  void push (unsigned idx) { heap.push (idx, larger_score (scores)); }
  void pop () { heap.pop (larger_score (scores)); }
  void remove (unsigned idx) { heap.remove (idx, larger_score (scores)); }
};

}; // namespace TabularaSAT

#endif
