#include "references.hpp"
#include "solver.hpp"

namespace TabularaSAT {

void Solver::assign (unsigned lit, Reference reason) {

  unsigned assignment_level;

#ifndef NCHRONO

  // With chronological backtracking enabled variables are assigned at an
  // assignment level computed from the reason which might be smaller than
  // the current decision level (and thus the assignment out-of-order). In
  // the following part we compute the assignment level explicitly.

  // Furthermore we try to reduce the length of binary clause implication
  // chains by 'reason jumping' and in essence generating transitive
  // implicit hyper binary clauses on-the-fly, which however are only used
  // as reasons and thus are not really added.

  if (reason == DECISION_REASON)
    assignment_level = level;
  else if (reason == UNIT_REASON)
    assignment_level = 0;
  else if (tagged (reason)) {
    unsigned other = untag (reason);
    assert (values[other] < 0);
    unsigned other_idx = index (other);
    assignment_level = levels[other_idx];
#ifndef NJUMPREASONS
    if (assignment_level > 0 && OPTION (jump_reasons)) {
      Reference other_reason = reasons[other_idx];
      if (binary_reference (other_reason)) {
#ifndef QUIET
        statistics.jumped_reasons++;
#endif
        debug_binary (lit, other, "jumping reason");
        assert (values[untag (other_reason)] < 0);
        reason = other_reason;
      }
    }
#endif
  } else {
    assignment_level = 0;
    Clause *c = dereference_clause (reason);
    for (auto other : *c) {
      if (other == lit)
        continue;
      unsigned other_level = levels[index (other)];
      if (other_level > assignment_level)
        assignment_level = other_level;
    }
  }
#else
  assignment_level = level; // Without chronological backtracking.
#endif

  // Here comes the usual assignment logic.

  unsigned idx = index (lit);
  levels[idx] = assignment_level;
  if (!assignment_level) {
    Flags &f = flags[idx];
#ifndef NSUBSTITUTE
    if (f.substituted) {
      assert (!f.active);
#ifndef QUIET
      statistics.substituted_variables--;
#endif
      f.substituted = false;
    } else
#endif
    {
      assert (f.active);
      assert (!f.fixed);
      assert (!f.eliminated);
      deactivate_variable (idx);
    }
    f.fixed = true;
    INC_COUNT (fixed_variables);
    reason = UNIT_REASON;
#if !defined(NCHRONO) && !defined(QUIET)
    if (level) {
      out_of_order_units++;
      assert (out_of_order_units);
    }
#endif
  } else {
    assert (unassigned);
    unassigned--;
    debug ("now %u variables unassigned", unassigned);
  }

  reasons[idx] = reason;

#ifndef QUIET
  if (reason == DECISION_REASON && !flags[idx].active) {
    assigned_eliminated++;
    assert (assigned_eliminated);
  }
#endif

  bool save_phase = true;
#ifndef NVIVIFY
  if (vivifying)
    save_phase = !assignment_level;
#endif
#ifndef NPROBE
  assert (!probing || !assignment_level || reason == DECISION_REASON);
#endif
  if (save_phase)
    phases[idx].saved = sign (lit) ? -1 : 1;

  unsigned not_lit = negate (lit);
  assert (!values[lit]);
  assert (!values[not_lit]);
  values[lit] = 1;
  values[not_lit] = -1;

  assert (trail.size () <= UINT_MAX);
  position[idx] = trail.size ();
  trail.push_back (lit);

#ifdef LOGGING
  if (logging ()) {
    const char *s = debug_literal (lit);
    if (reason == DECISION_REASON)
      debug ("assign %s as decision", s);
    else if (reason == UNIT_REASON)
      debug ("assign %s as unit", s);
    else if (tagged (reason))
      debug_binary (lit, untag (reason), "assign %s with reason", s);
    else
      debug_reference (reason, "assign %s reason", s);
#ifndef NCHRONO
    if (assignment_level != level)
      debug ("out-of-order assigned %s", s);
#endif
  }
#endif
}

#ifndef NPROBE

void Solver::failed_literal_assign (unsigned lit, Reference reason) {

  if (level) {
    assert (level == 1);
    unsigned idx = index (lit);
    levels[idx] = level;
    assert (unassigned);
    unassigned--;
    debug ("now %u variables unassigned", unassigned);

    unsigned not_lit = negate (lit);
    assert (!values[lit]);
    assert (!values[not_lit]);
    values[lit] = 1;
    values[not_lit] = -1;

    assert (trail.size () <= UINT_MAX);
    trail.push_back (lit);
  } else
    assign (lit, UNIT_REASON);

  debug ("assign %s during failed literal probing", debug_literal (lit));
}

#endif

}; // namespace TabularaSAT
