#include "clause.hpp"
#include "inline.hpp"
#include "references.hpp"
#include "solver.hpp"
#include "watch.hpp"

#include <cinttypes>

namespace TabularaSAT {

void Solver::reset_positions () {
  first_redundant = INVALID_POSITION;
#ifndef NGLUE
  first_tier2 = INVALID_POSITION;
#endif
  last_irredundant = 0;
}

void Solver::finalize_position_update () {
  if (first_redundant == INVALID_POSITION)
    message (2, "no redundant clause left");
  else
    message (2, "first redundant clause position %zu", first_redundant);
#ifndef NGLUE
  if (first_tier2 == INVALID_POSITION)
    message (2, "no redundant tier2 clause left");
  else
    message (2, "first redundant tier2 clause position %zu", first_tier2);
#endif
  message (2, "last irredundant clause position %zu", last_irredundant);
}

void Solver::mark_clause_garbage (Clause *c, const char *reason) {
  debug_clause (c, "marking garbage and deleting %s", reason);
  assert (!c->garbage);
  c->garbage = true;
  INC_COUNT (garbage_clauses);
  trace_clause ('d', c);
  if (c->redundant)
    DEC_COUNT (redundant_clauses);
  else {
    DEC_COUNT (irredundant_clauses);
    mark_removed (c);
  }
  (void) reason;
}

void Solver::mark_binary_garbage (unsigned lit, unsigned other,
                                  const char *reason) {
  debug_binary (lit, other, "marking garbage %s", reason);
  mark_removed (lit), mark_removed (other);
  trace_binary ('d', lit, other);
  DEC_COUNT (binary_clauses);
  (void) reason;
}

void Solver::mark_reason_clauses (bool mark) {
  assert (level);
  size_t begin = trail_root_level ();
  size_t end = trail.size ();
  assert (begin <= end);
  size_t marked = 0;
  for (size_t i = begin; i != end; i++) {
    unsigned lit = trail[i];
    unsigned idx = index (lit);
    Reference reason = reasons[idx];
    if (reason == DECISION_REASON)
      continue;
    if (reason == UNIT_REASON)
      continue;
    if (tagged (reason))
      continue;
    Clause *c = dereference_clause (reason);
    assert (c->reason != mark);
    c->reason = mark;
    marked++;
  }
#ifndef QUIET
  if (mark)
    message (2, "protected %zu reason clauses", marked);
  else
    message (2, "released protection of %zu reason clauses", marked);
#else
  (void) marked;
#endif
}

void Solver::map_reasons (Position first_swept,
                          const std::vector<Position> &map) {
  assert (level);
  size_t begin = trail_root_level ();
  size_t end = trail.size ();
  assert (begin <= end);
  size_t mapped = 0;
  for (size_t i = begin; i != end; i++) {
    unsigned lit = trail[i];
    unsigned idx = index (lit);
    Reference reason = reasons[idx];
    if (reason == DECISION_REASON)
      continue;
    if (reason == UNIT_REASON)
      continue;
    if (tagged (reason))
      continue;
    Position src_pos = untag (reason);
    if (src_pos < first_swept)
      continue;
    Position delta = src_pos - first_swept;
    assert (delta / 4 < map.size ());
    Position dst_pos = map[delta / 4];
    if (dst_pos == INVALID_POSITION)
      continue;
    reasons[idx] = tag_large (dst_pos);
    mapped++;
  }
#ifndef QUIET
  message (2, "mapped %zu reason clause references", mapped);
#else
  (void) mapped;
#endif
}

void Solver::mark_satisfied_and_shrunken_clauses (Position &first_marked) {

  size_t new_garbage, old_garbage, marked_shrink;
  size_t first_garbage, first_shrunken;

  new_garbage = old_garbage = marked_shrink = 0;
  first_garbage = first_shrunken = INVALID_POSITION;

  for (auto c : clauses) {

    Position pos = clauses.position (c);

    if (c->garbage) {
      old_garbage++;
      if (pos < first_garbage)
        first_garbage = pos;
      continue;
    }

    bool shrink = false, satisfied = false, eliminated = false;

    for (auto lit : *c) {
      unsigned idx = index (lit);
      if (eliminating && flags[idx].eliminated) {
        eliminated = true;
        break;
      }
      signed char value = values[lit];
      if (!value)
        continue;
      if (levels[idx])
        continue;
      if (value > 0) {
        satisfied = true;
        break;
      }
      if (value < 0)
        shrink = true;
    }

    if (!satisfied && !shrink && !eliminated)
      continue;

    if (satisfied) {
      if (pos < first_garbage)
        first_garbage = pos;
      mark_clause_garbage (c, "satisfied");
      assert (c->garbage);
      new_garbage++;
    } else if (eliminated) {
      assert (!watching);
      assert (eliminating);
      if (pos < first_garbage)
        first_garbage = pos;
      mark_clause_garbage (c, "eliminated");
      assert (c->garbage);
      new_garbage++;
    } else if (!eliminating) {
      if (pos < first_shrunken)
        first_shrunken = pos;
      debug_clause (c, "marked to shrink");
      assert (shrink);
      assert (!c->shrink);
      c->shrink = true;
      marked_shrink++;
    } else {

      // Without watches during variable elimination we do not want to
      // shrink clauses as it would require to disconnect and reconnect
      // such clauses which is complex, costly and not really needed.

      assert (!watching);
    }
  }

  if (first_garbage != INVALID_POSITION) {
    if (first_garbage < first_marked)
      first_marked = first_garbage;
    message (2, "first garbage clause at position %zu", first_garbage);
  }

  if (first_shrunken != INVALID_POSITION) {
    if (first_shrunken < first_marked)
      first_marked = first_shrunken;
    message (2, "first shrunken clause at position %zu", first_shrunken);
  }
#ifndef QUIET
  size_t total_garbage = old_garbage + new_garbage;
  message (1,
           "marked %zu clauses as garbage "
           " (%zu old %.0f%%, %zu new %.0f%%)",
           total_garbage, old_garbage, percent (old_garbage, total_garbage),
           new_garbage, percent (new_garbage, total_garbage));
  message (1, "marked %zu clauses to be shrunken", marked_shrink);
#else
  (void) old_garbage;
  (void) new_garbage;
  (void) marked_shrink;
#endif
}

void Solver::mark_simplified_clauses (Position &first_simplified) {

  unsigned last_fixed = last.fixed_variables_at_reduce_or_collect;
  unsigned now_fixed = GET_COUNT (fixed_variables);
  if (eliminating || last_fixed < now_fixed) {
    if (eliminating)
      message (2, "full simplification during variable elimination");
    else
      message (2, "full simplification due to %" PRIu64 " new units",
               statistics.fixed_variables -
                   last.fixed_variables_at_reduce_or_collect);
#ifndef QUIET
    statistics.simplifications++;
#endif
    mark_satisfied_and_shrunken_clauses (first_simplified);

  } else if (GET_COUNT (garbage_clauses)) {

    message (2,
             "no new unit so only taking care of %" PRIu64
             " previously marked garbage clauses",
             GET_COUNT (garbage_clauses));
#ifndef QUIET
    statistics.simplifications++;
#endif
    for (auto c : clauses)
      if (c->garbage) {
        size_t pos = clauses.position (c);
        if (pos < first_simplified) {
          message (2, "first garbage clause at position %zu", pos);
          first_simplified = pos;
        }
        break;
      }
  } else
    message (2, "no new units nor previously marked garbage clauses");
}

// This is the actual compacting garbage collection of the clause arena.
// It removes garbage clauses and shrinks clauses previously marked to be
// shrunken since they contain a root-level falsified literal.

void Solver::copy_and_collect_clauses (Position first_swept,
                                       std::vector<Position> &map) {
  Clause *begin = clauses[first_swept];
  Clause *end = *clauses.end ();

  assert (clauses.size () <= UINT_MAX);
  assert ((unsigned *) end - (unsigned *) begin <= UINT_MAX);

  size_t traversed = 0, mapped = first_swept;
  size_t collected_clauses = 0, copied_clauses = 0;
  size_t shrunken_clauses = 0, shrunken_literals = 0;

  Clause *src = begin, *dst = begin;

  if (first_swept <= first_redundant)
    first_redundant = INVALID_POSITION;
#ifndef NGLUE
  if (first_swept <= first_tier2)
    first_tier2 = INVALID_POSITION;
#endif
  if (first_swept <= last_irredundant)
    last_irredundant = first_swept;

  while (src != end) {

    assert (clauses[first_swept + traversed] == src);
    if (map.size () < traversed / 4)
      map.resize (traversed / 4);

    Clause *next_src = src->next ();
    unsigned src_size = src->size;
    bool collect;

    if (src->garbage) // Just collect.
      collect = true;
    else if (!src->shrink) {

    COPY:

      debug_clause (src, "copying");
      size_t bytes = Clause::bytes (src_size);
      memmove (dst, src, bytes);
      copied_clauses++;
      collect = false;

    } else { // Shrink clause.

      // See discussion in 'mark_simplified_clauses'.  We do not want to
      // shrink clauses when garbage collecting eliminated clauses during
      // variable eliminations as it is awkward, complex and redundant.

      assert (!eliminating);
      assert (watching);

      assert (clause.empty ());

      // Copy root-level unassigned literals to temporary 'clause' and if
      // the clause is late satisfied, i.e., by a unit obtained from
      // shrinking a previous clause to size one, then mark and delete the
      // clause and treat as if it was just collected.

      unsigned *begin_src_literals = src->literals;
      unsigned *end_src_literals = begin_src_literals + src_size;
      unsigned *p = begin_src_literals;

      while (p != end_src_literals) {
        unsigned lit = *p++;
        signed value = values[lit];
        if (value) {
          unsigned idx = index (lit);
          if (!levels[idx]) {
            assert (value <= 0);
            if (value < 0)
              continue; // Skip root-level falsified.
          }
        }
        clause.push_back (lit); // Keep root-level unassigned.
      }

      debug_clause (src, "shrinking");

      // At least one root-level literal should be there and therefore the
      // size of the clause should shrink.  Thus instead of copying the
      // 'src' literals as is we copy only the unassigned ones from the
      // temporary clause which already has the size of the shrunken
      // clause.

      unsigned dst_size = clause.size ();
      assert (dst_size);
      assert (dst_size < src_size);

      // Without chronological backtracking the shrunken clause is always
      // at least binary, but with it the clause might be a hidden unit
      // clause not detected yet, e.g., as the root-level unit
      // '22(12)@1=1' in '22(12)@1=1 20(11)@0=-1 19(-10)@0=-1'.  We tried
      // to reassign the resulting unit on the root-level and mark such a
      // clause as garbage, but that resulted in additional new 'late'
      // satisfied clauses and other complexities (such as a form of
      // required propagation somewhere here or during flushing of
      // watches), which we now simply avoid by delaying the detection of
      // this root level unit until we backtrack to decision level zero
      // and propagate.

      if (dst_size == 1) {

        debug_clause (src, "unit %s due to chronological backtracking in",
                      debug_literal (clause[0]));

        assert (!src->garbage);
        src->shrink = false;

        clause.clear ();
        goto COPY;
      }

      // Literals of shrunken clauses are marked as added such that other
      // (larger) clauses with at least two of these literals are checked
      // for subsumption later.

      for (auto lit : clause)
        mark_added (lit);

      shrunken_literals += src_size - dst_size;

      trace_temporary ('a');
      trace_clause ('d', src);

      if (dst_size == 2) { // Shrunken clause binary.

        debug_binary (clause[0], clause[1], "reduced");

        if (src->reason) {
          if (values[clause[0]] < 0)
            std::swap (clause[0], clause[1]);
          assert (values[clause[0]] > 0);
          assert (values[clause[1]] < 0);
          Reference reason = new_binary ();
          debug_binary (clause[0], clause[1], "updated reason of %s",
                        debug_literal (clause[0]));
          unsigned idx = index (clause[0]);
          reasons[idx] = reason;
        } else
          new_binary ();

        debug_clause (src, "delete");

        if (src->redundant)
          DEC_COUNT (redundant_clauses);
        else
          DEC_COUNT (irredundant_clauses);

        collect = true;

      } else { // Shrunken at least ternary.

        // Reestablish watch invariants if watched literals are removed.

        // Reestablish watch invariants if watched literals are removed.

        Reference src_ref = reference_clause (src);
        if (clause[0] != begin_src_literals[0] &&
            clause[0] == begin_src_literals[1]) {
          unwatch_literal (begin_src_literals[0], src_ref);
          watch_literal (clause[1], clause[0], src_ref);
        } else if (clause[0] == begin_src_literals[0] &&
                   clause[1] != begin_src_literals[1]) {
          unwatch_literal (begin_src_literals[1], src_ref);
          watch_literal (clause[1], clause[0], src_ref);
        } else if (clause[0] != begin_src_literals[0] &&
                   clause[1] != begin_src_literals[1]) {
          unwatch_literal (begin_src_literals[0], src_ref);
          unwatch_literal (begin_src_literals[1], src_ref);
          watch_literal (clause[0], clause[1], src_ref);
          watch_literal (clause[1], clause[0], src_ref);
        } else {
          assert (clause[0] == begin_src_literals[0]);
          assert (clause[1] == begin_src_literals[1]);
        }

        // Finally copy the shrunken clause header and literals.

        *dst = *src; // Just bulk copy the header first.

        // Then update the necessary bits.

        dst->shrink = false;
        dst->size = dst_size;
#ifndef NGLUE
        adjust_glue_to_updated_size (dst);
#endif
#ifndef NREPLACEMENTCACHE
        dst->replaced = 2;
#endif
        memcpy (dst->literals, &*clause.begin (), 4 * dst_size);

        debug_clause (dst, "reduced");

        shrunken_clauses++;
        collect = false;
      }

      clause.clear ();

    } // End of third branch of '!src->garbage'.

    // It remains to either map all still remaining references to the
    // source clause as reasons of literals on the trail or as references
    // in watcher stacks to its new position unless the clause is not
    // copied and needs to be mapped or otherwise marked as being not
    // mapped as it was garbage collected or shrunken to a unit or binary
    // clause.

    if (collect) {

      debug_clause (src, "collecting");

      map.push_back (INVALID_POSITION);
      assert (map[traversed / 4] == INVALID_POSITION);
#ifndef QUIET
      statistics.collected_clauses++;
#endif
      collected_clauses++;

    } else {

      debug_clause (dst, "copied and keeping");

      map.push_back (mapped);
      assert (map[traversed / 4] == mapped);

      update_position (dst);

      assert (!dst->shrink);
      mapped += Clause::words (dst->size);
      dst->shortened = false;
      dst = dst->next ();
    }

    traversed += (unsigned *) next_src - (unsigned *) src;
    src = next_src;

  } // End of 'src != end'.

  clauses.resize (dst);
  if (clauses.shrink_to_fit ())
    message (1, "clauses arena capacity decreased to %zu words",
             clauses.capacity ());

  finalize_position_update ();
#ifndef QUIET
  message (2, "copied %zu large non-garbage clauses", copied_clauses);

  message (2, "reduced %zu large non-garbage clauses (%zu literals)",
           shrunken_clauses, shrunken_literals);
  size_t bytes = 4 * (traversed - (mapped - first_swept));
  message (1, "collected %zu large clauses with %.0f MB (%zu bytes)",
           collected_clauses, bytes / (double) (1 << 20), bytes);
#else
  (void) copied_clauses;
  (void) collected_clauses;
  (void) shrunken_clauses;
  (void) shrunken_literals;
#endif
}

// The following function flushes references to large clauses, maps
// references to large non-garbage clauses to their new position and
// potentially flushes satisfied binary watches (if 'simplify' is 'true').

// It was a hot-spot during reduction of redundant clauses as updating
// watches required to go over all watches in the references stacks
// of all literals.  These are located arbitrarily in memory.  Accessing
// them has poor memory locality if the references are traversed linearly
// along the literal order.  It is more efficient to place all references
// in an arena and traverse them linearly.  This is the main reason we
// place 'References' into an arena.

// At the same time this sweep of the references arena flushes garbage and
// empty references stacks from the arena.  Thus this function also
// serves as a compacting garbage collector. It is even a bit more
// aggressive as it shrinks the capacity of the references to its minimum
// size (in earlier versions we used 'shrink_to_fit' on each references
// stack instead).

void Solver::flush_and_map_watches (Position first_swept, bool simplify,
                                    const std::vector<Position> &map) {
  // When flushing and moving binary watches we save large clause watches
  // (after their blocking literal) on a temporary stack and then put them
  // immediately back to the end of the watcher stack.

  std::vector<Reference> large;

  size_t flushed = 0, mapped = 0;

  // Traversing the arena is way more cache friendly than going over all
  // literals and fixing and flushing stale watches.  At the same time we
  // garbage collect empty references and compactify their data arrays.

  References *begin_arena = *matrix.arena.begin ();
  References *end_arena = *matrix.arena.end ();

  // The first arena references element is the shared empty references
  // stacks (with literal 'INVALID_LITERAL'), which is always kept.

  assert (begin_arena != end_arena);
  assert (begin_arena->lit == INVALID_LITERAL);
  References *dst_refs = begin_arena->next ();
  assert (dst_refs <= end_arena);

  for (References *src_refs = dst_refs, *next_src_refs;
       src_refs != end_arena; src_refs = next_src_refs) {
    next_src_refs = src_refs->next ();
    unsigned lit = src_refs->lit;
    if (lit == INVALID_LITERAL)
      continue;

    assert (lit < literals);
    auto begin_src_watches = src_refs->begin ();
    auto end_src_watches = src_refs->end ();
    auto p = begin_src_watches;

    auto begin_dst_watches = dst_refs->begin ();
    auto q = begin_dst_watches;

    assert (large.empty ());

    // Flush, map or copy watches from source to destination.

    signed char lit_fixed = simplify ? fixed (lit) : 0;

    while (p != end_src_watches) {
      Reference first = *p++;
      if (!tagged (first)) {
        Reference second;
        size_t src_pos;
        if (watching) {
          second = *p++;
          src_pos = untag (second);
        } else {
          second = INVALID_REFERENCE;
          src_pos = untag (first);
        }
        if (first_swept <= src_pos) {
          Position delta = src_pos - first_swept;
          assert (delta / 4 <= map.size ());
          Position dst_pos = map[delta / 4];
          if (dst_pos == INVALID_POSITION)
            continue;
          if (watching)
            large.push_back (first);
          large.push_back (tag_large (dst_pos));
          mapped++;
        } else {
          *q++ = first;
          if (watching)
            *q++ = second;
        }
      } else if (simplify) {
        unsigned other = untag (first);
        auto other_fixed = fixed (other);
        if (lit_fixed <= 0 && other_fixed <= 0)
          *q++ = first;
        else {
          flushed++;
          if (lit > other)
            mark_binary_garbage (lit, other, "satisfied");
        }
      } else
        *q++ = first;
    }

    // We do not need 'LiteralBlockingReference' here as the literal is
    // fixed and 'blocking' and 'ref' can be treated in the same way.

    auto l = large.begin ();
    auto end_large = large.end ();
    while (l != end_large)
      *q++ = *l++;
    large.clear ();

    // Set the destination header shrinking it capacity to fit the size.

    size_t new_size = q - begin_dst_watches;
    size_t new_offset;
    if (new_size) {
      dst_refs->lit = lit;
      dst_refs->_capacity = dst_refs->_size = new_size;
      new_offset = (unsigned *) dst_refs - (unsigned *) begin_arena;
      dst_refs = dst_refs->next ();
    } else
      new_offset = 0; // Flush empty destination references.

    // The following elegant line of code is the reason for explicitly
    // mapping back from each references stack to its literal by adding
    // the 'lit' field.  We have to make sure that we can update the
    // mapping from literals to offsets on-the-fly during our sweep of the
    // arena flushing and collecting garbage references stacks. In Kissat
    // we sort the literals with respect to their offset instead and
    // traverse the sorted list of literals in parallel.  Having the
    // literal in arena avoids sorting and simplifies code slightly.

    matrix.offsets[lit] = new_offset;
  }

  assert (!flushed || simplify);
  message (1, "flushed %zu binary clause watches", flushed);
  message (1, "mapped %zu large clause watches", mapped);
  (void) flushed;
  (void) mapped;

  resize_references_arena (dst_refs);
  matrix.garbage = 0;
}

void Solver::copy_collect_map_and_flush (Position first_swept) {
  std::vector<Position> map;
  copy_and_collect_clauses (first_swept, map);
  bool simplify = last.fixed_variables_at_reduce_or_collect <
                  GET_COUNT (fixed_variables);
  flush_and_map_watches (first_swept, simplify, map);
  if (level)
    map_reasons (first_swept, map);
}

// Very similar to 'reduce' except that we do not reduce learned clauses
// but only collect garbage clauses and simplify clauses.  We also account
// this to be part of simplification and not search.  This function is
// triggered by 'subsume' and 'eliminate' in case clauses are subsumed or
// eliminated.

void Solver::collect_clauses () {

  assert (OPTION (preprocessing) || OPTION (inprocessing));
  assert (OPTION (subsume_clauses) || OPTION (eliminate_variables));

  assert (!level);

  check_all ();

  start ("collect", 2);
#ifndef QUIET
  statistics.garbage_collections++;
  message (2, "collection %" PRIu64 " at conflict %" PRIu64,
           statistics.garbage_collections, statistics.conflicts);
#endif
  Position first_marked = INVALID_POSITION;
  mark_simplified_clauses (first_marked);

  if (first_marked != INVALID_POSITION) {
    debug ("first marked clause at position %" POSFMT, first_marked);
    copy_collect_map_and_flush (first_marked);
  }

  if (!eliminating) {
    assert (watching);
    last.fixed_variables_at_reduce_or_collect = GET_COUNT (fixed_variables);
  } else
    assert (!watching);

  SET_COUNT (garbage_clauses, 0);
  collecting_clauses = false;

  stop ("collect");
  report_progress ('c');

  check_all ();
}

void Solver::garbage_collect_clauses_during_variable_elimination () {

  size_t garbage = GET_COUNT (garbage_clauses);
  double factor = OPTION (elimination_garbage_limit) / 100.0;
  size_t irredundant = GET_COUNT (irredundant_clauses);
  size_t limit = factor * irredundant;
  if (garbage <= limit)
    return;

  message (1, "garbage collection during variable elimination triggered");
  message (1,
           "%zu garbage clauses > limit %zu = %.0f%% of irredundant "
           "clauses %zu",
           garbage, limit, 100 * factor, irredundant);

  assert (!watching);
  flush_eliminated_binary_clauses ();
  collect_clauses ();
}

}; // namespace TabularaSAT
