#ifndef NOPTIONS

#include "solver.hpp"

namespace TabularaSAT {

// List of internal configurations (option sets).

// Note that 'options' is of type 'SolverOptions' and not 'Options' as all
// the configurations work with the generic OPTION ()

struct SolverConfiguration : Configuration {
  SolverOptions &options;
  SolverConfiguration (SolverOptions &o) : options (o) {}
};

#ifndef VANILLA

struct PlainConfiguration : SolverConfiguration {

  const char *name () { return "plain"; }

  const char *description () {
    return "default variant without preprocessing nor inprocessing";
  }

  void set () {
    OPTION (inprocessing) = 0;
    OPTION (preprocessing) = 0;
  }

  PlainConfiguration (SolverOptions &o) : SolverConfiguration (o) {}
};

#endif

#if defined(DEFAULT) || defined(GENERIC)

struct DefaultConfiguration : SolverConfiguration {
  const char *name () { return "default"; }
  const char *description () { return "default variant"; }
  void set () {}
  DefaultConfiguration (SolverOptions &o) : SolverConfiguration (o) {}
};

#endif

#ifndef NACTIVITY

// The MiniSAT style 'vanilla' configuration requires activities instead of
// glues and thus is only available for the 'generic' solver, but not for
// the 'default' solver which saves the space to store activities in clauses
// by defining 'NACTIVITY' in the default 'variants.hpp' file.

struct VanillaConfiguration : SolverConfiguration {

  const char *name () { return "vanilla"; }

  const char *description () {
    return "mimics most recent MiniSAT from 2015/2012/2009";
  }

  void set () {

    OPTION (activity_relevant) = 1;
#ifndef NALLUIPS
    OPTION (all_uips) = 0;
#endif
    OPTION (bump_learned) = 1;
#ifndef NBUMPREASONS
    OPTION (bump_reason_side) = 0;
#endif
#ifndef NCHRONO
    OPTION (chronological_backtracking) = 0;
#endif
#ifndef NELIMINATIONHEAP
    OPTION (eliminate_irrelevant) = 0;
    OPTION (elimination_heap) = 1;
    OPTION (elimination_scores) = 1;
    OPTION (elimination_scoring) = 0;
#endif
#ifndef NGATES
    OPTION (extract_gates) = 0;
#endif
#ifndef NPROBE
    OPTION (failed_literal_probing) = 0;
#endif
#ifndef NGLUE
    OPTION (glue_relevant) = 0;
#endif
#ifndef NSWITCH
    OPTION (initial_mode_stable) = 1;
#endif
    OPTION (initial_phase) = 0;
    OPTION (inprocessing) = 0;
#ifndef NJUMPREASONS
    OPTION (jump_reasons) = 0;
#endif
#ifndef NGLUE
    OPTION (keep_tier1) = 0;
#endif
#ifndef NUSED
    OPTION (keep_used) = 0;
#endif
    OPTION (maximum_elimination_bound) = 0;
    OPTION (minimum_learned) = 0;
#ifndef NOTFS
    OPTION (on_the_fly_strengthen) = 0;
#endif
#ifndef NPROMOTE
    OPTION (promote_clauses) = 0;
#endif
    OPTION (reduce_fraction) = 50;
    OPTION (reluctant_restart_interval) = 100;
#ifndef VANILLA
    OPTION (reduce_vanilla) = 1;
#endif
#ifndef NREUSETRAIL
    OPTION (rephase) = 0;
#endif
#ifndef NREUSETRAIL
    OPTION (reuse_trail) = 0;
#endif
#ifndef NSHRINK
    OPTION (shrink_learned_clause) = 0;
#endif
#ifndef NSIZERELEVANT
    OPTION (size_relevant) = 0;
#endif
#ifndef NGLUE
    OPTION (stable_restart_filter) = 0;
#endif
#ifndef NSUBSTITUTE
    OPTION (substitute_variables) = 0;
#endif
#ifndef NSWITCH
    OPTION (switch_search_mode) = 0;
#endif
#ifndef NTARGET
    OPTION (target_phases) = 0;
#endif
#ifndef NTRANSITIVE
    OPTION (transitive_reduction) = 0;
#endif
#ifndef NVIVIFY
    OPTION (vivify_clauses) = 0;
#endif
  }

  VanillaConfiguration (SolverOptions &o) : SolverConfiguration (o) {}
};

struct CoreConfiguration : VanillaConfiguration {

  const char *name () { return "core"; }
  const char *description () {
    return "same as '--vanilla' but without preprocessing";
  }
  void set () {
    VanillaConfiguration::set ();
    OPTION (preprocessing) = 0;
  }
  CoreConfiguration (SolverOptions &o) : VanillaConfiguration (o) {}
};

struct ParfaitConfiguration : VanillaConfiguration {
  const char *name () { return "parfait"; }
  const char *description () {
    return "same as '--vanilla' with inprocessing (no preprocessing)";
  }
  void set () {
    VanillaConfiguration::set ();
    OPTION (inprocessing) = 1;
    OPTION (preprocessing) = 0;
  }
  ParfaitConfiguration (SolverOptions &o) : VanillaConfiguration (o) {}
};

struct SemifreddoConfiguration : VanillaConfiguration {
  const char *name () { return "semifreddo"; }
  const char *description () {
    return "same as '--vanilla' with preprocessing and inprocessing";
  }
  void set () {
    VanillaConfiguration::set ();
    OPTION (inprocessing) = 1;
  }
  SemifreddoConfiguration (SolverOptions &o) : VanillaConfiguration (o) {}
};

struct StraciatellaConfiguration : VanillaConfiguration {
  const char *name () { return "straciatella"; }
  const char *description () {
    return "same as '--semifreddo' with all simplifiers except "
           "'vivification'";
  }
  void set () {
    VanillaConfiguration::set ();
    OPTION (inprocessing) = 1;
    OPTION (substitute_variables) = 1;
    OPTION (transitive_reduction) = 1;
  }
  StraciatellaConfiguration (SolverOptions &o) : VanillaConfiguration (o) {}
};

#endif

SolverConfigurations::SolverConfigurations (SolverOptions &o)
    : options (o) {
#ifdef VANILLA
  {
    VanillaConfiguration configuration (o);
    configuration.set ();
  }
#endif
}

void SolverConfigurations::initialize () {
  assert (!initialized);
#ifndef NACTIVITY
  list.push_back (new CoreConfiguration (options));
  list.push_back (new ParfaitConfiguration (options));
  list.push_back (new SemifreddoConfiguration (options));
  list.push_back (new StraciatellaConfiguration (options));
  list.push_back (new VanillaConfiguration (options));
#endif
#ifndef VANILLA
  list.push_back (new PlainConfiguration (options));
#endif
#if defined(DEFAULT) || defined(GENERIC)
  list.push_back (new DefaultConfiguration (options));
#endif
  initialized = true;
}

SolverConfigurations::~SolverConfigurations () {
  if (initialized)
    for (auto configuration : list)
      delete configuration;
}

}; // namespace TabularaSAT

#endif
