#include "references.hpp"
#include "solver.hpp"
#include "ticks.hpp"
#include "watch.hpp"

#include <cassert>
#include <cstdint>

namespace TabularaSAT {

// We have four dedicated propagation functions.  Earlier we used one
// generic 'propagate' function with arguments and inlined it here four
// times trusting the compiler to do its job and really instantiate them by
// constant propagation.  Instead we now have put the propagation code into
// 'propagate.hpp' in four versions controlled by macros.  This makes sure
// that already the compiler sees four versions and constant propagation
// and thus specialization is ensured (and also does not depend on
// the 'vivifying', 'probing' and 'warming' flags).

#define SEARCH_PROPAGATE
#include "propagate.hpp"
#undef SEARCH_PROPAGATE

#ifndef NVIVIFY
#define VIVIFY_PROPAGATE
#include "propagate.hpp"
#undef VIVIFY_PROPAGATE
#undef PROPAGATE
#endif

#ifndef NWALK
#define WARMING_PROPAGATE
#include "propagate.hpp"
#undef WARMING_PROPAGATE
#endif

#ifndef NPROBE
#define PROBING_PROPAGATE
#include "propagate.hpp"
#undef PROBING_PROPAGATE
#endif

// As a result of strengthening in subsumption we might produce unit
// clauses, which can not be propagated during subsumption (easily).
// Therefore we just learn them implicitly (and trace them already for
// instance) but instead of assigning them eagerly during subsumption push
// them on a stack for delayed assignment and propagation, which is
// triggered then in the next round of the CDCL propagation loop.

void Solver::assign_and_propagate_units () {
  assert (!level);
  assert (watching);
  assert (!inconsistent);
  assert (!units.empty ());
  message (2, "assigning and propagated %zu delayed units", units.size ());
  unsigned assigned = 0;
  for (auto unit : units) {
    unsigned idx = index (unit);
    if (flags[idx].eliminated) {
      debug ("already eliminated delayed unit %s", debug_literal (unit));
      continue;
    }
    signed char value = values[unit];
    if (value > 0) {
      debug ("already satisfied delayed unit %s", debug_literal (unit));
      continue;
    }
    if (value < 0)
      debug ("falsified delayed unit %s", debug_literal (unit));
    else {
      debug ("assigning delayed unit %s", debug_literal (unit));
      assign (unit, UNIT_REASON);
      assigned++;
      if (search_propagate ())
        continue;
    }
    inconsistent = true;
    trace_empty ('a');
    return;
  }
  units.clear ();
#ifndef QUIET
  if (assigned || verbosity > 0)
    report_progress ('a');
#else
  (void) assigned;
#endif
}

void Solver::propagate_after_simplification () {
  assert (!level);
  if (inconsistent)
    return;
  if (!search_propagate ()) {
    inconsistent = true;
    trace_empty ('a');
  } else if (!units.empty ())
    assign_and_propagate_units ();
}
}; // namespace TabularaSAT
