#ifndef _tabularasat_schedule_hpp_INCLUDED
#define _tabularasat_schedule_hpp_INCLUDED

#include "inline.hpp"
#include "solver.hpp"

#include <cstdint>

namespace TabularaSAT {

struct Wrapper {
  Solver &solver;
  Wrapper (Solver *s) : solver (*s) {}
  virtual const char *name () = 0;
};

struct Scheduler : virtual Wrapper {

  virtual uint64_t &count () = 0;   // How often it was called.
  virtual uint64_t &limit () = 0;   // The conflict limit reference.
  virtual unsigned interval () = 0; // The base conflict interval.

  // How often called during preprocessing (returing a zero pointer means
  // that this scheduler can never be called during preprocessing).

  virtual uint64_t *preprocessed () { return 0; }

  virtual unsigned initial () { // The initial conflict interval.
    return interval ();         // Default is to use the base interval.
  }

  virtual double function (uint64_t) = 0;

  // Scale the conflict limit both initially and during updates by the
  // 'Solver::scale ()' function which takes the formula size into account.
  // Default is to not to scale while inprocessing does it.

  virtual double scale () { return 1.0; }

  void increment_counter (const char *prefix = "");
  void set_next_conflict_limit ();
  void set_initial_conflict_limit ();
};

struct PowerLogScheduler : Scheduler {

  // This is the exponent aka second argument to 'nlogpow' which is used to
  // stretch the base conflict interval based on the number of times this
  // function was scheduled (the 'count').  With default zero we linearly
  // increase the conflict interval, i.e., for 'reduce_redundant_clauses',
  // and in general it is multiplied by 'count * log^exponent (count)'.

  virtual int exponent () = 0;

  virtual double function (uint64_t c) { return nlogpow (c, exponent ()); }
};

// Overwrite 'scale ()' by calling the solver's 'scale ()' function.

struct PowerLogScalingScheduler : PowerLogScheduler {
  double scale () { return solver.scale (); }
};

// Used for routines that need to determine the number of 'ticks' invested.

struct Effort : virtual Wrapper {
  virtual uint64_t &last () = 0;
  virtual unsigned effort () = 0;
  virtual uint64_t get_ticks_limit ();
};

#ifndef VANILLA

struct ReduceScheduler : Scheduler {
  const char *name () { return "clause reduction"; }
  uint64_t &count () { return GET_SOLVER_COUNT (reductions); }
  uint64_t &limit () { return solver.limits.reduce.conflicts; }
  unsigned interval () { return SOLVER_OPTION (reduce_interval); }
  double function (uint64_t c) { return sqrt (c); }
  ReduceScheduler (Solver *s) : Wrapper (s) {}
};

#endif

#ifndef NREPHASE

struct RephaseScheduler : PowerLogScheduler {
  const char *name () { return "rephasing"; }
  uint64_t &count () { return GET_SOLVER_COUNT (rephasings); }
  uint64_t &limit () { return solver.limits.rephase; }
  unsigned interval () { return SOLVER_OPTION (rephase_interval); }
  int exponent () { return 5; }
  RephaseScheduler (Solver *s) : Wrapper (s) {}
};

#endif

struct SubsumptionScheduler : PowerLogScalingScheduler, Effort {

  const char *name () { return "irredundant clause subsumption"; }

  // Scheduling part.

  uint64_t &count () { return GET_SOLVER_COUNT (subsumptions); }
  uint64_t &limit () { return solver.limits.subsume; }
  unsigned interval () { return SOLVER_OPTION (subsumption_interval); }
  unsigned initial () { return SOLVER_OPTION (initial_subsumption_limit); }
  int exponent () { return 3; }

  uint64_t *preprocessed () {
    return &GET_SOLVER_COUNT (preprocessing_subsumptions);
  }

  // Effort part.

  uint64_t &last () { return solver.last.search_ticks_at_subsumption; }
  unsigned effort () { return SOLVER_OPTION (subsumption_effort); }

  SubsumptionScheduler (Solver *s) : Wrapper (s) {}
};

struct EliminationScheduler : PowerLogScalingScheduler, Effort {

  const char *name () { return "variable elimination"; }

  // Scheduling part.

  uint64_t &count () { return GET_SOLVER_COUNT (eliminations); }
  uint64_t &limit () { return solver.limits.eliminate; }
  unsigned interval () { return SOLVER_OPTION (elimination_interval); }
  unsigned initial () { return SOLVER_OPTION (initial_elimination_limit); }
  int exponent () { return 4; }

  uint64_t *preprocessed () {
    return &GET_SOLVER_COUNT (preprocessing_eliminations);
  }

  // Effort part.

  uint64_t &last () { return solver.last.search_ticks_at_elimination; }
  unsigned effort () { return SOLVER_OPTION (elimination_effort); }

  EliminationScheduler (Solver *s) : Wrapper (s) {}
};

#ifndef NSUBSTITUTE

struct SubstitutionScheduler : PowerLogScalingScheduler {

  const char *name () { return "variable substitution"; }

  uint64_t &count () { return GET_SOLVER_COUNT (substitutions); }
  uint64_t &limit () { return solver.limits.substitute; }
  unsigned interval () { return SOLVER_OPTION (substitution_interval); }
  unsigned initial () { return SOLVER_OPTION (initial_substitution_limit); }
  int exponent () { return 2; }

  uint64_t *preprocessed () {
    return &GET_SOLVER_COUNT (preprocessing_substitutions);
  }

  SubstitutionScheduler (Solver *s) : Wrapper (s) {}
};

#endif

#ifndef NPROBE

struct FailedLiteralScheduler : PowerLogScalingScheduler, Effort {

  const char *name () { return "probing"; }

  // Scheduling part.

  uint64_t &count () { return GET_SOLVER_COUNT (failed_literal_probings); }
  uint64_t &limit () { return solver.limits.probe; }
  unsigned interval () { return SOLVER_OPTION (failed_literal_interval); }
  unsigned initial () {
    return SOLVER_OPTION (initial_failed_literal_limit);
  }
  int exponent () { return 2; }

  uint64_t *preprocessed () {
    return &GET_SOLVER_COUNT (preprocessing_probings);
  }

  // Effort part.

  unsigned effort () { return SOLVER_OPTION (failed_literal_effort); }
  uint64_t &last () {
    return solver.last.search_ticks_at_failed_literal_probing;
  }

  FailedLiteralScheduler (Solver *s) : Wrapper (s) {}
};

#endif

#ifndef NVIVIFY

struct VivificationScheduler : PowerLogScalingScheduler, Effort {

  const char *name () { return "vivification"; }

  // Scheduling part.

  uint64_t &count () { return GET_SOLVER_COUNT (vivifications); }
  uint64_t &limit () { return solver.limits.vivify; }
  unsigned interval () { return SOLVER_OPTION (vivification_interval); }
  unsigned initial () { return SOLVER_OPTION (initial_vivification_limit); }
  int exponent () { return 1; }

  uint64_t *preprocessed () {
    return &GET_SOLVER_COUNT (preprocessing_vivifications);
  }

  // Effort part.

  unsigned effort () { return SOLVER_OPTION (vivification_effort); }
  uint64_t &last () { return solver.last.search_ticks_at_vivification; }

  VivificationScheduler (Solver *s) : Wrapper (s) {}
};

#endif

#ifndef NTRANSITIVE

struct TransitiveReductionScheduler : PowerLogScalingScheduler, Effort {
  const char *name () { return "transitive binary clause reduction"; }

  // Scheduling part.

  uint64_t &count () { return GET_SOLVER_COUNT (transitive_reductions); }
  uint64_t &limit () { return solver.limits.transitive; }
  unsigned interval () { return SOLVER_OPTION (transitive_interval); }
  unsigned initial () { return SOLVER_OPTION (initial_transitive_limit); }
  int exponent () { return 1; }

  uint64_t *preprocessed () {
    return &GET_SOLVER_COUNT (preprocessing_transitive_reductions);
  }

  // Effort part.

  unsigned effort () { return SOLVER_OPTION (transitive_effort); }
  uint64_t &last () {
    return solver.last.search_ticks_at_transitive_reduction;
  }

  TransitiveReductionScheduler (Solver *s) : Wrapper (s) {}
};

#endif

}; // namespace TabularaSAT

#endif
